(*
	interpreter: 
		ocaml -I ../xml-light str.cma xml-light.cma fls.ml
*)

include Keyword_list;;

exception Invalid_expression of string;;
exception Empty_expression;;

let remove_spaces = Str.global_replace (Str.regexp " ") "";;
let remove_tabs = Str.global_replace (Str.regexp "\t") "";;
let remove_newlines = Str.global_replace (Str.regexp "\n") "";;
let concat_string_list = List.fold_left (^) "" ;;

(*return list of pairs (a,b) with positions of all parentheses in string s;
a: position,
b: 1 for '(', (-1) for ')'*)
let parentheses s =
	let rec search_from c l i =
		if i <= String.length s then
		(
			try
			let p = String.index_from s i c in
			search_from c ((p, c)::l) (p+1)
			with
			Not_found -> l
		)
		else
			l
	in
	let leftp = search_from '(' [] 0
	and rightp = search_from ')' [] 0 in
	let l =
		List.sort
			(
				fun a b ->
					if (fst a) < (fst b) then 
						(-1)
					else
						1
			)
			(leftp @ rightp)
	in
	List.map
		(
			fun t ->
				(
					fst t, 
					if snd t = '(' then
						1
					else
						(-1)
				)
		)
		l
;;

(*list of XML tags which have only attribute "Identifier":*)
let idtaglist = 
[
	"Declaration"; 
	"DeclaredContents"; 
	"DisplayObjectList";
	"DisplayObject";
	"FromList";
	"ListLength";
	"ListIndex";
	"ListMember";
	"StimulusListItem";
	"ImageListItem";
	"ItemWithIdentifier";
	"InsertList";
	"SetVariable";
	"SetVariableByEvent";
	"SetStringVariableByEvent";
	"SetMatrixElementByEvent";
	"SetMatrixElementAtSceneStart";
	"DeclaredList";
	"ListFromDeclaredVector";
	"DeclaredFunction";
	"DeclaredScalarFunction";
	"DeclaredMatrixFunction";
	"DeclaredStringFunction";
	"DeclaredString";
	"DeclaredMatrix";
	"DeclaredValue";
	"Symbol"
];;

let drop_first_and_last_char st =
	let s = Str.string_before st ((String.length st) - 1) in
	Str.string_after s 1
;;

(** [preprocess_fls_string flsstring]: 
strip comments and replace strings in string [flsstring];
return Hashtable with strings and preprocessed string*)
let preprocess_fls_string flsstring =
	let stringhash = Hashtbl.create 10 in
	(*substitute "internal" strings (strings within strings) with \n:*)
	let st1 = Str.global_replace (Str.regexp_string "\\\"") "\n#\n" flsstring in
	let r = Str.regexp "\"[^\"]*\"" in
	let rec fill_hash s i pos = 
		try
		let new_pos = Str.search_forward r s pos in
		let ms = 
			let s1 = 
				Str.global_replace 
					(Str.regexp_string "\n#\n") 
					"\""
					(Str.matched_string s) 
			in
			drop_first_and_last_char s1
		in
		Hashtbl.add stringhash i ms;
		let new_s = Str.replace_first r ("@" ^ (string_of_int i) ^ "@") s in
		fill_hash new_s (i+1) new_pos
		with
			Not_found -> s
	in
	let st2 = fill_hash st1 1 0 in
	let comments_removed =
		(*everything after ;, like in common lisp*)
		let st3 = Str.global_replace (Str.regexp ";.*$") "" st2 in
		(*everything between #| and |#, like in common lisp*)
		Str.global_replace (Str.regexp "#|[^|]*|#") "" st3
	in
	(stringhash, comments_removed)
;;

(** [xml_list_of_fls_string flsstring]: convert string [flsstring] to list of Xml.xml *)
let xml_list_of_fls_string flsstring =
	try
	let symbols = 
	[
		("'", "Quote"); 
		("!!", "UnquoteSplicing"); 
		("!", "Unquote");
		("&", "Eval")
	] in
	let listlast l = List.hd (List.rev l) in
	(*add quotes to functions' argument lists:*)
	let flsstring1 = 
		let s1 = Str.global_replace 
			( Str.regexp ("(Function[^[A-Za-z]*\\[\\([^]=\"]*\\)\\]" ) )
			" (Function [\"\\1\"]" 
			flsstring
		in
		let s2 = Str.global_replace 
			( Str.regexp ("(MatrixFunction[^[]*\\[\\([^]=\"]*\\)\\]" ) )
			" (MatrixFunction [\"\\1\"]" 
			s1
		in
		let s3 = Str.global_replace 
			( Str.regexp ("(StringFunction[^[]*\\[\\([^]=\"]*\\)\\]" ) )
			" (StringFunction [\"\\1\"]" 
			s2
		in
		Str.global_replace 
			( Str.regexp ("(SourceFunction[^[]*\\[\\([^]=\"]*\\)\\]" ) )
			" (SourceFunction [\"\\1\"]" 
			s3
	in
	let (stringhash, flsstring2) = preprocess_fls_string flsstring1 in
	(*if empty string:*)
	let comments_removed =
		if remove_newlines (remove_spaces (remove_tabs flsstring2)) = "" then
			raise Empty_expression
		else flsstring2
	in
	(*substitute all strings with @n@ where n is number of string:*)
	let flsstr = 
		(*substitute the "syntactic sugar", i.e. single numbers, IDs, $arg etc.:*)
		let synt_sugar =
			let sp = "[ \t\n)']"
			and ap = "[_A-Za-z0-9.-]"
			and symbs = (List.fold_left (fun a b -> a ^ (fst b)) "" symbols) in
			let sy = ("[" ^ symbs ^ "]")
			and apl = "[_A-Za-z0-9. \t\n,-]" in
			let ops = 
			[
				("\\+", "Plus");
				("-", "Subtract");
				("\\*", "Times");
				("\\*\\*", "Power");
				("/", "Divide");
				("<", "Less");
				(">", "Greater");
				("<>", "Unequal");
				("=", "Equal");
				("<=", "LessEqual");
				(">=", "GreaterEqual");
				(":=", "Declaration");
				("#", "SourceList")
			] in
			let do_op sourcestring t =
				let s1 = Str.global_replace 
					( Str.regexp ("([ \t\n]*" ^ (fst t) ^ "\\([( \t\n]+\\)" ) )
					("("^ (snd t) ^"\\1") 
					sourcestring
				in
				Str.global_replace 
					( Str.regexp ("([ \t\n]*" ^ (fst t) ^ "\\([( \t\n]+\\)" ) )
					("("^ (snd t) ^"\\1") 
					s1
			in
			let simp_op1 =
				List.fold_left do_op comments_removed ops
			in
			(*(Declaration x ...) -> (Declaration [x] ...):*)
			let simp_op = 
				Str.global_replace 
					( Str.regexp ("(Declaration" ^ sp ^ "*\\(" ^ ap ^ "+\\)" ^ sp ^ "*" ) )
					"(Declaration [\\1] " 
					simp_op1
			in
			let wo_single_numbers_and_ids = 
				(*if whole string is only single number or ID (i. e. no parentheses):*)
				let simp_op_no_spaces = remove_newlines (remove_spaces (remove_tabs simp_op)) in
				if Str.string_match (Str.regexp ("[^()"^ symbs ^"]+$")) simp_op_no_spaces 0 then
					let t = 
						try
						ignore (float_of_string simp_op);
						"(NumericValue [Value = "
						with Failure _ ->
						(
							if simp_op_no_spaces.[0] == '@' then
								"(String [Value = "
							else
								"(DeclaredValue [Identifier = "
						)
					in
					t ^ simp_op_no_spaces ^ "])"
				else
				(
					(*numeric values; this is needed to be performed twice:*)
					let st1' = Str.global_replace 
						( Str.regexp ("\\([^[,=]" ^ sp ^ "+\\)\\([0-9-.]+\\)\\(" ^ sp ^ "\\)" ) )
						"\\1(NumericValue [Value = \\2])\\3" 
						simp_op 
					in
					let st1 = Str.global_replace 
						( Str.regexp ("\\([^[,=]" ^ sp ^ "+\\)\\([0-9-.]+\\)\\(" ^ sp ^ "\\)" ) )
						"\\1(NumericValue [Value = \\2])\\3" 
						st1' 
					in
					(*strings; this is needed to be performed twice:*)
					let process_strings1 =Str.global_replace 
						( Str.regexp ("\\([^[,=(\t\n ]" ^ sp ^ "+\\)\\(@[0-9]+@\\)\\(" ^ sp ^ "\\)" ) )
						"\\1(String [Value = \\2])\\3" 
						st1 
					in
					let process_strings =Str.global_replace 
						( Str.regexp ("\\([^[,=(\t\n ]" ^ sp ^ "+\\)\\(@[0-9]+@\\)\\(" ^ sp ^ "\\)" ) )
						"\\1(String [Value = \\2])\\3" 
						process_strings1 
					in
					(*IIDs; this is needed to be performed twice:*)
					let single_id_with_symbol = Str.global_replace 
						( Str.regexp ("^\\(" ^ sp ^ "*" ^ sy ^ "*\\)\\(" ^ ap ^ "+\\)\\(" ^ sp ^ "*\\)$" ) ) 
						"\\1(DeclaredValue [Identifier = \\2])\\3"
						process_strings
					in
					let st2 = Str.global_replace 
						( Str.regexp ("\\([^[,=(\t\n ]" ^ sp ^ "+" ^ sy ^ "*\\)\\(" ^ ap ^ "+\\)\\(" ^ sp ^ "\\)" ) )
						"\\1(DeclaredValue [Identifier = \\2])\\3" 
						single_id_with_symbol 
					in
					Str.global_replace 
						( Str.regexp ("\\([^[,=(\t\n ]" ^ sp ^ "+" ^ sy ^ "*\\)\\(" ^ ap ^ "+\\)\\(" ^ sp ^ "\\)" ) )
						"\\1(DeclaredValue [Identifier = \\2])\\3" 
						st2
				)
			in
			let wo_args =
				Str.global_replace 
					( Str.regexp ("\\$\\(" ^ ap ^ "+\\)\\(" ^ sp ^ "\\)" ) )
					" (Argument [Identifier = \\1])\\2" 
					wo_single_numbers_and_ids
			in
			(*(idtag [x]) -> (idtag [Identifier = "x"]);
			use global list idtaglist for this:*)
			let process_id_tag sourcestring t =
				Str.global_replace 
					( Str.regexp ("(" ^ t ^ sp ^ "*\\[" ^ sp ^ "*\\(@?" ^ ap ^ "+@?\\)" ^ sp ^ "*\\]" ) )
					(" (" ^ t ^" [Identifier = \\1]") 
					sourcestring
			in
			let simp =
				List.fold_left process_id_tag wo_args idtaglist
			in

			(*(Function [x]) -> (Function [ArgumentList = "x"]):*)
			let simp_funs =
				let s1 = Str.global_replace 
					( Str.regexp ("(Function" ^ sp ^ "*\\[" ^ sp ^ "*\\(@?" ^ apl ^ "+@?\\)" ^ sp ^ "*\\]" ) )
					" (Function [ArgumentList = \\1]" 
					simp
				in
				let s2 = Str.global_replace 
					( Str.regexp ("(MatrixFunction" ^ sp ^ "*\\[" ^ sp ^ "*\\(@?" ^ apl ^ "+@?\\)" ^ sp ^ "*\\]" ) )
					" (MatrixFunction [ArgumentList = \\1]" 
					s1
				in
				let s3 = Str.global_replace 
					( Str.regexp ("(StringFunction" ^ sp ^ "*\\[" ^ sp ^ "*\\(@?" ^ apl ^ "+@?\\)" ^ sp ^ "*\\]" ) )
					" (StringFunction [ArgumentList = \\1]" 
					s2
				in
				Str.global_replace 
					( Str.regexp ("(SourceFunction" ^ sp ^ "*\\[" ^ sp ^ "*\\(@?" ^ apl ^ "+@?\\)" ^ sp ^ "*\\]" ) )
					" (SourceFunction [ArgumentList = \\1]" 
					s3
			in
			simp_funs
		in
		remove_newlines (remove_spaces (remove_tabs synt_sugar))
	in
	
	let rec fls_string_list_of_fls_string flsstringlist = function
		| "" -> flsstringlist
		| fs ->
		(
			match parentheses fs with
			[] -> raise (Invalid_expression fs)
			| ph::parenths -> 
			(
				let split_fls_string parenth s =
					let rec stacklist = function
						| ([], _) -> raise (Invalid_expression s)
						| (h::t, state) ->
							let nstack = state + (snd h) in
							if nstack = 0 then
								(fst h)
							else
								stacklist (t, nstack)
					in
					let pos = ((stacklist (parenth, 1)) + 1) in
					let newstring = Str.first_chars s pos in
					let reststring = Str.string_after s pos in
					(newstring, reststring)
				in
				let rec check_symbols_or_procede = function
					| [] -> 
					(
						if (fst ph) > 0 || (snd ph) < 1 then
							failwith ("'" ^ fs ^ "' does not start with '('");
						let ll = listlast parenths in
						if (fst ll) < ((String.length fs) - 1) || (snd ll) > (-1) then
							failwith ("'" ^ fs ^ "' does not end with ')'");
						let (newstring, reststring) = split_fls_string parenths fs in
						fls_string_list_of_fls_string (flsstringlist@[newstring]) reststring
					)
					| (symb, keyword)::t ->
						if Str.string_match (Str.regexp ("[(]?[^(]*" ^ symb ^ "\\(.*\\)")) fs 0 then	
						(
							let rest = Str.matched_group 1 fs in
							match parentheses rest with
							[] -> raise (Invalid_expression fs)
							| ph::ps -> 
							let first_expression_after = (fst (split_fls_string ps rest)) in
							let quotestring = 
								Str.replace_first 
									(Str.regexp_string (symb ^ first_expression_after)) 
									("(" ^ keyword ^ first_expression_after ^ ")") 
									fs 
							in
							fls_string_list_of_fls_string flsstringlist quotestring
						)
						else
							check_symbols_or_procede t
				in
				check_symbols_or_procede symbols
			)	
		)
	in
	let tag_and_arglist_of_xml_string xmlst =
		try
		let _ = Str.search_forward (Str.regexp "\\[[^]]*\\]") xmlst 0 in
		let st = drop_first_and_last_char (Str.matched_string xmlst) in
		let xmltag = List.hd (Str.split (Str.regexp_string "[") xmlst) in
		let argstringlist = Str.split (Str.regexp ",") st in
		let argpair_of_string argstr =
			match Str.split (Str.regexp "=") argstr with
			[] -> failwith ("Argument string '[" ^ argstr ^ "]' is not valid.")
			| h::t -> 
			(
				match t with
				[] -> failwith ("Argument string '[" ^ argstr ^ "]' is not valid.")
				| q::t1 -> 
					if Str.string_match (Str.regexp "@.*") q 0 then
						let q1 =
							Hashtbl.find 
								stringhash 
								(int_of_string (drop_first_and_last_char q))
						in
						(h, q1)
					else
						(h, q)
			)
		in
		(xmltag, List.map argpair_of_string argstringlist)
		with
			Not_found -> (xmlst, [])
	in
	
	let rec xml_of_fls fstr =
		let node =
			let st = drop_first_and_last_char fstr in
			let rec new_xml_element xmltag arglist new_flsstring =
				if List.mem xmltag keyword_list then
					Xml.Element
					(
						xmltag,
						arglist,
						List.map xml_of_fls (fls_string_list_of_fls_string [] new_flsstring)
					)
				else
					new_xml_element "DeclaredFunction" ([("Identifier", xmltag)] @ arglist) new_flsstring
			in
			match parentheses st with
			[] -> 
				let (xmltag, arglist) = tag_and_arglist_of_xml_string st in
				if Str.string_match (Str.regexp "@.*") xmltag 0 then
					Xml.PCData
					(
						Hashtbl.find 
							stringhash 
							(int_of_string (drop_first_and_last_char xmltag))
					)
				else
					new_xml_element xmltag arglist ""
					(*Xml.Element (xmltag, arglist, [])*)
			| h::t ->
				let (xmltag, arglist) = 
					tag_and_arglist_of_xml_string (Str.string_before st (fst h))
				in
				let new_flsstring = Str.string_after st (fst h) in
				new_xml_element xmltag arglist new_flsstring
		in
		(*convert quoted source to symbols:*)
		let rec check_for_quotes = function
			| Xml.Element ("Quote", _, cld) ->
				let node_of_attrib = fun (k, t) ->
					Xml.Element 
					(
						"SymbolAttribute", 
						[("AttributeTag", k); ("AttributeValue", t)],
						[]
					)
				in
				let rec quote_xml x =
				(
					match Xml.tag x with
					"Symbol" -> x
					| "Unquote" | "UnquoteSplicing" -> 
						Xml.Element 
						(
							Xml.tag x, 
							Xml.attribs x, 
							(List.map check_for_quotes (Xml.children x))
						)
					| "DeclaredValue" ->
						Xml.Element 
						(
							"Symbol", 
							["Identifier", Xml.attrib x "Identifier"], 
							[]
						)
					| "DeclaredFunction" ->
						let att = List.filter (fun t -> fst t <> "Identifier") (Xml.attribs x) in
						Xml.Element 
						(
							"Symbol", 
							["Identifier", Xml.attrib x "Identifier"], 
							(List.map node_of_attrib att) 
								@ (List.map quote_xml (Xml.children x))
						)
					| tg ->
						Xml.Element 
						(
							"Symbol", 
							["Identifier", tg], 
							(List.map node_of_attrib (Xml.attribs x)) 
								@ (List.map quote_xml (Xml.children x))
						)
				) in
				if cld = [] then
					failwith "Quote is called without arguments."
				else
					quote_xml (List.hd cld)
			| t -> t
		in
		check_for_quotes node
	in
	List.map 
		xml_of_fls
		(fls_string_list_of_fls_string [] flsstr)
	
	with Empty_expression -> [Xml.Element ("String", ["Value", ""], [])]
;;

(** [convert_fls_tags elem]: convert all Fls tags in XML element [elem] to pure XML *)
let rec convert_fls_tags = function
	| Xml.Element (xmltag, arglist, cld) ->
		if xmltag = "Fls" then
		(
			let flscont = 
				match List.hd cld with
				Xml.PCData s -> s
				| _ -> failwith ""
			in
			let xl = xml_list_of_fls_string flscont in
			if (List.length xl) <> 1 then
				failwith (flscont ^ " does not denote a single XML element.");
			List.hd xl
		)
		else
			Xml.Element (xmltag, arglist, (List.map convert_fls_tags cld))
	| p -> p
;;

(** [fls_string_of_xml elem]: Convert Xml.xml element [elem] to an fls string *)
let fls_string_of_xml elem =
	let argstring_of_arglist = function
		| [] -> ""
		| al -> 
		(
			let f = function
				| (a, b) ->
					let s = 
					(
						try
						let t = (string_of_float (float_of_string b)) in
						if Str.last_chars t 1 = "." then
							Str.string_before t ((String.length t) - 1)
						else
							t
						with Failure _ -> 
							let s1 = Str.global_replace (Str.regexp_string "\"") "\\\"" b in
							"\"" ^ s1 ^ "\""
					) in
					a ^ " = " ^ s ^ ", "
			in
			let slist = List.map f al in
			let st = concat_string_list slist in
			let last_dropped =  Str.string_before st ((String.length st) - 2) in
			" [" ^ last_dropped ^ "]"
		)
	in
	let rec fsox init_string = function
		| Xml.Element (xmltag, arglist, cld) ->
			init_string
			^ "(" 
			^ xmltag 
			^ (argstring_of_arglist arglist) 
			^ (concat_string_list (List.map (fsox (init_string ^ "  ")) cld)) 
			^ ")"
		| Xml.PCData pcd -> "(\"" ^ pcd ^ "\")"
	in
	(*convert symbols to Quote:*)
	let quoted =
		let rec check_for_symbols = function
			| Xml.Element ("Symbol", atts, cld) as symb->
				let rec node_of_symbol = function
					| Xml.Element ("Symbol", sy_atts, sy_cld) as sy ->
						let (nodeattlist, restlist) = 
							List.partition (fun t -> Xml.tag t = "SymbolAttribute") sy_cld
						in
						let attlist =
							let att_of_attnode an =
								(Xml.attrib an "AttributeTag", Xml.attrib an "AttributeValue")
							in
							List.map att_of_attnode nodeattlist
						in
						let ident = Xml.attrib sy "Identifier" in
						if List.mem ident keyword_list then 
							Xml.Element (ident, attlist, List.map node_of_symbol restlist)
						else
						(
							if restlist = [] then
								Xml.Element ("DeclaredValue", ["Identifier", ident], [])
							else
								Xml.Element 
								(
									"DeclaredFunction", 
									("Identifier", ident) :: attlist,
									(*["Identifier", ident],*)
									List.map node_of_symbol restlist
								)
						)
					| Xml.Element _ as el -> check_for_symbols el
					| t -> t
				in
				Xml.Element 
				(
					"Quote",
					[],
					[node_of_symbol symb]
				)
			| Xml.Element (key, atts, cld) -> Xml.Element (key, atts, List.map check_for_symbols cld)
			| t -> t
		in
		check_for_symbols elem
	in
	let long_string = Str.string_after (fsox "\n" quoted) 1 in
	(*** "syntactic sugar": ****)
	(*NumericValue [Value = "n"] -> n*)
	let wo_numval = 
		Str.global_replace 
			( Str.regexp ("(NumericValue \\[Value = \"?\\([^)\"]*\\)\"?\\])" ) )
			"\\1"
			long_string
	in
	let wo_decl =
		Str.global_replace 
			( Str.regexp ("(DeclaredValue \\[Identifier = \"?\\([^)\"]*\\)\"?\\])" ) )
			"\\1"
			wo_numval
	in
	let wo_str =
		Str.global_replace 
			( Str.regexp ("(String \\[Value = \"?\\([^)\"]*\\)\"?\\])" ) )
			"\"\\1\""
			wo_decl
	in
	let wo_arg =
		Str.global_replace 
			( Str.regexp ("(Argument \\[Identifier = \"?\\([^)\"]*\\)\"?\\])" ) )
			"$\\1"
			wo_str
	in
	let process_id_tag sourcestring t =
		Str.global_replace 
			( Str.regexp ("(" ^ t ^ " \\[Identifier = \"?\\([^)\",]*\\)\"?\\]" ) )
			("(" ^ t ^ " [\\1]")
			sourcestring
	in
	let simp =
		List.fold_left process_id_tag wo_arg idtaglist
	in

	let wo_fun =
		let wo_declared_fun_with_attributes =
			Str.global_replace 
				( Str.regexp ("(DeclaredFunction \\[Identifier = \"\\([^\"]*\\)\", \\([^]]*\\)\\]" ) )
				"(\\1 [\\2]"
				simp
		in
		let wo_declared_fun =
			Str.global_replace 
				( Str.regexp ("(DeclaredFunction \\[\\([^)]*\\)\\]" ) )
				"(\\1"
				wo_declared_fun_with_attributes
		in
		Str.global_replace 
			( Str.regexp ("(Function \\[ArgumentList = \"?\\([^)\"]*\\)\"?\\]" ) )
			"(Function [\\1]"
			wo_declared_fun
	in
	let wo_strfun =
		Str.global_replace 
			( Str.regexp ("(StringFunction \\[ArgumentList = \"?\\([^)\"]*\\)\"?\\]" ) )
			"(StringFunction [\\1]"
			wo_fun
	in
	let wo_srcfun =
		Str.global_replace 
			( Str.regexp ("(SourceFunction \\[ArgumentList = \"?\\([^)\"]*\\)\"?\\]" ) )
			"(SourceFunction [\\1]"
			wo_strfun
	in
	let wo_matfun =
		Str.global_replace 
			( Str.regexp ("(MatrixFunction \\[ArgumentList = \"?\\([^)\"]*\\)\"?\\]" ) )
			"(MatrixFunction [\\1]"
			wo_srcfun
	in
	let ops = 
	[
		("+", "Plus");
		("-", "Subtract");
		("*", "Times");
		("**", "Power");
		("/", "Divide");
		("<", "Less");
		(">", "Greater");
		("<>", "Unequal");
		("=", "Equal");
		("<=", "LessEqual");
		(">=", "GreaterEqual");
		(":=", "Declaration");
		("#", "SourceList")
	] in
	let do_op sourcestring t =
		Str.global_replace 
			( Str.regexp ("(" ^ (snd t) ^ "\\([( \t\n]+\\)" ) )
			("("^ (fst t) ^"\\1") 
			sourcestring
	in
	let simp_op =
		let so1 = List.fold_left do_op wo_matfun ops in
		List.fold_left do_op so1 ops
	in
	(*print concisely: (has to be performed twice)*)
	let con1 = Str.global_replace 
		( Str.regexp ("\n[\t ]*\\([^(\n]+\\)\n" ) )
		" \\1\n"
		simp_op
	in
	let con2 = Str.global_replace 
		( Str.regexp ("\n[\t ]*\\([^(\n]+\\)\n" ) )
		" \\1\n"
		con1
	in
	Str.global_replace 
		( Str.regexp ("\\(\n[\t ]*([-+*/] ?[_A-Za-z0-9.-]*\\)\n[\t ]*" ) )
		"\\1 "
		con2
;;

(* [xml_of_file filename]: open file [filename] and convert it to Xml-light xml;
it is NOT validated against a DTD!*)
let xml_of_file filename =
	let inp = open_in filename in
	let rec find_start ch linestart line =
		let sl = String.length linestart in
		if (String.length line >= sl && String.sub line 0 sl = linestart) then
			line
		else
			find_start ch linestart (input_line ch)
	in
	let doctypeline = find_start inp "<!DOCTYPE" "" in
	(*internal DTD:*)
	if not (Str.string_match (Str.regexp ".*SYSTEM.*") doctypeline 0) then
		ignore (find_start inp "]>" "");
	let x = Xml.parse_in inp in
	close_in inp;
	x
;;

(** [string_of_file filename]: read file [filename] and return its contents as string *)
let string_of_file filename =
	let inp = open_in filename in
	let l = in_channel_length inp 
	and s = String.create 1024 in
	let res = Buffer.create l in
	let rec fill_buff () =
		match input inp s 0 1024 with
		0 -> Buffer.contents res
		| n -> 
			Buffer.add_substring res s 0 n;
			fill_buff ()
	in
	let s = fill_buff () in
	close_in inp;
	s
;;

(** [xml_file_of_fls_file ~encoding ~internal_dtd filename]: convert xml file [filename] 
to a corresponding fls file with extension fls;
[encoding]: XML encoding, optional, default: "UTF-8"
[internal_dtd]: string containing internal DTD, default: None
*)
let xml_file_of_fls_file ?(encoding="UTF-8") ?(internal_dtd=None) filename =
	let flsstring = string_of_file filename in
	let x = xml_list_of_fls_string flsstring in
	let first_tag = Xml.tag (List.hd x) in
	let xmlstring = concat_string_list (List.map Xml.to_string_fmt x) in
	let new_name = (Filename.chop_extension filename) ^ ".fld" in
	let dtd_string =
		match internal_dtd with
		None -> "<!DOCTYPE " ^ first_tag ^ " SYSTEM \"experiments.dtd\">\n"
		| Some dtdstring -> 
			"<!DOCTYPE " ^ first_tag ^ "[\n"
			^ dtdstring ^ "\n]>\n"
	in
	let oc = open_out new_name in
	output_string 
		oc 
		("<?xml version=\"1.0\" encoding=\"" ^ encoding ^ "\"?>\n" ^ dtd_string);
	output_string oc xmlstring;
	close_out oc
;;

(** [fls_file_of_xml_file filename]: convert fls file [filename] to a corresponding XML file with extension xml *)
let fls_file_of_xml_file filename =
	let x = xml_of_file filename in
	let flsstring =  fls_string_of_xml x in
	let new_name = (Filename.chop_extension filename) ^ ".fls" in
	let oc = open_out new_name in
	output_string oc flsstring;
	close_out oc
;;

(*
fls_file_of_xml_file "t.fld";;
xml_file_of_fls_file "t.fls";;
*)
