type output_format = Full_output | Partial_output | No_output;;
type shell_declaration = 
{
	identifier: string;
	content: Xml.xml;
	internal_declarations: shell_declaration list;
};;

exception Interrupt_evaluation of string;;
exception Flashdot_error of string;;

let flashdot_binary =
	let exname = Sys.executable_name in
	let flashdot_dir = Filename.dirname exname in
	Filename.concat flashdot_dir "flashdot"
;;
let windows = (Sys.os_type = "Win32");;
let flashdot_binary = if windows then "flashdot" else "./flashdot";;

(*input, output, and error files for communicating with flashdot:*)
let flashdot_input_file = Filename.temp_file "flashdot_input" ".fld";;
let flashdot_output_file = Filename.temp_file "flashdot_output" ".out";;
let flashdot_error_file = Filename.temp_file "flashdot_error" ".err";;

(*to be evaluated graphically:*)
let gui_objects =
[
	"Experiment";
	"Scene";
	"Stimulus";
	"Button";
	"Image"
];;

(*function tags:*)
let functions =
[
	"Function";
	"StringFunction";
	"MatrixFunction";
	"SourceFunction"
];;

let output_format = ref Partial_output;;

let input_channel = ref stdin;;

let complete_string_start = ref "> ";;
let partial_string_start = ref "   +";;

(*association list of global declarations:*)
let global_declarations = ref [];;
let old_global_declarations = ref [];;

(*delete list of files, if they exist:*)
let rec delete_if_exists = function
	| [] -> ()
	| h::t -> 
		if Sys.file_exists h then Sys.remove h;
		delete_if_exists t
;;

(*call flashdot with command [cmd] and return result as string:*)
let call_flashdot cmd =
	delete_if_exists [flashdot_output_file; flashdot_error_file];
	let to_flashdot = open_out flashdot_input_file in
	output_string to_flashdot cmd;
	close_out to_flashdot;
	let command_string =
		flashdot_binary 
		^ " -o " ^ flashdot_output_file 
		(*^ (if windows then "" else " -o_stderr " ^ flashdot_error_file )*)
		^ " -o_stderr " ^ flashdot_error_file 
		^ " " ^ flashdot_input_file
	in
	(
		match (Unix.system command_string) with
		Unix.WEXITED 127 -> failwith "No shell can be executed, so FlashDot cannot be called."
		| Unix.WSIGNALED i -> prerr_endline ("FlashDot was killed by signal " ^ (string_of_int i));
		| Unix.WSTOPPED i -> prerr_endline ("FlashDot was stopped by signal " ^ (string_of_int i));
		| _ -> ()
	);
	if Sys.file_exists flashdot_output_file then
		Fls.string_of_file flashdot_output_file
	else 
		if Sys.file_exists flashdot_error_file then
			let errstring = Fls.string_of_file flashdot_error_file in
			if errstring = "" then ""
			else raise (Flashdot_error errstring)
		else ""
;;

(*this does not work because terminating flashdot crashes the shell*)
let call_flashdot_old cmd =
	(*distinguish here between Expression and rest; for rest maybe call Unix.system*)
	let (flin, flout) = Unix.open_process (flashdot_binary ^ " -stdin_fld -o -") in
	output_string flout cmd;
	flush_all ();
	close_out flout;
	let flstring = input_line flin in
	close_in flin;
	ignore (Unix.close_process (flin, flout));
	print_endline flstring
;;

(*include_declarations [elem] [decllist]: insert all declarations of decllist into elem*)
let rec include_declarations elem = function
	| [] -> elem
	| h::t -> 
		let rec declaration_of_list_item = function
			| {identifier=ident; content=content; internal_declarations=_} ->
				Xml.Element 
				(
					"Declaration", 
					["Identifier", ident], 
					[content]
				)
		in
		include_declarations (Xml.Element ("Let", [], (declaration_of_list_item h) :: [elem])) (h.internal_declarations @ t)
;;


(*convert Xml.xml x to xml expression for use with flashdot:*)
let make_expression x =
	if List.mem (Xml.tag x) gui_objects then
		x
	else
		Xml.Element
		(
			"Expression",
			[],
			[include_declarations x !global_declarations]
		)
;;

let exit_shell () =
	delete_if_exists [flashdot_input_file; flashdot_output_file; flashdot_error_file];
	exit 0
;;

let update_global_declarations ident content = 
	let content_tag = Xml.tag content in
	let declarations_without =
		List.filter
			(fun t -> t.identifier <> ident)
			!global_declarations
	in
	let evaluated_content =
		let old_content =
			match Xml.children content with
			h::t -> h
			| [] -> 
				raise (Interrupt_evaluation ("ERROR in declaration \"" ^ ident ^ "\":\n\nNo contents or missing declaration type (e.g. Constant, Matrix, ...)."));
		in
		match content_tag with
		| "Source" -> old_content
		| "Constant" | "Variable" | "Matrix" | "StringVariable" ->
			let x = Xml.Element ("Unquote", [], [old_content]) in
			let expr = make_expression x in
			let xmlstring =
				"<!DOCTYPE Experiment SYSTEM \"experiments.dtd\">\n"
				^ (Xml.to_string expr)
			in
			let result = call_flashdot xmlstring in
			let result_xml = List.hd (Fls.xml_list_of_fls_string result) in
			Xml.Element ("Eval", [], [result_xml])
		| fs ->
			if List.mem fs functions then
				(*include_declarations old_content declarations_without*) old_content
			else
				raise (Interrupt_evaluation ("No valid declaration type."))
			
	in
	let include_internally =	(*for functions; all currently declared expressions must be included*)
		if List.mem content_tag functions then declarations_without else []
	in
	let new_content =
		Xml.Element (content_tag, (Xml.attribs content), [evaluated_content])
	in
	let new_decl = 
	{
		identifier=ident; 
		content=new_content; 
		internal_declarations = include_internally
	} in
	global_declarations := new_decl::declarations_without;
	print_endline (content_tag ^ " " ^ ident ^ ": \n");
;;

(*print help:*)
let help () =
	print_endline "Welcome to the FlashDot shell.\nThe following top level commands exist:\n\n";
	let help_list =
	[
		"help", "show this help";
		"full_output", "do not abbreviate output";
		"partial_output", "abbreviate long output (default)";
		"no_output", "do not show output";
		"stop", "break erroneous inputs (useful if the shell does not recognize end of input)";
		"global_declarations", "show all top level declarations";
		"clear_all", "clear all top level declarations";
		"clear", "clear decl: clear declaration decl";
		"load", "load filename: load file filename and execute it";
		"save", "save expr filename: execute expression expr and save results to file filename";
		"quit", "exit shell"
	] in
	let print_help = function
		| (keyword, entry) -> 
			let delim =
				let rec spaces s = function
					| 0 -> s
					| n -> spaces (s^" ") (n-1)
				in
				spaces "" (20 - String.length keyword)  
			in
			print_endline ("#" ^ keyword ^ delim ^ entry)
	in
	let sorted =
		let cmp = fun a b -> compare (fst a) (fst b) in
		List.sort cmp help_list
	in
	List.iter print_help sorted
;;

(*core function:*)
let rec read_prompt complete current_string =
	flush_all ();
	print_string (if complete then !complete_string_start else !partial_string_start);
	flush_all ();
	let sp = "[ \t]+"
	and nsp = "[^ \t]+"
	and ap = "[_A-Za-z0-9.-]+" in
	let new_input =
		(*strip preceding and successing spaces:*)
		let strip s = 
			let s1 = Str.global_replace ( Str.regexp ("^[ \t]*\\(.*[^ \t]\\)[ \t]*$" ) )"\\1" s in
			Str.global_replace ( Str.regexp ("^[ \t]*$" ) )"" s1
		in
		strip (input_line !input_channel)
	in
	let new_string = current_string ^ "\n" ^ new_input in
	let preprocessed_input = (*strip comments*)
		let (_, st) = Fls.preprocess_fls_string new_string in
		Str.global_replace (Str.regexp "\n*") "" st
	in
	if preprocessed_input = "" then
		read_prompt true "";
	(*check for unclosed comments:*)
	if Str.string_match (Str.regexp ".*#|.*") preprocessed_input 0 then
		read_prompt false new_string;
	match new_input with
	"#quit" -> exit_shell ()
	| "" | "#" -> read_prompt complete current_string
	| inp -> 
		let rec process_string out_format out_channel st =
			let output_formatted_string result =
				(	
					match out_format with
					No_output -> ()
					| Partial_output ->
						if String.length result > 5000 then
							output_string out_channel ((String.sub result 0 5000) ^ "\n...")
						else
							output_string out_channel result;
					| Full_output -> output_string out_channel result;
				);
				output_string out_channel "\n";
			in
			match List.hd (Fls.xml_list_of_fls_string st) with
			Xml.Element ("Declaration", ident, h::t) ->
				let identifier = 
					match ident with
					(a, b)::t -> b
					| _ -> 
						raise (Interrupt_evaluation ("ERROR:\n\nIncorrect declaration, missing Identifier.")); 
				in
				old_global_declarations := !global_declarations;
				update_global_declarations identifier h;
				if List.mem (Xml.tag h) functions then
					let decl = List.find (fun t -> t.identifier = identifier) !global_declarations in
					output_string 
						out_channel 
						((Fls.fls_string_of_xml decl.content) ^ "\n")
				else
				(
					try
					process_string out_format out_channel identifier;
					with Flashdot_error s |  Interrupt_evaluation s-> 
						(global_declarations := !old_global_declarations;
						print_endline identifier;
						raise (Flashdot_error s))
				)
					
			| x ->
			(
				let expr = make_expression x in
				let xmlstring =
					"<!DOCTYPE Experiment SYSTEM \"experiments.dtd\">\n"
					^ (Xml.to_string expr)
				in
				let result = call_flashdot xmlstring in
				output_formatted_string result;
			)
		in
		try
		if inp.[0] == '#' && inp.[1] <> '|' then	(*top level keywords with single arguments*)
		(
			let process_keyword = function
				| ("help", []) -> help ()
				| ("full_output", []) -> output_format := Full_output;
				| ("partial_output", []) -> output_format := Partial_output;
				| ("no_output", []) -> output_format := No_output;
				| ("stop", []) -> ()
				| ("global_declarations", []) -> 
					let rec print_declarations = function
						| decl::t -> 
							print_endline (decl.identifier ^ ":\t" ^ (Xml.tag decl.content));
							print_declarations t
						| _ -> ()
					in
					print_declarations (List.rev !global_declarations);
				| ("clear_all", []) -> global_declarations := [];
				| ("clear", decl::[]) -> 
				(
					match List.partition (fun t -> t.identifier = decl) !global_declarations with
					([], _ ) -> 
						raise (Interrupt_evaluation ("Cannot clear \"" ^ decl ^ "\": not declared."));
					| (_, rest) -> global_declarations := rest
				)
				| ("load", filename::[]) ->	
					print_endline ("Loading file " ^ filename);
					let old_input_channel = !input_channel in
					let (old_complete_string_start, old_partial_string_start) =
						(!complete_string_start, !partial_string_start)
					in
					if Sys.file_exists filename then
					(
						let inp_ch = 
							try open_in filename
							with _ -> raise (Interrupt_evaluation ("Cannot open file \"" ^ filename ^ "\"."));
						in
						input_channel := inp_ch;
						complete_string_start := "";
						partial_string_start := "";
						try
						read_prompt true "";
						with End_of_file -> 
						(
							input_channel := old_input_channel;
							complete_string_start := old_complete_string_start;
							partial_string_start := old_partial_string_start;
							close_in inp_ch;
						)
					)
					else
						raise (Interrupt_evaluation ("Cannot open file \"" ^ filename ^ "\": does not exist."));
				| ("save", expr::filename::[]) ->	(*give expression in quotes*)
				(
					try
					let out_ch = open_out filename in
					process_string Full_output out_ch expr;
					close_out out_ch;
					print_endline ("Expression saved to file \"" ^ filename ^ "\".");
					with _ -> raise (Interrupt_evaluation ("Cannot open file \"" ^ filename ^ "\"."));
				)
				| _ -> raise (Interrupt_evaluation ("Unknown command \"" ^ inp ^ "\" or wrong number of arguments."));
			in
			let rec extract_keyword_and_arguments argument_list s =
				(*argument is string:*)
				if Str.string_match (Str.regexp ("^\\(.*[^ \t]\\)" ^ sp ^ "\"\\(.*\\)\"$"))  s 0 then
					extract_keyword_and_arguments
						((Str.matched_group 2 s)::argument_list)
						(Str.matched_group 1 s)
				else
					(*argument is ID:*)
					if Str.string_match (Str.regexp ("^\\(.*[^ \t]\\)" ^ sp ^ "\\(" ^ nsp ^ "\\)$"))  s 0 then
						extract_keyword_and_arguments
							((Str.matched_group 2 s)::argument_list)
							(Str.matched_group 1 s)
					else
						if Str.string_match (Str.regexp ("^#\\(" ^ ap ^ "\\)$"))  s 0 then
							((Str.matched_group 1 s), argument_list)
						else
							raise (Interrupt_evaluation ("Invalid input sequence \"" ^ inp ^ "\""))
			in
			let (keyword, argument_list) = extract_keyword_and_arguments [] inp in
			process_keyword (keyword, argument_list);
			read_prompt true "";
		)
		else
		(
			try
			(*print_newline ();*)
			(
				try
				process_string !output_format stdout new_string;
				with Flashdot_error s | Interrupt_evaluation s -> prerr_endline s
			);
			read_prompt true "";
			with 
				Failure s  | Fls.Invalid_expression s ->
					( (*print_endline ("failstring: " ^ s);*) read_prompt false new_string)
		)
		with Interrupt_evaluation s -> (prerr_endline s; read_prompt true "";) 
;;

let main () =
	print_newline ();
	ignore (Unix.system (flashdot_binary ^ " -version"));
	print_endline "Type #help to see top level commands.\n\n";
	try
	read_prompt true ""
	with End_of_file -> 
	(
		print_newline ();
		exit_shell ();
	)
;;

main ();;
