(*
	interpreter: 
		ocaml -I ../xml-light str.cma xml-light.cma macros.ml
*)

type attribute_macro =
{
	is_recursive: bool;
	macro_expression: Str.regexp;
	macro_expression_string: string;
	target_string: string
};;


(*auxiliary functions for processing xml:*)
let aod elem attr default =
	let attrib_declared attr elem =
		List.mem attr (fst (List.split (Xml.attribs elem)))
	in
	if attrib_declared attr elem then
		Xml.attrib elem attr
	else
		default
;;

(** convert xml expression [x] to attribute macro *)
let attribute_macro_of_xml x =
	let bos s1 s2 = bool_of_string (aod x s1 s2) in
	let rs = Xml.attrib x "Expression" in
	let r =
		if bos "RegularExpression" "false" then
		(
			if bos "CaseSensitive" "true" then
				Str.regexp rs
			else
				Str.regexp_case_fold rs
		)
		else
		(
			if bos "CaseSensitive" "true" then
				Str.regexp_string rs
			else
				Str.regexp_string_case_fold rs
		)
	in
	{
		is_recursive = bos "Recursive" "false";
		macro_expression = r;
		macro_expression_string = rs;
		target_string = Xml.attrib x "TargetString"
	}
;;

(** [expand_attribute_macro mac elem]: 
expand all attributes of xml element [elem] 
and all of its children recursively
according to attribute macro [mac];
if [elem] is AttributeMacros node then it won't be expanded
*)
let rec expand_attribute_macro mac elem =
	(*process all attributes of x:*)
	let process_attributes x =
		(*return true if a substring of s matches:*)
		let matches_regexp s =
			try
				ignore (Str.search_forward mac.macro_expression (snd s) 0);
				true
			with Not_found -> false
		in
		let rec process_atlist l =
		(
			let (unexpanded, expanded) =
				List.partition matches_regexp l
			in
			match unexpanded with
			[] -> l
			| _ -> 
				let replace_exp = function
					(key, cont) -> 
					(
						key,
						Str.global_replace 
							mac.macro_expression 
							mac.target_string 
							cont
					)
				in
				if mac.is_recursive then
					process_atlist ((List.map replace_exp unexpanded) @ expanded)
				else
					((List.map replace_exp unexpanded) @ expanded)
		) in
		(*let new_atlist = process_atlist (Xml.attribs x) in*)
		process_atlist (Xml.attribs x)
	in
	try
	(
		let cld = Xml.children elem in
		let tg = Xml.tag elem in
		if tg = "AttributeMacros" then 
			elem
		else
			Xml.Element
			(
				tg,
				process_attributes elem,
				List.map (expand_attribute_macro mac) cld
			)
	)
	with Xml.Not_element x -> x
;;

(** [expand_attribute_macros argv_array elem]: 
find AttributeMacros in xml element [elem]
(if exists) and expand all attribute macros;
treat strings in argv_array (length n) as attribute macros
of the form ("ARGV1;", argv_array.[0]) to ("ARGVn", argv_array.[n-1])
which are expanded after all other attribute macros
*)
let expand_attribute_macros argv_array elem =
	try
	let argv_macros =
		let macro_of_argstring i s =
		let ex_string = "ARGV" ^ (string_of_int (i + 1)) ^ ";" in
		{
			is_recursive = false;
			macro_expression = (Str.regexp ex_string);
			macro_expression_string = ex_string;
			target_string = s
		} in
		Array.to_list (Array.mapi macro_of_argstring argv_array)
	in
	let maclist =
		if List.exists (fun t -> Xml.tag t = "AttributeMacros") (Xml.children elem) then
			let ams = List.find (fun t -> Xml.tag t = "AttributeMacros") (Xml.children elem) in
			List.map attribute_macro_of_xml (Xml.children ams) 
		else
			[]
	in
	let rec expand x = function
		| [] -> x
		| h::t -> expand (expand_attribute_macro h x) t
	in
	expand elem (maclist @ argv_macros)
	with 
		Xml.Not_element x -> x
;;


(*Tests:
let x = Xml.parse_file "../macrotest.fld" in
let xexp = expand_attribute_macros x in
print_endline (Xml.to_string_fmt xexp);;*)
