(*
	generating resp. maintaining documentation
*)

type function_entry =
{
	name: string;
	syntactic_sugar: string list;
	attributes: attribute list;
	parent: string list;
	children: Dtd.dtd_element_type;
	category: string option;
	description: string option;
	examples: example list
}
and attribute =
{
	attr_name: string;
	attr_contents: Dtd.dtd_attr_type * Dtd.dtd_attr_default;
	attr_description: string option;
}
and example =
{
	ex_filename: string option;
	ex_contents: string
};;

let default_dtd_file = "../experiments.dtd";;
let default_flashdot_doc_file = "flashdotdoc.xml";;

(*auxiliary function:*)
let contains_string str contained =
	let p = try Str.search_forward (Str.regexp contained) str 0
		with Not_found -> -1
	in
	p>=0 
;;
(*convert DTDs from file (handler) to strings; 
return dtdstring and hashtables with entities and categories (i. e. inverse entity hashtbl)*)
let dtd_string_and_hashtables_from_file dtd_filename =
	let entityqueue = Queue.create () in
	let inp_dtd = open_in dtd_filename in
	let rec file_to_string () =
		try 
			let line = input_line inp_dtd in
			if contains_string line "]>" then
			(
				failwith "DTD_END";
			);
			if contains_string line "!ENTITY" then	(*check for parameter entities*)
				let _ = Str.search_forward (Str.regexp "%[ ]+\\([^ ]+\\) [^\"]*\"\\(.*\\)\"") line 0 in
				let ent = Str.matched_group 1 line in
				let subs = Str.matched_group 2 line in
				Queue.push (ent, subs) entityqueue;
				file_to_string ()
			else
				(line ^ (file_to_string ()))
		with 
			End_of_file -> close_in inp_dtd; ""
			| Failure "DTD_END" -> ""
	in
	let dstr = file_to_string () in
	let category_hashtbl = Hashtbl.create 200 in
	let entity_hashtbl = Hashtbl.create 10 in
	let category_list =
	[
		"funs", "number";
		"matfuns", "matrix";
		"stringfuns", "string";
		"sourcefuns", "source";
		"listfuns", "list";
		"stimuluscontents", "stimulus_content";
		"imagecontents", "image_content";
		"rng", "rng"
	] in
	(*substitute parameter entities and fill hashtbl:*)
	let rec subs_ent str q =
		match Queue.is_empty q with
		true -> (str, category_hashtbl, entity_hashtbl)
		| _ -> 
			let (entity, conts) = Queue.pop q in
			let category = List.assoc entity category_list in
			let entries_list =
				let parenths_stripped = Str.global_replace (Str.regexp "^(\\(.*\\))$") "\\1" conts in
				let no_spaces = Str.global_replace (Str.regexp "[ ]+") "" parenths_stripped in
				Str.split (Str.regexp_string "|") no_spaces
			in
			Hashtbl.add entity_hashtbl category entries_list;
			List.iter (fun t -> Hashtbl.add category_hashtbl t category) entries_list;
			let str_repl = Str.global_replace (Str.regexp ("%" ^ entity ^ ";")) conts str in
			subs_ent str_repl q
	in
	subs_ent dstr entityqueue
;;

(*read Xml.xml from file inputfile:*)
let xml_of_file inputfile =
	let (xmlstructure, dtd) =
			let inp = open_in inputfile in
			let rec find_start ch linestart line =
				if (String.length line >= 9 && String.sub line 0 9 = linestart) then
					line
				else
					find_start ch linestart (input_line ch)
			in
			let doctypeline = find_start inp "<!DOCTYPE" "" in
			let dtd =
				let dtd_file =
					let _ = 
						if not (Str.string_match (Str.regexp ".*\"\\(.*dtd\\).*") doctypeline 0) then
						failwith "No DTD file declared."
					in
					Str.matched_group 1 doctypeline
				in
					let (dtdstring, _, _) = dtd_string_and_hashtables_from_file dtd_file in
					let dtdstring_replaced = 
						Str.global_replace (Str.regexp " IDR?E?F? ") " CDATA " dtdstring
					in
					Dtd.check (Dtd.parse_string dtdstring_replaced)
			in
			let x = Xml.parse_in inp in
			close_in inp;
			(x, dtd) 
	in
	let first_tag = Xml.tag xmlstructure in
	let _ = Dtd.prove dtd first_tag xmlstructure in
	xmlstructure
;;

let (dtd, category_hashtbl, entity_hashtbl) = 
	match dtd_string_and_hashtables_from_file default_dtd_file with
	(dtd_string, a, b) -> (Dtd.parse_string dtd_string, a, b)
;;

let function_names =
	let dtd_elements = List.filter (function | Dtd.DTDElement _ -> true | _ -> false) dtd in
	List.map (function Dtd.DTDElement (name, _) -> name | _ -> "") dtd_elements
;;

(*hashtbl containing all attributes:*)
let attribute_hashtbl =
	let atthash = Hashtbl.create 200 in
	List.iter 
		(
			function 
			| Dtd.DTDAttribute (element, attname, atttype, attdefault) ->
				Hashtbl.add atthash element (attname, atttype, attdefault)
			| _ -> ()
		) 
		dtd;
	atthash
;;
(*hashtbl containing all parents:*)
let parent_hashtbl =
	let parhash = Hashtbl.create 200 in
	List.iter
		(
			function
			| Dtd.DTDElement (element, (Dtd.DTDChild cld)) ->
				let rec add_parent = function
					| Dtd.DTDTag tg -> Hashtbl.add parhash tg element
					| Dtd.DTDOptional c -> add_parent c
					| Dtd.DTDZeroOrMore c -> add_parent c
					| Dtd.DTDOneOrMore c -> add_parent c
					| Dtd.DTDChoice cl -> List.iter add_parent cl
					| Dtd.DTDChildren cl -> List.iter add_parent cl
					| _ -> ()
				in
				add_parent cld
			| _ -> ()
		)
		dtd;
	parhash
;;

(*write xml structure doc_xml to standard flashdot doc file:*)
let write_doc_file doc_xml =
	let out = open_out default_flashdot_doc_file in
	output_string out "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
	output_string out "<!DOCTYPE FlashDotDoc SYSTEM \"flashdotdoc.dtd\">\n";
	output_string out (Xml.to_string_fmt doc_xml);
	close_out out
;;

(*initially generate default_flashdot_doc_file:
(NOT REQUIRED ANYMORE)*)
(*
let xml_of_function_name funname = 
	let xml_of_attrib = function
		| (attname, _, _) ->
			Xml.Element ("Attribute", ["Name", attname], [])
	in
	let attlist = 
		List.map xml_of_attrib (Hashtbl.find_all attribute_hashtbl funname) 
	in
	Xml.Element 
	(
		"Function",
		[("Name", funname); ("Include", "true")],
		attlist
	)
in
let xml_fun_list = List.map xml_of_function_name function_names in
let fld_doc = Xml.Element ("FlashDotDoc", [], xml_fun_list) in
write_doc_file fld_doc;;*)

(*contents of docfile; update the file if necessary:*)
let docfile_funlist_xml =
	let fun_list = 
		Xml.children (xml_of_file default_flashdot_doc_file) 
	in
	(*update doc file if newer functions in experiments.dtd:*)
	let new_functions =
		let old_fun_names = List.map ( fun t -> Xml.attrib t "Name" ) fun_list in
		(*check if all names are contained:*)
		List.filter (fun t -> not (List.mem t old_fun_names)) function_names
	in
	match new_functions with
	[] -> fun_list
	| nf -> 
		let xml_of_function_name funname = 
			let xml_of_attrib = function
				| (attname, _, _) ->
					Xml.Element ("Attribute", ["Name", attname], [])
			in
			let attlist = 
				List.map xml_of_attrib (Hashtbl.find_all attribute_hashtbl funname) 
			in
			Xml.Element 
			(
				"Function",
				[("Name", funname); ("Include", "true")],
				attlist
			)
		in
		let new_fun_list =
			(List.map xml_of_function_name nf) @ fun_list
		in
		let fld_doc = Xml.Element ("FlashDotDoc", [], new_fun_list) in
		write_doc_file fld_doc;
		new_fun_list
;;

(*merge information from experiments.dtd and flashdotdoc.xml to single list:*)
let function_entry_list =
	(*check whether attribute attr is declared in element elem:*)
	let attrib_declared attr elem =
		List.mem attr (fst (List.split (Xml.attribs elem)))
	in
	(*get attribute attr of element elem or default if attribute is not declared*)
	let aod elem attr default =
		if attrib_declared attr elem then
			Xml.attrib elem attr
		else
			default
	in
	let entry_of_xml x =
		(*filter children of an xml element elem the tags of which match childtag:*)
		let children childtag =
			List.filter (fun t -> (Xml.tag t) = childtag) (Xml.children x)
		in
		let funname = Xml.attrib x "Name" in
		let syntactic_sugar_of_xml = function
			| Xml.Element (_, _, (Xml.PCData s)::_) -> s
			| _ -> failwith "incorrect SyntacticSugar"
		in
		let synt_sug = List.map syntactic_sugar_of_xml (children "SyntacticSugar") in
		let attributes = 
			let attribs_descr = 
				List.map 
					(
						function 
						| Xml.Element (_, ("Name", atn)::_, (Xml.Element (_, _, (Xml.PCData descr)::_))::_) -> 
							(atn, Some descr)
						| _ -> ("", None)
					) 
					(children "Attribute") 
			in
			let attribute_of_dtd_attrib = function
				| (attr_name, attr_type, attr_default) ->
					let descr =
						try
						List.assoc attr_name attribs_descr
						with Not_found -> None
					in
					{
						attr_name= attr_name;
						attr_contents = (attr_type, attr_default);
						attr_description = descr;
					}
			in
			List.map attribute_of_dtd_attrib (Hashtbl.find_all attribute_hashtbl funname)
		in
		let descr =
			match children "Description" with
			[] -> None
			| (Xml.Element (_, _, (Xml.PCData s)::_))::_ -> Some s
			| _ -> failwith "incorrect Description"
		in
		let cldrn =
			match List.find (function | Dtd.DTDElement (n, _) -> n=funname | _ -> false) dtd with
			| Dtd.DTDElement (_, a) -> a
			| _ -> failwith ""
		in
		let categ =
			match aod x "Category" "Miscellaneous" with
			"Miscellaneous" -> None
			| s -> Some s
		in
		let example_of_xml = function
			| Xml.Element (_, atl, (Xml.PCData expl)::_) ->
				let filename = 
					match atl with ("Filename", fn)::_ -> Some fn | _ -> None
				in
				{
					ex_filename = filename;
					ex_contents = expl
				}
			| _ -> failwith "incorrect Example"
		in
		let examples = List.map example_of_xml (children "Example") in
		{
			name = funname;
			syntactic_sugar = synt_sug;
			attributes = attributes;
			parent = Hashtbl.find_all parent_hashtbl funname;
			children = cldrn;
			category = categ;
			description = descr;
			examples = examples
		}
	in
	let funs_to_include = 
		List.filter (fun t -> bool_of_string (aod t "Include" "true")) docfile_funlist_xml
	in
	List.map entry_of_xml funs_to_include
;;

(*convert entry to string containing latex:*)
let latex_of_entry e =
	let stl str = "\\subsubsection*{" ^ str ^ "}\n" in	(*"subtitle"*)
	(*let tt str = "{\\tt " ^ str ^ "}" in*)
	let it str = "\\textit{" ^ str ^ "}" in
	let verb str = "\\begin{verbatim}\n" ^ str ^ "\n\\end{verbatim}" in
	let rec string_join c = function
		| [] -> ""
		| [s] -> s
		| (h1::h2::t) -> h1 ^ c ^ (string_join c (h2::t))
	in
	let synt_sug = 
		let rec syntactic_sugar_of_list = function
			| [] -> ""
			| h::t ->
				(stl "Syntactic sugar") ^ h ^"\n" ^ (syntactic_sugar_of_list t)
		in
		syntactic_sugar_of_list e.syntactic_sugar
	in
	let attributes =
		let rec attribs_of_list = function
			| [] -> ""
			| h::t -> 
				let (atype, adef) = h.attr_contents in
				let typestring =
					match atype with
					| Dtd.DTDCData -> it "string"
					| Dtd.DTDNMToken -> it "name token"
					| Dtd.DTDEnum sl -> string_join " | " sl
					| _ -> "not yet implemented" (*not yet implemented*)
				in
				let defaultstring =
					match adef with
					| Dtd.DTDDefault s | Dtd.DTDFixed s -> s
					| Dtd.DTDRequired -> it "required"
					| Dtd.DTDImplied -> it "implied"
				in
				h.attr_name ^ " & " ^ typestring ^ " & " ^ defaultstring ^ "\\\\\n"
				^ (attribs_of_list t)
		in
		(stl "Attributes") 
		^ "\\begin{tabular}{lll}}\n\\textbf{attribute} & \\textbf{type} & \\textbf{properties}\\\\\\hline\n"
		^ (attribs_of_list e.attributes)
		^ "\\end{tabular}"
	in
	let parent = 
		match e.parent with
		[] -> ""
		| plist -> (stl "Parent") ^ (string_join " | " plist)	(*include links here?*)
	in
	let children = 
		let rec child_of_dtd =
			let ss symb c = (child_of_dtd c) ^ symb in
			function
			| Dtd.DTDTag s -> s 	(*include links here?*)
			| Dtd.DTDPCData -> it "string"
			| Dtd.DTDOptional c -> ss "?" c
			| Dtd.DTDZeroOrMore c -> ss "*" c
			| Dtd.DTDOneOrMore c -> ss "+" c
			| Dtd.DTDChoice sl -> "( " ^ (string_join " | " (List.map child_of_dtd sl)) ^ " )"
			| Dtd.DTDChildren sl -> "(" ^ (string_join " $\\times$ " (List.map child_of_dtd sl)) ^ ")"
		in
		match e.children with
		| Dtd.DTDEmpty -> ""
		| Dtd.DTDAny -> (stl "Children") ^ (it "any")
		| Dtd.DTDChild cld -> (stl "Children") ^ (child_of_dtd cld) 
	in
	(*let category = 
		match e.category with
		[] -> ""
	in*)
	let description = 
		match e.description with
		None -> ""
		| Some s -> (stl "Description") ^ s
	in
	let examples =
		let rec examples_of_list = function
			| [] -> ""
			| h::t ->
				let filename =
					match h.ex_filename with
					None -> ""
					| Some s -> "(" ^ (it ("file " ^ s)) ^ ")\n"
				in
				(stl "Example") ^ filename 
				^ (verb h.ex_contents) ^ "\n" 
				^ (examples_of_list t)
		in
		examples_of_list e.examples
	in
	"\\subsection*{" ^ e.name ^ "}\n\n"
	^ synt_sug ^ "\n"
	^ attributes ^ "\n"
	^ parent ^ "\n"
	^ children ^ "\n"
	^ description ^ "\n"
	^ examples ^ "\n"
;;

print_endline (latex_of_entry (List.hd function_entry_list));;
