(* compile with: 

bytecode:
	ocamlc -I +gsl str.cma unix.cma xml-light.cma bigarray.cma gsl.cma mathexpr.mli mathexpr.ml
	ocamlc -a -o mathexpr.cma mathexpr.cmo
	
native:
	ocamlopt -I +gsl str.cmxa unix.cmxa xml-light.cmxa bigarray.cmxa gsl.cmxa mathexpr.mli mathexpr.ml
	ocamlopt -a -o mathexpr.cmxa mathexpr.cmx
*)

include Keywords;;	(*list of all keywords*)
include Random_rng;;

(*number or matrix:*)
type nom = Scal of float | Mat of Gsl_matrix.matrix | Str of string | Xml of Xml.xml;;
type float' = Floatconst of float | Floatref of (float ref);;
type declared_function = 
{
	nargs: int; 
	fcontent: (nom array -> float)
};;
type declared_matrix_function = 
{
	mat_nargs: int; 
	mat_fcontent: (nom array -> Gsl_matrix.matrix)
};;
type declared_string_function = 
{
	str_nargs: int; 
	str_fcontent: (nom array -> string)
};;
type declared_xml_function = 
{
	xml_nargs: int; 
	xml_fcontent: (nom array -> Xml.xml)
};;
type declaration = 
	Constant of float 
	| Variable of (float ref) 
	| String_variable of (string ref)
	| Xml_variable of (Xml.xml ref)
	| Function of declared_function
	| Matrix of Gsl_matrix.matrix
	| Matrix_function of declared_matrix_function
	| String_function of declared_string_function
	| Xml_function of declared_xml_function
;;
type expr_content = 
	Fun of (nom array -> nom)
	| Const of nom 
	| Var of (float ref)
	| Str_var of (string ref)
	| Xml_var of (Xml.xml ref)
	| Arg of (bool * int)
	| Let of (nom array -> nom) array
	| Cond of 
	(
		((nom array -> nom) array)
		* ((nom array -> nom) array) 
		* (nom array -> nom) 
	)
;;
type expression = {head: expr_content; body: (nom array -> nom) list};;

exception Non_declared_argument of string;;
exception Undeclared_identifier of string;;
exception Invalid_declaration_type of string;;
exception Not_enough_arguments of string;;



(** [global_declarations]: hash table containing declarations for lookup *)
let global_declarations = Hashtbl.create 30;;

(** [local_declarations]: reference to a hash table which can contain local declarations *)
let local_declarations = ref (Hashtbl.create 10);;

(** [declare id decl]: create a global declaration [decl] with identifier "id" *)
let declare =
	Hashtbl.add global_declarations
;;
(** [replace id decl]: replace a global declaration [decl] with identifier "id" *)
let replace_declaration =
	Hashtbl.replace global_declarations
;;

(*some declarations:*)
declare "PI" (Constant (acos (-1.)));;
declare "PI_2" (Constant Gsl_math.pi_2);;
declare "PI_4" (Constant Gsl_math.pi_4);;
declare "I_1_PI" (Constant Gsl_math.i_1_pi);;
declare "I_2_PI" (Constant Gsl_math.i_2_pi);;
declare "E" (Constant Gsl_math.e);;
declare "LOG2E" (Constant Gsl_math.log2e);;
declare "LOG10E" (Constant Gsl_math.log10e);;
declare "SQRT2" (Constant Gsl_math.sqrt2);;
declare "SQRT1_2" (Constant Gsl_math.sqrt1_2);;
declare "LN10" (Constant Gsl_math.ln10);;
declare "LN2" (Constant Gsl_math.ln2);;
declare "LNPI" (Constant Gsl_math.lnpi);;
declare "EULER" (Constant Gsl_math.euler);;


(** [print_gsl_matrix m]: print GSL matrix m to stdio *)
let print_gsl_matrix m =
	let (r, c) = Gsl_matrix.dims m in
	for i=0 to r-1 do
		print_newline ();
		for j=0 to c-1 do
			Printf.printf "%f " m.{i, j}
		done;
	done;
	print_newline ()
;;


(** [float_of_nom xmltag n]: convert nom [n] to float;
xmltag: can be given to raise an appropriate exception in case of failure
(xmltag is optional, default: "")*)
let float_of_nom ?(xmltag="") = function
	| Scal f -> f
	(*| Str s -> try float_of_string s with Failure _ -> nan*)
	| Mat _ -> failwith ("Function " ^ xmltag ^ " was called with a matrix but a scalar was expected.")
	| Str s -> failwith ("Function " ^ xmltag ^ " was called with string \"" ^ s ^ "\" but a scalar was expected.")
	| Xml s -> failwith ("Function " ^ xmltag ^ " was called with string \"" ^ (Xml.to_string_fmt s) ^ "\" but a scalar was expected.")
;;
(** [matrix_of_nom xmltag n]: convert nom [n] to matrix;
xmltag: can be given to raise an appropriate exception in case of failure
(xmltag is optional, default: "")*)
let matrix_of_nom ?(xmltag="") = function
	| Mat m -> m
	| Scal s -> failwith ("Function " ^ xmltag ^ "was called with scalar " ^ (string_of_float s) ^ " but a matrix was expected.")
	| Str s -> failwith ("Function " ^ xmltag ^ " was called with string \"" ^ s ^ "\" but a matrix was expected.")
	| Xml s -> failwith ("Function " ^ xmltag ^ " was called with string \"" ^ (Xml.to_string_fmt s) ^ "\" but a matrix was expected.")
;;
(** [string_of_nom n]: convert nom [n] to string *)
let string_of_nom = function
	| Str s -> s
	| Scal f -> (string_of_float f)
	| Xml x -> Xml.to_string_fmt x
	| Mat m -> 
	(
		let (r, c) = Gsl_matrix.dims m in
		let s = ref "" in
		for i=0 to r-1 do
			s := !s ^ "\n";
			for j=0 to c-1 do
				s := !s ^ (string_of_float m.{i, j}) ^ ","
			done;
		done;
		s := !s ^ "\n";
		!s
	)
;;
(** [xml_of_nom xmltag n]: convert nom [n] to Xml.xml;
xmltag: can be given to raise an appropriate exception in case of failure
(xmltag is optional, default: "")*)
let xml_of_nom ?(xmltag="") = function
	| Xml x -> x
	| Mat m -> 
		let (r, c) = Gsl_matrix.dims m in
		Xml.Element
		(
			"NumericValueMatrix",
			["Columns", (string_of_int c); "Rows", (string_of_int r)],
			[Xml.PCData (string_of_nom (Mat m))]
		)
		(*failwith ("Function " ^ xmltag ^ " was called with a matrix but source was expected.")*)
	| Scal s ->
		Xml.Element ("NumericValue", ["Value", (string_of_float s)], [])
		(*failwith ("Function " ^ xmltag ^ "was called with scalar " ^ (string_of_float s) ^ " but source was expected.")*)
	| Str s -> 
		Xml.Element ("String", ["Value", s], [])
		(*failwith ("Function " ^ xmltag ^ " was called with string \"" ^ s ^ "\" but source was expected.")*)
;;

(*auxiliary functions for processing xml:*)
let aod elem attr default =
	let attrib_declared attr elem =
		List.mem attr (fst (List.split (Xml.attribs elem)))
	in
	if attrib_declared attr elem then
		Xml.attrib elem attr
	else
		default
;;

(*get declaration:*)
let gd identifier =
	try
	Hashtbl.find !local_declarations identifier
	with Not_found ->
	(
		try
		Hashtbl.find global_declarations identifier
		with Not_found ->raise (Undeclared_identifier identifier)
	)
;;

let argstringlist_of_string argstring = 
	let spaces_removed = Str.global_replace (Str.regexp " ") "" argstring in
	if spaces_removed = (*"VOID"*) "" then 
		[]
	else
		Str.split (Str.regexp ",") spaces_removed 
;;

(** [get_declaration identifier]: 
get declaration [identifier] from global or local declarations hash table
*)
let get_declaration = gd;;

let float'_of_string s = 
	try
	float_of_string s
	with Failure t ->
	(
		match gd s with
		Constant c -> c
		| Variable v -> !v
		| Function f -> 
			if f.nargs > 0 then
				raise (Invalid_declaration_type s)
			else
				f.fcontent [||]
		| _ -> raise (Invalid_declaration_type s)
	)
;;
let float'_list_of_string s =
	let stringlist = Str.split (Str.regexp ", *") s in
	List.map float'_of_string stringlist
;;

(*considers two lists s1 and s2 as sets and compares them, 
i. e. the series of elements is neglected:*)
let set_compare s1 s2 =
	(List.length s1 = List.length s2)
	&& (List.for_all (fun t -> List.mem t s2) s1)
;;
(*checks wether two Xml Elements are equal (currently unused):*)
let xml_equal x y =
	let rec xme = function
		| (Xml.Element (xt, xa, xc), Xml.Element (yt, ya, yc)) ->
			(xt = yt)
			&& (set_compare xa ya)
			&& (List.length xc = List.length yc)
			&& (List.for_all xme (List.combine xc yc))
		| (Xml.PCData s1, Xml.PCData s2) -> s1 = s2
		| _ -> false
	in
	xme (x, y)
;;

(** [transform_symbols elem]:
transform all "Symbol" tags from quoted expression [elem]
into an evaluable expression*)
let rec transform_symbols = function
	| Xml.Element ("Symbol", atts, cld) as symb->
		let rec node_of_symbol = function
			| Xml.Element ("Symbol", sy_atts, sy_cld) as sy ->
				let (nodeattlist, restlist) = 
					List.partition (fun t -> Xml.tag t = "SymbolAttribute") sy_cld
				in
				let (pcdatalist, attlist) =
					let att_of_attnode an =
						(Xml.attrib an "AttributeTag", Xml.attrib an "AttributeValue")
					in
					match List.map att_of_attnode nodeattlist with
					("PCData", pcd)::t -> ([Xml.PCData pcd], t)
					| l -> ([], l)
				in
				let ident = Xml.attrib sy "Identifier" in
				if List.mem ident keywords then 
					Xml.Element (ident, attlist, pcdatalist @ (List.map node_of_symbol restlist))
				else
				(
					if restlist = [] then
						Xml.Element ("DeclaredValue", ["Identifier", ident], [])
					else
						Xml.Element 
						(
							"DeclaredFunction", 
							["Identifier", ident], 
							List.map node_of_symbol restlist
						)
				)
			| Xml.Element _ as el -> transform_symbols el
			| t -> t
		in
		node_of_symbol symb
	| Xml.Element (key, atts, cld) -> Xml.Element (key, atts, List.map transform_symbols cld)
	| t -> t
;;


(** [random_function_of_xml elem]: convert [elem] to an ocaml function: *)
let random_function_of_xml elem =
	let rng = rng_of_xml elem in
	let rand_type = List.hd (Xml.children elem) in
	match (Xml.tag rand_type) with
	"Continuous" ->
	(
		let cont_type = (List.hd (Xml.children rand_type)) in
		match (Xml.tag cont_type) with
		"Uniform" ->
			let rmin = float'_of_string (aod cont_type "Min" "0.0")
			and rmax = float'_of_string (aod cont_type "Max" "1.0") in
			fun () -> Gsl_randist.flat rng rmin rmax
		| "Gaussian" ->
			let sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
			if (bool_of_string (aod cont_type "RatioMethod" "false")) then
				fun () -> Gsl_randist.gaussian rng sigma
			else
				fun () -> Gsl_randist.gaussian_ratio_method rng sigma
		| "GaussianTail" ->
			let sigma = float'_of_string (aod cont_type "Sigma" "1.0")
			and a = float'_of_string (aod cont_type "A" "1.0") in
			fun () -> Gsl_randist.gaussian_tail rng a sigma
		| "Exponential" -> 
			let mu = float'_of_string (aod cont_type "Mu" "1.0") in
			fun () -> Gsl_randist.exponential rng mu
		| "Laplace" -> 
			let a = float'_of_string (aod cont_type "A" "1.0") in
			fun () -> Gsl_randist.laplace rng a
		| "ExponentialPower" ->
			let a = float'_of_string (aod cont_type "A" "1.0")
			and b = float'_of_string (aod cont_type "B" "1.0") in
			fun () -> Gsl_randist.exppow rng a b
		| "Cauchy" -> 
			let a = float'_of_string (aod cont_type "A" "1.0") in
			fun () -> Gsl_randist.cauchy rng a
		| "Rayleigh" -> 
			let sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
			fun () -> Gsl_randist.rayleigh rng sigma
		| "RayleighTail" ->
			let sigma = float'_of_string (aod cont_type "Sigma" "1.0")
			and a = float'_of_string (aod cont_type "A" "1.0") in
			fun () -> Gsl_randist.rayleigh_tail rng a sigma
		| "Landau" ->
			fun () -> Gsl_randist.landau rng
		| "Levy" ->
			let c = float'_of_string (aod cont_type "C" "1.0")
			and alpha = float'_of_string (aod cont_type "Alpha" "1.0") in
			fun () -> Gsl_randist.levy rng c alpha
		| "LevySkew" ->
			let c = float'_of_string (aod cont_type "C" "1.0")
			and alpha = float'_of_string (aod cont_type "Alpha" "1.0")
			and beta = float'_of_string (aod cont_type "Beta" "1.0") in
			fun () -> Gsl_randist.levy_skew rng c alpha beta
		| "Gamma" ->
			let a = float'_of_string (aod cont_type "A" "1.0")
			and b = float'_of_string (aod cont_type "B" "1.0") in
			fun () -> Gsl_randist.gamma rng a b
		| "Lognormal" ->
			let zeta = float'_of_string (aod cont_type "Zeta" "1.0")
			and sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
			fun () -> Gsl_randist.lognormal rng zeta sigma
		| "ChiSquared" -> 
			let nu = float'_of_string (aod cont_type "Nu" "1.0") in
			fun () -> Gsl_randist.chisq rng nu
		| "FDist" ->
			let nu1 = float'_of_string (aod cont_type "Nu1" "1.0")
			and nu2 = float'_of_string (aod cont_type "Nu2" "1.0") in
			fun () -> Gsl_randist.fdist rng nu1 nu2
		| "TDist" -> 
			let nu = float'_of_string (aod cont_type "Nu" "1.0") in
			fun () -> Gsl_randist.tdist rng nu
		| "Beta" ->
			let a = float'_of_string (aod cont_type "A" "1.0")
			and b = float'_of_string (aod cont_type "B" "1.0") in
			fun () -> Gsl_randist.beta rng a b
		| "Logistic" -> 
			let a = float'_of_string (aod cont_type "A" "1.0") in
			fun () -> Gsl_randist.logistic rng a
		| "Pareto" ->
			let a = float'_of_string (aod cont_type "A" "1.0")
			and b = float'_of_string (aod cont_type "B" "1.0") in
			fun () -> Gsl_randist.pareto rng a b
		| "Weibull" ->
			let a = float'_of_string (aod cont_type "A" "1.0")
			and b = float'_of_string (aod cont_type "B" "1.0") in
			fun () -> Gsl_randist.weibull rng a b
		| "Gumbel1" ->
			let a = float'_of_string (aod cont_type "A" "1.0")
			and b = float'_of_string (aod cont_type "B" "1.0") in
			fun () -> Gsl_randist.gumbel1 rng a b
		| _ -> (*Gumbel2*)
			let a = float'_of_string (aod cont_type "A" "1.0")
			and b = float'_of_string (aod cont_type "B" "1.0") in
			fun () -> Gsl_randist.gumbel2 rng a b
		(*| _ -> 
			let alphalist = float'_list_of_string (Xml.attrib cont_type "Alphalist")
			and thetalist = float'_list_of_string (Xml.attrib cont_type "Thetalist") in
			let alpha = Array.of_list alphalist
			and theta = Array.of_list thetalist in
			fun () -> Gsl_randist.dirichlet rng alpha theta*)
	)
	| "Discrete" ->
	(
		let cont_type = (List.hd (Xml.children rand_type)) in
		let ios s = int_of_float (float'_of_string s) in
		let foi i = float_of_int i in
		match (Xml.tag cont_type) with
		"DiscreteUniform" ->
			let rmin = float'_of_string (aod cont_type "Min" "0.0")
			and rmax = float'_of_string (aod cont_type "Max" "1.0") in
			fun () -> floor (Gsl_randist.flat rng rmin (rmax +. 1.))
		| "Poisson" -> 
			let mu = float'_of_string (aod cont_type "Mu" "1.0") in
			fun () -> foi (Gsl_randist.poisson rng mu)
		| "Bernoulli" ->
			let p = float'_of_string (aod cont_type "P" "1.0") in
			fun () -> foi  (Gsl_randist.bernoulli rng p)
		| "Binomial" ->
			let p = float'_of_string (aod cont_type "P" "1.0") in
			let n = ios (aod cont_type "N" "1") in
			fun () -> foi  (Gsl_randist.binomial rng p n)
		| "NegativeBinomial" ->
			let p = float'_of_string (aod cont_type "P" "1.0") in
			let n = float'_of_string (aod cont_type "N" "1.0") in
			(*uncomment this for ocamlgsl > 0.5 ::::let n = int_of_float n in::::*)
			fun () -> foi  (Gsl_randist.negative_binomial rng p n)
		| "Pascal" ->
			let p = float'_of_string (aod cont_type "P" "1.0") in
			let k = ios (aod cont_type "K" "1") in
			fun () -> foi  (Gsl_randist.pascal rng p k)
		| "Geometric" ->
			let p = float'_of_string (aod cont_type "P" "1.0") in
			fun () -> foi  (Gsl_randist.geometric rng p)
		| "HyperGeometric" ->
			let n1 = ios (aod cont_type "N1" "1") in
			let n2 = ios (aod cont_type "N2" "1") in
			let t = ios (aod cont_type "T" "1") in
			fun () -> foi  (Gsl_randist.hypergeometric rng ~n1:n1 ~n2:n2 ~t:t)
		| _ -> (*"Logarithmic"*)
			let p = float'_of_string (aod cont_type "P" "1.0") in
			fun () -> foi  (Gsl_randist.logarithmic rng p)
	)
	| _ -> 
		let rlist = float'_list_of_string (Xml.attrib rand_type "ListOfValues") in
		if List.length rlist < 1 then
			failwith "ListOfValues must contain at least one entry."
		else
			fun () -> 
				let r = Gsl_rng.uniform_int rng (List.length rlist) in
				List.nth rlist r
;;


(* read an xml file and convert it to a string which can be processed by xml-light: *)
(* THIS FUNCTION IS NOT NEEDED ANYMORE! *)
let dtd_string_from_file dtd_filename =
	let contains_string str contained =
		let p = try Str.search_forward (Str.regexp contained) str 0
			with Not_found -> -1
		in
		if p>=0 then 
			true
		else
			false
	in
	let inp_dtd = open_in dtd_filename in
	let entityqueue = Queue.create () in
	let rec file_to_string () =
		try 
			let line = input_line inp_dtd in
			if contains_string line "!ENTITY" then	(*check for parameter entities*)
				let _ = Str.search_forward (Str.regexp "%[ ]+\\([^ ]+\\) [^\"]*\"\\(.*\\)\"") line 0 in
				let ent = Str.matched_group 1 line in
				let subs = Str.matched_group 2 line in
				Queue.push (ent, subs) entityqueue;
				file_to_string ()
			else
				(line ^ (file_to_string ()))
		with End_of_file -> close_in inp_dtd; ""
	in
	let dstr = file_to_string () in
	(*substitute parameter entities:*)
	let rec subs_ent str q =
		match Queue.is_empty q with
		true -> str
		| _ -> 
			let e = Queue.pop q in
			let str_repl = Str.global_replace (Str.regexp ("%" ^ (fst e) ^ ";")) (snd e) str in
			subs_ent str_repl q
	in
	subs_ent dstr entityqueue
;;

(*load the standard DTD for mathexpr:
let dst = dtd_string_from_file "mathexpr.dtd";;*) 

let dst =
	let contains_string str contained =
		let p = try Str.search_forward (Str.regexp contained) str 0
			with Not_found -> -1
		in
		if p>=0 then 
			true
		else
			false
	in
	let dtd_string_default = Mathexpr_dtd.dtd_string in
	let entityqueue = Queue.create () in
	let dstlist = Str.split (Str.regexp "\n") dtd_string_default in
	let rec string_of_list st l =
		match l with
		[] -> st
		| _ ->
			let line = List.hd l in
			if contains_string line "!ENTITY" then	(*check for parameter entities*)
				let _ = Str.search_forward (Str.regexp "%[ ]+\\([^ ]+\\) [^\"]*\"\\(.*\\)\"") line 0 in
				let ent = Str.matched_group 1 line in
				let subs = Str.matched_group 2 line in
				Queue.push (ent, subs) entityqueue;
				string_of_list st (List.tl l)
			else
				string_of_list (st ^ line) (List.tl l)
	in
	let dstr = string_of_list "" dstlist in
	(*substitute parameter entities:*)
	let rec subs_ent str q =
		match Queue.is_empty q with
		true -> str
		| _ -> 
			let e = Queue.pop q in
			let str_repl = Str.global_replace (Str.regexp ("%" ^ (fst e) ^ ";")) (snd e) str in
			subs_ent str_repl q
	in
	subs_ent dstr entityqueue
;;


let dtd = Dtd.check (Dtd.parse_string dst);;

(** [xml_of_file filename]: open file [filename] and convert it to Xml-light xml *)
let xml_of_file filename =
	let inp = open_in filename in
	let rec find_start ch linestart line =
		if (String.length line >= 9 && String.sub line 0 9 = linestart) then
			line
		else
			find_start ch linestart (input_line ch)
	in
	(*let doctypeline = find_start inp "<!DOCTYPE" "" in*)
	let x = Xml.parse_in inp in
	close_in inp;
	ignore (Dtd.prove dtd "MathematicalExpression" x);
	x
;;


let rec function_of_xml ?(internal_functions=[]) ?(assoc_list=[]) self_id xml_fun =
	let self_number_of_arguments =
		let tg = Xml.tag xml_fun in
		if tg = "Function" || tg = "MatrixFunction" || tg = "StringFunction" || tg = "SourceFunction" then
			let argstring = aod xml_fun "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			List.length argstringlist
		else
			0
	in
	(*pointer that points to the function itself; initialized to id:*)
	let self_pointer = ref (fun x -> x.(0)) in
	
	let is_matrix_function =
		(Xml.tag xml_fun) = "MatrixFunction"
	in
	
	let rec fun_of_xml internal_functions assoc_list elem =
		let fon = float_of_nom ~xmltag:(Xml.tag elem)
		and son = string_of_nom
		and xon = xml_of_nom ~xmltag:(Xml.tag elem)
		and mon = matrix_of_nom ~xmltag:(Xml.tag elem) in
		let fob relation x y =
			if relation x y then
					1.0
				else
					0.0
		in
		let matrix_min_max m =
			let nrows = Bigarray.Array2.dim1 m in
			let minvec = Gsl_vector.create nrows
			and maxvec = Gsl_vector.create nrows in
			for i = 0 to nrows - 1 do
				let curr_row = Bigarray.Array2.slice_left m i in
				minvec.{i} <- Gsl_vector.min curr_row;
				maxvec.{i} <- Gsl_vector.max curr_row;
			done;
			((Gsl_vector.min minvec), (Gsl_vector.max maxvec))
		in
		(*get rows/columns from matrix m; first: start, n: number of rows/colums:*)
		let get_rows =
			Bigarray.Array2.sub_left
		in
		let get_columns m first n =
			let transpose mat =
				let (r, c) = Gsl_matrix.dims mat in
				let mt =  Gsl_matrix.create c r in
				Gsl_matrix.transpose mt mat;
				mt
			in
			let newmat = Bigarray.Array2.sub_left (transpose m) first n in
			transpose newmat
		in
		(*matlab linspace and logspace:*)
		let linspace start stop len =
			if len <= 0 then
				failwith "linspace: invalid length. Check all calls of the functions SequenceVector, LinSpace, and LogSpace.";
			let v = Gsl_matrix.create 1 len in
			let d = 
				(stop -. start) /. (float_of_int (len - 1)) 
			in
			for i=0 to len-1 do
				v.{0, i} <- start +. (float_of_int i) *. d
			done;
			v
		in
		let logspace start stop len =
			let v = linspace start stop len in
			for i=0 to len-1 do
				v.{0, i} <- 10. ** v.{0, i}
			done;
			v
		in
		(*concatenate 2 matrices m1 and m2:*)
		let concat_matrices by_rows m1 m2 =
			let (dm1x, dm1y) = Gsl_matrix.dims m1
			and (dm2x, dm2y) = Gsl_matrix.dims m2 in
			if not by_rows then
			(
				if dm1y <> dm2y then
					failwith "Matrix concatanation failed: wrong dimensions.";
				let m = Gsl_matrix.create (dm1x + dm2x) dm1y in
				for j = 0 to dm1y - 1 do
					for i = 0 to dm1x - 1 do
						m.{i, j} <- m1.{i, j}
					done;
					for i = dm1x to dm1x + dm2x - 1 do
						m.{i, j} <- m2.{i - dm1x, j}
					done
				done;
				m
			)
			else
			(
				if dm1x <> dm2x then
					failwith "Matrix concatanation failed: wrong dimensions.";
				let m = Gsl_matrix.create dm1x (dm1y + dm2y) in
				for i = 0 to dm1x - 1 do
					for j = 0 to dm1y - 1 do
						m.{i, j} <- m1.{i, j}
					done;
					for j = dm1y to dm1y + dm2y - 1 do
						m.{i, j} <- m2.{i, j - dm1y}
					done
				done;
				m
			)
		in
		let function_of_expression e arg_array = 
			match e.head with
				Fun f ->  
				(
					f (
						let flist =
							List.map 
								(
									fun t -> t arg_array
								)
								e.body 
						in
						Array.append
							(Array.of_list flist)
							arg_array
					)
				)
				| Arg (in_situ, i) -> 
				(
					(*DELETE this:
					let str_of_nom = function
						| Scal f -> (string_of_float f) ^ ", "
						| Xml x -> (Xml.to_string x) ^ ", "
						| _ -> "M, "
					in
					Printf.printf "%i, [|" i; 
					Array.iter (fun t -> print_string (str_of_nom t)) arg_array; 
					print_endline "|]";*)
					
					let a = arg_array.(i) in
					match a with
					Mat m ->
						if in_situ then
							Mat m
						else
							Mat (Gsl_matrix.copy m)
					| _ -> a
				)
				| Const c -> c
				| Var v -> Scal !v
				| Str_var v -> Str !v
				| Xml_var v -> Xml !v
				| Let a -> 
					let scope = List.hd e.body in
					let new_args = 
						Array.map 
							(
								fun t -> t arg_array
							)
							a
					in
					(*DELETE this:
						let str_of_nom = function
							| Scal f -> (string_of_float f) ^ ", "
							| Xml x -> (Xml.to_string x) ^ ", "
							| _ -> "M, "
						in
						print_string "APPEND: [|"; 
						Array.iter (fun t -> print_string (str_of_nom t)) arg_array; 
						print_string "|],\t";
						print_string "[|"; 
						Array.iter (fun t -> print_string (str_of_nom t)) new_args; 
						print_endline "|],\t";*)
					
					scope (Array.append arg_array new_args)
				| Cond c ->
					let (conditions, results, default) = c in
					let bool_of_cond = function
						| Scal 0.0 -> false
						| _ -> true
					in
					let l = Array.length conditions in
					let rec find_condition i =
						if i < l then
						(
							if bool_of_cond (conditions.(i) arg_array) then
								results.(i) arg_array
							else
								find_condition (i+1)
						)
						else
							default arg_array
					in
					find_condition 0
		in
		let apply1 f =
			let nf x = Scal (f (fon x.(0))) in
			function_of_expression 
				{head = Fun nf; body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]}
		in
		let apply2 f =
			let nf x = Scal (f (fon x.(0)) (fon x.(1)) ) in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		in
		let apply2float_1int_to_mat f =
			let nf x = 
				Mat (f (fon x.(0)) (fon x.(1)) (int_of_float (fon x.(2))) ) 
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2)
				]
			}
		in
		let apply_float_of_mat f =
			let nf x = Scal (f (mon x.(0)) ) in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))
				]
			}
		in
		let apply_logical_to_matricies f =
			let nf x =
				let m1 = mon x.(0)
				and m2 = mon x.(1) in
				let (r, c) = Gsl_matrix.dims m1 in
				let result = ref 1.0 in
				if (r, c) = (Gsl_matrix.dims m2) then
				(
					for i = 0 to r-1 do
						for j = 0 to c-1 do
							if not (f m1.{i, j} m2.{i, j}) then
								result := 0.0;
						done
					done;
					Scal !result
				)
				else
					failwith "Matrix comparisons can only be performed on two matrices with the same dimensions."
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		in
		let apply_float_matrix f =
			let nf x =
				let s = fon x.(0)
				and m = mon x.(1) in
				f m s;
				Mat m
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		in
		let apply_int2_matrix f =
			let nf x =
				let s1 = int_of_float (fon x.(0))
				and s2 = int_of_float (fon x.(1))
				and m = mon x.(2) in
				f m s1 s2;
				Mat m
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 1)));
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2)
				]
			}
		in
		let apply_float_of_string2 f =
			let nf x =
				let s1 = son x.(0)
				and s2 = son x.(1) in
				Scal (f s1 s2)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		in
		(*let apply_int2_float_matrix f =
			let nf x =
				let s1 = int_of_float (fon x.(0))
				and s2 = int_of_float (fon x.(1))
				and s3 = fon x.(2)
				and m = mon x.(3) in
				f m s1 s2 s3;
				Mat m
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 1)));
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 3)
				]
			}
		in*)
		let apply_matrix f =
			let nf x = 
				let m = mon x.(0) in
				f m;
				Mat m
			in
			function_of_expression 
				{head = Fun nf; body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]}
		in
		let apply_matrix2 f = 
			let nf x =
				let m1 = mon x.(0)
				and m2 = mon x.(1) in
				f m1 m2;
				Mat m1
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		in
		
		let source_equal x y =
			let rec srceq = function
				| (Xml.Element (xt, xa, xc), Xml.Element (yt, ya, yc)) ->
					let split_attribs = List.partition (fun t -> Xml.tag t = "SymbolAttribute") in
					let (xatt, xcld) = split_attribs xc
					and (yatt, ycld) = split_attribs yc in
					(xt = yt)
					&& (set_compare xatt yatt)
					&& (set_compare xa ya)
					&& (List.length xcld = List.length ycld)
					&& (List.for_all srceq (List.combine xcld ycld))
				| (Xml.PCData s1, Xml.PCData s2) -> s1 = s2
				| _ -> false
			in
			srceq (x, y)

		in
		
		let condition () =
			let chl = Xml.children elem in
			let (clist, defaultlist) = 
				List.partition (fun t -> Xml.tag t <> "Default" && Xml.tag t <> "DefaultScalar" && Xml.tag t <> "DefaultMatrix" && Xml.tag t <> "DefaultString") chl 
			in
			let condlist = 
				List.map 
				(
					fun t -> 
						let tchl = Xml.children t in
						(
							fun_of_xml internal_functions assoc_list (List.hd tchl),
							fun_of_xml internal_functions assoc_list (List.hd (List.tl tchl))
						)
				)
				clist
			in
			let default = 
				fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.hd defaultlist)))
			in
			let (c1, c2) = List.split condlist in
			
			function_of_expression
			{
				head = 
					Cond 
					(
						Array.of_list c1,
						Array.of_list c2,
						default
					);
				body = []
			}
		in
		let if_then_else () =
			let chl = Xml.children elem in
			function_of_expression
			{
				head = 
					Cond 
					(
						[| fun_of_xml internal_functions assoc_list (List.hd chl) |],
						[| fun_of_xml internal_functions assoc_list (List.nth chl 1) |],
						fun_of_xml internal_functions assoc_list (List.nth chl 2)
					);
				body = []
			}
		in
		
		let internal_declarations () =
			let chl = List.rev (Xml.children elem) in
			let decls = 
				List.map 
					(
						fun t ->
						(
							Xml.attrib t "Identifier",
							List.hd (Xml.children t)
						)
					)
					(List.rev (List.tl chl))
			in
			let (fcts, nonfcts) = 
				List.partition
					(
						fun t -> Xml.tag (snd t) = "Function" || Xml.tag (snd t) = "MatrixFunction" || Xml.tag (snd t) = "StringFunction" || Xml.tag (snd t) = "SourceFunction"
					)
					decls
			in
			let decl_fcts =
				List.map 
					(
						fun t ->
							let l = List.length (argstringlist_of_string (aod elem "ArgumentList" "")) in
							match Xml.tag (snd t) with
							"Function" ->
							(
								let cont =
									let c = 
										function_of_xml 
											~internal_functions:internal_functions 
											~assoc_list:assoc_list 
											(fst t) 
											(snd t)
									in
									fun x -> fon (c x)
								in
								(
									(fst t),
									Function
									{
										nargs = l;
										fcontent = cont
									}
								)
							)
							| "MatrixFunction" ->
							(
								let cont =
									let c = 
										function_of_xml 
											~internal_functions:internal_functions 
											~assoc_list:assoc_list 
											(fst t) 
											(snd t)
									in
									fun x -> mon (c x)
								in
								(
									(fst t),
									Matrix_function
									{
										mat_nargs = l;
										mat_fcontent = cont
									}
								)
							)
							| "SourceFunction" ->
							(
								let cont =
									let c = 
										function_of_xml 
											~internal_functions:internal_functions 
											~assoc_list:assoc_list 
											(fst t) 
											(snd t)
									in
									fun x -> xon (c x)
								in
								(
									(fst t),
									Xml_function
									{
										xml_nargs = l;
										xml_fcontent = cont
									}
								)
							)
							| _ -> (*"StringFunction"*)
							(
								let cont =
									let c = 
										function_of_xml 
											~internal_functions:internal_functions 
											~assoc_list:assoc_list 
											(fst t) 
											(snd t)
									in
									fun x -> son (c x)
								in
								(
									(fst t),
									String_function
									{
										str_nargs = l;
										str_fcontent = cont
									}
								)
							)
					)
					(List.rev fcts)
			in
			let (nonfun_ids, nonfun_values) =
				let l = List.map
					(
						fun t ->
							let c = 
								fun_of_xml internal_functions assoc_list (snd t)
							in
							(
								(fst t),
								c
							)
					)
					nonfcts
				in
				List.split l
			in
			let new_assoc =
				let rec make_id_assoc_list number al = function
					| [] -> al
					| h::t -> make_id_assoc_list (number + 1) ((h, number)::al) t
				in
				let l = List.length assoc_list in
				make_id_assoc_list l [] nonfun_ids
			in
			function_of_expression
			{
				head = Let (Array.of_list nonfun_values);
				body =
				[
					fun_of_xml 
						(decl_fcts @ internal_functions) 
						(*(assoc_list @ new_assoc)*)
						(new_assoc @ assoc_list)
						(List.hd chl)
				]
			}
		in
		
		let first_child () = 
			fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))
		in
		match Xml.tag elem with
		"Function" ->
			let argstring = aod elem "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			let content = List.hd (Xml.children elem) in
			if Xml.tag content == "Identity" then
				fun x -> x.(0)
			else
			(
				(*if there is an existing argument list, the function is internal;
				it receives all the arguments of its parent function after its own arguments:*)
				let renumbered_old_assoc_list =
					let la = List.length argstringlist in
					let rec renumber l = function
						| [] -> l
						| h::t -> renumber ((fst h, la + (snd h))::l) t
					in
					List.rev
						(renumber [] assoc_list)
				in
				(*association list of arguments:*)
				let rec make_id_assoc_list number al l =
					match l with
					[] -> al
					| _ -> make_id_assoc_list (number + 1) (((List.hd l), number)::al) (List.tl l)
				in
				let id_assoc_list = make_id_assoc_list 0 [] argstringlist in				
				(*body:*)
				fun_of_xml internal_functions (id_assoc_list @ renumbered_old_assoc_list) content
			)
		| "MatrixFunction" | "StringFunction" | "SourceFunction" ->
			let argstring = aod elem "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			let content = List.hd (Xml.children elem) in
			
			(*if there is an existing argument list, the function is internal;
			it receives all the arguments of its parent function after its own arguments:*)
			let renumbered_old_assoc_list =
				let la = List.length argstringlist in
				let rec renumber l = function
					| [] -> l
					| h::t -> renumber ((fst h, la + (snd h))::l) t
				in
				List.rev
					(renumber [] assoc_list)
			in
			(*association list of arguments:*)
			let rec make_id_assoc_list number al l =
				match l with
				[] -> al
				| _ -> make_id_assoc_list (number + 1) (((List.hd l), number)::al) (List.tl l)
			in
			let id_assoc_list = make_id_assoc_list 0 [] argstringlist in
			(*body:*)
			fun_of_xml internal_functions (id_assoc_list @ renumbered_old_assoc_list) content
		| "Constant" -> first_child ()
		| "Variable" -> 
		(
			let cld = Xml.children elem in
			if cld = [] then
				fun x -> Scal 0.0
			else
			
				fun_of_xml internal_functions assoc_list ( List.hd cld ) 
		)
		| "StringVariable" -> 
		(
			let cld = Xml.children elem in
			if cld = [] then
				fun x -> Str ""
			else
			
				fun_of_xml internal_functions assoc_list ( List.hd cld ) 
		)
		| "Source" -> first_child ()
		| "Body" -> first_child ()	(*deprecated*)
		| "Matrix" -> first_child ()
		| "DefaultExpression" -> first_child ()
		| "DefaultString" -> first_child ()
		| "FunctionOfMatrix" -> first_child ()
		| "MatrixComparison" -> first_child ()
		| "Hyperbolic" -> first_child ()
		
		(*polymorphic w.r.t data type:*)
		| "Let" -> internal_declarations ()
		| "Case" -> condition ()
		| "IfThenElse" -> if_then_else ()
		| "Eval" ->
			let nf z = 
				let symb_exp = xon z.(0) in
				(*retransform "Symbol" nodes to expressions:*)
				let symbols_transformed =
					transform_symbols symb_exp
				in
				(function_of_xml ~internal_functions:internal_functions ~assoc_list:assoc_list "" symbols_transformed) (Array.sub z 1 ((Array.length z) - 1))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
				]
			}
		| "DeclaredValue" ->	(* corresponds to NumericValue, DeclaredMatrix, DeclaredString, ...*)
			let ident = (Xml.attrib elem "Identifier") in
			function_of_expression
			{
				head = 
				(
					try
					let arg_number = List.assoc ident assoc_list in
					Arg (false, arg_number)
					with Not_found -> 
					try
					let c = float_of_string ident in
					Const (Scal c)
					with Failure "float_of_string" ->
					(
						match gd ident with
						Constant c -> Const (Scal c)
						| Variable v -> Var v
						| Function f -> 
							if f.nargs > 0 then
								raise (Invalid_declaration_type ident)
							else 
								let nf x = Scal (f.fcontent x) in
								Fun nf
						| Matrix t -> 
							Const (Mat (Gsl_matrix.copy t))
						| Matrix_function f -> 
							if f.mat_nargs > 0 then
								raise (Invalid_declaration_type ident)
							else
								let nf x = Mat (f.mat_fcontent x) in
								Fun nf
						| String_variable v -> Str_var v
						| Xml_variable x -> Xml_var x
						| _ -> raise (Invalid_declaration_type ident)
					)
				);
				body = []
			}
		| "DeclaredFunction" ->	(*DeclaredScalarFunction, DeclaredMatrixFunction, DeclaredStringFunction*)
			let ident = Xml.attrib elem "Identifier" in
			let chl = Xml.children elem in
			let rlist = 
				List.map 
					(fun_of_xml internal_functions assoc_list)
					chl
			in
			(*** recursive case: ***)
			if ident = self_id then
			(
				match (Xml.tag xml_fun) with
				"Function" -> 
					let f = 
						let self_fun = 
							let sf = lazy (!self_pointer) in
							fun x -> fon ((Lazy.force sf) x)
						in
						{
							nargs = self_number_of_arguments;
							fcontent = self_fun
						}
					in
					if (List.length chl) < f.nargs then
						raise (Not_enough_arguments ident)
					else
						function_of_expression
						{
							head = Fun (fun x -> Scal (f.fcontent x));
							body = rlist
						}
				| "MatrixFunction" ->
					let f = 
						let self_fun = 
							let sf = lazy (!self_pointer) in
							fun x -> mon ((Lazy.force sf) x)
						in
						{
							mat_nargs = self_number_of_arguments;
							mat_fcontent = self_fun
						}
					in
					if (List.length chl) < f.mat_nargs then
						raise (Not_enough_arguments ident)
					else
						function_of_expression
						{
							head = Fun (fun x -> Mat (f.mat_fcontent x));
							body = rlist
						}
				| "StringFunction" -> 	(*"StringFunction"*)
					let f = 
						let self_fun = 
							let sf = lazy (!self_pointer) in
							fun x -> son ((Lazy.force sf) x)
						in
						{
							str_nargs = self_number_of_arguments;
							str_fcontent = self_fun
						}
					in
					if (List.length chl) < f.str_nargs then
						raise (Not_enough_arguments ident)
					else
						function_of_expression
						{
							head = Fun (fun x -> Str (f.str_fcontent x));
							body = rlist
						}
				| _ -> 	(*"SourceFunction"*)
					let f = 
						let self_fun = 
							let sf = lazy (!self_pointer) in
							fun x -> xon ((Lazy.force sf) x)
						in
						{
							xml_nargs = self_number_of_arguments;
							xml_fcontent = self_fun
						}
					in
					if (List.length chl) < f.xml_nargs then
						raise (Not_enough_arguments ident)
					else
						function_of_expression
						{
							head = Fun (fun x -> Xml (f.xml_fcontent x));
							body = rlist
						}
			)
			(*** non-recursive case: ***)
			else
			(
				match
				(
					try List.assoc ident internal_functions
					with Not_found -> gd ident
				)
				with
				Function f -> 
					if (List.length chl) < f.nargs then
						raise (Not_enough_arguments ident)
					else
						function_of_expression
						{
							head = Fun (fun x -> Scal (f.fcontent x));
							body = rlist
						}
				| Matrix_function f -> 
					if (List.length chl) < f.mat_nargs then
						raise (Not_enough_arguments ident)
					else
						function_of_expression
						{
							head = Fun (fun x -> Mat (f.mat_fcontent x));
							body = rlist
						}
				| String_function f -> 
					if (List.length chl) < f.str_nargs then
						raise (Not_enough_arguments ident)
					else
						function_of_expression
						{
							head = Fun (fun x -> Str (f.str_fcontent x));
							body = rlist
						}
				| Xml_function f -> 
					if (List.length chl) < f.xml_nargs then
						raise (Not_enough_arguments ident)
					else
						function_of_expression
						{
							head = Fun (fun x -> Xml (f.xml_fcontent x));
							body = rlist
						}
				| _ -> raise (Invalid_declaration_type ident)
			)
			
		
		(*return float:*)
		| "Plus" -> 
			let chl = Xml.children elem in
			if List.length chl == 2 then
				apply2 (+.)
			else
			(
				let rlist = 
					List.map 
					(
						fun t -> fun_of_xml internal_functions assoc_list t
					)
					chl
				in
				let nf z = 
					let s = ref 0.0 in
					for i=0 to (List.length rlist) -1 do
						s := !s +. (fon z.(i))
					done;
					Scal !s
				in
				function_of_expression
				{
					head = Fun nf;
					body = rlist
				}
			)
		| "Times" ->
			let chl = Xml.children elem in
			if List.length chl == 2 then
				apply2 ( *. )
			else
			(
				let rlist = 
					List.map 
					(
						fun t -> fun_of_xml internal_functions assoc_list t
					)
					chl
				in
				let nf z = 
					let s = ref 1.0 in
					for i=0 to (List.length rlist) -1 do
						s := !s *. (fon z.(i))
					done;
					Scal !s
				in
				function_of_expression
				{
					head = Fun nf;
					body = rlist
				}
			)
		| "And" ->
			let chl = Xml.children elem in
			if List.length chl == 2 then
				apply2 ( fun x y -> if x *. y <> 0. then 1. else 0. )
			else
			(
				let rlist = 
					List.map 
					(
						fun t -> fun_of_xml internal_functions assoc_list t
					)
					chl
				in
				let nf z = 
					let s = ref 1.0 in
					for i=0 to (List.length rlist) -1 do
						s := !s *. (fon z.(i))
					done;
					Scal ( if !s <> 0. then 1. else 0. )
				in
				function_of_expression
				{
					head = Fun nf;
					body = rlist
				}
			)
		| "Or" -> 
			let chl = Xml.children elem in
			let rlist = 
				List.map 
				(
					fun t -> fun_of_xml internal_functions assoc_list t
				)
				chl
			in
			let nf z = 
				let s = ref 0.0 in
				for i=0 to (List.length rlist) -1 do
					s := !s +. (abs_float (fon z.(i)))
				done;
				Scal ( if !s > 0. then 1. else 0.)
			in
			function_of_expression
			{
				head = Fun nf;
				body = rlist
			}
		| "ScalarCase" -> condition ()
		| "ScalarIfThenElse" -> if_then_else ()
		
		(*	let chl = Xml.children elem in
			let (clist, defaultlist) = List.partition (fun t -> Xml.tag t <> "Default") chl in
			let condlist = 
				List.map 
				(
					fun t -> 
						let tchl = Xml.children t in
						(
							fun_of_xml internal_functions assoc_list (List.hd tchl),
							fun_of_xml internal_functions assoc_list (List.hd (List.tl tchl))
						)
				)
				clist
			in
			let default = 
				fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.hd defaultlist)))
			in
			let (c1, c2) = List.split condlist in
			
			function_of_expression
			{
				head = 
					Cond 
					(
						Array.of_list c1,
						Array.of_list c2,
						default
					);
				body = []
			}
		*)
		| "Divide" -> apply2 (/.)
		| "Subtract" -> apply2 (-.) 
		| "Power" -> apply2 ( ** )
		| "Mod" -> apply2 mod_float
		| "ArcTan2" -> apply2 atan2
		| "Abs" -> apply1 abs_float
		| "Negate" -> apply1 (~-.)
		| "Exp" -> apply1 exp
		| "Log" -> apply1 log
		| "Log10" -> apply1 log10
		| "Sqrt" -> apply1 sqrt
		| "Sin" -> apply1 sin
		| "Cos" -> apply1 cos
		| "Tan" -> apply1 tan
		| "ArcSin" -> apply1 asin
		| "ArcCos" -> apply1 acos
		| "ArcTan" -> apply1 atan
		| "Sinh" -> apply1 sinh
		| "Cosh" -> apply1 cosh
		| "Tanh" -> apply1 tanh
		| "ArcSinh" -> apply1 Gsl_math.asinh
		| "ArcCosh" -> apply1 Gsl_math.acosh
		| "ArcTanh" -> apply1 Gsl_math.atanh
		| "Floor" -> apply1 floor
		| "Ceil" -> apply1 ceil
		| "Not" -> apply1 (fun x -> if x <> 0.0 then 0.0 else 1.0)
		| "Less" -> apply2 (fob (<))
		| "LessEqual" -> apply2 (fob (<=))
		| "Greater" -> apply2 (fob (>))
		| "GreaterEqual" -> apply2 (fob (>=))
		| "Equal" -> apply2 (fob (=))
		| "Unequal" -> apply2 (fob (<>))
		| "NumericValue" -> 
			function_of_expression
			{
				head = 
				(
					let s = (Xml.attrib elem "Value") in
					try
					Const (Scal (float_of_string s))
					with Failure t ->
					(
						try
						let arg_number = List.assoc s assoc_list in
						Arg (false, arg_number)
						with Not_found -> 
						(
							match gd s with
							Constant c -> Const (Scal c)
							| Variable v -> Var v
							| Function f -> 
								if f.nargs > 0 then
									raise (Invalid_declaration_type s)
								else 
									let nf x = Scal (f.fcontent x) in
									Fun nf
							| _ -> raise (Invalid_declaration_type s)
						)
					)
				);
				body = []
			}
		| "StringToCharNumber" ->
			let nf x =
				let s = son x.(0) in
				if String.length s < 1 then
					failwith "StringToCharNumber is called with empty string";
				Scal (float_of_int (int_of_char s.[0]))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "ListLength" ->	(*special declarations, start with "@LENGTH@ " and represent list lengths*)
			let listid = (Xml.attrib elem "Identifier") in
			let s = "@LENGTH@ " ^ listid in
			function_of_expression
			{
				head = 
				(
					try
					(
						match gd s with
						Constant c -> Const (Scal c)
						| _ -> failwith "no length"
					)
					with Undeclared_identifier _ -> raise (Undeclared_identifier listid)
				);
				body = []
			}
		| "ListMember" -> 
			let listid = (Xml.attrib elem "Identifier") in
			let nf x = 
				let item_id = (son x.(0)) in
				try
					(
						let list_length =
							match gd ("@LENGTH@ " ^ listid) with
							Constant c -> int_of_float c
							| _ -> failwith "no length"
						in
						let rec find_in_list = function
							| (true, _) -> Scal 1.0
							| (_, 0) -> Scal 0.0
							| (_, i) -> 
								find_in_list 
								(
									(
										match gd ("@LISTINDEX@ " ^ listid ^ " " ^ (string_of_int i)) with
										String_variable v -> (!v = item_id)
										| _ -> failwith "no index"
									),
									i - 1
								)
						in
						find_in_list (false, list_length)
					)
					with Undeclared_identifier _ -> raise (Undeclared_identifier listid)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
				]
			}
		| "NumberOfChildren" -> 
			let nf x = 
				let e = xon x.(0) in
				Scal (float_of_int (List.length (Xml.children e)))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceForAllChildren" -> 
			let boolfun_xml = List.hd (Xml.children elem) in
			let argstring = aod boolfun_xml "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			if List.length argstringlist <> 1 then
				failwith "A function applied to SourceForAllChildren must have exactly one argument.";
			let boolfun = function_of_xml ~internal_functions:internal_functions ~assoc_list:assoc_list "" boolfun_xml in
			let nf z =
				let e = xon z.(0) in
				let cld = List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (Xml.children e) in
				let boollist = 
					List.map 
						(
							fun t ->
								z.(0) <- Xml t;
								match (boolfun z) with
								| Scal s -> s <> 0.0
								| _ -> failwith "Wrong data type returned from function in SourceForAllChildren (scalar number expected)"
						)
						cld 
				in
				let true_for_all =
					not (List.mem false boollist)
				in
				Scal (if true_for_all then 1.0 else 0.0)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [ fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem))) ]
			}
		| "SourceExistsChild" -> 
			let boolfun_xml = List.hd (Xml.children elem) in
			let argstring = aod boolfun_xml "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			if List.length argstringlist <> 1 then
				failwith "A function applied to SourceExistsChild must have exactly one argument.";
			let boolfun = function_of_xml ~internal_functions:internal_functions ~assoc_list:assoc_list "" boolfun_xml in
			let nf z =
				let e = xon z.(0) in
				let cld = List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (Xml.children e) in
				let boollist = 
					List.map 
						(
							fun t ->
								z.(0) <- Xml t;
								match (boolfun z) with
								| Scal s -> s <> 0.0
								| _ -> failwith "Wrong data type returned from function in SourceExistsChild (scalar number expected)"
						)
						cld 
				in
				let cond =
					List.mem true boollist
				in
				Scal (if cond then 1.0 else 0.0)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [ fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem))) ]
			}
		| "SourceEquals" ->
			let nf x =
				let x1 = xon x.(0)
				and x2 = xon x.(1) in
				Scal (if (source_equal x1 x2) then 1.0 else 0.0)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		| "SourceIsChild" ->
			let nf x =
				let candidate = xon x.(0)
				and xlist = xon x.(1) in
				let is_mem = List.exists (source_equal candidate) (Xml.children xlist) in
				Scal (if is_mem then 1.0 else 0.0)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		| "SourceIsAttribute" ->
			let nf x =
				let attlist = 
					List.filter (fun t -> Xml.tag t = "SymbolAttribute") (Xml.children (xon x.(0))) 
				and attname = son x.(1) in
				let is_mem = List.exists (fun t -> Xml.attrib t "AttributeTag" = attname) attlist in
				Scal (if is_mem then 1.0 else 0.0)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		| "SourceIsAssocKey" ->
			let nf x =
				let assoc_list = Xml.children (xon x.(0))
				and key = xon x.(1) in
				let is_mem = 
					List.exists (fun t -> (source_equal (List.hd (Xml.children t)) key)) assoc_list 	
				in
				Scal (if is_mem then 1.0 else 0.0)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		| "Random" ->
			let randfun = random_function_of_xml elem in
			function_of_expression
			{
				head = Fun (fun x -> Scal (randfun ()));
				body = []
			}
		| "MatrixElement" ->
			let rown = List.hd (Xml.children (List.nth (Xml.children elem) 0))
			and coln = List.hd (Xml.children (List.nth (Xml.children elem) 1))
			and m = (List.nth (Xml.children elem) 2) in
			function_of_expression
			{
				head = Fun
				(
					fun a ->
						(*print_gsl_matrix (mon a.(2));*)
						let declared_matrix = mon a.(2) 
						and rn = ((int_of_float (fon a.(0))) - 1)
						and cn = ((int_of_float (fon a.(1))) - 1) in
						try
						Scal (Gsl_matrix.get declared_matrix rn cn)
						with 
							Invalid_argument "index out of bounds" -> 
								let (d1, d2) = Gsl_matrix.dims declared_matrix in
								failwith 
								(
									"Index out of bounds: Element (" 
									^ (string_of_int rn) ^ ","
									^ (string_of_int cn)
									^ ") is to be extracted from matrix \n"
									^ (Xml.to_string_fmt m)
									^ "\n which has dimensions ("
									^ (string_of_int d1) ^ ","
									^ (string_of_int d2) ^ ")."
								)
				);
				body =
				[
					fun_of_xml internal_functions assoc_list rown;
					fun_of_xml internal_functions assoc_list coln;
					fun_of_xml internal_functions assoc_list m
				]
			}
		| "G" -> 
			(* macro for MatrixElement:
			(G mat) -> (G mat 1); (G mat i) -> (G mat 1 i);
			(G mat row col) -> (MatrixElement row col)*) 
			let cld = Xml.children elem in
			let rec make_mat_elem = function
				| mat::[] -> 
					make_mat_elem [mat; Xml.Element ("NumericValue", ["Value", "1"], [])]
				| mat::col::[] -> 
					make_mat_elem [mat; Xml.Element ("NumericValue", ["Value", "1"], []); col]
				| mat::row::col::[] ->
					let r = Xml.Element ("NumberOfRows", [], [row])
					and c = Xml.Element ("NumberOfColumns", [], [col]) in
					Xml.Element ("MatrixElement", [], [r; c; mat])
				| _ -> failwith "G: wrong number of arguments"
			in
			fun_of_xml internal_functions assoc_list (make_mat_elem cld)
		| "MatrixNumberOfRows" -> apply_float_of_mat ( fun t -> float_of_int (fst (Gsl_matrix.dims t)) )
		| "MatrixNumberOfColumns" -> apply_float_of_mat ( fun t -> float_of_int (snd (Gsl_matrix.dims t)) )
		| "MatrixMin" -> apply_float_of_mat ( fun t -> fst (matrix_min_max t) )
		| "MatrixMax" -> apply_float_of_mat ( fun t -> snd (matrix_min_max t) )
		| "Determinant" ->
			let det m =
				let (nrows, ncols) = Gsl_matrix.dims m in
				if nrows <> ncols then
					failwith "Determinants can only be computed from square matrices.";
				let perm = Gsl_permut.make nrows in
				let signum = Gsl_linalg._LU_decomp (`M m) perm in
				Gsl_linalg._LU_det (`M m) signum
			in
			apply_float_of_mat det
		| "FoldVector" ->
			let left =
				match aod elem "ApplicationOrder" "Left" with
				"Left" -> true
				| _ -> false
			in
			let fun_xml = List.hd (Xml.children elem) in
			let argstring = aod fun_xml "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			if List.length argstringlist <> 2 then
				failwith "A function applied to FoldVector must have 2 arguments.";
			let foldfun = 
				function_of_xml 
					~internal_functions:internal_functions 
					~assoc_list:assoc_list 
					"" 
					fun_xml
			in
			let nf x =
				let v = Gsl_matrix.copy (mon x.(1)) in
				let (r, l) = Gsl_matrix.dims v in
				if r > 1 then
					prerr_endline "Warning: FoldVector is applied with a matrix instead of a vector, i.e. the first dimension of its first argument is greater than 1. Only the first row of the matrix will be used.";
				if left then
				(
					for i = 0 to l-1 do
						x.(1) <- Scal v.{0, i};
						x.(0) <- foldfun x
					done;
					x.(0)
				)
				else
				(
					x.(1) <- x.(0);
					for i = 1 to l do
						x.(0) <- Scal v.{0, l - i};
						x.(1) <- foldfun x
					done;
					x.(1)
				)
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2);
				]
			}
		| "MatrixEqual" -> apply_logical_to_matricies (=)
		| "MatrixLess" -> apply_logical_to_matricies (<)
		| "MatrixGreater" -> apply_logical_to_matricies (>)
		| "MatrixLessEqual" -> apply_logical_to_matricies (<=)
		| "MatrixGreaterEqual" -> apply_logical_to_matricies (>=)
		| "LetInExpression" -> internal_declarations ()
		| "DeclaredScalarFunction" ->
			let ident = Xml.attrib elem "Identifier" in
			let f = 
			(
				if ident = self_id then
				(
					if is_matrix_function then
						failwith ("DeclaredFunction was applied with identifier \"" ^ ident ^ "\" in a recursive call, but this is the identifier of a matrix function. Maybe you want to use DeclaredMatrixFunction instead.");
					let self_fun = 
						let sf = lazy (!self_pointer) in
						fun x -> fon ((Lazy.force sf) x)
					in
					{
						nargs = self_number_of_arguments;
						fcontent = self_fun
					}
				)
				else
				(
					try 
					(
						match List.assoc ident internal_functions with
						Function t -> t
						| _ -> raise (Invalid_declaration_type ident)
					)
					with
					Not_found ->
					(
						match gd ident with
						Function t -> t
						| _ -> raise (Invalid_declaration_type ident)
					)
				)
			) in
			let chl = Xml.children elem in
			if (List.length chl) < f.nargs then
				raise (Not_enough_arguments ident)
			else
			(
				let rlist = 
					List.map 
					(
						fun_of_xml internal_functions assoc_list
					)
					chl
				in
				function_of_expression
				{
					head = Fun (fun x -> Scal (f.fcontent x));
					body = rlist
				}
			)
		| "StringToNumber" ->
		(
			let nf x = 
				Scal 
				(
					try float_of_string (son x.(0)) 
					with Failure _ -> nan
				)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))
				]
			}
		)
		| "StringEquals" -> 
			apply_float_of_string2 (fun s1 s2 -> if s1 = s2 then 1.0 else 0.0)
		| "StringContains" ->
			apply_float_of_string2 
			(
				fun s1 s2 -> 
					try ignore (Str.search_forward (Str.regexp_string s2) s1 0); 1.0
					with Not_found -> 0.0
			)
		| "StringLength" ->
		(
			let nf x = 
				Scal (float_of_int (String.length (son x.(0))))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))
				]
			}
		)
		
		
		(*return matrix:*)
		| "MatrixAdd" -> apply_matrix2 Gsl_matrix.add
		| "MatrixSub" -> apply_matrix2 Gsl_matrix.sub
		| "MatrixMultElements" -> apply_matrix2 Gsl_matrix.mul_elements
		| "MatrixDivElements" -> apply_matrix2 Gsl_matrix.div_elements
		| "MatrixMult" ->
			function_of_expression
			{
				head = Fun 
				(
					fun x  ->
						let m1 = mon x.(0)
						and m2 = mon x.(1) in
						let (d1, c1) = Gsl_matrix.dims m1
						and (c2, d2) = Gsl_matrix.dims m2 in
						if c1 <> c2 then
							failwith "Error in matrix multiplication: invalid matrix dimensions."
						else
						(
							let p = Gsl_matrix.create d1 d2 in
							Gsl_linalg.matmult ~a:(`M m1) ~b:(`M m2) (`M p);
							Mat p
						)
				);
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		| "MatrixScale" -> apply_float_matrix Gsl_matrix.scale
		| "MatrixAddConstant" -> apply_float_matrix Gsl_matrix.add_constant
		| "MatrixScale01" -> 
			let matrix_scale01 m = 
				let (mn, mx) = matrix_min_max m in
				let r = mx -. mn in
				Gsl_matrix.add_constant m (-. mn);
				if r = 0.0 then
					(*m*)()
				else
				(
					Gsl_matrix.scale m (1./.r);
					(*m*)
				)
			in
			apply_matrix matrix_scale01
		| "Transpose" ->
			function_of_expression
			{
				head = Fun 
				(
					fun x ->
						let sourcemat = mon x.(0) in
						let (d1, d2) = Gsl_matrix.dims sourcemat in
						let destmat = Gsl_matrix.create d2 d1 in
						Gsl_matrix.transpose destmat sourcemat;
						Mat destmat
				); 
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SwapRows" -> apply_int2_matrix Gsl_matrix.swap_rows
		| "SwapColumns" -> apply_int2_matrix Gsl_matrix.swap_columns
		| "SwapRowCol" -> apply_int2_matrix Gsl_matrix.swap_rowcol
		| "ShuffleRows" -> 
			let rng = rng_of_xml elem in
			let shuffle_rows m =
				let r = Bigarray.Array2.dim1 m in
				for i = 0 to r-1 do
					let newrow = Gsl_rng.uniform_int rng r in
					Gsl_matrix.swap_rows m i newrow
				done
			in
			apply_matrix shuffle_rows
		| "ShuffleColumns" -> 
			let rng = rng_of_xml elem in
			let shuffle_cols m =
				let c = Bigarray.Array2.dim2 m in
				for i = 0 to c-1 do
					let newcol = Gsl_rng.uniform_int rng c in
					Gsl_matrix.swap_columns m i newcol
				done
			in
			apply_matrix shuffle_cols
		| "ApplyIndexVector" ->
			let to_rows =
				match aod elem "ApplyTo" "Rows" with
				"Rows" -> true
				| _ -> false
			in
			let nf x = 
				let v = mon x.(0) 
				and m = mon x.(1) in
				let (r, c) = Gsl_matrix.dims m 
				and (vr, vc) =  Gsl_matrix.dims v in
				if vr > 1 then
					prerr_endline "Warning: ApplyIndexVector is applied with a matrix instead of a vector, i.e. the first dimension of its first argument is greater than 1. The matrix is flattened before application.";
				let index_vector_list =
					let rec make_iv l = function
						| 0 -> l
						| i -> 
							let h = int_of_float v.{0, i-1} in
							make_iv (h::l) (i - 1)
					in
					make_iv [] vc
				in
				let ivl = List.length index_vector_list in
				let counter = ref 0 in
				if to_rows then
				(
					let out_of_bounds =
						List.exists (fun t -> t < 1 || t > r) index_vector_list
					in
					if out_of_bounds then
						failwith "ApplyIndexVector: out of bounds";
					let target_matrix = Gsl_matrix.create ivl c in
					List.iter
						(
							fun k ->
								for i = 0 to c-1 do
									target_matrix.{!counter, i} <- m.{k-1, i}
								done;
								incr counter
						)
						index_vector_list;
					Mat target_matrix
				)
				else
				(
					let out_of_bounds =
						List.exists (fun t -> t < 1 || t > c) index_vector_list
					in
					if out_of_bounds then
						failwith "ApplyIndexVector: out of bounds";
					let target_matrix = Gsl_matrix.create r ivl in
					List.iter
						(
							fun k ->
								for i = 0 to r-1 do
									target_matrix.{i, !counter} <- m.{i, k-1}
								done;
								incr counter
						)
						index_vector_list;
					Mat target_matrix
				)
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
				]
			}
		| "RowSums" ->
			let nf x =
				let m = mon x.(0) in
				let (r,c) = Gsl_matrix.dims m in
				let sum_vec = Gsl_matrix.create 1 r in
				for i = 0 to r-1 do
					let s = ref 0.0 in
					for j = 0 to c-1 do
						s := !s +. m.{i, j}
					done;
					sum_vec.{0, i} <- !s
				done;
				Mat sum_vec
			in
			function_of_expression
			{
				head = Fun nf; 
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "ColumnSums" ->
			let nf x =
				let m = mon x.(0) in
				let (r,c) = Gsl_matrix.dims m in
				let sum_vec = Gsl_matrix.create 1 c in
				for i = 0 to c-1 do
					let s = ref 0.0 in
					for j = 0 to r-1 do
						s := !s +. m.{j, i}
					done;
					sum_vec.{0, i} <- !s
				done;
				Mat sum_vec
			in
			function_of_expression
			{
				head = Fun nf; 
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "GetRows" ->
		(
			let cld = Xml.children elem in
			if List.exists (fun t -> Xml.tag t = "NumberOfRows") cld then
			(
				let nf x =
					let first = int_of_float (fon x.(0)) in
					let n = int_of_float (fon x.(1)) in
					let m = mon x.(2) in
					let (r, c) = Gsl_matrix.dims m in
					if first < 1 || first + n - 1 > r then
						failwith "GetRows: Invalid submatrix.";
					Mat (get_rows m (first-1) n)
				in
				function_of_expression
				{
					head = Fun nf;
					body = [
						fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
						fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 1)));
						fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2)
					]
				}
			)
			else
			(
				let nf x =
					let first = int_of_float (fon x.(0)) in
					let m = mon x.(1) in
					let (r, c) = Gsl_matrix.dims m in
					if first < 1 || first > r then
						failwith "GetRows: Invalid submatrix.";
					Mat (get_rows m (first-1) 1)
				in
				function_of_expression
				{
					head = Fun nf;
					body = [
						fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
						fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1)
					]
				}
			)
		)
		| "GetColumns" ->
		(
			let cld = Xml.children elem in
			if List.exists (fun t -> Xml.tag t = "NumberOfColumns") cld then
			(
				let nf x =
					let first = int_of_float (fon x.(0)) in
					let n = int_of_float (fon x.(1)) in
					let m = mon x.(2) in
					let (r, c) = Gsl_matrix.dims m in
					if first < 1 || first + n - 1 > c then
						failwith "GetColumns: Invalid submatrix.";
					Mat (get_columns m (first-1) n)
				in
				function_of_expression
				{
					head = Fun nf;
					body = [
						fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
						fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 1)));
						fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2)
					]
				}
			)
			else
			(
				let nf x =
					let first = int_of_float (fon x.(0)) in
					let m = mon x.(1) in
					let (r, c) = Gsl_matrix.dims m in
					if first < 1 || first > c then
						failwith "GetColumns: Invalid submatrix.";
					Mat (get_columns m (first-1) 1)
				in
				function_of_expression
				{
					head = Fun nf;
					body = [
						fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
						fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1)
					]
				}
			)
		)
		| "LinSpace" -> apply2float_1int_to_mat linspace
		| "LogSpace" -> apply2float_1int_to_mat logspace
		| "RestrictedShuffling" ->
			let rng = rng_of_xml elem in
			let forbidden_sequences = List.tl (Xml.children elem) in
			let forb_seq_of_xml x =
				let seqtype = aod x "SequenceType" "Abstract" in
				let seqstringlist =
					let s = aod elem "Sequence" "a, a, a" in
					let spaces_removed = Str.global_replace (Str.regexp " ") "" s in
					Str.split (Str.regexp ",") spaces_removed 
				in
				match seqtype with
				"Abstract" -> Sequences.Abstract seqstringlist
				| _ -> Sequences.Verbatim (List.map float_of_string seqstringlist)
			in
			let forb_seq_list = List.map forb_seq_of_xml forbidden_sequences in
			let find_valid = Sequences.find_valid_seq_gsl rng forb_seq_list in
			apply_matrix find_valid
		| "MatrixOfAllowedSequences" -> 
			let forbidden_sequences = List.tl (Xml.children elem) in
			let forb_seq_of_xml x =
				let seqtype = aod x "SequenceType" "Abstract" in
				let seqstringlist =
					let s = aod elem "Sequence" "a, a, a" in
					let spaces_removed = Str.global_replace (Str.regexp " ") "" s in
					Str.split (Str.regexp ",") spaces_removed 
				in
				match seqtype with
				"Abstract" -> Sequences.Abstract seqstringlist
				| _ -> Sequences.Verbatim (List.map float_of_string seqstringlist)
			in
			let forb_seq_list = List.map forb_seq_of_xml forbidden_sequences in
			apply_matrix (Sequences.allowed_sequences_gsl forb_seq_list)
		| "SequenceVector" ->	(*"colon" in matlab*)
		(
			let cld = Xml.children elem in
			if List.length cld < 3 then
			(
				let nf x = 
					let start = fon x.(0)
					and stop = fon x.(1) in
					let range = stop -. start in
					Mat ( linspace start stop ((int_of_float range) + 1) )
				in
				function_of_expression
				{
					head = Fun nf;
					body = [
						fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
						fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
					]
				}
			)
			else
			(
				let nf x = 
					let start = fon x.(0)
					and stop = fon x.(1)
					and stepwidth = fon x.(2) in
					let d = 
						(floor ((stop -. start) /. stepwidth))
					in
					let newstop = 
						start +. stepwidth *. d
					in
					Mat ( linspace start newstop ((int_of_float d) + 1) )
				in
				function_of_expression
				{
					head = Fun nf;
					body = [
						fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
						fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
						fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2)
					]
				}
			)
		)
		| "Submatrix" -> 
			let nf x =
				let firstrow = int_of_float (fon x.(0)) 
				and nrows = int_of_float (fon x.(1)) 
				and firstcol = int_of_float (fon x.(2))
				and ncols = int_of_float (fon x.(3)) 
				and m = mon x.(4) in
				let (r, c) = Gsl_matrix.dims m in
				if firstrow < 1 || firstrow + nrows - 1 > r || firstcol < 1 || firstcol + ncols - 1 > c then
					failwith "Submatrix: Invalid submatrix.";
				let m1 = get_rows m (firstrow-1) nrows in
				Mat (get_columns m1 (firstcol-1) ncols)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 1)));
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 2)));
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 3)));
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 4)
				]
			}
		| "MapFunction" ->
			let matmapfun_xml = List.hd (Xml.children elem) in
			let argstring = aod matmapfun_xml "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			if List.length argstringlist <> 3 then
				failwith "A function applied to MapFunction must have 3 arguments. The first two arguments represent row i and column j, the third argument stands for the corresponding matrix element (i, j).";
			let matmapfun = function_of_xml ~internal_functions:internal_functions ~assoc_list:assoc_list "" matmapfun_xml in
			let mapfun z =
				let m = mon z.(0) in
				let (r, c) = Gsl_matrix.dims m in
				let x = Array.append [|Scal 0.0; Scal 0.0|] z in
				for i = 0 to r-1 do
					for j = 0 to c-1 do
						x.(0) <- Scal (float_of_int (i+1));
						x.(1) <- Scal (float_of_int (j+1));
						x.(2) <- Scal (Gsl_matrix.get m i j);
						let newval = 
							fon ( matmapfun x ) 
						in
						Gsl_matrix.set m i j newval
					done;
				done;
				Mat m
			in
			function_of_expression
			{
				head = Fun mapfun;
				body = [ fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem))) ]
				(*body = 
				[ 
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem))) ;
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem))) ;
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem))) 
				]*)
			}
		| "DeclaredMatrixFunction" ->
			let ident = Xml.attrib elem "Identifier" in
			let f = 
			(
				if ident = self_id then
				(
					if not is_matrix_function then
						failwith ("DeclaredMatrixFunction was applied with identifier \"" ^ ident ^ "\" in a recursive call, but this is the identifier of a scalar function. Maybe you want to use DeclaredFunction instead.");
					let self_fun = 
						let sf = lazy (!self_pointer) in
						fun x -> mon ((Lazy.force sf) x)
					in
					{
						mat_nargs = self_number_of_arguments;
						mat_fcontent = self_fun
					}
				)
				else
				(
					try
					(
						match List.assoc ident internal_functions with
						Matrix_function t -> t
						| _ -> raise (Invalid_declaration_type ident)
					)
					with
					Not_found ->
					(
						match gd ident with
						Matrix_function t -> t
						| _ -> raise (Invalid_declaration_type ident)
					)
				)
			) in
			let chl = Xml.children elem in
			if (List.length chl) < f.mat_nargs then
				raise (Not_enough_arguments ident)
			else
			(
				let rlist = 
					List.map 
					(
						fun_of_xml internal_functions assoc_list
					)
					chl
				in
				function_of_expression
				{
					head = Fun (fun x -> Mat (f.mat_fcontent x));
					body = rlist
				}
			)
		| "Concatenate" ->
			let chl = Xml.children elem in
			let by_rows =
				match aod elem "Direction" "Rows" with
				"Rows" -> true
				| _ -> false
			in
			let matlist = 
				List.map 
					(fun_of_xml internal_functions assoc_list)
					chl
			in
			if List.length matlist < 2 then
				List.hd matlist
			else
				let concfun z =
					let x = Array.sub z 0 (List.length matlist) in
					let mlist =
						Array.to_list 
						(
							Array.map mon x
						)
					in
					Mat
					(
						List.fold_left
							(concat_matrices by_rows)
							(List.hd mlist)
							(List.tl mlist)
					)
				in
				function_of_expression
				{
					head = Fun concfun;
					body = matlist
				}
		| "MatrixFromGeneratingFunction" ->
		(
			let genfun_xml = List.hd (Xml.children elem) in
			let argstring = aod genfun_xml "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			if List.length argstringlist <> 2 then
				failwith "A function applied to MatrixFromGeneratingFunction must have 2 arguments.";
			let genfun = function_of_xml ~internal_functions:internal_functions ~assoc_list:assoc_list "" genfun_xml in
			let nf x =
				let r = int_of_float (fon x.(0))
				and c = int_of_float (fon x.(1)) in
				let m = Gsl_matrix.create r c in
				for i = 0 to r-1 do
					for j = 0 to c-1 do
						x.(0) <- Scal (float_of_int (i+1));
						x.(1) <- Scal (float_of_int (j+1));
						let newval = fon ( genfun x ) in
						Gsl_matrix.set m i j newval
					done;
				done;
				Mat m
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 1)));
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 2)));
				]
			}
		)	(*MatrixFromGeneratingFunction*)
		| "MatrixFromIteratedFunction" ->
		(
			let by_rows =
				match aod elem "Direction" "Rows" with
				"Rows" -> true
				| _ -> false
			in
			let iterfun_xml = List.hd (Xml.children elem) in
			let argstring = aod iterfun_xml "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			if List.length argstringlist <> 2 then
				failwith "A function applied to MatrixFromIteratedFunction must have 2 arguments. The first argument represents the number of iterations, the second argument stands for the seed matrix.";
			let iterfun = function_of_xml ~internal_functions:internal_functions ~assoc_list:assoc_list "" iterfun_xml in
			let nf x =
				let steps = int_of_float (fon x.(0)) in
				if steps < 1 then
					failwith "Step in MatrixFromIteratedFunction is applied with a value < 1.";
				let m_seed = Gsl_matrix.copy (mon x.(1)) in
				let last_mat =  mon (iterfun x) in
				let rec concat_repeatedly m = function
					| 0 -> m
					| i ->
					(
						x.(0) <- Scal (float_of_int i);
						x.(1) <- Mat (Gsl_matrix.copy m_seed);
						concat_repeatedly ((concat_matrices by_rows) (mon (iterfun x)) m) (i-1)
					)
				in
				Mat (concat_repeatedly last_mat (steps - 1))
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 1)));
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2);
				]
			}
		)	(*MatrixFromIteratedFunction*)
		| "RepeatMatrix" ->
		(
			let by_rows =
				match aod elem "Direction" "Rows" with
				"Rows" -> true
				| _ -> false
			in
			let nf x =
				let steps = int_of_float (fon x.(0)) in
				if steps < 1 then
					failwith "Step in RepeatMatrix is applied with a value < 1.";
				let mt = mon x.(1) in
				(*too slow:*)
				(*let rec concat_repeatedly m = function
					| 0 -> m
					| i ->
						concat_repeatedly ((concat_matrices by_rows) mt m) (i-1)
				in
				Mat (concat_repeatedly mt (steps - 1))*)
				let (source_r, source_c) = Gsl_matrix.dims mt in
				if by_rows then
				(
					let target_c = steps * source_c in
					let m = Gsl_matrix.create source_r target_c in
					for i = 0 to source_r - 1 do
						for j = 0 to target_c - 1 do
							m.{i, j} <- mt.{i, j mod source_c}
						done;
					done;
					Mat m
				)
				else
				(
					let target_r = steps * source_r in
					let m = Gsl_matrix.create target_r source_c in
					for i = 0 to target_r - 1 do
						for j = 0 to source_c - 1 do
							m.{i, j} <- mt.{i mod source_r, j}
						done;
					done;
					Mat m
				)
				
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
				]
			}
		)	(*RepeatMatrix*)
		| "ReshapeMatrix" ->
			let rows =
				match Xml.tag (List.hd (Xml.children elem)) with
				"NumberOfRows" -> true
				| _ -> false
			in
			let nf x =
				let i = int_of_float (fon x.(0))
				and m = mon x.(1) in
				let (r_old, c_old) = Gsl_matrix.dims m in
				let (r_new, c_new) = 
					if rows then
						(i, r_old * c_old / i)
					else
						(r_old * c_old / i, i)
				in
				if r_new < 1 || c_new < 1 || r_old * c_old <> r_new * c_new then
					failwith "ReshapeMatrix: invalid dimensions.";
				let ga = Bigarray.genarray_of_array2 m in
				Mat (Bigarray.reshape_2 ga r_new c_new)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1)
				]
			}
		| "SetMatrixElement" ->
			let nf x =
				let r = int_of_float (fon x.(0))
				and c = int_of_float (fon x.(1)) 
				and matval = fon x.(2) 
				and m = mon x.(3) in
				let (nr, nc) = Gsl_matrix.dims m in
				if r < 1 || c < 1 || r > nr || c > nc then
					failwith "SetElement: out of bounds.";
				(*print_gsl_matrix m;*)
				m.{r-1, c-1} <- matval;
				Mat m
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 1)));
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 3);
				]
			}
		| "ConstantValueMatrix" ->
		(
			let nf x =
				let r = int_of_float (fon x.(0))
				and c = int_of_float (fon x.(1)) 
				and matval = fon x.(2) in
				if r < 1 || c < 1 then
					failwith "ConstantValueMatrix must not have rows or columns < 1.";
				Mat (Gsl_matrix.create ~init:matval r c)
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 1)));
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2);
				]
			}
		)
		| "IdentityMatrix" ->
			let nf x =
				let r = int_of_float (fon x.(0)) in
				if r < 1 then
					failwith "IdentityMatrix must not have rows < 1.";
				let m = Gsl_matrix.create r r in
				Gsl_matrix.set_id m;
				Mat m
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 0)));
				]
			}
		| "RandomMatrix" ->
			let randfun = random_function_of_xml elem in
			let nf x =
				let r = int_of_float (fon x.(0)) 
				and c = int_of_float (fon x.(1)) in
				if r < 1 || c < 1 then
					failwith "RandomMatrix must not have rows or columns < 1.";
				let m = Gsl_matrix.create r c in
				for i = 0 to r-1 do
					for j = 0 to c-1 do
						let newval = randfun () in
						Gsl_matrix.set m i j newval
					done;
				done;
				Mat m
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 1)));
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children (List.nth (Xml.children elem) 2)));
				]
			}
		| "GaborPatch" -> failwith "Gabor patch is not yet implemented..."
		| "DeclaredMatrix" -> 
			let ident = Xml.attrib elem "Identifier" in
			let in_situ = bool_of_string (aod elem "ModifyInSitu" "false") in
			function_of_expression
			{
				head = 
				(
					try
					let arg_number = List.assoc ident assoc_list in
					Arg (in_situ, arg_number)
					with Not_found -> 
					(
						match gd ident with
						Matrix t -> 
							let m = 
								if in_situ then
									t
								else
									Gsl_matrix.copy t
							in
							Const (Mat m)
						| Matrix_function f -> 
							if f.mat_nargs > 0 then
								raise (Invalid_declaration_type ident)
							else
								let nf x = Mat (f.mat_fcontent x) in
								Fun nf
						| _ -> raise (Invalid_declaration_type ident)
					)
				);
				body = []
			}
		| "NumericValueMatrix" ->
			let nrows = int_of_float (float'_of_string (aod elem "Rows" "1"))
			and ncols = int_of_float (float'_of_string (aod elem "Columns" "1")) 
			and delimiter = aod elem "Delimiter" ","
			and row_delimiter = aod elem "RowDelimiter" "," in
			if nrows < 1 || ncols < 1 then
				failwith "NumericValueMatrix must not have rows or columns < 1.";
			let matlist =
				let matstring = 
					let spaces_removed = Str.global_replace (Str.regexp " ") "" (Xml.pcdata (List.hd (Xml.children elem))) in
					let tabs_removed = Str.global_replace (Str.regexp "\t") "" spaces_removed in
					let nl_cut = Str.global_replace (Str.regexp "\n") "" tabs_removed in
					if delimiter = row_delimiter then
						nl_cut
					else
						Str.global_replace (Str.regexp_string row_delimiter) delimiter nl_cut
				in
				Str.split (Str.regexp_string delimiter) matstring
			in
			if nrows * ncols <> List.length matlist then
				failwith "The number of values in NumericValueMatrix does not correspond to the given dimensions.";
			let m = Gsl_matrix.create nrows ncols in
			let counter = ref 0 in
			List.iter 
				(
					fun t ->
						Gsl_matrix.set m (!counter / ncols) (!counter mod ncols) (float'_of_string t); incr counter
				) 
				matlist;
			function_of_expression
			{
				head = Const (Mat m);
				body = []
			}
		| "MatrixFromTextFile" ->
		(
			let replace_in_delimiters d =
				let d1 = Str.global_replace (Str.regexp_string "\\n") "\n" d in
				Str.global_replace (Str.regexp_string "\\t") "\t" d1
			in
			let col_delim = replace_in_delimiters (aod elem "Delimiter" ",") 
			and row_delim = replace_in_delimiters (aod elem "RowDelimiter" ",\\n") in
			let rowstringlist =
				let matrixstring =	(*read file to a string*)
					let f = open_in (aod elem "Filename" "") in
					let l = in_channel_length f in
					let strbuf = String.create 1024 in
					let resbuf = Buffer.create l in
					try
					while true do
						let n = input f strbuf 0 1024 in
						if n = 0 then raise Exit else
						Buffer.add_substring resbuf strbuf 0 n
					done;
					Buffer.contents resbuf
					with Exit -> Buffer.contents resbuf
				in
				let cleaned_matrixstring = 
					if Str.last_chars matrixstring 1 = "\n" then
						Str.first_chars matrixstring ((String.length matrixstring)  - 1)
					else
						matrixstring
				in
				Str.split (Str.regexp_string row_delim) cleaned_matrixstring
			in
			let rowlist =
				List.map (Str.split (Str.regexp_string col_delim)) rowstringlist
			in
			let (nrows, ncols) = (List.length rowlist, List.length (List.hd rowlist)) in
			(*check whether all columns have equal lengths:*)
			if List.exists (fun t -> List.length t <> ncols) rowlist then
				failwith ("Cannot convert file " ^ (aod elem "Filename" "") ^ " to a matrix: Not all columns have equal lengths.");
			let m = Gsl_matrix.create nrows ncols in
			let colcounter = ref 0
			and rowcounter = ref 0 in
			let fill_row =
				List.iter 
					(
						fun t -> 
							m.{!rowcounter, !colcounter} <- float_of_string t; 
							incr colcounter
					)
			in
			List.iter
				(fun t -> fill_row t; colcounter := 0; incr rowcounter)
				rowlist;
			function_of_expression
			{
				head = Const (Mat m);
				body = []
			}
		)
		| "SampleMultibin" -> 
			(*SampleMultibin successes failures nsamp: implementation of Multibin.multibin_samples*)
			let rng = rng_of_xml elem in
			let prec = float'_of_string (aod elem "Precision" "0.000001") in
			let nf x =
				let successes = Array.map int_of_float (Gsl_matrix.to_array (mon x.(0)))
				and failures = Array.map int_of_float (Gsl_matrix.to_array (mon x.(1))) 
				and nsamp = int_of_float (fon x.(2)) in
				if Array.length successes <> Array.length failures then
					failwith "SampleMultibin: the two vectors given as elements have different lengths.";
				if nsamp < 1 then 
					failwith "SampleMultibin: third element (number of samples) must be >= 1.";
				let q = 1. +. prec in
				let indices = Multibin.multibin_samples ~rng:rng ~q:q successes failures nsamp in
				Mat (Gsl_matrix.of_arrays [|(Array.map (fun e -> float_of_int (e+1)) indices)|])
			in
			
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2)
				]
			}
		| "LetInMatrix" -> internal_declarations ()
		| "MatrixCase" -> condition ()
		| "MatrixIfThenElse" -> if_then_else ()
		
		
		| "Argument" ->		(*has the same effect as "DeclaredValue"! (except for MatrixInSitu)*) 
		(
			let key = aod elem "Identifier" "x" in
			let in_situ = bool_of_string (aod elem "MatrixInSitu" "false") in
			try
			let arg_number = List.assoc key assoc_list in
			function_of_expression {head = Arg (in_situ, arg_number); body = []}
			with Not_found -> raise (Non_declared_argument key)
		)
		
		(*return string:*)
		| "String" ->
			let s = (Xml.attrib elem "Value") in
			function_of_expression
			{
				head = Const (Str s);
				body = []
			}
		| "EmptyString" -> 
			function_of_expression
			{
				head = Const (Str "");
				body = []
			}
		| "ConcatenateStrings" ->
			let chl = Xml.children elem in
			let strlist = 
				List.map 
					(fun_of_xml internal_functions assoc_list)
					chl
			in
			let concfun z =
				let x = Array.sub z 0 (List.length strlist) in
				let mlist =
					Array.to_list 
					(
						Array.map son x
					)
				in
				Str ( List.fold_left (^) "" mlist )
			in
			function_of_expression
			{
				head = Fun concfun;
				body = strlist
			}
		| "Substring" ->
			let nf x =
				Str ( String.sub (son x.(0)) (int_of_float (fon x.(1))) (int_of_float (fon x.(2))) )
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2)
				]
			}
		| "StringReplace" ->
			let bos s1 s2 = bool_of_string (aod elem s1 s2) in
			let r rs =
				if bos "RegularExpression" "false" then
				(
					if bos "CaseSensitive" "true" then
						Str.regexp rs
					else
						Str.regexp_case_fold rs
				)
				else
				(
					if bos "CaseSensitive" "true" then
						Str.regexp_string rs
					else
						Str.regexp_string_case_fold rs
				)
			in
			let nf x =
				Str ( Str.global_replace (r (son x.(0))) (son x.(1)) (son x.(2)) )
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2)
				]
			}
		| "NumberToString" ->
			let nf x = 
				let s = string_of_float (fon x.(0)) in
				let sl1 = (String.length s) - 1 in
				let s1 = 
					if s.[sl1] == '.' then String.sub s 0 sl1 else s
				in
				Str s1 
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
				]
			}
		| "CharNumberToString" -> (*ASCII character number to string*)
			let nf x = 
				let chnr = int_of_float (fon x.(0)) in
				if chnr < 0 then
					failwith "CharNumberToString is called with a number less than 0.";
				if chnr > 255 then
					failwith "CharNumberToString supports currently only ASCII characters with numbers from [0, 255] but is here called with a greater number";
				Str (String.make 1 (char_of_int chnr)) 
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
				]
			}
		| "SourceHeadName" -> 
			let el = List.hd (Xml.children elem) in
			function_of_expression
			{
				head = Const (Str (Xml.tag el));
				body = []
			}
		| "MatrixToString" ->
			let nf x = Str (son x.(0)) in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
				]
			}
		| "ListIndex" ->
			let listid = (Xml.attrib elem "Identifier") in
			let nf x = 
				let i = int_of_float (fon x.(0)) in
				
				try
					(
						let list_length =
							match gd ("@LENGTH@ " ^ listid) with
							Constant c -> int_of_float c
							| _ -> failwith "no length"
						in
						if i < 1 || i > list_length then
							failwith ("ListIndex tries to extract indext of item number " ^ string_of_int i ^ " of list " ^ listid ^ ", but it is out of range.");
						match gd ("@LISTINDEX@ " ^ listid ^ " " ^ (string_of_int i)) with
						String_variable v -> Str !v
						| _ -> failwith "no index"
					)
					with Undeclared_identifier _ -> raise (Undeclared_identifier listid)
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 0);
				]
			}
		| "LetInString" -> internal_declarations ()
		| "StringCase" -> condition ()
		| "StringIfThenElse" -> if_then_else ()
		
		(*return xml:*)
		| "Symbol" ->
			let cld = Xml.children elem in
			let n_cld = List.length cld in
			let nf x =
				let symbol_children = 
					let a = Array.sub x 0 n_cld in 
					let unspliced = Array.to_list (Array.map xon a) in
					if List.exists (fun t -> Xml.tag t = "UnquoteSplicing") unspliced then
						let children_list_list =
							List.map
								(
									fun t -> 
										if Xml.tag t = "UnquoteSplicing" then
											Xml.children t
										else
											[t]
								)
								unspliced
						in
						List.flatten children_list_list
					else
						unspliced
				in
				Xml 
				(
					Xml.Element ("Symbol", Xml.attribs elem, symbol_children)
				)
			in
			function_of_expression
			{
				head = Fun nf;
				body = List.map (fun_of_xml internal_functions assoc_list) cld
			}
		| "SymbolAttribute" -> 
			function_of_expression
			{
				head = Const (Xml elem);
				body = []
			}
		| "Unquote" | "ToSource" ->
			let nf z =
				let x = xon z.(0) in
				(*list of PCData, if existing:*)
				let pcdatalist = 
					match Xml.children x with
					h::t -> 
					(
						match h with 
						Xml.PCData pcd -> 
							[Xml.Element 
							(
								"SymbolAttribute",
								[("AttributeTag", "PCData"); ("AttributeValue", pcd)], 
								[]
							)]
						| _ -> []
					)
					| [] -> []
				in
				let symb =
					match x with
					| _ ->
					(
						match Xml.tag x with
						"Symbol" -> x
						| tg -> 
							let node_of_attrib = fun (k, t) ->
								Xml.Element 
								(
									"SymbolAttribute", 
									[("AttributeTag", k); ("AttributeValue", t)],
									[]
								)
							in
							Xml.Element 
							(
								"Symbol", 
								["Identifier", tg], 
								pcdatalist @ (List.map node_of_attrib (Xml.attribs x)) 
							)
					)
				in
				Xml symb
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "UnquoteSplicing" -> 
			let nf z =
				let x = xon z.(0) in
				let symb =
				(
					match Xml.tag x with
					"Symbol" -> x
					| tg -> 
						let node_of_attrib = fun (k, t) ->
							Xml.Element 
							(
								"SymbolAttribute", 
								[("AttributeTag", k); ("AttributeValue", t)],
								[]
							)
						in
						Xml.Element 
						(
							"Symbol", 
							["Identifier", tg], 
							(List.map node_of_attrib (Xml.attribs x)) 
						)
				) in
				Xml (Xml.Element ("UnquoteSplicing", [], Xml.children symb))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceMapOnChildren" ->
			let srcmapfun_xml = List.hd (Xml.children elem) in
			let argstring = aod srcmapfun_xml "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			if List.length argstringlist <> 2 then
				failwith "A function applied to SourceMapOnChildren must have 2 arguments. The first argument represents index i, the second argument stands for the corresponding source list item at position i.";
			let srcmapfun = function_of_xml ~internal_functions:internal_functions ~assoc_list:assoc_list "" srcmapfun_xml in
			let mapfun z =
				let x = xon z.(0) in
				let cld_arr = 
					Array.of_list 
						(List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (Xml.children x)) 
				in
				let a = Array.append [|Scal 0.0; Scal 0.0|] z in
				let ra = 
					Array.mapi
						(
							fun i el ->
								a.(0) <- Scal (float_of_int (i+1));
								a.(1) <- Xml el;
								xon ( srcmapfun a ) 
						)
						cld_arr
				in
				Xml (Xml.Element (Xml.tag x, Xml.attribs x, Array.to_list ra))
			in
			function_of_expression
			{
				head = Fun mapfun;
				body = [ fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem))) ]
			}
		| "SourceList" -> 
			let chl = Xml.children elem in
			let cld_list = 
				List.map 
					(fun_of_xml internal_functions assoc_list)
					chl
			in
			let nf z =
				let x = Array.sub z 0 (List.length cld_list) in
				let l = 
					Array.to_list (Array.map xon x) 
				in
				Xml (Xml.Element ("Symbol", ["Identifier", "SourceList"], l))
			in
			function_of_expression
			{
				head = Fun nf;
				body = cld_list
			}
		| "MatrixToSourceList" -> 
			let nf x =
				let m = mon x.(0) in
				let al = Array.to_list (Gsl_matrix.to_array m) in
				let l =
					List.map 
						(fun t -> 
							Xml.Element 
								("Symbol", 
								[("Identifier", "NumericValue")],
								[
									Xml.Element 
										("SymbolAttribute", 
										[
											"AttributeTag", "Value";
											"AttributeValue", (string_of_float t)
										],
										[])
								])) 
						al
				in
				Xml (Xml.Element ("Symbol", ["Identifier", "SourceList"], l))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceFoldChildren" -> 
			let left =
				match aod elem "ApplicationOrder" "Left" with
				"Left" -> true
				| _ -> false
			in
			(*let cld = Xml.children elem in*)
			let fun_xml = List.hd (Xml.children elem) in
			let argstring = aod fun_xml "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			if List.length argstringlist <> 2 then
				failwith "A function applied to SourceFoldChildren must have 2 arguments.";
			let foldfun = 
				(function_of_xml 
					~internal_functions:internal_functions 
					~assoc_list:assoc_list 
					"" 
					fun_xml
				)
			in
			let nf x =
				let startelem = xon x.(0) in
				let cld_to_fold = 
					(List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (Xml.children (xon x.(1))))
				in
				let ffun p q =
					let a = Array.append [|Xml p; Xml q|] x in
					xon (foldfun a)
				in
				let new_elem = 
					if left then
						List.fold_left ffun startelem cld_to_fold
					else
						List.fold_right ffun cld_to_fold startelem
				in
				Xml new_elem
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2);
				]
			}
		| "SourceConcatenateChildren" ->
			let chl = Xml.children elem in
			let cld_list = 
				List.map 
					(fun_of_xml internal_functions assoc_list)
					chl
			in
			let nf z =
				let x = Array.sub z 0 (List.length cld_list) in
				let l = 
					Array.to_list 
					(
						Array.map (fun t -> (Xml.children (xon t))) x
					) 
				in
				let new_children = 
					List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (List.flatten l)
				in
				Xml (Xml.Element ("Symbol", ["Identifier", "SourceList"], new_children))
			in
			function_of_expression
			{
				head = Fun nf;
				body = cld_list
			}
		| "SourceRevertChildren" ->
			let nf x =
				let e = xon x.(0) in
				let (att, cld) = List.partition (fun t -> Xml.tag t = "SymbolAttribute") (Xml.children e) in
				let new_cld = att @ (List.rev cld) in
				Xml ( Xml.Element (Xml.tag e, Xml.attribs e, new_cld) )
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceSetHead" -> 
			let nf x =
				let headstring = son x.(0)
				and e = xon x.(1) in
				let cld = Xml.children e in
				Xml (Xml.Element (headstring, [], cld))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		| "SourceSetAttribute" ->
			let nf x =
				let e = xon x.(0)
				and attname = son x.(1)
				and attval = son x.(2) in
				let (att, cld) = 
					List.partition (fun t -> Xml.tag t = "SymbolAttribute") (Xml.children e) 
				in
				let new_att = 
					Xml.Element 
					(
						"SymbolAttribute", 
						[("AttributeTag", attname); ("AttributeValue", attval)],
						[]
					)
				in
				let new_attlist = 
					let deleted = 
						List.filter (fun t -> Xml.attrib t "AttributeTag" <> attname) att
					in
					new_att :: deleted
				in
				let new_children = new_attlist @ cld in
				Xml (Xml.Element (Xml.tag e, Xml.attribs e, new_children))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)));
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2);
				]
			}
		| "SourceDeleteAttribute" ->
			let nf x =
				let e = xon x.(0)
				and attname = son x.(1) in
				let new_children = 
					let not_to_delete = function
						| Xml.Element ("SymbolAttribute", name::_, []) ->
							snd name <> attname
						| _ -> true
					in
					List.filter not_to_delete (Xml.children e)
				in
				Xml (Xml.Element (Xml.tag e, Xml.attribs e, new_children))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)));
				]
			}
		| "SourceAttributesToAssocList" ->
			let nf x =
				let e = xon x.(0) in
				let attlist = List.filter (fun t -> Xml.tag t = "SymbolAttribute") (Xml.children e) in
				let pair_of_attr = function
					| Xml.Element (_, atttag::attval::[], []) ->
						let pair = [
							Xml.Element 
							(
								"Symbol", 
								[("Identifier", "String")],
								[
									Xml.Element 
									(
										"SymbolAttribute", 
										[("AttributeTag", "Value"); ("AttributeValue", (snd atttag))], 
										[]
									)
								]
							);
							Xml.Element 
							(
								"Symbol", 
								[("Identifier", "String")],
								[
									Xml.Element 
									(
										"SymbolAttribute", 
										[("AttributeTag", "Value"); ("AttributeValue", (snd attval))], 
										[]
									)
								]
							)
						] in
						Xml.Element ("Symbol", ["Identifier", "SourceList"], pair)
					| _ -> failwith ""
				in
				let pairlist = List.map pair_of_attr attlist in
				Xml ( Xml.Element ("Symbol", ["Identifier", "SourceList"], pairlist) )
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceSetAttributesFromAssocList" ->
			let nf x =
				let e = xon x.(0)
				and assoc_list = Xml.children (xon x.(1)) in
				let cld = List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (Xml.children e) in
				let attr_of_pair = function
					| Xml.Element ("Symbol", ["Identifier", "SourceList"], attname::attval::[]) ->
						let get_string = function
							| Xml.Element ("Symbol", ["Identifier", "String"], s::[]) ->
							(	
								match s with
								Xml.Element ("SymbolAttribute", an::av::[], []) -> snd av
								| _ -> failwith ""
							)
							| k -> failwith ("Function SourceSetAttributesFromAssocList was applied with argument \"" ^ (Xml.to_string_fmt k) ^ "\" that is not a string.")
						in
						let attname_string = get_string attname
						and attval_string = get_string attval in
						Xml.Element 
						(
							"SymbolAttribute", 
							[("AttributeTag", attname_string); ("AttributeValue", attval_string)], 
							[]
						)
					| k -> failwith ("Function SourceSetAttributesFromAssocList was applied with argument \"" ^ (Xml.to_string_fmt k) ^ "\" that is not an association list element.")
				in
				let new_attribs = List.map attr_of_pair assoc_list in
				
				Xml (Xml.Element (Xml.tag e, Xml.attribs e, new_attribs @ cld))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)));
				]
			}
		| "SourceSplitList" ->
			let nf x =
				let assoc_list = Xml.children (xon x.(0)) in
				let rec make_new_list l1 l2 = function
					| [] -> 
						let wrap_to_source_list l =
							Xml.Element ("Symbol", ["Identifier", "SourceList"], l)
						in
						wrap_to_source_list
						[
							wrap_to_source_list (List.rev l1);
							wrap_to_source_list (List.rev l2)
						]
					| (Xml.Element ("Symbol", ["Identifier", "SourceList"], h1::h2::[]))::t -> 
						make_new_list (h1::l1) (h2::l2) t
					| k::t -> failwith ("Function SourceSplitList was applied with argument \"" ^ (Xml.to_string_fmt k) ^ "\" that is not a pair list element.")
				in
				
				Xml ( make_new_list [] [] assoc_list )
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceCombineLists" ->
			let nf x =
				let l1 = Xml.children (xon x.(0))
				and l2 = Xml.children (xon x.(1)) in
				if List.length l1 <> List.length l2 then
					failwith "Function SourceCombineLists is called with two lists of different lengths.";
				let wrap_to_source_list l =
					Xml.Element ("Symbol", ["Identifier", "SourceList"], l)
				in
				let rec combine_lists comblist = function
					| (h1::t1, h2::t2) ->
						combine_lists ((wrap_to_source_list [h1; h2])::comblist) (t1, t2)
					| _ -> wrap_to_source_list (List.rev comblist)
				in
				Xml (combine_lists [] (l1, l2))
			in
			
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)));
				]
			}
		| "SourceGetAssociation" ->	(*List.assoc*)
			let nf x =
				let assoc_list = Xml.children (xon x.(0)) in
				let key = xon x.(1) in
				let rec find_assoc = function
					| [] -> failwith "SourceGetAssociation: No association found."
					| (Xml.Element ("Symbol", ["Identifier", "SourceList"], h1::h2::[]))::t -> 
						if source_equal h1 key then
							h2
						else
							find_assoc t
					| k::t -> failwith ("Function SourceGetAssociation was applied with argument \"" ^ (Xml.to_string_fmt k) ^ "\" that is not a pair list element.")
				in
				
				Xml ( find_assoc assoc_list )
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)));
				]
			}
		| "SourceDeleteAssociation" ->	(*List.assoc*)
			let nf x =
				let assoc_list = Xml.children (xon x.(0)) in
				let key = xon x.(1) in
				let new_list =
					List.filter 
						(fun t -> not (source_equal (List.hd (Xml.children t)) key))
						assoc_list
				in
				(*let rec find_assoc = function
					| [] -> assoc_list
					| (Xml.Element ("Symbol", ["Identifier", "SourceList"], h1::h2::[]))::t -> 
						if source_equal h1 key then
							h2
						else
							find_assoc t
					| k::t -> failwith ("Function SourceGetAssociation was applied with argument \"" ^ (Xml.to_string_fmt k) ^ "\" that is not a pair list element.")
				in*)
				
				Xml ( Xml.Element ("Symbol", ["Identifier", "SourceList"], new_list) )
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)));
				]
			}
		| "SourceApply" ->
			let nf x =
				let head = xon x.(0)	
				and cld = 
					let c = Xml.children (xon x.(1)) in
					(List.filter (fun t -> Xml.tag t <> "SymbolAttribute") c)
				in
				let new_attr = 
					List.filter (fun t -> Xml.tag t = "SymbolAttribute") (Xml.children head) 
				in
				Xml (Xml.Element (Xml.tag head, Xml.attribs head, new_attr @ cld))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		| "SourceHead" -> 
			let nf x =
				let e = xon x.(0) in
				let new_cld = 
					List.filter (fun t -> Xml.tag t = "SymbolAttribute") (Xml.children e)
				in
				Xml (Xml.Element (Xml.tag e, Xml.attribs e, new_cld))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceChildren" -> 
			let nf x =
				let e = xon x.(0) in
				let cld = List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (Xml.children e) in
				Xml (Xml.Element ("Symbol", ["Identifier", "SourceList"], cld))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceFirstChild" ->
			let nf x =
				let e = xon x.(0) in
				let cld = List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (Xml.children e) in
				if cld = [] then
					failwith ("SourceFirstChild is applied to element " ^ (Xml.tag e) ^ " which has no children.");
				Xml (List.hd (Xml.children e))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceLastChild" ->
			let nf x =
				let e = xon x.(0) in
				let cld = List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (Xml.children e) in
				if cld = [] then
					failwith ("SourceLastChild is applied to element " ^ (Xml.tag e) ^ " which has no children.");
				let cld_list = Xml.children e in
				Xml (List.nth cld_list ((List.length cld_list) - 1))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceChildrenTail" ->	(*all children except first one*) 
			let nf x =
				let e = xon x.(0) in
				let cld = List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (Xml.children e) in
				if cld = [] then
					failwith ("SourceChildrenTail is applied to element " ^ (Xml.tag e) ^ " which has no children.");
				Xml (Xml.Element ("Symbol", ["Identifier", "SourceList"], List.tl (Xml.children e)))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceNthChild" ->
			let nf x =
				let e = xon x.(0) in
				let n = int_of_float (fon x.(1)) in
				let cld = List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (Xml.children e) in
				if n < 1 || n > List.length cld then
					failwith ("SourceNthChild is applied to child number " ^ (string_of_int n) ^ " of " ^ (Xml.tag e) ^ ", but this is out of range.");
				Xml (List.nth (Xml.children e) (n-1))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		| "SourceFlattenChildren" ->	(*flatten one level*) 
			let nf x =
				let e = xon x.(0) in
				let (att, cld) = List.partition (fun t -> Xml.tag t = "SymbolAttribute") (Xml.children e) in
				let cldcld = 
					List.map 
						(fun t -> 
							let cl = Xml.children t in 
							if cl = [] then [t] else cl)
						cld
				in
				let new_children = 
					List.filter (fun t -> Xml.tag t <> "SymbolAttribute") (List.flatten cldcld)
				in
				Xml (Xml.Element (Xml.tag e, Xml.attribs e, att @ new_children))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
		| "SourceRepeat" ->
			let nf x =
				let e = xon x.(0)
				and number_of_repetitions = int_of_float (fon x.(1)) in
				let rec repeat el xml_list = function
					| 0 -> xml_list
					| n -> repeat el (el::xml_list) (n-1)
				in
				let xlist = repeat e [] number_of_repetitions in
				Xml (Xml.Element ("Symbol", ["Identifier", "SourceList"], xlist))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [
					fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem));
					fun_of_xml internal_functions assoc_list (List.hd (List.tl (Xml.children elem)))
				]
			}
		| "SourceListFromGeneratingFunction" -> 
			let genfun_xml = List.hd (Xml.children elem) in
			let argstring = aod genfun_xml "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			if List.length argstringlist <> 1 then
				failwith "A function applied to SourceListFromGeneratingFunction must have exactly one argument.";
			let genfun = function_of_xml ~internal_functions:internal_functions ~assoc_list:assoc_list "" genfun_xml in
			let nf x =
				let n = int_of_float (fon x.(0)) in
				let cld_list =
					Array.to_list
					(
						Array.init 
						n
						(fun i -> x.(0) <- Scal (float_of_int (i+1)); xon (genfun x))
					)
				in
				Xml (Xml.Element ("Symbol", ["Identifier", "SourceList"], cld_list))
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
				]
			}
		| "SourceListFromIteratedFunction" ->
			let iterfun_xml = List.hd (Xml.children elem) in
			let argstring = aod iterfun_xml "ArgumentList" "" in
			let argstringlist = argstringlist_of_string argstring in
			if List.length argstringlist <> 1 then
				failwith "A function applied to SourceListFromIteratedFunction must have exactly one argument. It represents the seed source element.";
			let iterfun = function_of_xml ~internal_functions:internal_functions ~assoc_list:assoc_list "" iterfun_xml in
			let nf x =
				let steps = int_of_float (fon x.(0)) in
				if steps < 1 then
					failwith "Step in SourceListFromIteratedFunction is applied with a value < 1.";
				let source_seed = xon x.(1) in
				(*let last_source =  xon (iterfun x) in*)
				let rec concat_repeatedly = function
					| (l, 0) -> l
					| ([], _) -> []
					| (h::t, i) ->
					(
						(*x.(0) <- Scal (float_of_int i);*)
						x.(0) <- Xml h;
						let next_element = xon (iterfun x) in
						concat_repeatedly ((next_element::h::t), (i-1))
					)
				in
				let new_children =  
					List.rev
						(concat_repeatedly ([source_seed], (steps-1))) 
				in
				Xml (Xml.Element ("Symbol", ["Identifier", "SourceList"], new_children))
			in
			function_of_expression
			{
				head = Fun nf;
				body = 
				[
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 1);
					fun_of_xml internal_functions assoc_list (List.nth (Xml.children elem) 2);
				]
			}
		| "EmptyExperiment" ->
			let ex =
				Xml.Element
				(
					"Experiment",
					[],
					[
						Xml.Element ("SimpleDisplayMode", [], []);
						Xml.Element ("Scene", ["SceneID", "blank"], []);
						Xml.Element 
						(
							"Sequence", 
							[], 
							[Xml.Element ("Presentation", [("Duration", "1"); ("SceneID", "blank")], [])]
						)
					]
				)
			in
			let rec quote_xml = function
				| Xml.Element (tg, attlist, chld) ->
					let node_of_attrib = fun (k, t) ->
						Xml.Element 
						(
							"SymbolAttribute", 
							[("AttributeTag", k); ("AttributeValue", t)],
							[]
						)
					in
					Xml.Element 
					(
						"Symbol", 
						["Identifier", tg], 
						(List.map node_of_attrib attlist) 
							@ (List.map quote_xml chld)
					)
				| _ -> failwith ""
			in
			let exq = quote_xml ex in
			function_of_expression
			{
				head = Const (Xml exq);
				body = []
			}
		| "BindToExperiment" ->
			let chl = Xml.children elem in
			let cld_list = 
				List.map 
					(fun_of_xml internal_functions assoc_list)
					chl
			in
			let repl = bool_of_string (aod elem "Replace" "false") in
			let nf z =
				let x = Array.sub z 0 (List.length cld_list) in
				let initial_experiment = xon x.(0)
				and list_of_bindings =
					Array.to_list 
					(
						Array.map
							xon
							(Array.sub x 1 ((Array.length x) - 1))
					)
				in

				let rec bind_to_ex ex = function
					| [] -> ex
					| h::t ->
						let ident x = aod x "Identifier" "" in
						let ih = ident h in
						let exp_taglist = List.rev
						[
							"DocumentInformation";
							"AttributeMacros";
							"GlobalDeclarations";
							"GlobalDisplayObjects";
							"SimpleDisplayMode";
							"EnhancedDisplayMode";
							"LogEvents"
						] in
						let cld = Xml.children ex in
						(*replace element with tag tag_to_match in experiment by new_elem:*)
						let exp_repl_elem tag_to_match new_elem =
							List.map
								(fun t -> 
									if ident t = tag_to_match then
										new_elem
									else
										t)
								cld
						in
						(*append elem after all elements with the same tag:*)
						let exp_add_elem tag_to_match new_elem =
							let rec make_unique_list targetlist = function
								| [] -> targetlist
								| lh::lt -> 
									if (List.length targetlist)>0 && ident lh = ident (List.hd (List.hd targetlist)) then
										make_unique_list (((List.hd targetlist) @ [lh])::(List.tl targetlist)) lt
									else
										make_unique_list ([lh]::targetlist) lt
							in
							let ll_new = List.map
								(fun t -> 
									if ident (List.hd t) = tag_to_match then
										t @ [new_elem]
									else t)
								(make_unique_list [] cld)
							in
							List.flatten ll_new
						in
						let new_cld =
						(
							match ih with
							"SimpleDisplayMode" | "EnhancedDisplayMode" ->
							(
								if List.exists (fun t -> ident t = "SimpleDisplayMode") cld then
									exp_repl_elem "SimpleDisplayMode" h
								else
									exp_repl_elem "EnhancedDisplayMode" h
							)
							| "DocumentInformation" | "AttributeMacros" | "GlobalDeclarations" | "GlobalDisplayObjects" | "LogEvents" -> 
								if (List.exists (fun t -> ident t = ih) cld) then
									exp_repl_elem (ih) h
								else
									let list_before_tag tg =
										let rec make_list = function
											| [] -> failwith ""
											| lh::lt -> 
												if lh = tg then
													lt
												else
													make_list lt
										in
										make_list exp_taglist
									in
									let rec find_and_add = function
										| [] -> h::cld
										| lh::lt -> 
											if (List.exists (fun t -> ident t = lh) cld) then
												exp_add_elem lh h
											else find_and_add lt
									in
									find_and_add (list_before_tag ih)
							| "Scene" | "Sequence" -> 
								if repl && (List.exists (fun t -> ident t = ih) cld) then
									exp_repl_elem ih h
								else exp_add_elem ih h
							| k -> failwith ("BindToExperiment: Argument \"" ^ k ^ "\" cannot be bound to an Experiment.")
						) in
						let new_ex = Xml.Element (Xml.tag ex, Xml.attribs ex, new_cld) in
						bind_to_ex new_ex t
				in
				Xml (bind_to_ex initial_experiment list_of_bindings)
			in
			function_of_expression
			{
				head = Fun nf;
				body = cld_list
			}
		| "SourceIfThenElse" -> if_then_else ()
		| "SourceCase" -> condition ()
		
		| "ProbabilityFunction" -> (*ProbabiltyFunction*)
		(
			let pdf_or_cdf = (List.hd (Xml.children elem)) in
			let cont_type = (List.hd (Xml.children pdf_or_cdf)) in
			let pfun =
			(
				let ios s = int_of_float (float'_of_string s) in
				(*let iof = float_of_int in*)
				match (Xml.tag cont_type) with
				"GaussianPDF" ->
					let sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
					fun x -> Gsl_randist.gaussian_pdf x sigma
				| "GaussianTailPDF" ->
					let sigma = float'_of_string (aod cont_type "Sigma" "1.0")
					and a = float'_of_string (aod cont_type "A" "1.0") in
					fun x -> Gsl_randist.gaussian_tail_pdf x a sigma
				| "ExponentialPDF" -> 
					let mu = float'_of_string (aod cont_type "Mu" "1.0") in
					fun x -> Gsl_randist.exponential_pdf x mu
				| "LaplacePDF" -> 
					let a = float'_of_string (aod cont_type "A" "1.0") in
					fun x -> Gsl_randist.laplace_pdf x a
				| "ExponentialPowerPDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_randist.exppow_pdf x a b
				| "CauchyPDF" -> 
					let a = float'_of_string (aod cont_type "A" "1.0") in
					fun x -> Gsl_randist.cauchy_pdf x a
				| "RayleighPDF" -> 
					let sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
					fun x -> Gsl_randist.rayleigh_pdf x sigma
				| "LandauPDF" ->
					fun x -> Gsl_randist.landau_pdf x
				| "GammaPDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_randist.gamma_pdf x a b
				| "LognormalPDF" ->
					let zeta = float'_of_string (aod cont_type "Zeta" "1.0")
					and sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
					fun x -> Gsl_randist.lognormal_pdf x zeta sigma
				| "ChiSquaredPDF" -> 
					let nu = float'_of_string (aod cont_type "Nu" "1.0") in
					fun x -> Gsl_randist.chisq_pdf x nu
				| "FDistPDF" ->
					let nu1 = float'_of_string (aod cont_type "Nu1" "1.0")
					and nu2 = float'_of_string (aod cont_type "Nu2" "1.0") in
					fun x -> Gsl_randist.fdist_pdf x nu1 nu2
				| "TDistPDF" -> 
					let nu = float'_of_string (aod cont_type "Nu" "1.0") in
					fun x -> Gsl_randist.tdist_pdf x nu
				| "BetaPDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_randist.beta_pdf x a b
				| "LogisticPDF" -> 
					let a = float'_of_string (aod cont_type "A" "1.0") in
					fun x -> Gsl_randist.logistic_pdf x a
				| "ParetoPDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_randist.pareto_pdf x a b
				| "WeibullPDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_randist.weibull_pdf x a b
				| "Gumbel1PDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_randist.gumbel1_pdf x a b
				| "Gumbel2PDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_randist.gumbel2_pdf x a b
				(*discrete:*)
				| "PoissonPDF" -> 
					let mu = float'_of_string (aod cont_type "Mu" "1.0") in
					fun x -> Gsl_randist.poisson_pdf (int_of_float x) mu
				| "BernoulliPDF" ->
					let p = float'_of_string (aod cont_type "P" "1.0") in
					fun x -> Gsl_randist.bernoulli_pdf (int_of_float x) p
				| "BinomialPDF" ->
					let p = float'_of_string (aod cont_type "P" "1.0") in
					let n = ios (aod cont_type "N" "1") in
					fun x -> Gsl_randist.binomial_pdf (int_of_float x) p n
				| "NegativeBinomialPDF" ->
					let p = float'_of_string (aod cont_type "P" "1.0") in
					let n = float'_of_string (aod cont_type "N" "1.0") in
					(*uncomment this for ocamlgsl > 0.5 ::::let n = int_of_float n in::::*)
					fun x -> Gsl_randist.negative_binomial_pdf (int_of_float x) p n
				| "PascalPDF" ->
					let p = float'_of_string (aod cont_type "P" "1.0") in
					let k = ios (aod cont_type "K" "1") in
					fun x -> Gsl_randist.pascal_pdf (int_of_float x) p k
				| "GeometricPDF" ->
					let p = float'_of_string (aod cont_type "P" "1.0") in
					fun x -> Gsl_randist.geometric_pdf (int_of_float x) p
				| "HyperGeometricPDF" ->
					let n1 = ios (aod cont_type "N1" "1") in
					let n2 = ios (aod cont_type "N2" "1") in
					let t = ios (aod cont_type "T" "1") in
					fun x -> Gsl_randist.hypergeometric_pdf (int_of_float x) ~n1:n1 ~n2:n2 ~t:t
				| "LogarithmicPDF" ->
					let p = float'_of_string (aod cont_type "P" "1.0") in
					fun x -> Gsl_randist.logarithmic_pdf (int_of_float x) p
				
				(*** CDFs: ***)
				| "GaussianCDF" ->
					let sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
					fun x -> Gsl_cdf.gaussian_P x sigma
				| "ExponentialCDF" -> 
					let mu = float'_of_string (aod cont_type "Mu" "1.0") in
					fun x -> Gsl_cdf.exponential_P x mu
				| "LaplaceCDF" -> 
					let a = float'_of_string (aod cont_type "A" "1.0") in
					fun x -> Gsl_cdf.laplace_P x a
				| "CauchyCDF" -> 
					let a = float'_of_string (aod cont_type "A" "1.0") in
					fun x -> Gsl_cdf.cauchy_P x a
				| "RayleighCDF" -> 
					let sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
					fun x -> Gsl_cdf.rayleigh_P x sigma
				| "GammaCDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_cdf.gamma_P x a b
				| "LognormalCDF" ->
					let zeta = float'_of_string (aod cont_type "Zeta" "1.0")
					and sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
					fun x -> Gsl_cdf.lognormal_P x zeta sigma
				| "ChiSquaredCDF" -> 
					let nu = float'_of_string (aod cont_type "Nu" "1.0") in
					fun x -> Gsl_cdf.chisq_P x nu
				| "FDistCDF" ->
					let nu1 = float'_of_string (aod cont_type "Nu1" "1.0")
					and nu2 = float'_of_string (aod cont_type "Nu2" "1.0") in
					fun x -> Gsl_cdf.fdist_P x nu1 nu2
				| "TDistCDF" -> 
					let nu = float'_of_string (aod cont_type "Nu" "1.0") in
					fun x -> Gsl_cdf.tdist_P x nu
				| "BetaCDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_cdf.beta_P x a b
				| "LogisticCDF" -> 
					let a = float'_of_string (aod cont_type "A" "1.0") in
					fun x -> Gsl_cdf.logistic_P x a
				| "ParetoCDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_cdf.pareto_P x a b
				| "WeibullCDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_cdf.weibull_P x a b
				| "Gumbel1CDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_cdf.gumbel1_P x a b
				| "Gumbel2CDF" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_cdf.gumbel2_P x a b
				
				(*** inverse CDFs: ***)
				| "GaussianCDF_inv" ->
					let sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
					fun x -> Gsl_cdf.gaussian_Pinv x sigma
				| "ExponentialCDF_inv" -> 
					let mu = float'_of_string (aod cont_type "Mu" "1.0") in
					fun x -> Gsl_cdf.exponential_Pinv x mu
				| "LaplaceCDF_inv" -> 
					let a = float'_of_string (aod cont_type "A" "1.0") in
					fun x -> Gsl_cdf.laplace_Pinv x a
				| "CauchyCDF_inv" -> 
					let a = float'_of_string (aod cont_type "A" "1.0") in
					fun x -> Gsl_cdf.cauchy_Pinv x a
				| "RayleighCDF_inv" -> 
					let sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
					fun x -> Gsl_cdf.rayleigh_Pinv x sigma
				| "GammaCDF_inv" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_cdf.gamma_Pinv x a b
				| "LognormalCDF_inv" ->
					let zeta = float'_of_string (aod cont_type "Zeta" "1.0")
					and sigma = float'_of_string (aod cont_type "Sigma" "1.0") in
					fun x -> Gsl_cdf.lognormal_Pinv x zeta sigma
				| "ChiSquaredCDF_inv" -> 
					let nu = float'_of_string (aod cont_type "Nu" "1.0") in
					fun x -> Gsl_cdf.chisq_Pinv x nu
				| "TDistCDF_inv" -> 
					let nu = float'_of_string (aod cont_type "Nu" "1.0") in
					fun x -> Gsl_cdf.tdist_Pinv x nu
				| "LogisticCDF_inv" -> 
					let a = float'_of_string (aod cont_type "A" "1.0") in
					fun x -> Gsl_cdf.logistic_Pinv x a
				| "ParetoCDF_inv" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_cdf.pareto_Pinv x a b
				| "WeibullCDF_inv" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_cdf.weibull_Pinv x a b
				| "Gumbel1CDF_inv" ->
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_cdf.gumbel1_Pinv x a b
				| _ ->	(*"Gumbel2CDF_inv"*)
					let a = float'_of_string (aod cont_type "A" "1.0")
					and b = float'_of_string (aod cont_type "B" "1.0") in
					fun x -> Gsl_cdf.gumbel2_Pinv x a b
			) in
			let nf x =
				Scal (pfun (fon x.(0)))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children cont_type))]
				(*body = [ocaml_exp_of_xml internal_declarations assoc_list (List.hd (Xml.children cont_type))]*)
			}
		)
		
		| "Declaration" -> (*for global declarations; just returned; LAZY!!!*)
			function_of_expression
			{
				head = Const (Xml elem);
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}
			(*does not work, must be lazy:*)
			(*let nf x =
				Xml (xon x.(0))
			in
			function_of_expression
			{
				head = Fun nf;
				body = [fun_of_xml internal_functions assoc_list (List.hd (Xml.children elem))]
			}*)
		
		(*to be evaluated graphically:*)
		| "Experiment" | "Scene" | "Stimulus" | "Image" | "Button" -> 
			let x =
				Xml (Xml.Element ("EvaluateGraphically", [], [elem]))
			in
			function_of_expression
			{
				head = Const x;
				body = []
			}
		
		| s -> failwith ("Function " ^ s ^ " is not defined.")
		
	in	(*fun_of_xml*)
	
	let f = fun_of_xml internal_functions assoc_list xml_fun in
	self_pointer := f;
	f
;;

(** [declaration_of_xml decl]: convert xml declaration [decl] to ocaml declaration *)
let declaration_of_xml decl =
	let declaration_content = (List.hd (Xml.children decl)) in
	let argstring = aod declaration_content "ArgumentList" "" in
	let argstringlist = argstringlist_of_string argstring
	and ident = Xml.attrib decl "Identifier" in
	(
		ident,
		match Xml.tag declaration_content with
		"Constant" -> Constant
			(
				float_of_nom ((function_of_xml "" declaration_content) [||])
			)
		| "Variable" -> Variable 
			(
				let cld = Xml.children declaration_content in
				if cld = [] then
					ref 0.0
				else
					ref
					(
						float_of_nom ((function_of_xml "" (List.hd cld)) [||])
					)
			)
		| "Function" ->
			let decl_fun = function_of_xml ident declaration_content in
			Function 
			{
				nargs = (List.length argstringlist);
				fcontent = fun a -> float_of_nom (decl_fun a)
			}
		| "Matrix" -> Matrix
			(
				matrix_of_nom ((function_of_xml "" declaration_content) [||])
			)
		| "StringVariable" -> String_variable 
			(
				let cld = Xml.children declaration_content in
				if cld = [] then
					ref ""
				else
					ref
					(
						string_of_nom ((function_of_xml "" (List.hd cld)) [||])
					)
			)
		| "Source" -> Xml_variable
			(
				ref (xml_of_nom ((function_of_xml "" declaration_content) [||]))
			)
		| "StringFunction" ->
			let decl_fun = function_of_xml ident declaration_content in
			String_function 
			{
				str_nargs = (List.length argstringlist);
				str_fcontent = fun a -> string_of_nom (decl_fun a)
			}
		| "SourceFunction" ->
			let decl_fun = function_of_xml ident declaration_content in
			Xml_function 
			{
				xml_nargs = (List.length argstringlist);
				xml_fcontent = fun a -> xml_of_nom (decl_fun a)
			}
		| _ -> (*"MatrixFunction"*)
			let decl_fun = function_of_xml ident declaration_content in
			Matrix_function
			{
				mat_nargs = (List.length argstringlist);
				mat_fcontent = fun a -> matrix_of_nom (decl_fun a)
			}
	)
;;

(** [generate_declarations elem]: extract declarations of xml element [elem] 
and add them to the global declaration hash table *)
let generate_declarations elem =
	let xml_decl = 
		List.filter (fun t -> (Xml.tag t) = "Declaration") (Xml.children elem) 
	in
	List.iter
		(
			fun t ->
				let decl = declaration_of_xml t in
				Hashtbl.add global_declarations (fst decl) (snd decl)
		)
		xml_decl
;;

(*TESTS:
let print_gsl_matrix m =
	let (r, c) = Gsl_matrix.dims m in
	for i=0 to r-1 do
		print_newline ();
		for j=0 to c-1 do
			Printf.printf "%f " m.{i, j}
		done;
	done;
	print_newline ()
;;

let ocaml_sinepatch npoints width orientation freq =
	let transpose mat =
		let (r, c) = Gsl_matrix.dims mat in
		let mt =  Gsl_matrix.create c r in
		Gsl_matrix.transpose mt mat;
		mt
	in
	let linspace start stop len =
		if len <= 0 then
			failwith "linspace: invalid length. Check all calls of the functions SequenceVector, LinSpace, and LogSpace.";
		let v = Gsl_matrix.create 1 len in
		let d = 
			(stop -. start) /. (float_of_int (len - 1)) 
		in
		for i=0 to len-1 do
			v.{0, i} <- start +. (float_of_int i) *. d
		done;
		v
	in
	let repeat by_rows steps mt =
		if steps < 1 then
			failwith "Step in RepeatMatrix is applied with a value < 1.";
		let (source_r, source_c) = Gsl_matrix.dims mt in
		if by_rows then
		(
			let target_c = steps * source_c in
			let m = Gsl_matrix.create source_r target_c in
			for i = 0 to source_r - 1 do
				for j = 0 to target_c - 1 do
					m.{i, j} <- mt.{i, j mod source_c}
				done;
			done;
			m
		)
		else
		(
			let target_r = steps * source_r in
			let m = Gsl_matrix.create target_r source_c in
			for i = 0 to target_r - 1 do
				for j = 0 to source_c - 1 do
					m.{i, j} <- mt.{i mod source_r, j}
				done;
			done;
			m
		)
	in
	let concat_matrices by_rows m1 m2 =
		let (dm1x, dm1y) = Gsl_matrix.dims m1
		and (dm2x, dm2y) = Gsl_matrix.dims m2 in
		if not by_rows then
		(
			if dm1y <> dm2y then
				failwith "Matrix concatanation failed: wrong dimensions.";
			let m = Gsl_matrix.create (dm1x + dm2x) dm1y in
			for j = 0 to dm1y - 1 do
				for i = 0 to dm1x - 1 do
					m.{i, j} <- m1.{i, j}
				done;
				for i = dm1x to dm1x + dm2x - 1 do
					m.{i, j} <- m2.{i - dm1x, j}
				done
			done;
			m
		)
		else
		(
			if dm1x <> dm2x then
				failwith "Matrix concatanation failed: wrong dimensions.";
			let m = Gsl_matrix.create dm1x (dm1y + dm2y) in
			for i = 0 to dm1x - 1 do
				for j = 0 to dm1y - 1 do
					m.{i, j} <- m1.{i, j}
				done;
				for j = dm1y to dm1y + dm2y - 1 do
					m.{i, j} <- m2.{i, j - dm1y}
				done
			done;
			m
		)
	in
	let reshape_matrix rows i m =
		let (r_old, c_old) = Gsl_matrix.dims m in
		let (r_new, c_new) = 
			if rows then
				(i, r_old * c_old / i)
			else
				(r_old * c_old / i, i)
		in
		if r_new < 1 || c_new < 1 || r_old * c_old <> r_new * c_new then
			failwith "ReshapeMatrix: invalid dimensions.";
		let ga = Bigarray.genarray_of_array2 m in
		(Bigarray.reshape_2 ga r_new c_new)
	in
	let rowsums m =
		let (r,c) = Gsl_matrix.dims m in
		let sum_vec = Gsl_matrix.create 1 r in
		for i = 0 to r-1 do
			let s = ref 0.0 in
			for j = 0 to c-1 do
				s := !s +. m.{i, j}
			done;
			sum_vec.{0, i} <- !s
		done;
		sum_vec
	in
	let mapfun matmapfun m =
		let (r, c) = Gsl_matrix.dims m in
		for i = 0 to r-1 do
			for j = 0 to c-1 do
				let newval = matmapfun m.{i, j} in
				Gsl_matrix.set m i j newval
			done;
		done;
		m
	in
	let matrix_mult m1 m2 = 
		let (d1, c1) = Gsl_matrix.dims m1
		and (c2, d2) = Gsl_matrix.dims m2 in
		if c1 <> c2 then
			failwith "Error in matrix multiplication: invalid matrix dimensions."
		else
		(
			let p = Gsl_matrix.create d1 d2 in
			Gsl_linalg.matmult ~a:(`M m1) ~b:(`M m2) (`M p);
			p
		)
	in
	let get_rows =
		Bigarray.Array2.sub_left
	in
	let get_columns m first n =
		let newmat = Bigarray.Array2.sub_left (transpose m) first n in
		transpose newmat
	in
	let alpha = orientation *. Gsl_math.pi /. 180. in
	let cotrans = Gsl_matrix.create ~init:(cos alpha) 2 2 in
	cotrans.{0,1} <- sin alpha;
	cotrans.{1,0} <- -. (sin alpha);
	let x = transpose (linspace (-. Gsl_math.pi) Gsl_math.pi npoints) in
	let coords = 
		let col1 = repeat false npoints x
		and col2 = 
			reshape_matrix false 1 (repeat true npoints x)
		in
		concat_matrices true col1 col2
	in
	let distmat =
		let coords1 = Gsl_matrix.copy coords in
		let d1 = 
			reshape_matrix 
				true 
				npoints 
				(rowsums (Gsl_matrix.mul_elements coords1 coords1; coords1))
		in
		mapfun sqrt d1
	in
	let ncoords = matrix_mult coords cotrans in
	let sinepat =
		let s1 =
			get_columns 
				(Gsl_matrix.scale ncoords freq; ncoords)
				0
				1
		in
		let s2 = mapfun sin s1 in
		reshape_matrix true npoints s2
	in
	let hull =
		let f x = 
			if x <= width *. Gsl_math.pi then
				0.5 *. (1. +. (cos (x /. width)))
			else
				0.0
		in
		mapfun f distmat
	in
	Gsl_matrix.mul_elements hull sinepat;
	transpose hull
;;
print_gsl_matrix (ocaml_sinepatch 10 0.8 45. 15.);;

let t1 = Unix.gettimeofday ();;
for i = 1 to 100 do
	ocaml_sinepatch 2500 0.4 90. 15.
done;;
let t2 = Unix.gettimeofday ();;
Printf.printf "patch: %f\n" (t2 -. t1);;



let x = xml_of_file "mathtest_timm.xml";;
generate_declarations x;;

let testfun = 
	let dcl = get_declaration "sinepatch" in
	match dcl with
	Matrix_function mf -> mf.mat_fcontent
	| _ -> failwith ""
;;

let sinepatch npoints width orientation freq =
	Gsl_matrix.to_arrays 
	(
		testfun [|Scal npoints; Scal width; Scal orientation; Scal freq|]
	)
;;
let t1 = Unix.gettimeofday ();;
for i = 1 to 100 do
	sinepatch 256. 0.4 90. 15.
done;;
let t2 = Unix.gettimeofday ();;
Printf.printf "patch: %f\n" (t2 -. t1);;
*)
(*let ft x = Gsl_matrix.to_arrays (testfun [|Scal x|]);;
let f1 x  = float_of_nom (testfun [|Scal x|]);;
let rec f2 = function
	| 0. -> 1.
	| 1. -> 1.
	| n -> (f2 (n-.1.)) +. (f2 (n-.2.))
;;*)

(*
let t1 = Unix.gettimeofday ();;
f1 30.;;
let t2 = Unix.gettimeofday ();;
Printf.printf "f1: %f\n" (t2 -. t1);;

f2 30.;;
Printf.printf "f2: %f\n" ((Unix.gettimeofday ()) -. t2);;
*)
