(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*  A modul for converting xml trees to functions written by Tobias Elze    *)
(*  requires lablgl and lablglut    *)
(*  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file ../LICENSE.     *)
(*                                                                     *)
(***********************************************************************)

type nom = Scal of float | Mat of Gsl_matrix.matrix | Str of string | Xml of Xml.xml;;
type float' = Floatconst of float | Floatref of (float ref);;
type declared_function = 
{
	nargs: int; 
	fcontent: (nom array -> float)
};;
type declared_matrix_function = 
{
	mat_nargs: int; 
	mat_fcontent: (nom array -> Gsl_matrix.matrix)
};;
type declared_string_function = 
{
	str_nargs: int; 
	str_fcontent: (nom array -> string)
};;
type declared_xml_function = 
{
	xml_nargs: int; 
	xml_fcontent: (nom array -> Xml.xml)
};;
type declaration = 
	Constant of float 
	| Variable of (float ref) 
	| String_variable of (string ref)
	| Xml_variable of (Xml.xml ref)
	| Function of declared_function
	| Matrix of Gsl_matrix.matrix
	| Matrix_function of declared_matrix_function
	| String_function of declared_string_function
	| Xml_function of declared_xml_function
;;
type expr_content = 
	Fun of (nom array -> nom)
	| Const of nom 
	| Var of (float ref)
	| Str_var of (string ref)
	| Xml_var of (Xml.xml ref)
	| Arg of (bool * int)
	| Let of (nom array -> nom) array
	| Cond of 
	(
		((nom array -> nom) array)
		* ((nom array -> nom) array) 
		* (nom array -> nom) 
	)
;;
type expression;;

exception Non_declared_argument of string;;
exception Undeclared_identifier of string;;
exception Invalid_declaration_type of string;;
exception Not_enough_arguments of string;;

val global_declarations : (string, declaration) Hashtbl.t
(** [global_declarations]: hash table containing declarations for lookup *)

val local_declarations : (string, declaration) Hashtbl.t ref
(** [local_declarations]: reference to a hash table which can contain local declarations *)

val declare : string -> declaration -> unit
(** [declare id decl]: create a global declaration [decl] with identifier "id" *)

val replace_declaration : string -> declaration -> unit
(** [replace id decl]: replace a global declaration [decl] with identifier "id" *)

val get_declaration : string -> declaration
(** [get_declaration identifier]: 
get declaration [identifier] from global or local declarations hash table
*)

val print_gsl_matrix : Gsl_matrix.matrix -> unit
(** [print_gsl_matrix m]: print GSL matrix m to stdio *)

val float_of_nom : ?xmltag:string -> nom -> float
(** [float_of_nom xmltag n]: convert nom [n] to float;
xmltag: can be given to raise an appropriate exception in case of failure
(xmltag is optional, default: "") *)

val matrix_of_nom : ?xmltag:string -> nom -> Gsl_matrix.matrix
(** [matrix_of_nom xmltag n]: convert nom [n] to matrix;
xmltag: can be given to raise an appropriate exception in case of failure
(xmltag is optional, default: "")*)

val string_of_nom : nom -> string
(** [string_of_nom n]: convert nom [n] to string *)

val xml_of_nom : ?xmltag:string -> nom -> Xml.xml
(** [xml_of_nom xmltag n]: convert nom [n] to Xml.xml;
xmltag: can be given to raise an appropriate exception in case of failure
(xmltag is optional, default: "")*)

val rng_of_xml : Xml.xml -> Gsl_rng.t
(** [rng_of_xml elem]: extract random number generater of an XML node [elem];
set random seed in addition*)

val random_function_of_xml : Xml.xml -> unit -> float
(** [random_function_of_xml elem]: convert [elem] to an ocaml function *)

val xml_of_file : string -> Xml.xml
(** [xml_of_file filename]: open file [filename] and convert it to Xml-light xml *)

val function_of_xml :
 ?internal_functions:(string * declaration) list ->
 ?assoc_list:(string * int) list -> string -> Xml.xml -> nom array -> nom
(** [function_of_xml ~internal_functions ~assoc_list id elem]:
generate an ocaml function (nom array -> nom) from an Xml element [elem];
[internal_functions]: internally declared functions; optional, default: [];
[assoc_list]: list of the names of the arguments with their corresponding values; optional, default: [];
[id]: identifier (for recursion);
raise Non_declared_argument if an argument is not in assoc_list,
raise Invalid_declaration_type if Function is used when non functional expressions are expected
and vice versa
raise Failure specific to several internal functions
*)

val declaration_of_xml : Xml.xml -> string * declaration
(** [declaration_of_xml decl]: convert xml declaration [decl] to ocaml declaration *)

val generate_declarations : Xml.xml -> unit
(** [generate_declarations elem]: extract declarations of xml element [elem] 
and add them to the global declaration hash table *)



