(*
	compile with:
	ocamlopt -o multibin -I +gsl unix.cmxa bigarray.cmxa gsl.cmxa multibin.ml
	link statically:
	ocamlopt -o multibin -I +gsl -ccopt -static unix.cmxa bigarray.cmxa gsl.cmxa multibin.ml
	interpreter:
	ocaml -I +gsl unix.cma bigarray.cma gsl.cma
*)

(*non functional implementation of array sum, for speed reasons*)
let sum arr = 
	let n = Array.length arr in
	let s = ref 0 in
	for i = 0 to n-1 do
		s := !s + arr.(i);
	done;
	!s
;;

let float_sum arr = 
	let n = Array.length arr in
	let s = ref 0. in
	for i = 0 to n-1 do
		s := !s +. arr.(i);
	done;
	!s
;;

let normalize arr =
	let s = float_sum arr in
	if s <> 0. then Array.map (fun e -> e /. s) arr
	else 
		let r = 1. /. (float_of_int (Array.length arr)) in
		Array.map (fun _ -> r) arr
;;

(*cumulative sum of float array:*)
let cumsum arr result = 
	let s = ref 0. in
	for i = 0 to (Array.length arr)-1 do
		s := !s +. arr.(i);
		result.(i) <- !s;
	done;
;;


(*same as Array.copy, but copies a1 to a2 in situ:*)
let copy_in_situ a1 a2 =
	Array.blit a1 0 a2 0 (Array.length a1)
;;

let print_arr arr =
	print_endline (String.concat "," (Array.to_list (Array.map (string_of_float) arr)))
;;
let print_int_arr arr =
	print_endline (String.concat "," (Array.to_list (Array.map (string_of_int) arr)))
;;

(*auxiliary function; returns lnbeta of cumulative sum from left to right 
in the successes and failures arrays:*)
let lnbet left right successes failures =
	Gsl_sf.lnbeta
		(float_of_int ( (sum (Array.sub successes left (right-left+1))) +1 ))
		(float_of_int ( (sum (Array.sub failures left (right-left+1))) +1 ))
;;

(*same as lnbet, but for float vectors:*)
let lnbet_float left right successes failures =
	Gsl_sf.lnbeta
		( (float_sum (Array.sub successes left (right-left+1))) +. 1. )
		( (float_sum (Array.sub failures left (right-left+1))) +. 1. )
;;

let array_max arr =
	let m = ref (arr.(0)) in
	for i = 0 to (Array.length arr)-1 do
		if arr.(i) > !m then m := arr.(i);
	done;
	!m
;;

(*return log(sum(exp(v))):*)
let log_sum_exp arr =
	let m = array_max arr in
	let s = ref 0. in
	for i = 0 to (Array.length arr)-1 do
		s := !s +. exp (arr.(i) -. m);
		(*Printf.printf "%g, %g, %g, %g\n" !s arr.(i) (arr.(i)-.m) (exp (arr.(i)-.m));*)
	done;
	log(!s) +. m
;;

(*result = log(exp(v)*exp(m)), where v is a vector, m a triangular matrix,
result the vector which will contain the result;
tmpvec: same length as result, for temporary storage*)
let triag_log_mult v m result tmpvec = 
	let n = Array.length v in
	for i = 0 to n-1 do
		for j = 0 to i do
			tmpvec.(j) <- v.(j) +. m.(j).(i-j);
		done;
		result.(i) <- log_sum_exp (Array.sub tmpvec 0 (i+1));
	done;
;;


(** [bayes_factor successes failures]: transform two int arrays containing 
successes and failures to a triangular matrix containing lnbeta(succ+1, fail+1)
*)
let bayes_factor successes failures =
	let n = Array.length successes in
	if n <> Array.length failures then 
		failwith "bayes_factor: successes and failures have different lengths";
	
	let bf = 
		Array.init n 
			(fun i -> Array.init (n-i) 
					(fun j -> lnbet i (i+j) successes failures))
	in
	bf
;;

(** same as bayes_factor, but for float vectors *)
let bayes_factor_float successes failures =
	let n = Array.length successes in
	if n <> Array.length failures then 
		failwith "bayes_factor_float: successes and failures have different lengths";
	
	let bf = 
		Array.init n 
			(fun i -> Array.init (n-i) 
					(fun j -> lnbet_float i (i+j) successes failures))
	in
	bf
;;

(** [bayes_factor_q successes failures]: same as bayes_factor, but multiplies with q
*)
let bayes_factor_q q successes failures =
	let n = Array.length successes in
	if n <> Array.length failures then 
		failwith "bayes_factor_q: successes and failures have different lengths";
	
	let bf = 
		Array.init n 
			(fun i -> Array.init (n-i) 
					(fun j -> q *. (lnbet i (i+j) successes failures)))
	in
	bf
;;

let marginal_m p bayesfactors result tmpvec =
	let n = Array.length result in
	copy_in_situ bayesfactors.(0) p;
	copy_in_situ p result;
	(*Array.blit p (n-1) result (n-1) 1;*)
	for i = 1 to n - 1 do
		triag_log_mult (Array.sub p 0 (n-i)) (Array.sub bayesfactors i (n-i)) result tmpvec;
		Array.blit result 0 p 0 (n-i);
	done;
	(*multiply with log of prior 1/2^(n-1):*)
	(*let lnpr = log (2. ** (float_of_int n)) in
	for i=0 to n-1 do
		result.(i) <- result.(i) -. lnpr;
	done*)
;;

let marginal_likelihood p margm tmpvec successes failures =
	let bf = bayes_factor successes failures in
	
	marginal_m p bf margm tmpvec;
	log_sum_exp margm;
;;

let marginal_likelihood_float p margm tmpvec successes failures =
	let bf = bayes_factor_float successes failures in
	
	marginal_m p bf margm tmpvec;
	log_sum_exp margm;
;;

(** [bernoulli_matrix pvec successes failures]: calculate a "confidence matrix"
(Bernoulli plot) with [pvec] p values for each (success, failure) pair*)
let bernoulli_matrix pvec successes failures =
	let n = Array.length successes in
	let bf = bayes_factor successes failures in
	let pv = Array.make n 0. in
	let result = Array.make n 0. in
	let tmpvec = Array.make n 0. in
	marginal_m pv bf result tmpvec;
	let lnp0 = log_sum_exp result in
	
	let tbf = Array.map (fun e -> Array.map (fun a -> a) e) bf in
	let cumsumtriang vec =	(*triangular cumsum (float)*)
		let n = Array.length vec in
		Array.init n (fun i -> let a = Array.map float_of_int (Array.sub vec i (n-i)) in cumsum a a; a);
	in
	let cumsucc = cumsumtriang successes in
	let cumfail = cumsumtriang failures in
	let tpp = Array.map (fun e -> Array.map (fun a -> a) e) bf in
	
	let ps = pvec in
	Array.map
		(fun p ->
			if p <= 0. || p >= 1. then
				Array.make n 0.
			else
			(
				(*calculate tpp for current p:*)
				for i = 0 to n-1 do
					for j=0 to n-i-1 do
						tpp.(i).(j) <- cumsucc.(i).(j) *. (log p) +. cumfail.(i).(j) *. (log (1. -. p));
					done;
				done;
				Array.init n 
					(fun i -> 
						for i = 0 to n-1 do
							for j=0 to n-i-1 do
								tbf.(i).(j) <- bf.(i).(j);
							done;
						done;
						let len = n - i in
						for k=0 to i do
							Array.blit tpp.(k) (i-k) tbf.(k) (i-k) len;
						done;
						marginal_m pv tbf result tmpvec;
						exp ((log_sum_exp result) -. lnp0); )
			))
		ps
;;

(** take n samples from a given pdf: *)
let pdf_sample ?(rng=(Gsl_rng.make Gsl_rng.MT19937)) pdf n =
	let l = Array.length pdf in
	let cdf = Array.make l 0. in
	cumsum pdf cdf;
	let sample r =
		let rec find_index ind =
			if cdf.(ind)>r then ind
			else find_index (ind+1)
		in
		find_index 0
	in
	(*print_int_arr (Array.init n ( fun _ -> sample ((Gsl_rng.uniform rng)*.cdf.(l-1)) ));*)
	Array.init n ( fun _ -> sample ((Gsl_rng.uniform rng)*.cdf.(l-1)) );
;;

(** calculate entropy from lnp=: log_sum_exp marginal_m:*)
let entropy ?(q=1.000001) successes failures lnp =
	let n = Array.length successes in
	let bfq = bayes_factor_q q successes failures in
	let p = Array.make n 0. in
	let margmq = Array.make n 0. in
	let tmpvec = Array.make n 0. in
	marginal_m p bfq margmq tmpvec;
	(*let qpows = Array.mapi (fun i e -> e *. (q ** (float_of_int (n-i)))) margm in*)
	let est = exp ((log_sum_exp margmq) -. q*.lnp) in
	(1. -. est) /. (q -. 1.)
;;

(** 
[expected_gain_local ?(q=1.000001) ?(nbins=2) successes failures]:
a localized version to calculate expected info gain (for performance reasons);
nbins: number of bins to the left and to the right to take for localized calculations*)
let expected_gain_local ?(q=1.000001) ?(nbins=2) successes failures =
	let n = Array.length successes in
	let len = min (2*nbins + 1) n in
	(*temporary vectors*)
	let margm = Array.make len 0. in
	let tmpvec = Array.make len 0. in
	let pv = Array.make len 0. in
	let succq = (Array.map (fun e ->  (float_of_int e) *. q) successes)
	and failq = (Array.map (fun e ->  (float_of_int e) *. q) failures) in
	Array.init 
		n
		(fun i -> 
			let left = min (n-len) (max 0 (i-nbins)) in
			let ml = marginal_likelihood pv margm tmpvec (Array.sub successes left len) (Array.sub failures left len) in
			let mlq = marginal_likelihood_float pv margm tmpvec (Array.sub succq left len) (Array.sub failq left len) in
			successes.(i) <- successes.(i)+1;
			let oldsuccq = succq.(i) in
			succq.(i) <- (float_of_int successes.(i)) *. q;
			let mls = 
				marginal_likelihood 
					pv 
					margm 
					tmpvec 
					(Array.sub successes left len) 
					(Array.sub failures left len) 
			in
			let mlsq = 
				marginal_likelihood_float
					pv
					margm 
					tmpvec
					(Array.sub succq left len)
					(Array.sub failq left len)
			in
			successes.(i) <- successes.(i)-1;
			succq.(i) <- oldsuccq;
			
			failures.(i) <- failures.(i)+1;
			let oldfailq = failq.(i) in
			failq.(i) <- (float_of_int failures.(i)) *. q;
			let mlf = 
				marginal_likelihood 
					pv 
					margm 
					tmpvec 
					(Array.sub successes left len) 
					(Array.sub failures left len) 
			in
			let mlfq = 
				marginal_likelihood_float
					pv
					margm 
					tmpvec
					(Array.sub succq left len)
					(Array.sub failq left len)
			in
			failures.(i) <- failures.(i)-1;
			failq.(i) <- oldfailq;
			
			let p_s = exp (mls -. ml)
			and p_f = exp (mlf -. ml) in
			let ent = ml -. (mlq -. ml) /. (q -. 1.) in
			let ents = mls -. (mlsq -. mls) /. (q -. 1.) in
			let entf = mlf -. (mlfq -. mlf) /. (q -. 1.) in

			ent -. p_s*.ents -. p_f*.entf);
;;

(*let entropy ?(q=1.0001) margm lnp =
	let n = Array.length margm in
	let qpows = Array.mapi (fun i e -> e *. (q ** (float_of_int (n-i)))) margm in
	let est = exp ((log_sum_exp qpows) -. q*.lnp) in
	(1. -. est) /. (q -. 1.)
;;*)

(** [(ev, delta, std) = multibin_estimator q successes failures]:
returns expectation values and max. expected information's gains (delta) 
according to the multibin method;
[successes, failures]: arrays containing successes and failures;
[q]: precision parameter for entropy calculation (see entropy), optional, default: 1.000001*)
let multibin_estimator ?(q=1.000001) successes failures =
	let n = Array.length successes in
	let bf = bayes_factor successes failures in
	(*print_int_arr successes;
	print_int_arr failures;
	print_endline "BF:";
	Array.iter print_arr bf;*)
	let p = Array.make n 0. in
	let result = Array.make n 0. in
	let tmpvec = Array.make n 0. in
	marginal_m p bf result tmpvec;
	
	let lnp0 = log_sum_exp result in
	let h0 = entropy ~q:q successes failures lnp0 in
	(*print_arr result;*)
	(*print_endline (string_of_float h0);*)
	
	let ev = Array.make n 0.
	and std = Array.make n 0.
	and delta = Array.make n 0. in

	for i=0 to n-1 do
		successes.(i) <- successes.(i) + 1;
		let bf = bayes_factor successes failures in
		marginal_m p bf result tmpvec;
		let lnp1 = log_sum_exp result in
		let h1 = entropy ~q:q successes failures lnp1 in
		successes.(i) <- successes.(i) - 1;	(*reset*)
		
		failures.(i) <- failures.(i) + 1;
		let bf = bayes_factor successes failures in
		marginal_m p bf result tmpvec;
		let lnp2 = log_sum_exp result in
		let h2 = entropy ~q:q successes failures lnp2 in
		failures.(i) <- failures.(i) - 1;	(*reset*)
		
		ev.(i) <- exp (lnp1 -. lnp0);
		delta.(i) <- h0 -. ev.(i)*.h1 -. (1. -. ev.(i))*.h2;
		
		successes.(i) <- successes.(i) + 2;
		let bf = bayes_factor successes failures in
		marginal_m p bf result tmpvec;
		let lnp_var = log_sum_exp result in
		std.(i) <- exp (lnp_var -. lnp0);
		std.(i) <- sqrt (std.(i) -. ev.(i)*.ev.(i));
		successes.(i) <- successes.(i) - 2;	(*reset*)
		(*std.(i) <- exp (lnp2 -. lnp0);
		std.(i) <- sqrt (std.(i) -. ev.(i)*.ev.(i));*)
	done;
	(ev, delta, std)
;;


(** 
[multibin_samples ?(rng=(Gsl_rng.make Gsl_rng.MT19937)) ?(q=1.000001) successes failures nsamp]:
return vector of nsamp indices for sampling according to multibin_estimator
[successes]: int vector of successes for each sample point,
[failures]: int vector of failures for each sample point;
[q]: precision parameter for entropy calculation (see entropy), optional, default: 1.000001
[rng]: random number generator
*)
let multibin_samples ?(rng=(Gsl_rng.make Gsl_rng.MT19937)) ?(q=1.000001) successes failures = 
	let (_, delta, _) = multibin_estimator ~q:q successes failures in
	(*print_arr (normalize (Array.map exp delta));*)
	(*let sample_pdf = normalize (Array.map exp delta) in*)
	let sample_pdf = normalize delta in
	pdf_sample ~rng:rng sample_pdf
;;


(** [multibin_samples_local ?(q=1.000001) ?(ratio_to_sample=0.1) successes failures]:
similar to multibin_samples, but sampling is calculated by a 
localized determination of maximal information's gain
(for speed reasons);

[ratio_to_sample]: percentage of length(successes) that determines the number of samples;
optional, default: 0.1;
n=ceil(ratio_to_sample*length(successes)) indices to sample are chosen from the maximal 
information's gains*)
let multibin_samples_local ?(q=1.000001) ?(ratio_to_sample=0.1) successes failures =
	if ratio_to_sample < 0. || ratio_to_sample >1. then 
		failwith "multibin_samples_local: ratio_to_sample must be within [0,1]";
	let nsamp = int_of_float (ceil (ratio_to_sample*.(float_of_int (Array.length successes)))) in
	let deltas = expected_gain_local ~q:q successes failures in
	let delt_ind = Array.mapi (fun i e -> (i, e)) deltas in
	Array.sort (fun a b -> -(compare (snd a) (snd b))) delt_ind;
	Array.map fst (Array.sub delt_ind 0 nsamp)
;;



let gibbs_step ?(rng=(Gsl_rng.make Gsl_rng.MT19937)) bayesfactors binning_scheme =
	let nborders = Array.length binning_scheme in
	let perm = Array.mapi (fun i _ -> i) binning_scheme in	(*permutation*)
	Gsl_randist.shuffle rng perm;
	(*print_int_arr perm;*)
	Array.iter
		(fun idx ->
			let rec find_left_border = function
				|0 | -1 -> 0
				| i -> if binning_scheme.(i-1) then i else find_left_border (i-1)
			in
			let rec find_right_border i = 
				if i>=nborders-1 then nborders else
					(if binning_scheme.(i+1) then i+1 else find_right_border (i+1))
			in
			let (lb, rb) = (find_left_border idx, find_right_border idx) in
			let prob_separated = 1. /. ( 1. +. exp (bayesfactors.(lb).(rb-lb) -. bayesfactors.(lb).(idx-lb) -. bayesfactors.(idx+1).(rb-(idx+1))) ) in
			binning_scheme.(idx) <- (Gsl_rng.uniform rng)<prob_separated; )
		perm;
	binning_scheme
;;

let gibbs_sampler ?(rng=(Gsl_rng.make Gsl_rng.MT19937)) initial_binning_scheme successes failures steps =
	let bf = bayes_factor successes failures in
	let rec gsample current_scheme binning_scheme_list = function
		| 0 -> current_scheme::binning_scheme_list
		| i -> 
			let new_scheme = Array.make (Array.length current_scheme) true in
			copy_in_situ (gibbs_step ~rng:rng bf current_scheme) new_scheme;
			gsample new_scheme (current_scheme::binning_scheme_list) (i-1)
	in
	Array.of_list (List.rev (gsample initial_binning_scheme [] steps))
;;



(*
let rng=Gsl_rng.make Gsl_rng.MT19937;;
(*let successes = [|30; 90; 30; 60; 60; 70; 60; 20; 20; 20|] in
let failures = Array.map ((-) 100) successes in*)
let successes = [|3; 16; 15; 15; 16; 16; 15; 13; 12; 11; 12; 11; 10; 12; 11; 11; 13; 10; 11; 9; 10; 13; 12; 9; 9; 10; 9; 8; 11; 7; 9; 12; 9; 8; 8; 10; 10; 3|]
and failures = [|13; 0; 1; 1; 0; 0; 1; 3; 4; 5; 4; 5; 6; 4; 5; 5; 3; 6; 5; 7; 6; 3; 4; 7; 7; 6; 7; 8; 5; 9; 7; 4; 7; 8; 8; 6; 6; 13|] in
let initial_binning_scheme = Array.init ((Array.length successes) - 1) (fun _ -> (Gsl_rng.uniform rng)<0.5) in
(*let initial_binning_scheme  = [|false; true; true; false; false; true; false; false; false|] in*)
let binning_schemes = gibbs_sampler initial_binning_scheme successes failures 100000 in
let bs_int = Array.map (Array.map (fun e -> if e then 1 else 0)) binning_schemes in
(*print_endline "Matrix of binning schemes:";*)
Array.iter print_int_arr bs_int;;
*)

(*
let rng=Gsl_rng.make Gsl_rng.MT19937;;
let successes = [|3; 16; 15; 15; 16; 16; 15; 13; 12; 11; 12; 11; 10; 12; 11; 11; 13; 10; 11; 9; 10; 13; 12; 9; 9; 10; 9; 8; 11; 7; 9; 12; 9; 8; 8; 10; 10; 3|]
and failures = [|13; 0; 1; 1; 0; 0; 1; 3; 4; 5; 4; 5; 6; 4; 5; 5; 3; 6; 5; 7; 6; 3; 4; 7; 7; 6; 7; 8; 5; 9; 7; 4; 7; 8; 8; 6; 6; 13|] in
(*let successes = [|1; 1; 1; 1; 1; 1; 400; 400; 400; 400; 400; 400; 400|]
and failures = [|400; 400; 400; 400; 400; 400; 1; 1; 1; 1; 1; 1; 1|] in*)

(*let indices = multibin_samples ~rng:rng ~q:1.000001 successes failures ((Array.length successes)*10) in*)
let indices = multibin_samples_local successes failures in
 indices;;
*)

