(*
	compile with:
	ocamlopt -o sequences -I +gsl unix.cmxa bigarray.cmxa gsl.cmxa sequences.ml
	link statically:
	ocamlopt -o sequences -I +gsl -ccopt -static unix.cmxa bigarray.cmxa gsl.cmxa sequences.ml
	interpreter:
	ocaml -I +gsl unix.cma bigarray.cma gsl.cma
*)

type forbidden_sequence = 
	| Abstract of (string list)
	| Verbatim of (float list)
;;

(*auxiliary functions:*)
let rec unique = function 
	| [] -> []
	| h :: t -> [h] @ (unique (List.filter (fun a -> a <> h) t))
;;


(*just for debugging:*)
let print_list l = 
	List.iter (Printf.printf "%.0f, ") l;
	print_newline ()
;;

(** [check_sequence forbidden seq]: check whether sequence [seq] is "valid",
i.e. does not match any of the forbidden sequences in list [forbidden]*)
let check_sequence forbidden seq =
	let matches_abstract abstract_sequence =
		let l =  List.length abstract_sequence in
		let seqhash = Hashtbl.create (List.length (unique abstract_sequence)) in
		let rec matches_forbidden = function
			| (true, _) -> true
			| (_, sq) -> 
				List.iter 
					(fun t -> Hashtbl.replace seqhash t (ref None))
					(unique abstract_sequence);
				let given_entries = ref [] in
				let rec chk_first = function
					| (_, []) -> true
					| ([], _) -> false
					| (h_sq::t_sq, h_forb::t_forb) -> 
					(
						let r = Hashtbl.find seqhash h_forb in
						match !r with
						None -> 
							if List.mem h_sq !given_entries then
								false
							else
							(
								r := Some h_sq;
								given_entries := h_sq::!given_entries;
								chk_first (t_sq, t_forb)
							)
						| Some v -> 
							if v = h_sq then
								chk_first (t_sq, t_forb)
							else
								false
					)
				in
				if List.length sq < l - 1 then
					false
				else
					matches_forbidden (chk_first (sq, abstract_sequence), List.tl sq) 
		in
		matches_forbidden (false, seq)
	in
	
	let matches_verbatim verbatim_sequence =
		let l =  List.length verbatim_sequence in
		let rec matches_forbidden = function
			| (true, _) -> true
			| (_, sq) -> 
				let rec chk_first = function
					| (_, []) -> true
					| ([], _) -> false
					| (h_sq::t_sq, h_forb::t_forb) -> 
						if h_sq = h_forb then
							chk_first (t_sq, t_forb)
						else
							false
				in
				if List.length sq < l then
					false
				else
					matches_forbidden (chk_first (sq, verbatim_sequence), List.tl sq) 
		in
		matches_forbidden (false, seq)
	in
	
	let rec seq_matches_forbidden_list = function
		| (true, _) -> true
		| (_, []) -> false
		| (_, (Abstract a)::t) ->
			seq_matches_forbidden_list ((matches_abstract a), t)
		| (_, (Verbatim v)::t)  -> 
			seq_matches_forbidden_list ((matches_verbatim v), t)
	in
	not (seq_matches_forbidden_list (false, forbidden))
;;

(*tail recursive and more efficient flatten, 
but does not keep the order of the arguments*)
let flatten_unordered l = 
	let rec fu targetlist =  function
		| [] -> targetlist
		| h::t -> fu (List.rev_append h targetlist) t
	in
	fu [] l
;;

let shuffle rng l =
	let a = Array.of_list l in
	Gsl_randist.shuffle rng a;
	Array.to_list a
;;

(** [find_valid_seq rng forbidden_seqs l]: permutate list [l] randomly until it is valid 
due to the list of forbidden sequences [forbidden_seqs];
[rng]: Random number generator*)
let rec find_valid_seq rng forbidden_seqs l = 
	if check_sequence forbidden_seqs l then
		l
	else
		find_valid_seq rng forbidden_seqs (shuffle rng l)
;;

(** [allowed_sequences forbidden_seqs seq]:
compute all permutations of [seq] and
return a list of all allowed sequences
due to the list of forbidden sequences [forbidden_seqs]*)
let allowed_sequences forbidden_seqs seq =
	let (abstract_forbidden_seqs, verbatim_forbidden_seqs) =
		let (l1, l2) = 
			List.partition
				(function | Abstract _ -> true | _ -> false)
				forbidden_seqs
		in
		(
			List.map (function | Abstract a -> a | _ -> failwith "") l1,
			List.map (function | Verbatim v -> v | _ -> failwith "") l2
		)
	in
	let classes = unique seq in
	(*frequency hash table:*)
	let freqs = Hashtbl.create (List.length classes) in
	List.iter
		(fun t -> Hashtbl.add freqs t (ref 0))
		classes;
	(*count:*)
	List.iter
		(fun t -> incr (Hashtbl.find freqs t))
		seq;
	let sorted_abstract_forbidden_sequences =
		List.sort 
			(fun a b -> (List.length a) - (List.length b))
			abstract_forbidden_seqs
	in
	(*this is actually a "take first", because the lists are in reversed order:*)
	let take_last =
		let rec tklst targetlist sourcelist = function
			| 0 -> targetlist
			| i ->
				match sourcelist with
				[] -> targetlist
				| h::t -> tklst (h::targetlist) t (i-1)
		in
		tklst [] 
	in
	(*check if seq does not match the abstract forbidden sequence a_forb_seq:*)
	let validate_abstr a_forb_seq seq =
		let l = List.length a_forb_seq in
		if List.length seq < l then
			true
		else
		(
			let tseq = take_last seq l 
			and u_forbidden = unique a_forb_seq in
			let seqhash = Hashtbl.create l in
			List.iter 
				(fun t -> Hashtbl.add seqhash t (ref None))
				u_forbidden;
			let given_entries = ref [] in
			let valid_pair = function (key, entry) -> 
				let r = Hashtbl.find seqhash key in
				match !r with
				None -> 
				(
					if List.mem entry !given_entries then
						false
					else
					(
						r := Some entry;
						given_entries := entry::!given_entries;
						true
					)
				)
				| Some v ->
				(
					v = entry
				)
			in
			let rec check_seq = function
				| (false, _) -> true
				| (_, []) -> false
				| (_, h::t) -> check_seq (valid_pair h, t)
			in
			check_seq (true, (List.combine a_forb_seq tseq))
		)
	in
	let validate_all seq =
		let rec va = function
			| (false, _) -> false
			| (_, []) -> true
			| (_, h::t) -> va ((validate_abstr h seq), t)
		in
		let abstr = 
			va (true, sorted_abstract_forbidden_sequences) 
		in
		let verbat =
			let rec va_verb = function
				| (false, _) -> false
				| (_, []) -> true
				| (_, h::t) -> 
					let l = List.length h in
					if List.length seq < l then
						true
					else
						let tseq = take_last seq l in
						va_verb ((h <> tseq), t)
			in
			va_verb (true, verbatim_forbidden_seqs)
		in
		abstr && verbat
	in
	let make_new_list old_list =
		let nlist newelem = function (restfreqs, oldseq) ->
			let new_elem_freq = (Hashtbl.find restfreqs newelem) - 1 in
			if new_elem_freq < 0 then
				None
			else
			(
				let newseq = newelem::oldseq in
				if validate_all newseq then
					Some
					(
						(
							let new_restfreqs = Hashtbl.copy restfreqs in
							Hashtbl.replace new_restfreqs newelem new_elem_freq;
							new_restfreqs
						),
						newseq
					)
				else
					None
			)
		in
		let t_new_lists old_entry = 
			List.map 
				(fun t -> nlist t old_entry)
				classes
		in
		let new_list_nested =
			List.rev_map 
				(
					fun t -> 
						List.filter 
							((<>) None)
							(t_new_lists t)
				) 
				old_list
		in
		let new_list_options = flatten_unordered new_list_nested in
		List.rev_map
			(
				function
				| Some el -> el
				| _ -> failwith ""	(*cannot happen*)
			)
			new_list_options;
	in
	let restfreqs = Hashtbl.create (List.length classes) in
	Hashtbl.iter
		(fun key entry -> Hashtbl.add restfreqs key !entry)
		freqs;
	let ap = ref [(restfreqs, [])] in
	(*iterative, maybe faster:*)
	for i = 1 to (List.length seq) do
		(*Printf.printf "%i: %i" ((List.length seq) - i + 1) (List.length !ap);
		print_newline ();*)
		ap := make_new_list !ap
	done;

	List.rev_map (fun t -> List.rev (snd t)) !ap
;;

let list_of_gsl_matrix m =
	let a = Gsl_matrix.to_array m in
	Array.to_list a
;;

let gsl_matrix_of_list l =
	let a = Array.of_list l in
	Gsl_matrix.of_arrays [|a|]
;;

(** [find_valid_seq_gsl rng forbidden_seqs mat]:
same as [find_valid_seq], but takes a GSL matrix instead of a list as an argument
and returns a GSL matrix, too*)
let find_valid_seq_gsl rng forbidden_seqs mat =
	 let l = find_valid_seq rng forbidden_seqs (list_of_gsl_matrix mat) in
	 gsl_matrix_of_list l
;;

(** [allowed_sequences_gsl forbidden_seqs seq]:
same as [allowed_sequences], but takes a GSL matrix instead of a list as an argument
and returns a GSL matrix, too*)
let allowed_sequences_gsl forbidden_seqs seq =
	let l = allowed_sequences forbidden_seqs (list_of_gsl_matrix seq) in
	let la = Array.of_list l in
	let a = Array.map Array.of_list la in
	Gsl_matrix.of_arrays a
;;

(*
let input_vector =
	let a = Array.init 3 (fun x -> (float_of_int x)+.1.) in
	let l = Array.to_list a in
	let rec list_rep targetlist l = function
		| 0 -> targetlist
		| i -> list_rep (l@targetlist) l (i-1)
	in
	list_rep [] l 5
;;
let forbidden_seqs = 
[
	Abstract ["a"; "a"];
	Abstract ["a"; "b"; "a"];
	Verbatim [1.; 1.; 1.];
	Verbatim [2.; 2.; 2.];
];;

allowed_sequences forbidden_seqs input_vector;;*)


(*let rng = Gsl_rng.make (Gsl_rng.default ());;
Gsl_rng.set rng (Nativeint.of_float (Unix.time ()));;
let shuffle l =
	let a = Array.of_list l in
	Gsl_randist.shuffle rng a;
	Array.to_list a
;;
*)
