
(* compile with: 
	ocamlc -I +lablGL -I +site-lib/sdl lablgl.cma lablglut.cma str.cma bigarray.cma sdl.cma sdlloader.cma objects2D.mli objects2D.ml
	ocamlc -a -I +lablGL -I +site-lib/sdl lablgl.cma lablglut.cma str.cma bigarray.cma sdl.cma sdlloader.cma -o objects2D.cma objects2D.cmo
*)

(*
The color argument is made an option type with default None since setting the color once in a display list will set it for the whole rest of the list no matter what color you specify when calling the list items.
*)

(* auxiliary function for drawing: *)
let draw_it draw_fun ?(color=None) ?(position=(0.0,0.0)) () =
	(
		match color with
		None -> ()
		| Some col -> GlDraw.color col; 
	);
	draw_fun position;
;;

(** [filled_rectangle ~width ~height ~color ~position ()]: draw a filled rectangle with lower left corner [position] (optional arguments, default: (0.0,0.0)), width [width] and height [height], in color [color] (optional, default: None) *)
let filled_rectangle ~width ~height =
	let draw_fun position =
		let (x,y) = position in
		GlDraw.rect position ((x+.width), (y+.height))
	in
	draw_it draw_fun
;;

(** [rectangle ~width ~height ~color ~position ()]: draw a rectangle with lower left corner [position] (optional arguments, default: (0.0,0.0)), width [width] and height [height], in color [color] (optional, default: None) *)
let rectangle ~width ~height =
	let draw_fun position = 
		let (x,y) = position in
		GlDraw.begins `line_loop;
			GlDraw.vertex2 position;
			GlDraw.vertex ~x ~y:(y +. height) ();
			GlDraw.vertex ~x:(x +. width) ~y:(y +. height) ();
			GlDraw.vertex ~x:(x +. width) ~y ();
		GlDraw.ends ()
	in
	draw_it draw_fun
;;

(**
[polygon_edges ~edges ~radius ~angle ~center ()]
return the edges of a regular polygon as an array of point2 (float*float)
[edges]: number of edges
[radius]: distance between the center and each of the edges
[angle]: rotation angle
angle = 0: lowest edge of the polygon is vertical
[center]: center point of the polygon (optional, default: (0.0,0.0))
*)
let polygon_edges ~edges:n ~radius ?(angle=0.0) ?(center=(0.0,0.0)) () = 
	if n>2 && radius>=0.0 then
	(
		let pi = acos (-1.0) in
		let a = pi *. (360.0/.(float_of_int n)) /. 180.0 in
		let b = -.(pi -. a) /. 2.0 -. pi*.angle /. 180.0 in
		Array.init n
		(
			fun i -> 
			(
				(fst center) +. radius*.(cos (b -. (float_of_int i)*. a)), 
				(snd center) +. radius*.(sin (b -. (float_of_int i)*. a)) 
			)
		)
	)
	else raise (Invalid_argument "polygon_edges")
;;

(**
[regular_polygon ~edges ~radius ~angle ~color ~position ()]
draw a regular polygon
[edges]: number of edges
[radius]: distance between the center and each of the edges
[angle]: rotation angle
angle = 0: lowest edge of the polygon is vertical
[color]: drawing color (optional, default: None)
[position]: center point of the polygon (optional, default: (0.0,0.0))
*)
let regular_polygon ~edges:n ~radius ?(angle=0.0) = 
	let draw_fun center =
		let points = polygon_edges ~edges:n ~radius ~angle ~center:center () in
		GlDraw.begins `line_loop;
			Array.iter GlDraw.vertex2 points;
		GlDraw.ends ();
	in
	draw_it draw_fun
;;

(**
[filled_regular_polygon ~edges ~radius ~angle ~color ~position ()]
draw a filled regular polygon
[edges]: number of edges
[radius]: distance between the center and each of the edges
[angle]: rotation angle
angle = 0: lowest edge of the polygon is vertical
[color]: drawing color (optional, default: None)
[position]: center point of the polygon (optional, default: (0.0,0.0))
*)
let filled_regular_polygon ~edges:n ~radius ?(angle=0.0) = 
	let draw_fun center =
		let points = polygon_edges ~edges:n ~radius ~angle ~center () in
		GlDraw.begins `polygon;
			Array.iter GlDraw.vertex2 points;
		GlDraw.ends ()
	in
	draw_it draw_fun
;;

(** [circle ~radius ~color ~position ()]: draw a circle. 
[color]: drawing color (optional, default: None)
[position] center, is optional, default: (0.0,0.0) *)
let circle = 
	regular_polygon ~edges:100 ~angle:0.0
;;

(** [filled_circle ~radius ~color ~position ()]: draw a filled circle. 
[color]: drawing color (optional, default: None)
[position] center, is optional, default: (0.0,0.0) *)
let filled_circle = 
	filled_regular_polygon ~edges:100 ~angle:0.0
;;

(** [partial_disk ~inner ~outer ~start ~sweep ~color ~position ()]: draw a partial disk
with inner radius [inner],
outer radius [outer],
start angle [start],
and sweep angle [sweep]
in color [color] (optional, default: None).
[position]: center, is optional, default: (0.0,0.0)
 *)
let partial_disk ~inner ~outer ~start ~sweep =
	let draw_fun center =
		GlMat.push ();
		GlMat.translate ~x:(fst center) ~y:(snd center) ();
		GluQuadric.partial_disk ~inner:inner ~outer:outer ~slices:100 ~loops:1 ~start:start ~sweep:sweep ();
		GlMat.pop ()
	in
	draw_it draw_fun
;;

(** [primitive color shp points ()]: draw a geometric primitive
in color [color] (optional, default: None);
use [GlDraw.vertex2] to draw the array of points2 [points]
applying shape [shp] ({`line_loop|`line_strip|`lines|`points|`polygon|`quad_strip|`quads
  |`triangle_fan|`triangle_strip|`triangles})
*)
let primitive color shp points =
	let draw_fun dummy =
		GlDraw.begins shp;
			Array.iter GlDraw.vertex2 points;
		GlDraw.ends ();
	in
	draw_it draw_fun ~color:color ~position:(0.0,0.0) ()
;;

(** [text_line ~text ~font ~color ~position]: display text string [text] with font [font]
in color [color] (optional, default: None)
at position [position] (optional, default: (0.0,0.0)) *)
let text_line ~text ~font =
	let draw_fun position =
		GlPix.raster_pos (fst position) (snd position) ();
		String.iter (fun t -> Glut.bitmapCharacter ~font:font ~c:(int_of_char t)) text
	in
	draw_it draw_fun
;;

(** [text_field ~text ~font ~width ~height ~pixel_width ~pixel_height ~alignment ~draw_frame ~position ()]:
display text string [text] in font [font] formatted in a textfield with width [width] and height [height];
[pixel_width]: width of a single pixel, corresponds to (width of internal scale) / (width in pixel);
[pixel_height]: height of a single pixel, corresponds to (height of internal scale) / (height in pixel);
[alignment]: alignment of the text, possible values: "Left", "Right", "Center";
[draw_frame]: whether to draw a frame around the field or not;
[color]: text color (optional, default: None).
[position]: position of the lower left corner of the text field; optional value, default: (0.0, 0.0);
text string [text] may contain line break symbols (\n).
*)
let text_field ~text ~font ~width ~height ~pixel_width ~pixel_height ~alignment ~draw_frame ?(color=None) ?(position=(0.0,0.0)) () =
	if draw_frame then rectangle ~width:width ~height:height ~color:color ~position:position ();
	let font_height = 
		let height_in_pixel =
		(
			match font with
			Glut.BITMAP_8_BY_13 -> 13
			| Glut.BITMAP_9_BY_15 -> 15
			| Glut.BITMAP_TIMES_ROMAN_10 -> 10
			| Glut.BITMAP_TIMES_ROMAN_24 -> 24
			| Glut.BITMAP_HELVETICA_10 -> 10
			| Glut.BITMAP_HELVETICA_12 -> 12
			| _ -> 18
		) in
		(float_of_int height_in_pixel) *. pixel_height
	in
	let horizontal_border = width /. 20. in
	let vertical_border = height /. 20. in
	let textwidth = width -. 2.*.horizontal_border in
	let textheight = height -. 2.*.vertical_border in
	let string_length s = (float_of_int (Glut.bitmapLength ~font:font ~str:s)) *. pixel_width in
	let check_width w = string_length w <= textwidth in
	let linebreaks_list = 
		let ll = Str.split (Str.regexp_string "\\n") text in
		List.map (fun t -> if (t = "" || t = " ") then "  " else t) ll
	in
	let lines_list_of_string text =
		let word_list = Str.split (Str.regexp " ") text in
		let rec generate_lines lineslist wordlist =
			match wordlist with
			[] -> List.map (fun x -> String.sub x 0 ((String.length x) - 1)) (List.rev lineslist)
			| _ -> 
				let current_line = List.hd lineslist in
				let current_word = List.hd wordlist in
				let append_word = (current_line ^ current_word ^ " ") in
				if check_width append_word || current_line = "" then
					generate_lines (append_word :: (List.tl lineslist)) (List.tl wordlist)
				else
					generate_lines ("" :: lineslist) wordlist
		in
		generate_lines [""] word_list
	in
	let lines_list =
		let ll = List.map lines_list_of_string linebreaks_list in
		List.flatten ll
	in
	let line_distance = 
		let l = float_of_int (List.length lines_list) in
		textheight /. l 
	in
	let x_distance line =
		match alignment with
		"Right" -> textwidth -. (string_length line)
		| "Center" -> (textwidth -. (string_length line))/.2.
		| _ -> 0.
	in
	let distance_list = List.map x_distance lines_list in
	let rec print_lines line_start_x line_start_y lplist =
		match lplist with
		[] -> ()
		| _ ->
			let (dist, line) = List.hd lplist in
			let pos = ( line_start_x +. dist, line_start_y ) in
			text_line ~text:line ~font:font ~color:color ~position:pos ();
			let ny = line_start_y -. line_distance in
			print_lines line_start_x ny (List.tl lplist)
	in
	let dist_lines_list = List.combine distance_list lines_list in
	let left_text_start = (fst position) +. horizontal_border
	and upper_text_start = (snd position) +. textheight +. vertical_border -. font_height in
	print_lines left_text_start upper_text_start dist_lines_list
;;

(** [button ~width ~height ~text ~font ~pixel_width ~pixel_height ~color_range ~shift_text_pos ~position ()]: 
draw button of height [height] and width [width] 
with textstring [text] in font [font]
[pixel_width]: width of a single pixel, corresponds to (width of internal scale) / (width in pixel);
[pixel_height]: height of a single pixel, corresponds to (height of internal scale) / (height in pixel);
[color_range]: pair representing the darkest and the brightest gray value of the button; 
optional value, default: (0.7, 0.95)
[shift_text_pos]: pair representing possible shifts in x and y directions
from automatically calculated text position; optional value, default: (0.0, 0.0)
[position]: position of the lower left corner of the text field; optional value, default: (0.0, 0.0)
*)
let button ~width ~height ~text ~font ~pixel_width ~pixel_height ?(color_range=(0.7, 0.95)) ?(shift_text_pos=(0.0,0.0)) ?(position=(0.0,0.0)) () =
	let steps = 50 in
	let inner = height /. 9. in
	let stepwidth = (height -. inner) /. (float_of_int steps) in
	let colstepwidth = ((snd color_range) -. (fst color_range)) /. (float_of_int steps) in
	let rec rectangles h col lower_x n =
		if n > 0 then
		(
			GlDraw.color (col, col, col);
			let left_corner = (fst position, lower_x) in
			filled_rectangle ~width:width ~height:h ~position:left_corner ();
			rectangles (h -. stepwidth) (col +. colstepwidth) (lower_x +. stepwidth /. 2.) (n-1)
		)
	in
	rectangles height (fst color_range) (snd position) steps;
	GlDraw.color (0., 0., 0.);
	let textpos = ((fst position) +. (fst shift_text_pos), (snd position) +. (snd shift_text_pos)) in
	text_field ~text:text ~font:font ~width:width ~height:height ~pixel_width:pixel_width ~pixel_height:pixel_height ~alignment:"Center" ~draw_frame:false ~position:textpos ();
;;

(** [pressed_button ~width ~height ~text ~font ~pixel_width ~pixel_height ~color_range ~position ()]: 
same as button, but with inverse default color_range and shifted text position
*)
let pressed_button ~width ~height ~text ~font ~pixel_width ~pixel_height ?(color_range=(0.95, 0.7)) ?(position=(0.0,0.0)) () =
	let text_vertical_shift = -.(height/.10.) in
	button ~width:width ~height:height ~text:text ~font:font ~pixel_width:pixel_width ~pixel_height:pixel_height ~color_range:color_range ~shift_text_pos:(0.0, text_vertical_shift) ~position:position ()
;;

(** [mouseover_button ~width ~height ~text ~font ~pixel_width ~pixel_height ~color_range ~position ()]: 
same as button, but with brighter color as default (0.75, 1.0)
*)
let mouseover_button ~width ~height ~text ~font ~pixel_width ~pixel_height ?(color_range=(0.75, 1.0)) ?(position=(0.0,0.0)) () =
	let text_vertical_shift = -.(height/.10.) in
	button ~width:width ~height:height ~text:text ~font:font ~pixel_width:pixel_width ~pixel_height:pixel_height ~color_range:color_range ~position:position ()
;;

(** [sdl_surface_of_image filename]: convert image from file [filename] to SDL surface, either in rgb or in rgba
*)
let sdl_surface_of_image filename =
	let s = Sdlloader.load_image filename in
	let (w, h, _) = Sdlvideo.surface_dims s in
	let sf = Sdlvideo.surface_format s in
	let surf = 
		if sf.Sdlvideo.bits_pp < 3 then
		(
			let s1 = 
				Sdlvideo.create_RGB_surface
					(*(Sdlvideo.surface_flags s)*)[`SWSURFACE]
					~w:w
					~h:h
					~bpp:24
					~rmask:sf.Sdlvideo.rmask
					~gmask:sf.Sdlvideo.gmask
					~bmask:sf.Sdlvideo.bmask
					~amask:sf.Sdlvideo.amask
			in
			for i = 0 to w-1 do
				for j = 0 to h-1 do
					let (b, g, r) = Sdlvideo.get_pixel_color s ~x:i ~y:j in
					Sdlvideo.put_pixel_color s1 ~x:i ~y:j (r, g, b)
				done;
			done;
			s1
		)
		else
			s
	in
	Sdlvideo.lock surf;
	(*convert bmp colors from bgr to rgb:*)
	let strl = String.length filename in
	if String.sub filename (strl - 3) 3 = "bmp" || String.sub filename (strl - 3) 3 = "BMP" then
	(
		for i = 0 to w-1 do
			for j = 0 to h-1 do
				let (b, g, r) = Sdlvideo.get_pixel_color surf ~x:i ~y:j in
				Sdlvideo.put_pixel_color surf ~x:i ~y:j (r, g, b)
			done;
		done;
	);
	(*turning upside down:*)
	for i = 0 to w-1 do
		for j = 0 to (h-1)/2 do
			let p1 = Sdlvideo.get_pixel surf ~x:i ~y:j in
			let p2 = Sdlvideo.get_pixel surf ~x:i ~y:(h-j-1) in
			Sdlvideo.put_pixel surf ~x:i ~y:j p2;
			Sdlvideo.put_pixel surf ~x:i ~y:(h-j-1) p1;
		done;
	done;
	surf
;;

(** [gl_pix_of_surface ~clip8 ~fill_color surf]: convert SDL surface [surf] to GlPix.t, 
either RGB or RGBA
[clip8]: if width no multiple of 8, increase width by filling the image with [fill_color];
this might be necessary on MS Windows so only effective on Windows;
optional, default: false
[fill_color]: see [clip8]; optional, default: (0.0,0.0,0.0)
*)
let gl_pix_of_surface ?(clip8=false) ?(fill_color=(0.0,0.0,0.0)) srf =
	let surf =
		let (width, height, _) = Sdlvideo.surface_dims srf in
		(*widths under windows must be multiples of 8;
		if not: fill with fill_color:*)
		if Sys.os_type = "Win32" && clip8 && width mod 8 > 0  then
		(
			let new_width = width + (8 - (width mod 8)) in
			let sf = Sdlvideo.surface_format srf in
			let s1 = 
				Sdlvideo.create_RGB_surface
					[`SWSURFACE]
					~w:new_width
					~h:height
					~bpp:24
					~rmask:sf.Sdlvideo.rmask
					~gmask:sf.Sdlvideo.gmask
					~bmask:sf.Sdlvideo.bmask
					~amask:sf.Sdlvideo.amask
			in
			Sdlvideo.blit_surface ~src:srf ~dst:s1 ();
			let colint = 
				let iocf c =  int_of_float (c *. 255.) in
				let (rc, gc, bc) = fill_color in
				Sdlvideo.map_RGB srf (iocf rc, iocf gc, iocf bc) 
			in
			Sdlvideo.fill_rect 
				~rect:{Sdlvideo.r_x=width; Sdlvideo.r_y=0; Sdlvideo.r_w=new_width - width; Sdlvideo.r_h=height}
				s1 
				colint;
			s1
		)
		else srf
	in
	let r = Sdlgl.to_raw surf in
	let format = 
		if Sdlvideo.surface_bpp surf > 24 then	(*just an attempt! maybe rather check for > 3 BYTES?*)
			`rgba
		else
			`rgb
	in
	let (w, h, _) = Sdlvideo.surface_dims surf in
	GlPix.of_raw r ~format:format ~width:w ~height:h
;;


(** [image ~filename ~lower_left_corner ()]:
draw an image loaded from file [filename];
[lower_left_corner]: lower left corner, optional, default: (0.0, 0.0)
*)
let image ~filename ?(lower_left_corner=(0.0,0.0)) () =
	let surf = sdl_surface_of_image filename in
	let p = gl_pix_of_surface surf in
	GlPix.raster_pos (fst lower_left_corner) (snd lower_left_corner) ();
	GlPix.draw p;
;;

(** [matrix_of_surface chosen_color s]: 
convert SDL video surface [s] to Bigarray matrix
[chosen_color]: what color to extract ([`blue | `gray | `green | `red ])
*)
let matrix_of_surface chosen_color s =
	let (r, c, _) = Sdlvideo.surface_dims s in
	let m = 	(*this is Gsl_matrix.matrix*)
		Bigarray.Array2.create 
			Bigarray.float64
			Bigarray.c_layout 
			c 
			r
	in
	let col_of_pix i j =
		let (ri, gi, bi) = Sdlvideo.get_pixel_color s ~x:i ~y:j in
		let sc01 x = 
			(float_of_int x) /. 255.
		in
		let (r, g, b) = (sc01 ri, sc01 gi, sc01 bi) in
		match chosen_color with
		[`red] -> r
		| [`green] -> g
		| [`blue] -> b
		| [`gray] -> (r +. g +. b) /. 3.
		| _ -> failwith "not yet implemented"
	in
	for i = 0 to c-1 do
		for j = 0 to r-1 do
			m.{c-1-i, j} <- col_of_pix j i
		done;
	done;
	m
;;

let texture_of_surface s =
	let (sw, sh, _) = Sdlvideo.surface_dims s in
	let format = 
		if Sdlvideo.surface_bpp s > 24 then	(*just an attempt! maybe rather check for > 3 BYTES?*)
			`rgba
		else
			`rgb
	in
	let image = GlPix.of_raw (Sdlgl.to_raw s) ~format:format ~width:sw ~height:sh in
	(*texture dimensions must be powers of 2:*)
	let log2ceil x =
		ceil ((log (float_of_int x)) /. (log 2.0))
	in
	let newdim x =
		int_of_float (2.0 ** (log2ceil x))
	in
	let (tw, th) = (newdim sw, newdim sh) in
	let newimage = GluMisc.scale_image ~width:tw ~height:th image in	(*SEGFAULT!!!*)
	GlPix.store (`unpack_alignment 1);
	let tex = GlTex.gen_texture () in
	GlTex.bind_texture ~target:`texture_2d tex;
	GlTex.image2d newimage;
	List.iter 
		(GlTex.parameter ~target:`texture_2d)
		[ 
			`wrap_s `clamp;
			`wrap_t `clamp;
			`mag_filter `nearest;
			`min_filter `nearest 
		];
	GlTex.env (`mode `decal);
	GlDraw.shade_model `flat;
	tex
;;

(** [texture_image_of_surface ~surface ~pixel_width ~pixel_height ~disable_texture ~lower_left_corner ()]:
draw an sdl surface [surface] as a texture on a rectangle with width and height of the surface;
the advantage over function [image] is that the resulting plot can be rotated;
disadvantageous is the less efficient representation, moreover, the surface has to be rescaled so that
width and height are powers of 2. Do not use it extensively, especially not with large surfaces.
[pixel_width]: width of a single pixel, corresponds to (width of internal scale) / (width in pixel);
[pixel_height]: height of a single pixel, corresponds to (height of internal scale) / (height in pixel);
[disable_texture]: disable the texture after plotting it, optional, default: true;
[lower_left_corner]: lower left corner, optional, default: (0.0, 0.0)
*)
let texture_image_of_surface ~surface ~pixel_width ~pixel_height ?(disable_texture=true) ?(lower_left_corner=(0.0,0.0)) () =
	Gl.enable `texture_2d;
	let tex = texture_of_surface surface in ();
	GlTex.bind_texture ~target:`texture_2d tex;
	let (c1, c2) = lower_left_corner in
	let (sw, sh, _) = Sdlvideo.surface_dims surface in
	let width = (float_of_int sw) *. pixel_width
	and height = (float_of_int sh) *. pixel_height in
	GlDraw.begins `quads;
	GlTex.coord2(0.0, 0.0); GlDraw.vertex2(c1, c2);
	GlTex.coord2(0.0, 1.0); GlDraw.vertex2(c1, c2 +. height);
	GlTex.coord2(1.0, 1.0); GlDraw.vertex2(c1 +. width, c2 +. height);
	GlTex.coord2(1.0, 0.0); GlDraw.vertex2(c1 +. width, c2);
	GlDraw.ends ();
	if disable_texture then
		Gl.disable `texture_2d;
;;

(** [draw_pos list_item position]: call [list_item] from display list and draw it on [position] *)
let draw_pos list_item position =
	GlMat.push ();
	GlMat.translate ~x:(fst position) ~y:(snd position) ();
	GlList.call list_item;
	GlMat.pop ();
;;

