(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*  A modul for 2D plotting written by Tobias Elze    *)
(*  requires lablgl and lablglut    *)
(*  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file ../LICENSE.     *)
(*                                                                     *)
(***********************************************************************)

val filled_rectangle : width:float -> height:float -> ?color:Gl.rgb option -> ?position:Gl.point2 -> unit -> unit
(** [filled_rectangle ~width ~height ~color ~position ()]: draw a filled rectangle with lower left corner [position] (optional arguments, default: (0.0,0.0)), width [width] and height [height], in color [color] (optional, default: None) *)

val rectangle : width:float -> height:float -> ?color:Gl.rgb option -> ?position:Gl.point2 -> unit -> unit
(** [rectangle ~width ~height ~color ~position ()]: draw a rectangle with lower left corner [position] (optional arguments, default: (0.0,0.0)), width [width] and height [height], in color [color] (optional, default: None) *)

val polygon_edges :
  edges:int ->
  radius:float ->
  ?angle:float -> 
  ?center:float * float -> unit -> (float * float) array 
(**
[polygon_edges ~edges ~radius ~angle ~color ~center ()]
return the edges of a regular polygon as an array of point2 (float*float)
[edges]: number of edges
[radius]: distance between the center and each of the edges
[angle]: rotation angle
angle = 0: lowest edge of the polygon is vertical
[center]: center point of the polygon (optional, default: (0.0,0.0))
*)
  
val regular_polygon :
  edges:int ->
  radius:float -> 
  ?angle:float ->
  ?color:Gl.rgb option ->
  ?position:float * float -> unit -> unit
(**
[regular_polygon ~edges ~radius ~angle ~color ~position ()]
draw a regular polygon
[edges]: number of edges
[radius]: distance between the center and each of the edges
[angle]: rotation angle
angle = 0: lowest edge of the polygon is vertical
[color]: drawing color (optional, default: None)
[position]: center point of the polygon (optional, default: (0.0,0.0))
*)
  
val filled_regular_polygon :
  edges:int ->
  radius:float -> 
  ?angle:float ->
  ?color:Gl.rgb option ->
  ?position:float * float -> unit -> unit
(**
[filled_regular_polygon ~edges ~radius ~angle ~color ~position ()]
draw a filled regular polygon
[edges]: number of edges
[radius]: distance between the center and each of the edges
[angle]: rotation angle
angle = 0: lowest edge of the polygon is vertical
[color]: drawing color (optional, default: None)
[position]: center point of the polygon (optional, default: (0.0,0.0))
*)

val circle : radius:float -> ?color:Gl.rgb option -> ?position:float * float -> unit -> unit
(** [circle ~radius ~color ~position ()]: draw a circle. 
[color]: drawing color (optional, default: None)
[position] center, is optional, default: (0.0,0.0) *)

val filled_circle : radius:float -> ?color:Gl.rgb option -> ?position:float * float -> unit -> unit
(** [filled_circle ~radius ~color ~position ()]: draw a filled circle. 
[color]: drawing color (optional, default: None)
[position] center, is optional, default: (0.0,0.0) *)

val partial_disk : inner:float -> outer:float -> start:float -> sweep:float -> ?color:Gl.rgb option -> ?position:float * float -> unit -> unit
(** [partial_disk ~inner ~outer ~start ~sweep ~color ~position ()]: draw a partial disk
with inner radius [inner],
outer radius [outer],
start angle [start],
and sweep angle [sweep]
in color [color] (optional, default: None).
[position]: center, is optional, default: (0.0,0.0)
 *)

val primitive : Gl.rgb option -> GlDraw.shape -> Gl.point2 array  -> unit
(** [primitive color shp points]: draw a geometric primitive
in color [color] (optional, default: None);
use [GlDraw.vertex2] to draw the array of points2 [points]
applying shape [shp] ([`line_loop|`line_strip|`lines|`points|`polygon|`quad_strip|`quads
  |`triangle_fan|`triangle_strip|`triangles])
*)

val text_line : text:string -> font:Glut.font_t -> ?color:Gl.rgb option -> ?position:float * float -> unit -> unit
(** [text_line ~text ~font ~color ~position]: display text string [text] with font [font]
in color [color] (optional, default: None)
at position [position] (optional, default: (0.0,0.0)) *)

val text_field : text:string -> font:Glut.font_t -> width:float -> height:float -> pixel_width:float -> pixel_height:float -> alignment:string -> draw_frame:bool -> ?color:Gl.rgb option -> ?position:float * float -> unit -> unit
(** [text_field ~text ~font ~width ~height ~pixel_width ~pixel_height ~alignment ~draw_frame ~position]:
display text string [text] in font [font] formatted in a textfield with width [width] and height [height];
[pixel_width]: width of a single pixel, corresponds to (width of internal scale) / (width in pixel);
[pixel_height]: height of a single pixel, corresponds to (height of internal scale) / (height in pixel);
[alignment]: alignment of the text, possible values: "Left", "Right", "Center";
[draw_frame]: whether to draw a frame around the field or not;
[color]: text color (optional, default: None).
[position]: position of the lower left corner of the text field; optional value, default: (0.0, 0.0)
text string [text] may contain line break symbols (\n).

Remark: if you use fullscreen mode and the fullscreen resolution differs from your specified resolution 
then formatting problems can occor
*)

val button : width:float -> height:float -> text:string -> font:Glut.font_t -> pixel_width:float -> pixel_height:float -> ?color_range:float * float -> ?shift_text_pos:float * float -> ?position:float * float -> unit -> unit
(** [button ~width ~height ~text ~font ~pixel_width ~pixel_height ~color_range ~shift_text_pos ~position]: 
draw button of height [height] and width [width] 
with textstring [text] in font [font]
[pixel_width]: width of a single pixel, corresponds to (width of internal scale) / (width in pixel);
[pixel_height]: height of a single pixel, corresponds to (height of internal scale) / (height in pixel);
[color_range]: pair representing the darkest and the brightest gray value of the button; 
optional value, default: (0.7, 0.95)
[shift_text_pos]: pair representing possible shifts in x and y directions
from automatically calculated text position; optional value, default: (0.0, 0.0)
[position]: position of the lower left corner of the text field; optional value, default: (0.0, 0.0)
*)

val pressed_button : width:float -> height:float -> text:string -> font:Glut.font_t -> pixel_width:float -> pixel_height:float -> ?color_range:float * float -> ?position:float * float -> unit -> unit
(** [pressed_button ~width ~height ~text ~font ~color_range ~position]: 
same as button, but with inverse default color_range and shifted text position
*)

val mouseover_button : width:float -> height:float -> text:string -> font:Glut.font_t -> pixel_width:float -> pixel_height:float -> ?color_range:float * float -> ?position:float * float -> unit -> unit
(** [mouseover_button ~width ~height ~text ~font ~pixel_width ~pixel_height ~color_range ~position]: 
same as button, but with brighter color as default (0.75, 1.0)
*)

val sdl_surface_of_image : string -> Sdlvideo.surface
(** [sdl_surface_of_image filename]: convert image from file [filename] to SDL surface, either in rgb or in rgba
*)

val gl_pix_of_surface : ?clip8:bool -> ?fill_color:Gl.rgb -> Sdlvideo.surface -> ([< Gl.format > `rgb `rgba ], [ `ubyte ]) GlPix.t
(** [gl_pix_of_surface ~clip8 ~fill_color surf]: convert SDL surface [surf] to GlPix.t, 
either RGB or RGBA
[clip8]: if width no multiple of 8, increase width by filling the image with [fill_color];
this might be necessary on MS Windows so only effective on Windows;
optional, default: false
[fill_color]: see [clip8]; optional, default: (0,0,0)
*)

val image : filename:string -> ?lower_left_corner:float * float -> unit -> unit
(** [image ~filename ~lower_left_corner]:
draw an image loaded from file [filename];
[lower_left_corner]: lower left corner, optional, default: (0.0, 0.0)
*)

val matrix_of_surface : [> `blue | `gray | `green | `red ] list -> Sdlvideo.surface -> (float, Bigarray.float64_elt, Bigarray.c_layout) Bigarray.Array2.t
(** [matrix_of_surface chosen_color s]: 
convert SDL video surface [s] to Bigarray matrix
[chosen_color]: what color to extract ([`blue | `gray | `green | `red ])
*)

val texture_image_of_surface : surface:Sdlvideo.surface -> pixel_width:float -> pixel_height:float -> ?disable_texture:bool -> ?lower_left_corner:float * float -> unit -> unit
(** [texture_image_of_surface ~surface ~pixel_width ~pixel_height ~disable_texture ~lower_left_corner ()]:
draw an sdl surface [surface] as a texture on a rectangle with width and height of the surface;
the advantage over function [image] is that the resulting plot can be rotated;
disadvantageous is the less efficient representation, moreover, the surface has to be rescaled so that
width and height are powers of 2. Do not use it extensively, especially not with large surfaces.
[pixel_width]: width of a single pixel, corresponds to (width of internal scale) / (width in pixel);
[pixel_height]: height of a single pixel, corresponds to (height of internal scale) / (height in pixel);
[disable_texture]: disable the texture after plotting it, optional, default: true;
[lower_left_corner]: lower left corner, optional, default: (0.0, 0.0)
*)

val draw_pos : GlList.t -> float * float -> unit
(** [draw_pos list_item position]: call [list_item] from display list and draw it on [position] *)
