(*
	use the script ./run to execute this program in interpreter lablglut
*)


(*directory of flashdot executable:*)
let executable_dir = 
	let exname = Sys.executable_name in
	Filename.dirname exname
;;
let flashdot_version = 
	let versionfile = Filename.concat executable_dir "flashdot_version" in
	let vf = open_in versionfile in
	let version = input_line vf in
	close_in vf;
	version
;;
(*experiments.dtd in directory of the running binary:*)
let default_dtd_file =
	Filename.concat executable_dir "experiments.dtd"
;;

(**************************
	Type definitions: 
**************************)

(*constants or variables:*)
type number = C of float | V of (float ref) | F of (unit -> float);;	(*V superfluous?*)
type matrix = MC of Gsl_matrix.matrix | MF of (unit -> Gsl_matrix.matrix);;
type string_or_string_function = SC of string | SF of (unit -> string);;

type input_mode = Stdin_fls | Stdin_fld | Input_file of string;;
type error_mode = Err_graphical | Err_file_handler of out_channel;;

type simple_display_mode =
{
	fullscreen: bool;
	window_title: string;
	window_position: int * int
};;
type enhanced_display_mode =
{
	color_depth: int;
	refresh_rate: int
};;
type disp_mode = Simple_display_mode of simple_display_mode | Enhanced_display_mode of enhanced_display_mode;;
type display_properties =
{
	mode: disp_mode;
	mutable width: int;
	mutable height: int;
	horizontal_range: float * float;
	vertical_range: float * float
};;

type rotation_t =
{
	rotation_angle: number;
	rotation_center: number * number
};;
type stimulus =
{
	position: (number * number);(* V *)
	color: number list;
	rotation: rotation_t option;	
	scale_xyz: (number list) option;
	display_stimulus: number;
	content_list: (unit -> unit) list;
	mutable content_compiled: GlList.t	(*necessary since compilation has to be done relatively late*)
};;

type current_status = 
{
	sceneID: string;
	duration: number;
	duration_in_frames: bool;
	wake_up_on_event_mode: bool
};;

type action_t = 
	Nothing 
	| Exit 
	| Jump_to_next_presentation 
	| Insert_sequence of (unit -> (current_status Queue.t))
	| Insert_sequence_stop of (unit -> (current_status Queue.t))
;;
type event_component_to_set = 
	Set_by_mouse_position_x of (float -> unit)
	| Set_by_mouse_position_y of (float -> unit)
	| Set_by_time_from_onset of (float -> unit)
	| Set_to_number of (unit -> unit)
;;

type image =
{
	image_position: (number * number);
	image_scale: (number * number);
	image_width: float;
	image_height: float;
	image_flip_horizontally: bool;
	image_flip_vertically: bool;
	display_image: number;
	image_content_surface: (Sdlvideo.surface ref);
	image_content_gl_pix: ( (([`rgb | `rgba], [`ubyte]) GlPix.t) ref)
};;

type button_state_t = Pressed | Unpressed | Mouseover;;
type button = 
{
	buttonID: string;
	mutable button_state: button_state_t;
	button_log_event: number;
	button_event_logfile_id: string;
	button_action: action_t;
	button_variables_to_set: event_component_to_set list;	(*this is also for matrices to modify*)
	(*button_matrices_to_modify: ((unit -> (int * int)) * event_component_to_set) list;*)
	button_variables_to_log: (string * (unit -> string)) list;
	button_matrices_to_log: (string * (unit -> Gsl_matrix.matrix)) list;
	button_index_vectors_to_log: (string * (string list)) list;
	button_list_indices_to_log: (string * number * (number -> string)) list;
	button_log_before_set: bool;
	button_sensitive_area: float * float -> bool;
	button_content: button_state_t -> GlList.t;
	mutable button_pressed_compiled: GlList.t;
	mutable button_unpressed_compiled: GlList.t;
	mutable button_mouseover_compiled: GlList.t;
};;



type keyboard_event = Ascii_key of char | Special_key of Glut.special_key_t;;
type mouse_event = 
{
	button: Glut.button_t;
	up_or_down: Glut.mouse_button_state_t;
	sensitive_area: float * float -> bool
};;
type event_t = Keyboard_event of keyboard_event | Mouse_event of mouse_event | Button_event of string;;
type log_format_t = XML | CommaSeparatedValues | SpaceSeparatedValues;;
type log_t =
{
	event_logfile: string;
	event_logfile_header: string option;
	log_key: bool;
	log_time: bool;
	log_mouse_position: bool;
	log_format: log_format_t;
	log_user_declared_variables: bool;
	log_user_declared_matrices: bool;
	log_list_index_vectors: bool;
	log_any: bool	(*if anything is logged at all*)
};;

type event =
{
	event_logfile_id: string;
	event_contents: event_t;
	event_id: string option;
	log_event: number;
	action: action_t;
	variables_to_set: event_component_to_set list;	(*this is also for matrices to modify*)
	variables_to_log: (string * (unit -> string)) list;
	matrices_to_log: (string * (unit -> Gsl_matrix.matrix)) list;
	index_vectors_to_log: (string * (string list)) list;
	list_indices_to_log: (string * number * (number -> string)) list;
	log_before_set: bool
}


type display_object = 
	Stimulus_scenewide of stimulus 
	| Stimulus_uncompiled of (unit ->  stimulus)
	| Image of image		(*fix image*)
	| Image_scenewide of (unit -> image) * ((image option) ref)
	| Image_uncompiled of (unit -> image)
	| Button of button
	| Stimulus_list_reference of object_list_reference
	| Image_list_reference of object_list_reference
and object_list_reference =
{
	object_stub: display_object;
	frames_per_item: number;
	object_list_repeat: bool;
	object_reference_array: (display_object ref) array
};;
type declared_display_object =
	Display_object of display_object
	| Stimulus_list of (string * display_object) list
	| Image_list of (string * display_object) list
;;
type variation = 
{
	number_of_frames: number option;
	variation_function: int -> int -> unit
};;
type scene = 
{
	background_color: number list;
	cursor_type: Glut.cursor_t;
	display_object_list: display_object list;
	event_list: event list;
	local_declarations: (string, Mathexpr.declaration) Hashtbl.t;
	variation_list: variation list
};;

type deviation =
{
	measured_duration: int;
	global_frame_number: int;
	scene_frame_number: int;
	concerned_scene: string
};;
type logged_event = 
{
	logged_id: string option;
	logged_logfile_id: string;
	logged_contents: event_t;
	logged_time: int;
	logged_mouse_position: float * float;
	logged_variables: (string * string) list;
	logged_matrices: (string * Gsl_matrix.matrix) list;
	logged_index_vectors: (string * (string list)) list
};;

type experiment =
{
	display: display_properties;
	scene_list: ( string * scene ) list;
	sequence: current_status Queue.t;
	logfile: string;
	check_framerate: bool;
	event_log: (string * log_t) list;	(*ID * log_t*)
	deviation_queue: deviation Queue.t;
	event_queue: logged_event Queue.t
};;

type os = Linux | MacOSX | Windows;;

let operating_system = match Sys.os_type with
	| "Unix" ->
	(
		let p = Unix.open_process_in "uname" in
		let os = input_line p in
		ignore (Unix.close_process_in p);
		match os with | "Darwin" -> MacOSX | _ -> Linux
	)
	| _ -> Windows
;;

(*default setting of MacOS X sync to VBL:*)
let macosx_default_sync_var = ref "false";;

(* open external proces pstring, scan its output for pattern and return first matched group: *)
let pattern_of_input_process pstring pattern =
	try
	let p = Unix.open_process_in pstring in
	let rec find_pat () =
		let line = input_line p in
		if Str.string_match pattern line 0 then
			Str.matched_group 1 line
		else find_pat ()
	in
	let pat = find_pat () in
	ignore (Unix.close_process_in p);
	pat;
	with _ -> ""
;;

(*synchronization to VBLANK, currently only for Linux and MacOSX:*)
match operating_system with 
	| Linux -> 
		Unix.putenv "__GL_SYNC_TO_VBLANK" "1"; 
		Unix.putenv "LIBGL_SYNC_REFRESH" "1"
	| MacOSX ->
		(*read defaults if com.apple.glut exists:*)
		if pattern_of_input_process "defaults domains" (Str.regexp ".*\\(com.apple.glut\\).*") <> "" then
		(macosx_default_sync_var := 
		(
			let bstr = pattern_of_input_process 
				"defaults read com.apple.glut GLUTSyncToVBLKey" 
				(Str.regexp "\\([01]\\)")
			in
			string_of_bool (bstr = "1");
		));
		ignore (Sys.command "defaults write com.apple.glut GLUTSyncToVBLKey -bool true");
	| _ -> ()
;;



(*global flag, set true when glut init is called:*)
let glut_initialized = ref false;;
(*global flag for the idle function that is called if nothing is to do:*)
let idle_fun = ref None;;
(*reference to logfile; if not empty then overwrite logfile by command line argument -logfile*)
let log_file = ref "";;
(*global flag, set true if execution of external processes is permitted:*)
let allow_external_execution = ref false;;
(*standard output format for errors:*)
let standard_error_output = 
	ref (if Sys.os_type = "Unix" then Err_file_handler stderr else Err_graphical)
;;
(*ID for default list item identifiers*)
let global_list_item_id = ref 0;;

(*global display object hash table (entries: string * declared_display_object):*)
let global_display_objects = Hashtbl.create 10;;
let local_display_objects = ref (Hashtbl.create 10);;
let get_declared_contents identifier =
	try
	Hashtbl.find !local_display_objects identifier
	with Not_found ->
	(
		try
		Hashtbl.find global_display_objects identifier
		with Not_found -> raise (Mathexpr.Undeclared_identifier identifier)
	)
;;

(*
	Global declarations:
	
	(only for enhanced display mode:)
	( "ENHANCED_DISPLAY_MODE_COLOR_DEPTH", Mathexpr.Constant ),
	( "ENHANCED_DISPLAY_MODE_REFRESH_RATE", Mathexpr.Constant )
	
	(general:)
	( "PI", Mathexpr.Constant ),
	( "CANVAS_WIDTH", Mathexpr.Constant ),
	( "CANVAS_HEIGHT", Mathexpr.Constant ),
	( "HORIZONTAL_RANGE_MIN", Mathexpr.Constant ),
	( "HORIZONTAL_RANGE_MAX", Mathexpr.Constant ),
	( "VERTICAL_RANGE_MIN", Mathexpr.Constant ),
	( "VERTICAL_RANGE_MAX", Mathexpr.Constant ),
	
	(presentationwide:)
	( "FRAMES_LEFT", Mathexpr.Variable )
*)

(*create/replace a global declaration:*)
let declare =
	Hashtbl.add Mathexpr.global_declarations
;;
let replace_declaration =
	Hashtbl.replace Mathexpr.global_declarations
;;

declare "FLASHDOT_VERSION" (Mathexpr.String_variable (ref flashdot_version));;


(*set and unset the reference Mathexpr.local_delcarations:*)
let set_mathexpr_local_declarations locdec =
	Mathexpr.local_declarations := locdec
;;
let unset_mathexpr_local_declarations () =
	Mathexpr.local_declarations := (Hashtbl.create 1)
;;

(*declare "PI" (Mathexpr.Constant (acos (-1.)));;*)

let remove_spaces = Str.global_replace (Str.regexp " ") "";;
let remove_leading_and_trailing_spaces s =
	let no_leading_spaces = Str.global_replace (Str.regexp "^ +") "" s in
	Str.global_replace (Str.regexp " +$") "" no_leading_spaces
;;

let number_of_string s =
	let (st, is_lazy) =
		(*prefix '#' => create a lazy expression*)
		if s.[0] = '#' then
			(Str.string_after s 1, true)
		else
			(s, false)
	in
	let x = List.hd (Fls.xml_list_of_fls_string st) in
	let f = Mathexpr.function_of_xml "" x in
	if is_lazy then
		F (fun () -> Mathexpr.float_of_nom (f [||]))
	else
		C (Mathexpr.float_of_nom (f [||]))
(*
	let gd identifier =
		try
		Hashtbl.find Mathexpr.global_declarations identifier
		with Not_found -> raise (Mathexpr.Undeclared_identifier identifier)
	in
	try
	C (float_of_string s)
	with Failure t ->
	(
		match Mathexpr.get_declaration s with
		Mathexpr.Constant c -> C c
		| Mathexpr.Variable v -> V v
		| Mathexpr.Function f -> 
			if f.Mathexpr.nargs > 0 then
				raise (Mathexpr.Invalid_declaration_type s)
			else
				F (fun () -> f.Mathexpr.fcontent [||])
		| _ -> raise (Mathexpr.Invalid_declaration_type s)
	)
*)
;;
let string_or_string_fun_of_string s =
	let (st, is_lazy) =
		(*prefix '#' => create a lazy expression*)
		if s.[0] = '#' then
			(Str.string_after s 1, true)
		else
			(s, false)
	in
	let x = List.hd (Fls.xml_list_of_fls_string st) in
	let f = Mathexpr.function_of_xml "" x in
	if is_lazy then
		SF (fun () -> Mathexpr.string_of_nom (f [||]))
	else
		SC (Mathexpr.string_of_nom (f [||]))
;;
let matrix_of_string s =
	let (st1, is_lazy) =
		(*prefix '#' => create a lazy expression*)
		if s.[0] = '#' then
			(Str.string_after s 1, true)
		else
			(s, false)
	in
	(*if its only a matrix identifier, it has to be converted to fls first:*)
	let st = 
		if st1.[0] == '(' then
			st1
		else
			"(DeclaredMatrix [Identifier=\"" ^ st1 ^ "\", ModifyInSitu=\"true\"])"
	in
	let x = List.hd (Fls.xml_list_of_fls_string st) in
	let f = Mathexpr.function_of_xml "" x in
	if is_lazy then
		MF (fun () -> Mathexpr.matrix_of_nom (f [||]))
	else
		MC (Mathexpr.matrix_of_nom (f [||]))
;;

(*split string s into single expressions;
use ',' as separator unless it occurs inside an expression:*)
let separate_expressions s =
	let strlist = List.map remove_leading_and_trailing_spaces (Str.split (Str.regexp ",") s) in
	let correct_enclosing_parenths ss = 
		String.length ss < 2 
		|| (ss.[0] = '(' || ss.[1] = '(') = (ss.[(String.length ss) - 1] = ')')
	in
	let rec make_list testlist = 
		let rec check_list correct_list = function
			| [] -> List.rev correct_list
			| h1::h2::t -> 
				if (correct_enclosing_parenths h1) then
					check_list (h1::correct_list) (h2::t)
				else
					make_list (correct_list @ ((h1 ^ "," ^ h2) :: t))
			| h::t -> check_list (h::correct_list) t
		in
		check_list [] testlist
	in
	make_list strlist
;;

let number_list_of_string s =
	let stringlist = separate_expressions s in
	List.map number_of_string stringlist
;;
let string_or_string_fun_list_of_string s =
	let stringlist = separate_expressions s in
	List.map string_or_string_fun_of_string stringlist
;;
(*convert color names to RGB:*)
let color_string = function
	| "white" -> "1.0,1.0,1.0"
	| "black" -> "0.0,0.0,0.0"
	| "red" -> "1.0,0.0,0.0"
	| "green" -> "0.0,1.0,0.0"
	| "blue" -> "0.0,0.0,1.0"
	| "cyan" -> "0.0,1.0,1.0"
	| "magenta" -> "1.0,0.0,1.0"
	| "yellow" -> "1.0,1.0,0.0"
	| s -> s
;;
(*list of 3 numbers representing a color:*)
let color_list_of_string s =
	let l = 
		try number_list_of_string s with
		Mathexpr.Undeclared_identifier idf -> (*this can't happen anymore...*)
			failwith ("Color attribute " ^ s ^ " contains undeclared identifier: \"" ^ idf ^ "\". Check the spelling. Maybe you want to use a number instead?") 
		| Mathexpr.Invalid_declaration_type dt ->
			failwith ("Color attribute " ^ s ^ " contains invalid declaration type. Identifier \"" ^ dt ^ "\" is a function with more than zero arguments.") 
	in
	if (List.length l) = 3 then
		l
	else
		failwith ("Color attribute " ^ s ^ " is not a list of three elements")
;;
let float_of_number = function
	| C c -> c
	| V v -> !v
	| F f -> f ()
;;
let int_of_number n =
	int_of_float (float_of_number n)
;;
let float_of_string' s =
	float_of_number (number_of_string s)
;;
(*declared matrices:*)
let get_declared_matrix s =
	match Mathexpr.get_declaration s with
	Mathexpr.Matrix m -> MC m
	| Mathexpr.Matrix_function mf ->
		if mf.Mathexpr.mat_nargs > 0 then
			raise (Mathexpr.Invalid_declaration_type s)
		else
			MF (fun () -> mf.Mathexpr.mat_fcontent [||])
	| _ -> raise (Mathexpr.Invalid_declaration_type s)
;;


(*global duration counter until next switch of presentations:*)
declare "FRAMES_LEFT" (Mathexpr.Variable (ref 1.0));;

let frames_left () = 
	match (Hashtbl.find Mathexpr.global_declarations "FRAMES_LEFT") with
	Mathexpr.Variable v -> (int_of_float (floor !v))
	| _ -> 0
;;
let set_frames_left n =
	match (Hashtbl.find Mathexpr.global_declarations "FRAMES_LEFT") with
	Mathexpr.Variable v -> v := (float_of_int n)
	| _ -> ()
;;

(*set a declared variable with identifier var_id to value newval:*)
let set_declared_variable var_id newval =
	match (Hashtbl.find Mathexpr.global_declarations var_id) with
	Mathexpr.Variable v -> v := newval
	| _ -> ()
;;

module Events =
	struct
		let lastscene = ref None;;
		
		(*return current scene of experiment ex:*)
		let current_scene ex =
			if Queue.is_empty ex.sequence then
			(
				match !lastscene with
				Some e -> List.assoc e.sceneID ex.scene_list
				| _ -> failwith "No event left"
			)
			else
			(
				let curr_state = Queue.peek ex.sequence in
				List.assoc curr_state.sceneID ex.scene_list
			)
		;;

		(*transform mouse position to "internal" coordinate system:*)
		let transform_mouse_pos width height hor_range vert_range mp =
			let (xold, yold) = (float_of_int (fst mp), float_of_int (snd mp))
			and (w, h) = (float_of_int width, float_of_int height)
			and (xmin, xmax) = hor_range
			and (ymin, ymax) = vert_range in
			let xnew = xmin +. xold *. (xmax -. xmin) /. w
			and ynew = ymax +. (ymin -. ymax) *. yold /. h in
			(xnew, ynew)
		;;
		
		(*perform action act in experiment ex:*)
		let perform_action ex act =
			(*insert queue_to_insert into q after the first element of q;*)
			let insert_queue queue_to_insert q =
				let q1 = Queue.copy queue_to_insert in
				let first_element = Queue.pop q in
				Queue.transfer q q1;
				Queue.push first_element q;
				Queue.transfer q1 q
			in
			Glut.idleFunc !idle_fun;
			match act with
			Nothing -> ();
			| Exit -> 
				if (not (Queue.is_empty ex.sequence)) then
					lastscene := Some (Queue.peek ex.sequence);
				Queue.clear ex.sequence; 
			| Jump_to_next_presentation -> 
				(*let curr_state = Queue.peek ex.sequence in*)
				set_frames_left 1;	(*stop presentation after next frame*)
			| (Insert_sequence queue_to_insert) -> 
				insert_queue (queue_to_insert ()) ex.sequence
			| (Insert_sequence_stop queue_to_insert) -> 
				insert_queue (queue_to_insert ()) ex.sequence;
				set_frames_left 1;	(*stop presentation after next frame*)
		;;
		
		(*set variables which have to be set when events are elicited;
		local_decl: local declarations of current scene*)
		let set_variables_by_event varlist local_decl mouse_position duration =
			set_mathexpr_local_declarations local_decl;
			(*print_float !(fst (List.hd varlist));
			Hashtbl.iter 
				(
					fun k i ->
						let v = (match i with
						Mathexpr.Constant c -> c
						| Mathexpr.Variable v -> !v 
						| _ -> 0.) in
						Printf.printf "%s -> %f\n" k v
				) 
				Mathexpr.global_declarations;
			print_newline ();*)
			let set_var = function
				| Set_by_mouse_position_x f -> f (fst mouse_position)
				| Set_by_mouse_position_y f -> f (snd mouse_position)
				| Set_by_time_from_onset f -> f (float_of_int duration)
				| Set_to_number f -> f ()
			in
			List.iter set_var varlist;
			unset_mathexpr_local_declarations ()
		;;
		(*generate the list of variables to be logged by the event:*)
		let log_variables_by_event varlist local_decl =
			set_mathexpr_local_declarations local_decl;
			let log_var v =
				(fst v, (snd v) ())
			in
			let lv = List.map log_var varlist in
			unset_mathexpr_local_declarations ();
			lv
		;;
		(*generate the list of matrices to be logged by the event:*)
		let log_matrices_by_event matlist local_decl =
			set_mathexpr_local_declarations local_decl;
			let log_mat m =
				(fst m, Gsl_matrix.copy ((snd m) ()))
			in
			let lv = List.map log_mat matlist in
			unset_mathexpr_local_declarations ();
			lv
		;;
		(*generate the list of index vector strings from index_vectors and list_indices to be logged by the event:*)
		let log_index_vectors_by_event ivlist indices_list local_decl =
			set_mathexpr_local_declarations local_decl;
			let li_list =
				List.map
					(
						fun (listid, n, fct) -> (listid, [fct n])
					)
					indices_list
			in
			set_mathexpr_local_declarations local_decl;
			li_list @ ivlist
		;;
		
		let bool_of_num n =
			float_of_number n <> 0.0
		;;
		
		(*process this when key (not special key) pressed; x, y: mouse position:*)
		let keyboard_action ex onset ~key ~x ~y =
			let duration = (Glut.get ~gtype:Glut.ELAPSED_TIME) - !onset in
			let pressed_key = char_of_int key in
			let pressed_key_in_list e =
				match e.event_contents with
				Keyboard_event (Ascii_key k) -> k = pressed_key
				| _ -> false
			in
			let sc = current_scene ex in
			let a = List.filter pressed_key_in_list sc.event_list in
			if List.length a > 0 then 
			(
				if not (List.hd a).log_before_set then
					set_variables_by_event (List.hd a).variables_to_set sc.local_declarations (0.0, 0.0) duration;
				if bool_of_num (List.hd a).log_event then
				(
					let lg = {
						logged_id = (List.hd a).event_id;
						logged_logfile_id = (List.hd a).event_logfile_id;
						logged_contents = 
							Keyboard_event (Ascii_key pressed_key);
						logged_time = duration;
						logged_mouse_position = (0.0, 0.0);	(*mouse position irrelevant for keyboard events*)
						logged_variables = 
							log_variables_by_event (List.hd a).variables_to_log sc.local_declarations;
						logged_matrices = 
							log_matrices_by_event (List.hd a).matrices_to_log sc.local_declarations;
						logged_index_vectors = 
							log_index_vectors_by_event 
								(List.hd a).index_vectors_to_log 
								(List.hd a).list_indices_to_log
								sc.local_declarations
					} in
					Queue.push lg ex.event_queue;
				);
				perform_action ex (List.hd a).action;
				if (List.hd a).log_before_set then
					set_variables_by_event (List.hd a).variables_to_set sc.local_declarations (0.0, 0.0) duration;
			)
		;;
		(*process this when special key pressed; x, y: mouse position:*)
		let special_key_action ex onset ~key ~x ~y =
			let duration = (Glut.get ~gtype:Glut.ELAPSED_TIME) - !onset in
			let pressed_key_in_list e =
				match e.event_contents with
				Keyboard_event (Special_key k) -> k = key
				| _ -> false
			in
			let sc = current_scene ex in
			let a = List.filter pressed_key_in_list sc.event_list in
			if List.length a > 0 then 
			(
				if not (List.hd a).log_before_set then
					set_variables_by_event (List.hd a).variables_to_set sc.local_declarations (0.0, 0.0) duration;
				if bool_of_num (List.hd a).log_event then
				(
					let lg = {
						logged_id = (List.hd a).event_id;
						logged_logfile_id = (List.hd a).event_logfile_id;
						logged_contents = 
							Keyboard_event (Special_key key);
						logged_time = duration;
						logged_mouse_position = (0.0, 0.0);	(*mouse position irrelevant for keyboard events*)
						logged_variables = 
							log_variables_by_event (List.hd a).variables_to_log sc.local_declarations;
						logged_matrices = 
							log_matrices_by_event (List.hd a).matrices_to_log sc.local_declarations;
						logged_index_vectors = 
							log_index_vectors_by_event
								(List.hd a).index_vectors_to_log
								(List.hd a).list_indices_to_log
								sc.local_declarations
					} in
					Queue.push lg ex.event_queue;
				);
				perform_action ex (List.hd a).action;
				if (List.hd a).log_before_set then
					set_variables_by_event (List.hd a).variables_to_set sc.local_declarations (0.0, 0.0) duration;
			)
		;;
		
		(*process this when mouse button pressed or released:*)
		let mouse_action ex onset ~button ~state ~x ~y =
			let duration = (Glut.get ~gtype:Glut.ELAPSED_TIME) - !onset in
			let d = ex.display in
			let tpos = transform_mouse_pos d.width d.height d.horizontal_range d.vertical_range (x, y) in
			let button_pressed_inside_sensitive_area b =
				match b.event_contents with
				Mouse_event p -> 
					p.button = button && p.up_or_down = state && (p.sensitive_area tpos)
				| _ -> false
			in
			let sc = current_scene ex in
			let a = List.filter button_pressed_inside_sensitive_area sc.event_list in
			if List.length a > 0 then 
			(
				if not (List.hd a).log_before_set then
					set_variables_by_event (List.hd a).variables_to_set sc.local_declarations tpos duration;
				if bool_of_num (List.hd a).log_event then
				(
					let lg = {
						logged_id = (List.hd a).event_id;
						logged_logfile_id = (List.hd a).event_logfile_id;
						logged_contents = 
							Mouse_event {
								button = button;
								up_or_down = state;
								sensitive_area = fun x -> true	(*irrelevant*)
							};
						logged_time = duration;
						logged_mouse_position = tpos;
						logged_variables = 
							log_variables_by_event (List.hd a).variables_to_log sc.local_declarations;
						logged_matrices = 
							log_matrices_by_event (List.hd a).matrices_to_log sc.local_declarations;
						logged_index_vectors = 
							log_index_vectors_by_event 
								(List.hd a).index_vectors_to_log
								(List.hd a).list_indices_to_log
								sc.local_declarations
					} in
					Queue.push lg ex.event_queue;
				);
				perform_action ex (List.hd a).action;
				if (List.hd a).log_before_set then
					set_variables_by_event (List.hd a).variables_to_set sc.local_declarations tpos duration;
			);
			
			(**** button events: ****)
			
			if button = Glut.LEFT_BUTTON then
			(
				let buttonlist = List.filter (fun x -> match x with (Button b) -> true |_ -> false) sc.display_object_list in
				if List.length buttonlist > 0 then
				(
					let buttons = List.map (fun x -> match x with (Button b) -> b | _ -> failwith "Error") buttonlist in
					
					let pressed_buttons = List.filter (fun b -> b.button_sensitive_area tpos) buttons in
					if state = Glut.UP then
					(
						(*check for button actions (whether mouse is released on sensitive area):*)
						if List.length pressed_buttons > 0 then
						(
							let b = List.hd pressed_buttons in
							if b.button_state = Pressed then
							(
								if not b.button_log_before_set then
									set_variables_by_event 
										b.button_variables_to_set 
										sc.local_declarations 
										tpos 
										duration;
								if bool_of_num b.button_log_event then
								(
									let lg = {
										logged_id = None;
										logged_logfile_id = b.button_event_logfile_id;
										logged_contents = (Button_event b.buttonID);
										logged_time = duration;
										logged_mouse_position = tpos;
										logged_variables = 
											log_variables_by_event b.button_variables_to_log sc.local_declarations;
										logged_matrices = 
											log_matrices_by_event b.button_matrices_to_log sc.local_declarations;
										logged_index_vectors = 
											log_index_vectors_by_event
												b.button_index_vectors_to_log
												b.button_list_indices_to_log
												sc.local_declarations
									} in
									Queue.push lg ex.event_queue;
								);
								perform_action ex b.button_action;
								if b.button_log_before_set then
									set_variables_by_event 
										b.button_variables_to_set 
										sc.local_declarations 
										tpos 
										duration;
							);
							(*if mouse button released all displayed buttons are unpressed:*)
							List.iter (fun b -> b.button_state <- Unpressed) buttons;
						)
					)
					else
					(
						if List.length pressed_buttons > 0 then
						(
							let b = List.hd pressed_buttons in
							b.button_state <- Pressed
						)
					)
				)
			)
		;;
		
		(*run this if mouse is moved without a mouse button being pressed:*)
		let mouse_moving ex ~x ~y =
			let sc = current_scene ex in
			let buttonlist = List.filter (fun x -> match x with (Button b) -> true |_ -> false) sc.display_object_list in
			if List.length buttonlist > 0 then
			(
				let d = ex.display in
				let tpos = transform_mouse_pos d.width d.height d.horizontal_range d.vertical_range (x, y) in
				let buttons = List.map (fun x -> match x with (Button b) -> b | _ -> failwith "Error") buttonlist in
				let highlight_mouseover_button b =
					if (b.button_sensitive_area tpos) then
						b.button_state <- Mouseover
					else
						b.button_state <- Unpressed
				in
				List.iter highlight_mouseover_button buttons;	
			)
		;;
		
	end
;;

module Display =
	struct
		let frames = ref 0;;
		let scene_frames = ref 1;;
		let current_frame = ref 0;;
		let mean = ref 0.0;;
		let std = ref 0.0;;
		let scene_onset = ref 0;;
		
		
		let frame_duration () =
			let t = Glut.get ~gtype:Glut.ELAPSED_TIME in
			let d = t - !current_frame in
			current_frame := t;
			d
		;;
		
		(*estimate frame rate based on avarage duration of 100 frames:*)
		let estimate_frame_rate () =
			let a = Array.make 100 0.0 in
			let rec draw_blank_screens n =
				match n with
				100 -> ()
				| _ -> 
					incr frames;
					GlClear.color (0.0, 0.0, 0.0);
					GlClear.clear [`color];
					Gl.finish ();
					Glut.swapBuffers();
					let duration = frame_duration () in
					mean := (!mean *. (float_of_int n) +. (float_of_int duration)) /. (float_of_int (n+1));
					(*under some systems, the first frames are not plotted, so exclude frames with durations < 1;
					also exclude possible strong intitial variations due to inaccuarate time measurements:*)
					if ( duration < 2 || (abs_float (!mean -. (float_of_int duration))) > (!mean /. 2.0) ) && !frames < 200 then
						draw_blank_screens 0
					else
					(
						a.(n) <- (float_of_int duration);
						draw_blank_screens (n+1)
					)
			in
			draw_blank_screens 0;
			let sum = Array.fold_left (+.) 0.0 in
			(*mean := (sum a)/.100.0;*)
			std := sqrt ( (sum (Array.map (fun x -> (x -. !mean)**2.) a)) /. 100.0 );
			frames := 1;
		;;
		
		let make_logfiles ex =
			(*open file filename; if not possible open temporary file instead*)
			let open_out_checked filename =
				try
				open_out filename
				with
					Sys_error e ->
					(
						let tfile = Filename.temp_file filename "" in
						prerr_endline 
						(
							"Could not open file " ^ filename ^ ": " ^ e ^ ".\n"
							^ "Opening file " ^ tfile ^ " instead."
						);
						open_out tfile
					)
			in
			
			(***** general log file: ******)
			
			if ex.check_framerate then
			(
				let general_log = open_out_checked ex.logfile in
				let pgl =
					Printf.fprintf general_log "%s\n"
				in
				pgl "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";	
				(*pgl "<!DOCTYPE Logfile SYSTEM \"logfile.dtd\">";*)
				pgl "<!DOCTYPE Logfile [";
				pgl "<!ELEMENT Logfile (DurationEstimation, EnhancedDisplayModeProperties?, Deviation*)>\n\n<!ELEMENT DurationEstimation EMPTY>\n<!ATTLIST DurationEstimation\n\tFrameRate CDATA #REQUIRED\n\tDurationMean CDATA #REQUIRED\n\tDurationStandardDeviation CDATA #REQUIRED\n>\n\n<!ELEMENT EnhancedDisplayModeProperties EMPTY>\n<!ATTLIST EnhancedDisplayModeProperties\n\tEnhancedModeActive (true|false) #REQUIRED\n\tDisplayChanged (true|false) \"false\"\n\tWidth CDATA #IMPLIED\n\tHeight CDATA #IMPLIED\n\tRefreshRate CDATA #IMPLIED\n\tColorDepth CDATA #IMPLIED\n>\n<!ELEMENT Deviation EMPTY>\n<!ATTLIST Deviation\n\tMeasuredDuration CDATA #REQUIRED\n\tGlobalFrameNumber CDATA #REQUIRED\n\tSceneFrameNumber CDATA #REQUIRED\n\tConcernedScene CDATA #REQUIRED\n>\n]>";
				
				let duration_estimation =
					Xml.Element 
					(
						"DurationEstimation",
						[
							("DurationMean", string_of_float !mean); 
							("DurationStandardDeviation", string_of_float !std);
							("FrameRate", string_of_float (1000. /. !mean)); 
						],
						[]
					)
				in
				let xml_of_deviation dev =
					Xml.Element
					(
						"Deviation",
						[
							("MeasuredDuration", string_of_int dev.measured_duration);
							("GlobalFrameNumber", string_of_int dev.global_frame_number);
							("SceneFrameNumber", string_of_int dev.scene_frame_number);
							("ConcernedScene", dev.concerned_scene)
						],
						[]
					)
				in
				let rec queue_to_xml_list l q =
					match Queue.is_empty q with
					true -> l
					| false -> queue_to_xml_list ( (xml_of_deviation (Queue.pop q)) :: l ) q
				in
				let devlist = List.rev (queue_to_xml_list [] ex.deviation_queue) in
				let enhanced_mode =
				(
					let boolstring_of_int x =
						match x with
						0 -> "false"
						| _ -> "true"
					in
					let bool_of_int x =
						match x with
						0 -> false
						| _ -> true
					in
					match ex.display.mode with
					Simple_display_mode s -> []
					| Enhanced_display_mode e -> 
						[
							Xml.Element
							(
								"EnhancedDisplayModeProperties",
								( [("EnhancedModeActive", boolstring_of_int (Glut.gameModeGet Glut.GAME_MODE_ACTIVE))]
								@ if bool_of_int (Glut.gameModeGet Glut.GAME_MODE_ACTIVE) then
								(
									[("DisplayChanged", boolstring_of_int (Glut.gameModeGet Glut.GAME_MODE_DISPLAY_CHANGED))]
									@ if bool_of_int (Glut.gameModeGet Glut.GAME_MODE_DISPLAY_CHANGED) then
									[
										("Width", string_of_int (Glut.gameModeGet Glut.GAME_MODE_WIDTH));
										("Height", string_of_int (Glut.gameModeGet Glut.GAME_MODE_HEIGHT));
										("RefreshRate", string_of_int (Glut.gameModeGet Glut.GAME_MODE_REFRESH_RATE));
										("ColorDepth", string_of_int (Glut.gameModeGet Glut.GAME_MODE_PIXEL_DEPTH));
									]
									else []
								)
								else []
								),
								[]
							)
						]
				) in
				let log_structure = 
					Xml.Element
					(
						"Logfile",
						[],
						duration_estimation :: (enhanced_mode @ devlist)
					)
				in
				pgl (Xml.to_string_fmt log_structure);
				close_out general_log;
			);
			
			(***** event log file: ******)
			
			let log_format_list = ex.event_log in
			let event_list =
				let rec list_of_queue l q =
					match Queue.is_empty q with
					true -> List.rev l
					| false ->  list_of_queue ((Queue.pop q) :: l ) q
				in
				list_of_queue [] ex.event_queue
			in
			(*combined list of log formats and corresponding events:*)
			let log_format_and_event_list =
				List.map
					(
						fun (eventlog_id, log_properties) ->
						(
							log_properties,
							List.filter 
								(fun t -> t.logged_logfile_id = eventlog_id)
								event_list
						)
					)
					log_format_list
			in
			(*write the logfiles:*)
			let process_logfile = function
				| (_, []) -> ()
				| (logprops, eventlist) ->
					if logprops.log_any then
					(
						let event_log = open_out_checked logprops.event_logfile in
						let pe =
							Printf.fprintf event_log "%s\n"
						in
						(match  logprops.event_logfile_header with Some s -> pe s | _ -> ());
						(*if no event_id is specified, generate an automatic one:*)
						let string_of_event_contents conts = function
							| Some evid -> evid
							| _ ->
							(
								match conts with
								Keyboard_event k -> 
								(
									match k with
									Ascii_key a ->
									(
										match int_of_char a with
										13 -> "Enter"
										| 27 -> "Escape"
										| 32 -> "Space"
										| 8 -> "Backspace"
										| 9 -> "Tab"
										| _ -> String.make 1 a
									)
									| Special_key s -> Glut.string_of_special s
								)
								| Mouse_event m ->
									(Glut.string_of_button m.button) ^ "_" ^ (Glut.string_of_button_state m.up_or_down)
								| Button_event b -> b
							)
						in
						let string_of_matrix ?(delim=",") mat =
							let s = ref "\n" in
							let (r, c) = Gsl_matrix.dims mat in
							for i = 0 to r - 1 do
								for j = 0 to c - 1 do
									s := !s ^ (string_of_float mat.{i, j});
									if i < r-1 || j < c-1 then
										s := !s ^ delim
								done;
								s := !s ^ "\n"
							done;
							!s
						in
						let rec string_of_index_vectors ?(delim=",") str = function
							[] -> str
							| h::l ->
								let str_new = 
								(
									match l with
									[] -> str ^ h
									| _ -> str ^ h ^ delim
								) in
								string_of_index_vectors ~delim:delim str_new l
						in
						let sep_string_of_logged_event delim ev =
							let rec variable_string_of_xml varlist str =
								match varlist with
								[] -> str
								| _ ->
									let vstr v =
										let s1 = (snd v) in
										if Str.last_chars s1 1 = "." then
											Str.string_before s1 ((String.length s1) - 1)
										else
											s1
									in
									variable_string_of_xml (List.tl varlist) (str ^ delim ^ (vstr (List.hd varlist)))
							in
							let rec matrix_string_of_matlist str = function
								[] -> str
								| h::matlist -> 
									let mstr m =
										string_of_matrix ~delim:delim (snd m)
									in
									matrix_string_of_matlist (str ^ delim ^ (mstr h)) matlist
							in
							let rec index_vectors_string_of_list str = function
								[] -> str
								| h::l -> 
									let ivstr iv =
										string_of_index_vectors ~delim:delim "" (snd iv)
									in
									index_vectors_string_of_list (str ^ delim ^ (ivstr h)) l
							in
							let s =
								(if logprops.log_key then 
									(string_of_event_contents ev.logged_contents ev.logged_id)
								else "")
								^ (if logprops.log_time then 
									delim ^ (string_of_int ev.logged_time)
								else "")
								^ (if logprops.log_mouse_position then 
									let (x, y) = ev.logged_mouse_position in
									delim ^ (string_of_float x) ^ delim ^ (string_of_float y)
								else "")
								^ (if logprops.log_user_declared_variables then 
									variable_string_of_xml ev.logged_variables ""
								else "")
								^ (if logprops.log_user_declared_matrices then 
									matrix_string_of_matlist "" ev.logged_matrices
								else "")
								^ (if logprops.log_list_index_vectors then 
									index_vectors_string_of_list "" ev.logged_index_vectors
								else "")
							in
							let l = String.length delim in
							let ls = String.length s in
							if ls > 0 && String.sub s 0 l = delim then
								String.sub s l (ls - l)
							else
								s
						in
						let event_list_to_file delim = 
							List.iter
								(
									fun t ->
										let s = sep_string_of_logged_event delim t in
										pe s
								)
								eventlist
						in
						(
							match logprops.log_format with
							XML ->
								pe "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
								pe "<!DOCTYPE EventLogfile [ <!ELEMENT EventLogfile (Event*)>\n\n<!ELEMENT Event (Variable*, Matrix*, IndexVector*)>\n<!ATTLIST Event\n\tIdentifier CDATA #IMPLIED\n\tTimeFromSceneOnset CDATA #IMPLIED\n\tMousePosition CDATA #IMPLIED\n>\n<!ELEMENT Variable EMPTY>\n<!ATTLIST Variable\n\tIdentifier CDATA #REQUIRED\n\tValue CDATA #REQUIRED\n>\n<!ELEMENT Matrix (#PCDATA)>\n<!ATTLIST Matrix\n\tIdentifier CDATA #REQUIRED\n\tColumns CDATA #REQUIRED\n\tRows CDATA #REQUIRED\n>\n<!ELEMENT IndexVector EMPTY>\n<!ATTLIST IndexVector\n\tListID CDATA #REQUIRED\n\tIndices CDATA #REQUIRED\n>]>\n";
								let xml_of_logged_event ev =
									let xml_of_logged_variable lv =
										Xml.Element
										(
											"Variable",
											[
												("Identifier", (fst lv));
												("Value", (snd lv))
											],
											[]
										)
									in
									let xml_of_logged_matrix lm =
										let (r, c) = Gsl_matrix.dims (snd lm) in
										Xml.Element
										(
											"Matrix",
											[
												("Identifier", (fst lm));
												("Rows", string_of_int r);
												("Columns", string_of_int c)
											],
											[Xml.PCData (string_of_matrix (snd lm))]
										)
									in
									let xml_of_logged_index_vector liv =
										Xml.Element
										(
											"IndexVector",
											[
												("ListID", (fst liv));
												("Indices", string_of_index_vectors "" (snd liv))
											],
											[]
										)
									in
									let logged_variables_list = 
										List.map xml_of_logged_variable ev.logged_variables
									in
									let logged_matrices_list = 
										List.map xml_of_logged_matrix ev.logged_matrices
									in
									let logged_index_vectors_list =
										List.map xml_of_logged_index_vector ev.logged_index_vectors
									in
									Xml.Element
									(
										"Event",
										(
											(if logprops.log_key then
												[("Identifier", string_of_event_contents ev.logged_contents ev.logged_id)]
											else [])
											@ (if logprops.log_time then
												[("TimeFromSceneOnset", string_of_int ev.logged_time)]
											else [])
											@ (if logprops.log_mouse_position then
												let (x, y) = ev.logged_mouse_position in
												[( "MousePosition", ((string_of_float x) ^ ", " ^ (string_of_float y)) )]
											else [])
										),
										logged_variables_list @ logged_matrices_list @ logged_index_vectors_list
									)
								in
								let event_log_structure = 
									Xml.Element
									(
										"EventLogfile",
										[],
										(List.map xml_of_logged_event eventlist)
									)
								in
								pe (Xml.to_string_fmt event_log_structure);
							| CommaSeparatedValues -> 
								event_list_to_file ","
							| SpaceSeparatedValues ->
								event_list_to_file " "
						);
						
						close_out event_log;
					)
			in	(*process_logfile*)
			
			List.iter process_logfile log_format_and_event_list;
			
			
			exit 0
		;;
			
(*			let logprops = ex.event_log in
			
			if logprops.log_any then
			(
				let event_log = open_out_checked logprops.event_logfile in
				let pe =
					Printf.fprintf event_log "%s\n"
				in
				(match  logprops.event_logfile_header with Some s -> pe s | _ -> ());
				(*if no event_id is specified, generate an automatic one:*)
				let string_of_event_contents conts = function
					| Some evid -> evid
					| _ ->
					(
						match conts with
						Keyboard_event k -> 
						(
							match k with
							Ascii_key a ->
							(
								match int_of_char a with
								13 -> "Enter"
								| 27 -> "Escape"
								| 32 -> "Space"
								| 8 -> "Backspace"
								| 9 -> "Tab"
								| _ -> String.make 1 a
							)
							| Special_key s -> Glut.string_of_special s
						)
						| Mouse_event m ->
							(Glut.string_of_button m.button) ^ "_" ^ (Glut.string_of_button_state m.up_or_down)
						| Button_event b -> b
					)
				in
				let string_of_matrix ?(delim=",") mat =
					let s = ref "\n" in
					let (r, c) = Gsl_matrix.dims mat in
					for i = 0 to r - 1 do
						for j = 0 to c - 1 do
							s := !s ^ (string_of_float mat.{i, j});
							if i < r-1 || j < c-1 then
								s := !s ^ delim
						done;
						s := !s ^ "\n"
					done;
					!s
				in
				let rec string_of_index_vectors ?(delim=",") str = function
					[] -> str
					| h::l ->
						let str_new = 
						(
							match l with
							[] -> str ^ h
							| _ -> str ^ h ^ delim
						) in
						string_of_index_vectors ~delim:delim str_new l
				in
				let sep_string_of_logged_event delim ev =
					let rec variable_string_of_xml varlist str =
						match varlist with
						[] -> str
						| _ ->
							let vstr v =
								let s1 = (snd v) in
								if Str.last_chars s1 1 = "." then
									Str.string_before s1 ((String.length s1) - 1)
								else
									s1
							in
							variable_string_of_xml (List.tl varlist) (str ^ delim ^ (vstr (List.hd varlist)))
					in
					let rec matrix_string_of_matlist str = function
						[] -> str
						| h::matlist -> 
							let mstr m =
								string_of_matrix ~delim:delim (snd m)
							in
							matrix_string_of_matlist (str ^ delim ^ (mstr h)) matlist
					in
					let rec index_vectors_string_of_list str = function
						[] -> str
						| h::l -> 
							let ivstr iv =
								string_of_index_vectors ~delim:delim "" (snd iv)
							in
							index_vectors_string_of_list (str ^ delim ^ (ivstr h)) l
					in
					let s =
						(if logprops.log_key then 
							(string_of_event_contents ev.logged_contents ev.logged_id)
						else "")
						^ (if logprops.log_time then 
							delim ^ (string_of_int ev.logged_time)
						else "")
						^ (if logprops.log_mouse_position then 
							let (x, y) = ev.logged_mouse_position in
							delim ^ (string_of_float x) ^ delim ^ (string_of_float y)
						else "")
						^ (if logprops.log_user_declared_variables then 
							variable_string_of_xml ev.logged_variables ""
						else "")
						^ (if logprops.log_user_declared_matrices then 
							matrix_string_of_matlist "" ev.logged_matrices
						else "")
						^ (if logprops.log_list_index_vectors then 
							index_vectors_string_of_list "" ev.logged_index_vectors
						else "")
					in
					let l = String.length delim in
					let ls = String.length s in
					if ls > 0 && String.sub s 0 l = delim then
						String.sub s l (ls - l)
					else
						s
				in
				let rec event_queue_to_file q delim =
					match Queue.is_empty q with
					true -> ()
					| false -> 
						let s = sep_string_of_logged_event delim (Queue.pop q) in
						pe s;
						event_queue_to_file q delim
				in
				(
					match ex.event_log.log_format with
					XML ->
						pe "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
						(*pe "<!DOCTYPE EventLogfile SYSTEM \"eventlogfile.dtd\">";*)
						pe "<!DOCTYPE EventLogfile [ <!ELEMENT EventLogfile (Event*)>\n\n<!ELEMENT Event (Variable*, Matrix*, IndexVector*)>\n<!ATTLIST Event\n\tIdentifier CDATA #IMPLIED\n\tTimeFromSceneOnset CDATA #IMPLIED\n\tMousePosition CDATA #IMPLIED\n>\n<!ELEMENT Variable EMPTY>\n<!ATTLIST Variable\n\tIdentifier CDATA #REQUIRED\n\tValue CDATA #REQUIRED\n>\n<!ELEMENT Matrix (#PCDATA)>\n<!ATTLIST Matrix\n\tIdentifier CDATA #REQUIRED\n\tColumns CDATA #REQUIRED\n\tRows CDATA #REQUIRED\n>\n<!ELEMENT IndexVector EMPTY>\n<!ATTLIST IndexVector\n\tListID CDATA #REQUIRED\n\tIndices CDATA #REQUIRED\n>]>\n";
						let xml_of_logged_event ev =
							let xml_of_logged_variable lv =
								Xml.Element
								(
									"Variable",
									[
										("Identifier", (fst lv));
										("Value", (snd lv))
									],
									[]
								)
							in
							let xml_of_logged_matrix lm =
								let (r, c) = Gsl_matrix.dims (snd lm) in
								Xml.Element
								(
									"Matrix",
									[
										("Identifier", (fst lm));
										("Rows", string_of_int r);
										("Columns", string_of_int c)
									],
									[Xml.PCData (string_of_matrix (snd lm))]
								)
							in
							let xml_of_logged_index_vector liv =
								Xml.Element
								(
									"IndexVector",
									[
										("ListID", (fst liv));
										("Indices", string_of_index_vectors "" (snd liv))
									],
									[]
								)
							in
							let logged_variables_list = 
								List.map xml_of_logged_variable ev.logged_variables
							in
							let logged_matrices_list = 
								List.map xml_of_logged_matrix ev.logged_matrices
							in
							let logged_index_vectors_list =
								List.map xml_of_logged_index_vector ev.logged_index_vectors
							in
							Xml.Element
							(
								"Event",
								(
									(if logprops.log_key then
										[("Identifier", string_of_event_contents ev.logged_contents ev.logged_id)]
									else [])
									@ (if logprops.log_time then
										[("TimeFromSceneOnset", string_of_int ev.logged_time)]
									else [])
									@ (if logprops.log_mouse_position then
										let (x, y) = ev.logged_mouse_position in
										[( "MousePosition", ((string_of_float x) ^ ", " ^ (string_of_float y)) )]
									else [])
								),
								logged_variables_list @ logged_matrices_list @ logged_index_vectors_list
							)
						in
						let rec event_queue_to_xml_list l q =
							match Queue.is_empty q with
							true -> l
							| false -> event_queue_to_xml_list ( (xml_of_logged_event (Queue.pop q)) :: l ) q
						in
						let eventlist = List.rev (event_queue_to_xml_list [] ex.event_queue) in
						let event_log_structure = 
							Xml.Element
							(
								"EventLogfile",
								[],
								eventlist
							)
						in
						pe (Xml.to_string_fmt event_log_structure);
					| CommaSeparatedValues -> 
						event_queue_to_file ex.event_queue ","
					| SpaceSeparatedValues ->
						event_queue_to_file ex.event_queue " "
				);
				
				close_out event_log;
			);
			exit 0
		;;
*)
		
		(*called when resizing the window:*)
		let resizefun ex ~w:w ~h:h =
			GlMat.mode `projection;
			GlMat.load_identity ();
			GlDraw.viewport ~x:0 ~y:0 ~w:w ~h:h;
			ex.display.width <- w;
			ex.display.height <- h;
			replace_declaration "CANVAS_WIDTH" (Mathexpr.Constant (float_of_int w));
			replace_declaration "CANVAS_HEIGHT" (Mathexpr.Constant (float_of_int h));
		;;
		
		let rec displayfun ex () =
			if (Queue.is_empty ex.sequence) then
				make_logfiles ex
			else
				let curr_state = Queue.peek ex.sequence in
				if curr_state.wake_up_on_event_mode then
					Glut.idleFunc None;
				let sc = List.assoc curr_state.sceneID ex.scene_list in
				(*duration:*)
				let dur currstate =
					if currstate.duration_in_frames then
						int_of_number currstate.duration
					else	(*estimate number of frames corresponding to msec:*)
						let desired_msec = float_of_number currstate.duration in
						(*Printf.printf "%i %f, %f -> %i; %i\n" !frames desired_msec !mean (int_of_float (desired_msec /. !mean)) ((Glut.get ~gtype:Glut.ELAPSED_TIME) - !scene_onset);*)
						if !frames < 1 then
							1
						else
						(
							(*check whether desired duration over:*)
							if (Glut.get ~gtype:Glut.ELAPSED_TIME) - !scene_onset > (int_of_float desired_msec) + 1 then
								set_frames_left 1;
							(*it should stop 1 frame before desired duration:*)
							int_of_float (desired_msec /. !mean)
						)
				in
				Mathexpr.local_declarations := sc.local_declarations;
				let runtime_error () =
					Queue.clear ex.sequence;
					Queue.push 
						{
							sceneID = "INTERNAL_ERROR"; 
							duration = C (-1.); 
							duration_in_frames = true;
							wake_up_on_event_mode = false
						}
						ex.sequence;
				in
				let color_of_number_list nl =
					let cl = List.map float_of_number nl in
					if List.for_all ( fun t -> (t <= 1.0) && (t >= 0.0) ) cl then
						(List.nth cl 0, List.nth cl 1, List.nth cl 2)
					else
						(*let sl = List.map string_of_float cl in
						let combined_string = (List.nth sl 0) ^ ", " ^ (List.nth sl 1) ^ ", " ^ (List.nth sl 2) in
						failwith (combined_string ^ ": No valid color")*)
						( runtime_error (); (0., 0., 0.) )
				in
				if frames_left () <> 0 && dur curr_state <> 0 then
				(
					GlClear.color (color_of_number_list sc.background_color);
					GlClear.clear [`color];
					GlMat.mode `projection;
					GlMat.load_identity ();
					GlMat.ortho ~x:ex.display.horizontal_range ~y:ex.display.vertical_range ~z:(-1.0,1.0);
					Glut.setCursor sc.cursor_type;
					GlMat.mode `modelview;
					GlMat.load_identity ();
					
					(*estimate frame rate:*)
					if !frames < 1 then 
					(
						scene_onset := Glut.get ~gtype:Glut.ELAPSED_TIME;
						(*set_frames_left (dur curr_state);*)
						if ex.check_framerate then
							estimate_frame_rate ()
						else
							frames := 1;
						scene_onset := Glut.get ~gtype:Glut.ELAPSED_TIME;
						set_frames_left (dur curr_state);
					);
					
					(*set variables from variations in time:*)
					let set_variation v =
						let nframes = 
						(
							match v.number_of_frames with
							Some i -> int_of_number i
							| None -> 
								if (dur curr_state) > 0 then
									(dur curr_state)
								else	(*NOT defined! so set to 100*)
									100
						) in
						v.variation_function nframes !scene_frames;
						(*Printf.printf "nframes= %i, scene_frames= %i, val= %f\n" nframes !scene_frames !declvar;*)
					in
					List.iter set_variation sc.variation_list;
					
					(*get position of array for a display object list:*)
					let get_array_position frames_per_item arrlength repeat =
						let nframes = arrlength * frames_per_item in
						if  (not repeat) && !scene_frames > nframes - 1 then
							None
						else
						(
							let n = !scene_frames - 1 in
							Some ((n mod nframes) / frames_per_item)
						)
					in
					(*"compile" scenewide images:*)
					if !scene_frames == 1 then
					(
						let comp_image_scenewide = function
							| (fi, i) -> i := Some (fi ())
						in
						let rec comp_dispobj = function
							| Image_scenewide (fi, i) -> comp_image_scenewide (fi, i)
							| Image_list_reference ilr ->
								let il = 
									Array.to_list
									(
										Array.map 
											(fun t -> !t)
											ilr.object_reference_array 
									)
								in
								List.iter comp_dispobj il
							| _ -> ()
						in
						List.iter comp_dispobj sc.display_object_list;
					);
						
							

					(*actual plotting:*)
					let rec plot_object o =
						let init_plot plotfun s =
							let triple_float_of_number_list nl =
								let cl = List.map float_of_number nl in
								(List.nth cl 0, List.nth cl 1, List.nth cl 2)
							in
							
							if (float_of_number s.display_stimulus) <> 0.0 then
							(
								GlDraw.color (color_of_number_list s.color); 
								GlMat.load_identity ();
								GlMat.push ();
								GlMat.translate ~x:(float_of_number (fst s.position)) ~y:(float_of_number (snd s.position)) ();
								(
									match s.scale_xyz with
									Some sc -> 
										GlMat.scale3 (triple_float_of_number_list sc);
									| None -> ()
								);
								(
									match s.rotation with
									Some r -> 
										let (nrotx, nroty) = r.rotation_center in 
										let (rotx, roty) = (float_of_number nrotx, float_of_number nroty) in
										GlMat.translate ~x:rotx ~y:roty ();
										GlMat.rotate3 ~angle:(float_of_number r.rotation_angle) (0.0, 0.0, 1.0);
										GlMat.translate ~x:(-.rotx) ~y:(-.roty) ();
									| None -> ()
								);
								(
									try
									plotfun ();
									with Failure f -> runtime_error ();
								);
								GlMat.pop ();
							)
						in
						match o with
						Stimulus_scenewide s ->
							let plotfun () =
								GlList.call s.content_compiled
							in
							init_plot plotfun s

						| Stimulus_uncompiled su ->
							let s = su () in
							let plotfun () =
								List.iter (fun t -> t ()) s.content_list
							in
							init_plot plotfun s
						| Image i ->
							if (float_of_number i.display_image) <> 0.0 then
							(
								GlMat.push ();
								let (sx, sy) = 
									( (float_of_number (fst i.image_scale)), (float_of_number (snd i.image_scale)) ) 
								in
								let (scalex, scaley) =
									let son = function 
										| true -> (-. 1.0)
										| _ -> 1.0
									in
									( (son i.image_flip_horizontally) *. sx, (son i.image_flip_vertically) *. sy )
								in
								GlPix.zoom ~x:scalex ~y:scaley;
								let (px, py) = 
									( (float_of_number (fst i.image_position)), (float_of_number (snd i.image_position)) ) 
								in
								let posx =
									if i.image_flip_horizontally then
										px +. ( i.image_width *. sx )
									else
										px
								and posy =
									if i.image_flip_vertically then
										py +. ( i.image_height *. sy )
									else
										py
								in
								GlPix.raster_pos posx posy ();
								GlPix.draw !(i.image_content_gl_pix);
								GlMat.pop ();
							)
						| Image_uncompiled i ->
							let im = i () in
							plot_object (Image im)
						| Image_scenewide (fi, i) -> 
						(
							match !i with
							Some im -> plot_object (Image im)
							| None -> failwith "Image_scenewide undeclared"
						)
						| Button b -> 
							(
								match b.button_state with
								Unpressed -> Objects2D.draw_pos b.button_unpressed_compiled (0.,0.)
								| Pressed -> Objects2D.draw_pos b.button_pressed_compiled (0.,0.)
								| Mouseover -> Objects2D.draw_pos b.button_mouseover_compiled (0.,0.)
							)
						| Stimulus_list_reference slr ->
						(
							let curr_array_pos = 
								let fpi = int_of_number slr.frames_per_item in
								get_array_position fpi (Array.length slr.object_reference_array) slr.object_list_repeat
							in
							match curr_array_pos with
							None -> ()
							| Some i ->
							(
								let stub = 
								(
									match slr.object_stub with
									Stimulus_scenewide s -> s
									| Stimulus_uncompiled su -> su ()
									| _ -> failwith "Stub stimuli have to be compiled."
								) in
								let curr_stim =
								(
									match !(slr.object_reference_array.(i)) with
									Stimulus_scenewide s ->
										stub.content_compiled <- s.content_compiled;
										(Stimulus_scenewide stub)
									|  Stimulus_uncompiled su -> 
										(Stimulus_uncompiled 
											(fun () -> 
												let s = su () in 
												{stub with content_list = s.content_list}))
									| _ -> failwith "" 
								) in
								(* stub.content_compiled <- curr_stim.content_compiled;*)
								plot_object curr_stim
							)
						)
						| Image_list_reference ilr ->
						(
							let curr_array_pos = 
								let fpi = int_of_number ilr.frames_per_item in
								get_array_position fpi (Array.length ilr.object_reference_array) ilr.object_list_repeat
							in
							match curr_array_pos with
							None -> ()
							| Some i ->
								plot_object !(ilr.object_reference_array.(i))
						)
					in
					List.iter plot_object sc.display_object_list;
	
					
					Gl.finish ();
					Glut.swapBuffers();
					
					(**only log finite length scenes:**)
					if ex.check_framerate && (dur curr_state) >= 0 then
					(
						(*time logging:*)
						let duration = frame_duration () in
						(*deviation := more than half of mean*)
						if abs_float ( (float_of_int duration) -. !mean ) > !mean/.2. then
						(
							let dev = {
								measured_duration = duration;
								global_frame_number = !frames;
								scene_frame_number = !scene_frames;
								concerned_scene = curr_state.sceneID
							} in
							Queue.push dev ex.deviation_queue;
						);
						
						incr frames;
						(*incr scene_frames;*)
					);
					incr scene_frames;
					set_frames_left ((frames_left ()) - 1);
					if (frames_left ()) == 0 then
						Glut.idleFunc !idle_fun;
				)
				else
				(
					let _ = Queue.pop ex.sequence in ();
					scene_onset := Glut.get ~gtype:Glut.ELAPSED_TIME;
					scene_frames := 1;
					if not (Queue.is_empty ex.sequence) then
					(
						let curr_state = Queue.peek ex.sequence in
						set_frames_left (dur curr_state);
					);
					Glut.idleFunc !idle_fun;
				)

		;;
		
		(** OpenGL initialization for experiment ex: *)
		let init ex =
			(*ignore(Glut.init Sys.argv);*)
			glut_initialized := true;
			Glut.initDisplayMode ~alpha:true ~double_buffer:true ~depth:true ();
			(
				match ex.display.mode with
				Simple_display_mode s -> 
				(
					Glut.initWindowSize ~w:ex.display.width ~h:ex.display.height ;
					let (x, y) = s.window_position in
					Glut.initWindowPosition ~x:x ~y:y;
					ignore(Glut.createWindow ~title:s.window_title);
					if s.fullscreen then ignore(Glut.fullScreen ());
				)
				| Enhanced_display_mode e ->
				(
					let gamemodestring = 
						(string_of_int ex.display.width) ^
						"x" ^
						(string_of_int ex.display.height) ^
						":" ^
						(string_of_int e.color_depth) ^
						"@" ^
						(string_of_int e.refresh_rate)
					in
					Glut.gameModeString gamemodestring;
					Glut.enterGameMode ();
					(*declarations:*)
					declare "ENHANCED_DISPLAY_MODE_COLOR_DEPTH" (Mathexpr.Constant (float_of_int (Glut.gameModeGet Glut.GAME_MODE_PIXEL_DEPTH)));
					declare "ENHANCED_DISPLAY_MODE_REFRESH_RATE" (Mathexpr.Constant (float_of_int (Glut.gameModeGet Glut.GAME_MODE_REFRESH_RATE)));
				)
			);	
			
			(* Compilation (cannot be done before invoking OpenGL): *)
			let compile_scene sc =
				Mathexpr.local_declarations := sc.local_declarations;
				let rec compile_display_object d =
					let make_comp_list ob =	
						let t = GlList.create `compile in
						List.iter (fun x -> x ()) ob;
						GlList.ends ();
						t
					in
					match d with
					Stimulus_scenewide st ->
						st.content_compiled <- make_comp_list st.content_list
					| Button b -> 
						b.button_pressed_compiled <- b.button_content Pressed;
						b.button_unpressed_compiled <- b.button_content Unpressed;
						b.button_mouseover_compiled <- b.button_content Mouseover
					| Stimulus_list_reference slr ->
						Array.iter
							(fun t -> compile_display_object !t)
							slr.object_reference_array
					| _ -> ()
				in
				List.iter compile_display_object sc.display_object_list
			in
			List.iter compile_scene (snd (List.split ex.scene_list));

			Glut.displayFunc ~cb:(displayfun ex);
			Glut.keyboardFunc ~cb:(Events.keyboard_action ex scene_onset);
			Glut.specialFunc ~cb:(Events.special_key_action ex scene_onset);
			Glut.mouseFunc ~cb:(Events.mouse_action ex scene_onset);
			Glut.passiveMotionFunc ~cb:(Events.mouse_moving ex);
			(*Glut.idleFunc (Some (displayfun ex));*)
			idle_fun := Some (displayfun ex);
			Glut.idleFunc !idle_fun;
			Glut.reshapeFunc ~cb:(resizefun ex);
			Glut.mainLoop();
		;;

	end
;;

(* exception reporting with OpenGL output;
dummy is a lazy value which is provided to make the function return any desired type *)

let error_message_scene ~canvas_width ~canvas_height ~horizontal_range ~vertical_range failure_string =
	let to_displist ob () =	
		let t = GlList.create `compile in
		List.iter (fun x -> x ()) ob;
		GlList.ends ();
		t
	in
	let (w,h) = (float_of_int canvas_width, float_of_int canvas_height) in
	let (h1, h2) = horizontal_range
	and (v1, v2) = vertical_range in
	(*calculate x and y positions from a fractional value v from lower left corner
	(i. e. "norm" width and height to 1 and set lower left corner to (0.0, 0.0))*)
	let px v =
		h1 +. (h2 -. h1)*.v
	and py v =
		v1 +. (v2 -. v1)*.v
	in
	let pixel_width = (h2 -. h1) /. w
	and pixel_height = (v2 -. v1) /. h in
	let dispobjlist = [	
			Stimulus_scenewide {
				position = (C 0.0, C 0.0);
				color = [C 1.0; C 1.0; C 1.0];
				rotation = None;
				scale_xyz = None;
				display_stimulus = C 1.0;
				content_list = [ 
					( fun () ->
						let s = "The following error occured while processing your input file:" in
						Objects2D.text_line ~text:s ~font:Glut.BITMAP_HELVETICA_18 ~color:(Some (1.,1.,1.)) ~position:(px 0.1, py 0.8) () 
					);
					( fun () ->
						Objects2D.text_field ~text:failure_string ~font:Glut.BITMAP_HELVETICA_18 ~width:(h2 -. h1) ~height:(0.1*.(v2 -. v1)) ~pixel_width:pixel_width ~pixel_height:pixel_height ~alignment:"Left" ~draw_frame:false ~color:(Some (1.,0.,0.)) ~position:(px 0.0, py 0.2) () 
					);
					( fun () ->
						let s = "Look for the concerning place in your input file or read the documentation." in
						Objects2D.text_line ~text:s ~font:Glut.BITMAP_HELVETICA_18 ~color:(Some (1.,1.,1.)) ~position:(px 0.1, py 0.11) () 
					);
				];
				content_compiled = to_displist [] ()
			};
			(*ADAPT THIS to widh/height:*)
			(*let button_pos = (0.425, 0.02)
			and b_width = 0.15
			and b_height = 0.04 in*)
			let button_pos = (px 0.425, py 0.03)
			and b_width = 0.15*.(h2 -. h1)
			and b_height = 0.05*.(v2 -. v1) in
			let b_content b_state =
				let bcont () = 
				(
					match b_state with
					Unpressed -> 
						Objects2D.button ~width:b_width ~height:b_height ~text:"Exit" ~font:Glut.BITMAP_HELVETICA_18 ~pixel_width:pixel_width ~pixel_height:pixel_height ~position:button_pos ()
					| Pressed -> 
						Objects2D.pressed_button ~width:b_width ~height:b_height ~text:"Exit" ~font:Glut.BITMAP_HELVETICA_18 ~pixel_width:pixel_width ~pixel_height:pixel_height ~position:button_pos ()
					| Mouseover ->
						Objects2D.mouseover_button ~width:b_width ~height:b_height ~text:"Exit" ~font:Glut.BITMAP_HELVETICA_18 ~pixel_width:pixel_width ~pixel_height:pixel_height ~position:button_pos ()
				) in
				let t = GlList.create `compile in
					bcont ();
				GlList.ends ();
				t
			in
			
			let sens_area =
				let (lx, ly) = button_pos in
				fun (x, y) -> 
					0. <= x -. lx && x -. lx <= b_width && 0. <= y -. ly && y -. ly <= b_height
			in
			let act = Exit in
			Button {
				buttonID = "Exit";
				button_state = Unpressed;
				button_log_event = C 0.0;
				button_event_logfile_id = "standard_event_logfile";
				button_action = act;
				button_variables_to_set = [];
				button_variables_to_log = [];
				button_matrices_to_log = [];
				button_index_vectors_to_log = [];
				button_list_indices_to_log = [];
				button_log_before_set = false;
				button_sensitive_area = sens_area;
				button_content = b_content;
				button_pressed_compiled = b_content Pressed;
				button_unpressed_compiled = b_content Unpressed;
				button_mouseover_compiled = b_content Mouseover;
			}
		]
	in
	let eventlist = [{
			event_logfile_id = "standard_event_logfile";
			event_contents = Button_event "Exit";
			event_id = None;
			log_event = C 0.0;
			action = Exit;
			variables_to_set = [];
			variables_to_log = [];
			matrices_to_log = [];
			index_vectors_to_log = [];
			list_indices_to_log = [];
			log_before_set = false
		}]
	in
	(
			"INTERNAL_ERROR",
			{
				background_color = [C 0.0; C 0.0; C 0.0];
				cursor_type = Glut.CURSOR_INHERIT;
				display_object_list = dispobjlist;
				event_list = eventlist;
				local_declarations = Hashtbl.copy Mathexpr.global_declarations;
				variation_list = []
			}
	)
;;


let gl_failwith failure_string =
	match !standard_error_output with
	Err_graphical ->
	(
		let disp_props = 
			{
				mode = Simple_display_mode {
						fullscreen = false;
						window_title = "Error";
						window_position = (0, 0)
					};
				width = 800;
				height = 600;
				horizontal_range = (0.0, 1.0);
				vertical_range = (0.0, 0.75)
			}
		in
	
		let error_scene = 
			error_message_scene 
				~canvas_width:800 
				~canvas_height:600 
				~horizontal_range:(0.,1.) 
				~vertical_range:(0., 0.75) 
				failure_string
		in
		let scenelist = [ error_scene ] in
		let curr_stat = {
				sceneID = "INTERNAL_ERROR";
				duration = C (-1.);
				duration_in_frames = true;
				wake_up_on_event_mode = false
			}
		in
		let q = Queue.create () in
		Queue.push curr_stat q;
		let eventlog = 
			{
				event_logfile = "eventlog.xml";
				event_logfile_header = None;
				log_key = false;
				log_time = false;
				log_mouse_position = false;
				log_format = XML;
				log_user_declared_variables = false;
				log_user_declared_matrices = false;
				log_list_index_vectors = false;
				log_any = false
			}
		in
		let error_exp = {
				display = disp_props;
				scene_list = scenelist;
				sequence = q;
				logfile = "logfile.xml";
				check_framerate = false;
				event_log = [("standard_event_logfile", eventlog)];
				deviation_queue = Queue.create ();
				event_queue = Queue.create ()
			}
		in
		Display.init error_exp;
		exit 0
	)
	| Err_file_handler fh -> 
		output_string fh ("ERROR:\n\n" ^ failure_string ^ "\n");
		flush_all ();
		if fh <> stdout && fh <> stderr then
			close_out fh;
		exit 0
;;

(*auxiliary function:*)
let contains_string str contained =
	let p = try Str.search_forward (Str.regexp contained) str 0
		with Not_found -> -1
	in
	p>=0 
;;
(*convert DTDs from file (handler) to strings:*)
let dtd_string_from_file_handler inp_dtd =
	let entityqueue = Queue.create () in
	let rec file_to_string () =
		try 
			let line = input_line inp_dtd in
			if contains_string line "]>" then
			(
				failwith "DTD_END";
			);
			if contains_string line "!ENTITY" then	(*check for parameter entities*)
				let _ = Str.search_forward (Str.regexp "%[ ]+\\([^ ]+\\) [^\"]*\"\\(.*\\)\"") line 0 in
				let ent = Str.matched_group 1 line in
				let subs = Str.matched_group 2 line in
				Queue.push (ent, subs) entityqueue;
				file_to_string ()
			else
				(line ^ (file_to_string ()))
		with 
			End_of_file -> close_in inp_dtd; ""
			| Failure "DTD_END" -> ""
	in
	let dstr = file_to_string () in
	(*substitute parameter entities:*)
	let rec subs_ent str q =
		match Queue.is_empty q with
		true -> str
		| _ -> 
			let e = Queue.pop q in
			let str_repl = Str.global_replace (Str.regexp ("%" ^ (fst e) ^ ";")) (snd e) str in
			subs_ent str_repl q
	in
	subs_ent dstr entityqueue
;;
let dtd_string_from_file dtd_filename =
	let inp_dtd = 
		try
		open_in dtd_filename
		with
		Sys_error _ -> 
		(
			(*"default": experiments.dtd in directory of the running binary:*)
			open_in default_dtd_file
		)
	in
	dtd_string_from_file_handler inp_dtd
;;

(* parse input_mode inputfile and convert it to Xml.xml *)
let parse_xml inputfile =
	let (xmlstructure, dtd) =
		let (inp_file, is_fls, is_file) = 
			match inputfile with
			Stdin_fls -> ("<STDIN>", true, false)
			| Stdin_fld -> ("<STDIN>", false, false)
			| Input_file fn -> (fn, Filename.check_suffix fn "fls", true)
		in
		if is_fls then
		(
			let x =
			(
				try
				let flsstring = 
					if is_file then
						Fls.string_of_file inp_file
					else
					(
						let rec string_of_stdin s =
							try
							let new_line = read_line () in
							string_of_stdin (s ^ new_line ^ "\n")
							with End_of_file -> s
						in
						string_of_stdin ""
					)
				in
				List.hd (Fls.xml_list_of_fls_string flsstring) 
				with
					Sys_error e -> gl_failwith ("Could not open file " ^ inp_file)
					| Xml.Error e -> gl_failwith (Xml.error e) 
					| Fls.Invalid_expression e -> gl_failwith ("Invalid fls expression: " ^ e) 
					| Failure e -> gl_failwith e 
			) in
			let dtd =
				let dtdstring = dtd_string_from_file default_dtd_file in
				let dtdstring_replaced = 
					Str.global_replace (Str.regexp " IDR?E?F? ") " CDATA " dtdstring
				in
				Dtd.check (Dtd.parse_string dtdstring_replaced)
			in
			(x, dtd)
		)
		else
		(
			let inp =
				if is_file then
					try open_in inp_file
					with Sys_error e -> gl_failwith ("Could not open file " ^ inp_file) 
				else
					stdin
			in
			let rec find_start ch linestart line =
				if (String.length line >= 9 && String.sub line 0 9 = linestart) then
					line
				else
					find_start ch linestart (input_line ch)
			in
			let doctypeline = 
				try
				find_start inp "<!DOCTYPE" ""
				with End_of_file -> gl_failwith "fld file does not contain DOCTYPE" 
			in
			let dtd =
				(*external:*)
				let dtdstring =
					if contains_string doctypeline "SYSTEM" then
					(
						let dtd_file =
							let _ = 
								if not (Str.string_match (Str.regexp ".*\"\\(.*dtd\\).*") doctypeline 0) then
								gl_failwith "No DTD file declared." 
							in
							Str.matched_group 1 doctypeline
						in
						(
							(*here could be distinguished between different DTDs*)
							match (Filename.basename dtd_file) with
							"experiments.dtd" -> dtd_string_from_file dtd_file
							| _ ->
								(
									gl_failwith ("DTD file " ^ dtd_file ^ " is not supported.") 
								)
						)
					)
					else (*internal:*)
					(
						dtd_string_from_file_handler inp
					)
				in
				let dtdstring_replaced = 
					Str.global_replace (Str.regexp " IDR?E?F? ") " CDATA " dtdstring
				in
				Dtd.check (Dtd.parse_string dtdstring_replaced)
			in
			let x = try Xml.parse_in inp
				with Xml.Error e -> gl_failwith (Xml.error e) 
			in
			close_in inp;
			(x, dtd)
		) 
	in
	let first_tag = Xml.tag xmlstructure in
	let _ = try Dtd.prove dtd first_tag xmlstructure
		with Dtd.Prove_error e -> gl_failwith (Dtd.prove_error e) 
	in
	xmlstructure
;;

(* convert from Xml.xml x to experiment data structures: *)
let experiment_of_xml input_structure =
	
	if (Xml.tag input_structure) <> "Experiment" then
		gl_failwith "First tag of input XML file ist not Experiment";
	
	(*flag whether framerate must be checked:*)
	let frame_rate_must_be_checked = ref false in
	
	(***
	auxiliary functions:
	***)
	(*check whether attribute attr is declared in element elem:*)
	let attrib_declared attr elem =
		List.mem attr (fst (List.split (Xml.attribs elem)))
	in
	(*get attribute attr of element elem or default if attribute is not declared*)
	let aod elem attr default =
		if attrib_declared attr elem then
			Xml.attrib elem attr
		else
			default
	in
	(*filter children of an xml element elem the tags of which match childtag:*)
	let children elem childtag =
		List.filter (fun t -> (Xml.tag t) = childtag) (Xml.children elem)
	in
	(*return the (first) child with tag childtag from an xml element elem; 
	if no such children exist return an "empty" xml element with tag childtag*)
	let child_or_default elem childtag =
		let childrenlist = (children elem childtag) in
		match childrenlist with
		[] -> Xml.Element (childtag, [], [])
		| _ -> List.hd childrenlist
	in
	let int_list_of_string s =
		let spaces_removed = Str.global_replace (Str.regexp " ") "" s in
		let stringlist = Str.split (Str.regexp ",") spaces_removed in
		List.map int_of_string stringlist
	in
	let int_pair_of_string s =
		let l = 
			try int_list_of_string s with
			Failure _ -> gl_failwith ("Attribute " ^ s ^ " does not consist of integer elements") 
		in
		if (List.length l) = 2 then
			(List.hd l, List.hd (List.tl l))
		else
			gl_failwith ("Attribute " ^ s ^ " is not an int list of two elements") 
	in
	let float_list_of_string s =
		let stringlist = Str.split (Str.regexp ", *") s in
		List.map float_of_string' stringlist
	in
	let float_pair_of_string s =
		let l = 
			try float_list_of_string s with
			Failure _ -> gl_failwith ("Attribute " ^ s ^ " does not consist of float elements") 
		in
		if (List.length l) = 2 then
			(List.hd l, List.hd (List.tl l))
		else
			gl_failwith ("Attribute " ^ s ^ " is not a float list of two elements") 
	in
	let ordered_float_pair_of_string s =
		let (a,b) = float_pair_of_string s in
		if a < b then 
			(a, b)
		else
			gl_failwith ("Attribute " ^ s ^ " is not a valid range") 
	in
	let number_pair_of_string s =
		let l = 
			try number_list_of_string s with
			Mathexpr.Undeclared_identifier idf -> 
				gl_failwith ("Attribute " ^ s ^ " contains undeclared identifier: \"" ^ idf ^ "\"") 
			| Mathexpr.Invalid_declaration_type dt ->
				gl_failwith ("Attribute " ^ s ^ " contains invalid declaration type. Identifier \"" ^ dt ^ "\" is a function with more than zero arguments.") 
		in
		if (List.length l) = 2 then
			(List.hd l, List.hd (List.tl l))
		else
			gl_failwith ("Attribute " ^ s ^ " is not a list of two elements") 
	in
	(*let number_color_of_string s =
		let l = 
			try number_list_of_string s with
			Mathexpr.Undeclared_identifier idf -> 
				gl_failwith ("Color attribute " ^ s ^ " contains undeclared identifier: \"" ^ idf ^ "\". Check the spelling. Maybe you want to use a number instead?") 
			| Mathexpr.Invalid_declaration_type dt ->
				gl_failwith ("Color attribute " ^ s ^ " contains invalid declaration type. Identifier \"" ^ dt ^ "\" is a function with more than zero arguments.") 
		in
		if (List.length l) = 3 then
			(List.nth l 0, List.nth l 1, List.nth l 2)
		else
			gl_failwith ("Color attribute " ^ s ^ " is not a list of three elements") 
	in*)
	
	let int_of_version_string vs =
		int_of_string (Str.global_replace (Str.regexp_string ".") "" vs)
	in
	let vstring = aod input_structure "RequiresVersion" "0.4.0" in
	let required_version = 
		let vs = 
			if List.length (Str.split (Str.regexp_string ".") vstring) < 2 then (vstring ^ ".0")
			else vstring
		in
		try
		int_of_version_string vs
		with _ -> gl_failwith (vstring ^ " is not a valid version specification. Check the RequiredVersion tag.")
	in
	let current_version =
		try
		int_of_version_string flashdot_version
		with _ -> gl_failwith ("\"" ^ vstring ^ "\" is not a valid version string. Check file flashdot_version.")
	in
	if required_version >  current_version then
		gl_failwith ("The experiment requires FlashDot version " ^ vstring ^ " but the version of this program is " ^ flashdot_version ^ ". If you think the experiment works with the current version of flashdot, modify the RequiredVersion tag." ); 
	
	let displaymode =
		let d = 
			let tmode = (children input_structure "SimpleDisplayMode") @ (children input_structure "EnhancedDisplayMode") in
			match tmode with
			[] -> Xml.Element ("SimpleDisplayMode", [], [])
			| _ -> List.hd tmode
		in
		match Xml.tag d with
		"SimpleDisplayMode" -> 
			Simple_display_mode {
				fullscreen = bool_of_string (aod d "FullScreen" "true");
				window_title = (aod d "WindowTitle" "Experiment");
				window_position = int_pair_of_string (aod d "WindowPosition" "0, 0")
			}
		| _ -> 
			Enhanced_display_mode {
				color_depth = int_of_string (aod d "ColorDepth" "16");
				refresh_rate = int_of_string (aod d "RefreshRate" "75")
			}
	in
	let (width, height) =
		let guess_resolution () =
			try (match operating_system with
				| Linux -> 
					let resstring = 
						pattern_of_input_process "xdpyinfo" (Str.regexp "^[ ]*dimensions:[ ]*\\(.+\\) pixels .*")
					in
					let reslist = List.map int_of_string (Str.split (Str.regexp "x") resstring) in
					((List.hd reslist), List.nth reslist 1) 
				| MacOSX ->
					let resstring = 
						pattern_of_input_process "system_profiler" (Str.regexp "^[ ]*Resolution:[ ]*\\([^@]+\\) ?@?.*$")
					in
					let reslist = List.map int_of_string (Str.split (Str.regexp " x ") resstring) in
					((List.hd reslist), List.nth reslist 1) 
				| Windows ->
				(
					let (tempfile, out) = Filename.open_temp_file "getresolution" ".vbs" in
					output_string out "strComputer = \".\"\n\nSet objWMIService = GetObject(\"winmgmts:\\\\\" & strComputer & \"\\root\\cimv2\")\n\nSet colItems = objWMIService.ExecQuery _\n    (\"Select * From Win32_DisplayConfiguration\")\n\nFor Each objItem in colItems\n    Wscript.Echo \"Resolution: {\" & objItem.PelsWidth & \"x\" & objItem.PelsHeight & \"}\"\n    Wscript.Echo\nNext\n";
					close_out out;
					let resstring = 
						pattern_of_input_process 
							("cscript " ^ tempfile) 
							(Str.regexp "^Resolution: {\\(.+\\)}.*")
					in
					Sys.remove tempfile;
					let reslist = List.map int_of_string (Str.split (Str.regexp "x") resstring) in
					((List.hd reslist), List.nth reslist 1)
				)
			) with _ -> (800, 600)
		in
		if bool_of_string (aod input_structure "GuessResolution" "false") then
			guess_resolution ()
		else
			(int_of_string (aod input_structure "CanvasWidth" "800"), 
			int_of_string (aod input_structure "CanvasHeight" "600"))
	in
	declare "CANVAS_WIDTH" (Mathexpr.Constant (float_of_int width));
	declare "CANVAS_HEIGHT" (Mathexpr.Constant (float_of_int height));
	let (hor_range, vert_range) = 
		if bool_of_string (aod input_structure "SetRangesToResolution" "false") then
		(
			(1., float_of_int width),
			(1., float_of_int height)
		)
		else
			(ordered_float_pair_of_string (aod input_structure "HorizontalRange" "-1.0, 1.0"),
			ordered_float_pair_of_string (aod input_structure "VerticalRange" "-0.75, 0.75"))
	in
	let displ = 
	{
		mode = displaymode;
		width = width;
		height = height;
		horizontal_range = hor_range;
		vertical_range = vert_range
	} 
	in
	declare "HORIZONTAL_RANGE_MIN" (Mathexpr.Constant (fst displ.horizontal_range));
	declare "HORIZONTAL_RANGE_MAX" (Mathexpr.Constant (snd displ.horizontal_range));
	declare "VERTICAL_RANGE_MIN" (Mathexpr.Constant (fst displ.vertical_range));
	declare "VERTICAL_RANGE_MAX" (Mathexpr.Constant (snd displ.vertical_range));
	let logging =
		let eventloglist_xml = 
			match (children input_structure "LogEvents") with
			[] -> [Xml.Element ("LogEvents", [], [])]
			| l -> l
		in
		let event_log_of_xml eventlog =
			let elid = aod eventlog "EventLogID" "standard_event_logfile" in
			let eventlogfile = aod eventlog "EventLogFile" "eventlog.xml" in
			let logkey = bool_of_string (aod eventlog "Identifier" "true")
			and logtime = bool_of_string (aod eventlog "TimeFromSceneOnset" "true")
			and logmouseposition = bool_of_string (aod eventlog "MousePosition" "false")
			and 	logformat = 
			(
				if not ( attrib_declared "Format" eventlog ) then
					XML
				else
					match Xml.attrib eventlog "Format" with
					"XML" -> XML
					| "CommaSeparatedValues" -> CommaSeparatedValues
					| _ -> SpaceSeparatedValues
			)
			and loguserdeclaredvariables = bool_of_string (aod eventlog "UserDeclaredVariables" "true")
			and loguserdeclaredmatrices = bool_of_string (aod eventlog "UserDeclaredMatrices" "true")
			and loglistindexvectors = bool_of_string (aod eventlog "ListIndexVectors" "true") in
			let logany = 
				(logkey || logtime || logmouseposition || loguserdeclaredvariables || loguserdeclaredmatrices || loglistindexvectors) 
			in
			let eventlogfileheader =
				match (children eventlog "LogFileHeader") with
				[] -> None
				| h::t -> 
					let comment = bool_of_string (aod h "AsXMLComment" "false") in
					let (c1, c2) =
						if comment then ("<!-- ", " -->") else ("", "")
					in
					let s = (Xml.pcdata (List.hd (Xml.children h))) in
					Some (c1 ^ s ^ c2)
			in
			(*check, if files exists:*)
			
			(
				elid,
				{
					event_logfile = eventlogfile;
					event_logfile_header = eventlogfileheader;
					log_key = logkey;
					log_time = logtime;
					log_mouse_position = logmouseposition;
					log_format = logformat;
					log_user_declared_variables = loguserdeclaredvariables;
					log_user_declared_matrices = loguserdeclaredmatrices;
					log_list_index_vectors = loglistindexvectors;
					log_any = logany
				},
				(bool_of_string (aod eventlog "LogfileOverwriteProtection" "false"))
			)
		in
		let eventlog_list_permissions =
			let el = List.map event_log_of_xml eventloglist_xml in
			(*overwrite logfile in list head, if given at command line:*)
			if !log_file = "" then
				el
			else
				let new_head =
					let (oid, oev, op) = List.hd el in
					(
						oid,
						{
							event_logfile = !log_file;
							event_logfile_header = oev.event_logfile_header;
							log_key = oev.log_key;
							log_time = oev.log_time;
							log_mouse_position = oev.log_mouse_position;
							log_format = oev.log_format;
							log_user_declared_variables = oev.log_user_declared_variables;
							log_user_declared_matrices = oev.log_user_declared_matrices;
							log_list_index_vectors = oev.log_list_index_vectors;
							log_any = oev.log_any
						},
						op
					)
				in
				new_head :: (List.tl el)
		in
		let eventlog_list = 
			List.map (fun t -> let (t1, t2, _) = t in (t1, t2)) eventlog_list_permissions
		in
		(*check IDs and file names for uniqueness:*)
		let rec unique = function 
			| [] -> []
			| x :: xt -> [x] @ (unique (List.filter (fun y -> y<>x) xt))
		in
		let idlist = List.map fst eventlog_list
		and filenamelist = List.map (fun t -> (snd t).event_logfile) eventlog_list in
		if List.length idlist <> (List.length (unique idlist)) then
			gl_failwith "The EventLogIDs are not unique. At least one ID occurs twice.";
		if List.length filenamelist <> (List.length (unique filenamelist)) then
			gl_failwith "The file names of the event log files are not unique. At least one name occurs twice.";
		
		(*check, if files exists:*)
		let file_exists_and_is_writable = function (_, evlog, evperm) ->
			if evlog.log_any then
			(
				let eventlogfile = evlog.event_logfile in
				try 
				Unix.access eventlogfile [Unix.F_OK]; (*unit, if file exists, exception otherwise*)
				if evperm then
					gl_failwith ("Event logfile \"" ^ eventlogfile ^ "\" already exists and overwrite protection is set."); 
				(*check, if eventlogfile can be overwritten:*)
				(
					try 
					Unix.access eventlogfile [Unix.W_OK]; (*unit, if file can be overwritten, exception otherwise*)
					with Unix.Unix_error _ -> gl_failwith ("Event logfile \"" ^ eventlogfile ^ "\" already exists and has no write permissions.") 
				)
				with _ -> ()
			) 
		in
		List.iter file_exists_and_is_writable eventlog_list_permissions;
		
		eventlog_list
	in (*logging*)
	let xmlscenelist = 
		let l = children input_structure "Scene" in
		let find_scene scene_id =
			try
			List.find (fun t -> Xml.attrib t "SceneID" = scene_id) l
			with Not_found -> gl_failwith ("Impossible scene inheritance: SceneID \"" ^ scene_id ^ "\" does not exist.") 
		in
		(*process inherited scenes:*)
		let inh_counter = ref 0 in
		let rec clean_inherited sc =
			incr inh_counter;
			if !inh_counter > 100000 then
				gl_failwith ("Stack overflow in scene inheritance. Maybe looping inheritance?") ;
			let inh_scenes = children sc "InheritScene" in
			if List.length inh_scenes > 0 then
			(
				(*process last "InheritScene" first:*)
				let (isc, isc_rest) =
					let lr = List.rev inh_scenes in
					(List.hd lr, List.rev (List.tl lr))
				in
				let inherited_scene =
					let scid = Xml.attrib isc "InheritFromScene" in
					if scid = Xml.attrib sc "SceneID" then
						gl_failwith ("Error in scene inheritance: Scene \"" ^ scid ^ "\" cannot inherit itself."); 
					clean_inherited (find_scene scid)
				in
				let loc_decl =
					let ld_sc = 
						child_or_default sc "LocalDeclarations" in
					if bool_of_string (aod isc "InheritLocalDeclarations" "true") then
						let ld_inh = child_or_default inherited_scene "LocalDeclarations" in
						Xml.Element 
						(
							"LocalDeclarations", 
							[], 
							(Xml.children ld_inh)
							@ (Xml.children ld_sc)
						);
					else
						ld_sc
				and loc_disp_obj =	(*deprecated!*)
					let ld_sc = child_or_default sc "LocalDisplayObjects" in
					if bool_of_string (aod isc "InheritLocalDisplayObjects" "true") then
						let ld_inh = child_or_default inherited_scene "LocalDisplayObjects" in
						Xml.Element 
						(
							"LocalDisplayObjects", 
							[], 
							(Xml.children ld_inh)
							@ (Xml.children ld_sc)
						)
					else
						ld_sc
				and new_contents =
					let new_events_list =
						if bool_of_string (aod isc "InheritEvents" "true") then
							(children inherited_scene "Events") @ (children sc "Events")
						else
							(children sc "Events")
					in
					let objlist = 
					[
						"Stimulus";
						"Image";
						"Button";
						"ProcessStimulusList";
						"ProcessImageList"
					] in
					let sc_conts = 
						List.filter (fun t -> List.mem (Xml.tag t) objlist) (Xml.children sc)
					in
					if bool_of_string (aod isc "InheritSceneContents" "true") then
						let inh_conts = 
							List.filter 
								(fun t -> List.mem (Xml.tag t) objlist) 
								(Xml.children inherited_scene)
						in
						inh_conts @ new_events_list @ sc_conts
					else
						sc_conts @ new_events_list
				in
				let new_scene = Xml.Element 
				(
					"Scene", 
					Xml.attribs sc, 
					isc_rest @ (loc_decl :: loc_disp_obj :: new_contents)
				) in
				clean_inherited new_scene
			)
			else
				sc
		in
		List.map clean_inherited l
	and xmlseqlist = children input_structure "Sequence" in
	let color_of_string s =
		let l = float_list_of_string (color_string s) in
		if (List.length l) = 3 && ( List.for_all ( fun t -> (t <= 1.0) && (t >= 0.0) ) l ) then
			(List.nth l 0, List.nth l 1, List.nth l 2)
		else
			gl_failwith ("Attribute "^s^" is not a valid color. Valid colors are lists of three elements within the interval [0.0, 1.0].") 
	in
	
	let sequence_of_xml q x =
		let repetitions = int_of_string (aod x "Repeat" "1") in
		let current_status_of_xml s =
			{
				sceneID = (Xml.attrib s "SceneID");
				duration = number_of_string (aod s "Duration" "100");
				duration_in_frames =
				(
					match (aod s "Unit" "frames") with
					"frames" -> true
					| _ -> (frame_rate_must_be_checked := true; false)
				);
				wake_up_on_event_mode = bool_of_string (aod s "WakeUpOnEventMode" "false")
			}
		in
		let sd = (aod x "RandomSeed" "AUTO") in
		if sd <> "AUTO" then
		(
			let randseed = 
				if sd <> "TIME" then
					try int_of_string (Xml.attrib x "RandomSeed")
						with Failure _ -> gl_failwith "RandomSeed must be integer." 
				else
					int_of_float (Unix.time ()) 
			in
			Random.init randseed;
		)
		else
			Random.self_init ();
		let shuffle = List.sort (fun _ _ -> (Random.int 3) - 1) in
		let seqlist = Xml.map current_status_of_xml x in
		let rec add_to_current_status_queue slist number =
			if number > 0 then
			(
				let sl () = 
					if bool_of_string (aod x "Shuffle" "false") then
						shuffle slist
					else
						slist
				in
				ignore ( List.fold_left (fun k el -> Queue.push el k; k) q (sl ()) );	
				add_to_current_status_queue slist (number-1)
			)
		in
		add_to_current_status_queue seqlist repetitions;
	in
	(*return list of exactly three numbers, for colors or scalings:*)
	let number_triple_list_of_string s =
		let l = 
			try number_list_of_string s with
			Mathexpr.Undeclared_identifier idf -> (*this can't happen anymore...*)
				gl_failwith ("Color or scaling attribute " ^ s ^ " contains undeclared identifier: \"" ^ idf ^ "\". Check the spelling. Maybe you want to use a number instead?") 
			| Mathexpr.Invalid_declaration_type dt ->
				gl_failwith ("Color or scaling attribute " ^ s ^ " contains invalid declaration type. Identifier \"" ^ dt ^ "\" is a function with more than zero arguments.") 
		in
		if (List.length l) = 3 then
			l
		else
			gl_failwith ("Color or scaling attribute " ^ s ^ " is not a list of three elements") 
	in
	
	let  font_of_xml x () = match (aod x "Font" "Helvetica18") with
		"Fixed8x13" -> Glut.BITMAP_8_BY_13
		| "Fixed9x15" -> Glut.BITMAP_9_BY_15
		| "TimesRoman10" -> Glut.BITMAP_TIMES_ROMAN_10
		| "TimesRoman24" -> Glut.BITMAP_TIMES_ROMAN_24
		| "Helvetica10" -> Glut.BITMAP_HELVETICA_10
		| "Helvetica12" -> Glut.BITMAP_HELVETICA_12
		| "Helvetica18" -> Glut.BITMAP_HELVETICA_18
		| _ -> Glut.BITMAP_HELVETICA_18
	in
	let to_displist ob () =	
		let t = GlList.create `compile in
		List.iter (fun x -> x ()) ob;
		GlList.ends ();
		t
	in
	(*convert the numbers \{...} in string s:*)
	let convert_numbers_in_string (*?(compile_flag=(ref true))*) ?(float_of_string_fun=float_of_string') s =
		let rec numlist_from_string l st pos =
			try
			let np = Str.search_forward (Str.regexp "\\(\\\\{[^}]+}\\)") st pos in
			let nl = (Str.matched_group 1 st) :: l in
			numlist_from_string nl st (np+1)
			with Not_found -> l
		in
		let numlist = numlist_from_string [] s 0 in
		let idlist = 
			List.map (Str.global_replace (Str.regexp "\\\\{\\(.*\\)}") "\\1") numlist
		in
		(*let float_of_string_fun s =
			let check_compile_flag nu =
				compile_flag :=	
				(
					match nu with
					C c -> !compile_flag
					| _ -> false
				)
			in
			let n = number_of_string s in
			check_compile_flag n;
			Printf.printf "%s -> %f\n" s (float_of_number n);
			float_of_number n
		in*)
		
		let rec replace_numbers st l =
			match l with
			[] -> st
			| _ -> 
				let ident = snd (List.hd l) in
				let target_string =
					try	(*try if it's a number*)
					let fl = float_of_string_fun ident in
					let n1 = string_of_float fl in
					let n = 
						if Str.last_chars n1 1 = "." then 
							Str.first_chars n1 ((String.length n1) - 1)
						else
							n1
					in
					n
					with Failure _ -> (*assume a string*)
					(	
						let ss = string_or_string_fun_of_string ident in
						match ss with
						SF f -> f ()
						| SC sc -> sc
					)
				in
				replace_numbers 
					(Str.global_replace (Str.regexp_string (fst (List.hd l))) target_string st) 
					(List.tl l)
		in
		replace_numbers s (List.combine numlist idlist)
	in
	let rec get_declared_object object_type ob =
		let identifier = (Xml.attrib ob "Identifier") in
		let declared_object = 
		(
			try
			get_declared_contents identifier
			with 
				Mathexpr.Undeclared_identifier s -> gl_failwith 
					("Identifier "^ s ^ " is not declared.") 
					
		) in
		(
			match declared_object with
			Display_object d ->
			(
				match d with
				Image _ -> 
				(
					match object_type with
					[`image] ->
						d
					| _ -> gl_failwith ("display object " ^ identifier ^ " is not an image.")
				)
				| Image_scenewide _ -> 
				(
					match object_type with
					[`image] ->
						d
					| _ -> gl_failwith ("display object " ^ identifier ^ " is not an image.")
				)
				| Image_uncompiled _ -> 
				(
					match object_type with
					[`image] ->
						d
					| _ -> gl_failwith ("display object " ^ identifier ^ " is not an image.")
				)
				| _ -> 
				(
					match object_type with
					[`stimulus] ->
						d
					| _ -> gl_failwith ("display object " ^ identifier ^ " is not an image.")
				)
			)
			| _ -> gl_failwith ("display object " ^ identifier ^ " is a display object list but is here used as a single display object.")
		)
	in
	let get_declared_object_list ob =
		let identifier = (Xml.attrib ob "Identifier") in
		(
			try
			get_declared_contents identifier
			with 
				Mathexpr.Undeclared_identifier s -> gl_failwith 
					("Identifier "^ s ^ " is not declared.") 
					
		) 
	in
	
	let display_object_from_list elem =
		let dolist = 
			match get_declared_object_list elem with 
			Stimulus_list ls -> ls
			| Image_list ls -> ls
			| _ -> failwith ""
		in
		let ld = List.length dolist in
		let listfun = (List.hd (Xml.children elem)) in
		match Xml.tag listfun with
		"ItemAtPosition" -> 
			let pos = int_of_float (float_of_string' (aod listfun "PositionInList" "1")) in
			if 0 < pos && pos <= ld then
				snd (List.nth dolist (pos - 1))
			else
				gl_failwith ("Function ItemAtPosition tries to extract item number "^(string_of_int pos)^", but the list requires items in the range [1, "^(string_of_int ld)^"]") 
		| "ItemWithIdentifier" ->
		(
			let item_id = (Xml.attrib listfun "Identifier") in
			try
			List.assoc item_id dolist
			with 
				_ -> gl_failwith ("Function ItemWithIdentifier tries to extract item \""^item_id^"\", but there is no such item in the list. Maybe it is a file name and you forgot the extension?") 
		)
		| "ListFirst" ->
			snd (List.hd dolist)
		| "ListLast" ->
			snd (List.hd (List.rev dolist))
		| _ -> 	(*ChooseItem*)
			Random.self_init ();
			snd (List.nth dolist (Random.int ld))
	in
	
	(*return refrences (SDL surface, GlPix.t):*)
	let image_contents_of_xml ic =
		let matrix_of_xml m =
			Mathexpr.matrix_of_nom
			(
				(Mathexpr.function_of_xml "" m) [||]
			)
		in
		(*fill width to multiples of 8 (necessary on Windows)*)
		let (clip8, clip8color) =
		(
			bool_of_string (aod ic "SetWidth8nOnWindows" "true"),
			color_of_string (aod ic "SetWidthFillColor" "0.0, 0.0, 0.0")
		) in
		match Xml.tag ic with
		"DeclaredContents" -> 
		(
			match get_declared_object [`image] ic with
			Image i -> (i.image_content_surface, i.image_content_gl_pix)
			| Image_scenewide (fi, i) -> 
			(
				i := Some (fi ());
				match !i with
				Some im ->
					(im.image_content_surface, im.image_content_gl_pix)
				| _ -> failwith "Image_scenewide undeclared."
			)
			| Image_uncompiled iuc -> 
				let i = iuc () in
				(i.image_content_surface, i.image_content_gl_pix)
			| _ -> failwith "" (*this cannot happen*)
		)
		| "FromList" ->
		(
			match display_object_from_list ic with
			Image i -> (i.image_content_surface, i.image_content_gl_pix)
			| Image_scenewide (fi, i) ->
			(
				i := Some (fi ());
				match !i with
				Some im ->
					(im.image_content_surface, im.image_content_gl_pix)
				| _ -> failwith "Image_scenewide undeclared."
			)
			| Image_uncompiled iuc -> 
				let i = iuc () in
				(i.image_content_surface, i.image_content_gl_pix)
			| _ -> gl_failwith "FromList extracted a stimulus where an image was expected"
		)
		| "FromImageFile" ->
		(
			let filename = 
				let fn = (convert_numbers_in_string (Xml.attrib ic "Filename")) in
				if Sys.file_exists fn then
					fn
				else
					let fn_executable_dir = Filename.concat executable_dir fn in
					if Sys.file_exists fn_executable_dir then
						fn_executable_dir
					else
						gl_failwith ("Image file " ^ fn ^ " does not exist.") 
			in
			try
			(
				let surf = Objects2D.sdl_surface_of_image filename in
				(ref surf, ref (Objects2D.gl_pix_of_surface ~clip8:clip8 ~fill_color:clip8color surf))
				
			)
			with _ -> 
			(
				gl_failwith 
					("Could not open file \"" ^ filename ^ "\". The reason might be that the file was not readable, the file type is not supported, or the image dimensions were unreasonably large. You can also try to convert your image to a different format. It is suggested to use RGB mode for images.")
					
			)
		)
		| "RGBA" ->
		(
			(*let matrix_of_xml m =
				let expr = Mathexpr.ocaml_exp_of_xml [] m in
				(Mathexpr.matrix_function_of_expression expr) [||]
			in*)
			let matlist = List.map matrix_of_xml (Xml.children ic) in
			(*must have the same dimensions:*)
			let (r, c) = Gsl_matrix.dims (List.hd matlist) in
			if List.exists (fun t ->Gsl_matrix.dims t <> (r, c)) matlist then
				gl_failwith "RGBA matrices have different dimensions."; 
			let color_of_float rf gf bf =
				let f x =
					int_of_float (255.0 *. x)
				in
				(f rf , f gf, f bf)
			in
			let surf =
			(
				if List.length matlist < 4 then
				(
					let s = 
						Sdlvideo.create_RGB_surface
							[`SWSURFACE]
							~w:c
							~h:r
							~bpp:24
							~rmask:255l
							~gmask:65280l
							~bmask:16711680l
							~amask:0l
					in
					let rmat = List.nth matlist 0
					and gmat = List.nth matlist 1
					and bmat = List.nth matlist 2 in
					for i = 0 to r-1 do
						for j = 0 to c-1 do
							let col = color_of_float rmat.{i, j} gmat.{i, j} bmat.{i, j} in
							Sdlvideo.put_pixel_color s ~x:j ~y:(r-1-i) col
						done;
					done;
					s
				)
				else
				(
					let s = 
						Sdlvideo.create_RGB_surface
							[`SWSURFACE]
							~w:c
							~h:r
							~bpp:32
							~rmask:255l
							~gmask:65280l
							~bmask:16711680l
							~amask:(-16777216l)
					in
					let rmat = List.nth matlist 0
					and gmat = List.nth matlist 1
					and bmat = List.nth matlist 2 
					and amat = List.nth matlist 3 in
					for i = 0 to r-1 do
						for j = 0 to c-1 do
							let col = color_of_float rmat.{i, j} gmat.{i, j} bmat.{i, j}
							and alpha = int_of_float (255.0 *. amat.{i, j}) in
							let pixel = Sdlvideo.map_RGB s ~alpha:alpha col in
							Sdlvideo.put_pixel s ~x:j ~y:(r-1-i) pixel
						done;
					done;
					s
				)
			) in
			(ref surf, ref (Objects2D.gl_pix_of_surface ~clip8:clip8 ~fill_color:clip8color surf))
		)
		| "Gray" ->
			(*let matrix_of_xml m =
				let expr = Mathexpr.ocaml_exp_of_xml [] m in
				(Mathexpr.matrix_function_of_expression expr) [||]
			in*)
			let mat = matrix_of_xml (List.hd (Xml.children ic)) in
			(*DELETE this:
			let matarr = Gsl_matrix.to_arrays mat in
			Array.iter
				(
					fun t ->
						Array.iter
							(
								Printf.printf "%.0f ";
							)
							t;
							Printf.printf "\n";
				)
				matarr;*)
			let (r, c) = Gsl_matrix.dims mat in
			let s = 
				Sdlvideo.create_RGB_surface
					[`SWSURFACE]
					~w:c
					~h:r
					~bpp:24
					~rmask:255l
					~gmask:65280l
					~bmask:16711680l
					~amask:0l
			in
			for i = 0 to r-1 do
				for j = 0 to c-1 do
					let v = int_of_float (255.0 *. mat.{i, j}) in
					Sdlvideo.put_pixel_color s ~x:j ~y:(r-1-i) (v, v, v)
				done;
			done;
			(ref s, ref (Objects2D.gl_pix_of_surface ~clip8:clip8 ~fill_color:clip8color s))
		| _ -> (*FillWithColor*)
			let w = float_of_string' (Xml.attrib ic "Width")
			and h = float_of_string' (Xml.attrib ic "Height") in
			let hor_range = (snd displ.horizontal_range) -. (fst displ.horizontal_range)
			and vert_range = (snd displ.vertical_range) -. (fst displ.vertical_range) in
			let r =
				let pixel_width = hor_range /. (float_of_int displ.width) in
				int_of_float (w /. pixel_width)
			and c =
				let pixel_height = vert_range /. (float_of_int displ.height) in
				int_of_float (h /. pixel_height)
			in
			let col =
				let color_of_string s =
					let l = 
						List.map
							(fun x -> int_of_float (255.0 *. x))
							(float_list_of_string s) 
					in
					
					if (List.length l) = 3 && ( List.for_all ( fun t -> (t <= 255) && (t >= 0) ) l ) then
						(List.nth l 0, List.nth l 1, List.nth l 2)
					else
						gl_failwith ("Attribute "^s^" is not a valid color. Valid colors are lists of three elements within the interval [0.0, 1.0].") 
				in
				color_of_string (color_string (aod ic "ColorRGB" "1.0, 1.0, 1.0"))
			in
			let surf =
			(
				if (aod ic "Transparency" "0.0") = "0.0" then
				(
					let s = 
						Sdlvideo.create_RGB_surface
							[`SWSURFACE]
							~w:r
							~h:c
							~bpp:24
							~rmask:255l
							~gmask:65280l
							~bmask:16711680l
							~amask:0l
					in
					for i = 0 to r-1 do
						for j = 0 to c-1 do
							Sdlvideo.put_pixel_color s ~x:i ~y:j col
						done;
					done;
					s
				)
				else
				(
					let s = 
						Sdlvideo.create_RGB_surface
							[`SWSURFACE]
							~w:r
							~h:c
							~bpp:32
							~rmask:255l
							~gmask:65280l
							~bmask:16711680l
							~amask:(-16777216l)
					in
					let alphaf = 1.0 -. (float_of_string' (Xml.attrib ic "Transparency")) in
					let alpha = int_of_float (255.0 *. alphaf) in
					let pixel = Sdlvideo.map_RGB s ~alpha:alpha col in
					for i = 0 to r-1 do
						for j = 0 to c-1 do
							Sdlvideo.put_pixel s ~x:i ~y:j pixel
						done;
					done;
					s
				)
			) in
			(ref surf, ref (Objects2D.gl_pix_of_surface ~clip8:clip8 ~fill_color:clip8color surf))
	in
	let image_of_xml im =
		let im_fun () =
			let (cont_surface, cont_gl_pix) = 			
				image_contents_of_xml (List.hd (Xml.children im))
			in
			let pos = number_pair_of_string (aod im "LowerLeftCorner" "0.0, 0.0") in
			let (w, h, _) = Sdlvideo.surface_dims !cont_surface in
			let hor_range = (snd displ.horizontal_range) -. (fst displ.horizontal_range)
			and vert_range = (snd displ.vertical_range) -. (fst displ.vertical_range) in
			let image_width =
				let pixel_width = hor_range /. (float_of_int displ.width) in
				(float_of_int w) *. pixel_width
			and image_height =
				let pixel_height = vert_range /. (float_of_int displ.height) in
				(float_of_int h) *. pixel_height
			in
			let posx = 
				if bool_of_string (aod im "CenterHorizontally" "false") then
					C ( (hor_range -. image_width) /. 2.0 +. (fst displ.horizontal_range) )
				else 
					(fst pos)
			and posy =
				if bool_of_string (aod im "CenterVertically" "false") then
					C ( (vert_range -. image_height) /. 2.0 +. (fst displ.vertical_range) )
				else
					(snd pos)
			in
			{
				image_position = (posx, posy);
				image_scale = number_pair_of_string (aod im "Scale" "1.0, 1.0");
				image_width = image_width;
				image_height = image_height;
				image_flip_horizontally = bool_of_string (aod im "FlipHorizontally" "false");
				image_flip_vertically = bool_of_string (aod im "FlipVertically" "false");
				display_image = number_of_string (aod im "Display" "1.0");
				image_content_surface = cont_surface;
				image_content_gl_pix = cont_gl_pix
			}
		in
		let compile_type =
			let cont = (List.hd (Xml.children im)) in
			match Xml.tag cont with
			"DeclaredContents" ->
			(
				match get_declared_object [`image] cont with
				Image _ -> [`progstart]
				| Image_scenewide (_, _) -> [`scenestart]
				| Image_uncompiled _ -> [`never]
				| _ -> failwith "" (*this cannot happen*)	
			)
			| "FromList" ->
			(
				let inherit_flag () =
					match display_object_from_list cont with
					Image _ -> [`progstart]
					| Image_scenewide (_, _) -> [`scenestart]
					| Image_uncompiled _ -> [`never]
					| _ -> gl_failwith "FromList extracted a stimulus where an image was expected"
				in
				match aod cont "CompileIfImage" "AtProgramStart" with
				"AtProgramStart" -> inherit_flag ()
				| "AtSceneStart" -> if inherit_flag () <> [`never] then [`scenestart] else [`never]
				| _ -> [`never]
			)
			| _ ->
			(
				match aod cont "Compile" "AtProgramStart" with
				"AtProgramStart" -> [`progstart]
				| "AtSceneStart" -> [`scenestart]
				| _ -> [`never]
			)
		in
		(
			match compile_type with
			[`progstart] -> Image (im_fun ())
			| [`scenestart] -> Image_scenewide (im_fun, ref None)
			| _ -> Image_uncompiled im_fun
		)
	in
	
	let stimulus_of_xml stim =
		let compile_flag = ref true in	(*if true then compilation is possible*)
		let check_compile_flag n =
			compile_flag :=	(*only constants can be compiled*)
			(
				match n with
				C c -> !compile_flag
				| _ -> false
			)
		in
		let stimtype x () =
			let float_of_string_stim s =	(*LOCAL declarations here! *)
				let n = number_of_string s in
				check_compile_flag n;
				float_of_number n
			in
			let int_of_string_stim s =
				int_of_float (float_of_string_stim s)
			in
			let float_pair_of_string_stim s =
				let nl = number_list_of_string s in
				List.iter check_compile_flag nl;
				float_pair_of_string s;
			in
			let color_of_string_stim s =
				let cs = color_string s in
				let nl = number_list_of_string cs in
				List.iter check_compile_flag nl;
				color_of_string cs
			in
			let fos s =
				float_of_string_stim (Xml.attrib x s)
			and opt_fos s default =
				float_of_string_stim (aod x s default)
			and rp () = float_pair_of_string_stim (aod x "RelativePosition" "0.0, 0.0")
			and col () = 
				if attrib_declared "ColorRGB" x then
					Some ( color_of_string_stim (Xml.attrib x "ColorRGB") )
				else
					None	
			and poly_edges s =
				let ed = int_of_string_stim s in
				if ed > 2 then
					ed
				else
					gl_failwith "polygon edges must be an integer > 2" 
			and fnt = font_of_xml x in
			match (Xml.tag x) with
			"Rectangle" -> 
				Objects2D.rectangle ~width:(fos "Width") ~height:(fos "Height") ~color:(col ()) ~position:(rp ()) ()
			| "FilledRectangle" -> 
				Objects2D.filled_rectangle ~width:(fos "Width") ~height:(fos "Height") ~color:(col ()) ~position:(rp ()) ()
			| "RegularPolygon" -> 
				Objects2D.regular_polygon ~edges:(poly_edges (Xml.attrib x "Edges")) ~radius:(fos "Radius") ~angle:(opt_fos "Angle" "0.0") ~color:(col ()) ~position:(rp ()) ()
			| "FilledRegularPolygon" -> 
				Objects2D.filled_regular_polygon ~edges:(poly_edges (Xml.attrib x "Edges")) ~radius:(fos "Radius") ~angle:(opt_fos "Angle" "0.0") ~color:(col ()) ~position:(rp ()) ()
			| "Circle" -> 
				Objects2D.circle ~radius:(fos "Radius") ~color:(col ()) ~position:(rp ()) ()
			| "FilledCircle" -> 
				Objects2D.filled_circle ~radius:(fos "Radius") ~color:(col ()) ~position:(rp ()) ()
			| "PartialDisk" ->
				Objects2D.partial_disk ~inner:(fos "InnerRadius") ~outer:(fos "OuterRadius") ~start:(opt_fos "StartAngle" "0.0") ~sweep:(opt_fos "SweepAngle" "360.0") ~color:(col ()) ~position:(rp ()) ();
			| "Primitive" -> 
				let shape = match (Xml.attrib x "Shape") with
					"LineLoop" -> `line_loop
					| "LineStrip" -> `line_strip
					| "Lines" -> `lines
					| "Polygon" -> `polygon
					| "QuadStrip" -> `quad_strip
					| "Quads" -> `quads
					| "TriangleFan" -> `triangle_fan
					| "TriangleStrip" -> `triangle_strip
					| "Triangles" -> `triangles
					| _ -> `points
				in
				let points_x = float_list_of_string (Xml.attrib x "PointsXValues")
				and points_y = float_list_of_string (Xml.attrib x "PointsYValues") in 
				let points = Array.of_list (List.combine points_x points_y) in	(*NOT tail recursive!*)
				Objects2D.primitive (col ()) shape points
			| "TextLine" -> 
				let txt = convert_numbers_in_string (*~compile_flag:compile_flag*) ~float_of_string_fun:float_of_string_stim (Xml.attrib x "Text") in
				Objects2D.text_line ~text:txt ~font:(fnt ()) ~color:(col ()) ~position:(rp ()) ()
			| "TextField" ->
				let txt = convert_numbers_in_string (*~compile_flag:compile_flag*) ~float_of_string_fun:float_of_string_stim (Xml.attrib x "Text") in
				let pixel_width = ((snd displ.horizontal_range) -. (fst displ.horizontal_range)) /. (float_of_int displ.width)
				and pixel_height = ((snd displ.vertical_range) -. (fst displ.vertical_range)) /. (float_of_int displ.height) in
				Objects2D.text_field ~text:txt ~font:(fnt ()) ~width:(fos "Width") ~height:(fos "Height") ~pixel_width:pixel_width ~pixel_height:pixel_height ~alignment:(aod x "Alignment" "Left") ~draw_frame:(bool_of_string (aod x "DrawFrame" "true")) ~color:(col ()) ~position:(rp ()) ()
			| "ImageInsideStimulus" ->
				let pixel_width = ((snd displ.horizontal_range) -. (fst displ.horizontal_range)) /. (float_of_int displ.width)
				and pixel_height = ((snd displ.vertical_range) -. (fst displ.vertical_range)) /. (float_of_int displ.height) in
				(*if it contains a matrix function, it must not be compiled:*)
				if !compile_flag then
				(
					let rec tag_in_xml_tree t =
					(
						match Xml.children t with
						[] -> ()
						| l -> 
						(
							let e = List.exists (fun t -> Xml.tag t = "DeclaredMatrixFunction") l in
							compile_flag := !compile_flag && e;
							List.iter tag_in_xml_tree l
						)
					) in
					tag_in_xml_tree x
				);
				(*let pos = (float_pair_of_string (aod x "RelativeLowerLeftCorner" "0.0, 0.0")) in*)
				(*must not be compiled before glut is initialized!*)
				if !glut_initialized then
					let 
						surf = !(fst (image_contents_of_xml (List.hd (Xml.children x)))) 
					in 
					Objects2D.texture_image_of_surface ~surface:surf ~pixel_width:pixel_width ~pixel_height:pixel_height ();
			| _ -> ()
		in
		let contentlist = if List.length (Xml.children stim) < 1 then [] else
		(
			let firstchild = (List.hd (Xml.children stim)) in
			match Xml.tag firstchild with
			"DeclaredContents" ->
			(
				match get_declared_object [`stimulus] firstchild with
				Stimulus_scenewide st ->
					compile_flag := true;
					st.content_list
				| Stimulus_uncompiled st ->
					compile_flag := false;
					(st ()).content_list
				| _ -> failwith "" (*this cannot happen*)
			)
			| "FromList" ->
			(
				match display_object_from_list firstchild with
				Stimulus_scenewide st ->
					compile_flag := true;
					st.content_list
				| Stimulus_uncompiled st ->
					compile_flag := false;
					(st ()).content_list
				| _ -> 
					gl_failwith "FromList extracted an image where a stimulus was expected." 
			)
			| "InsertList" ->
			(
				compile_flag := true;
				let stimlist = 
					match get_declared_object_list firstchild with
					| Stimulus_list sl -> sl
					| _ -> gl_failwith "InsertList extracted an image where a stimulus was expected." 
				in
				let rec make_content_list cl = function
					| [] -> cl
					| h::t ->
						let get_contents = function
							| (_, Stimulus_scenewide st) -> st.content_list
							| (_, Stimulus_uncompiled st) -> 
								compile_flag := false;
								(st ()).content_list
							| _ -> failwith "" (*cannot happen*)
						in
						make_content_list ((get_contents h) @ cl) t
				in
				make_content_list [] stimlist
			)
			| _ -> List.map stimtype (Xml.children stim)
		)
		in
		let st =
		{
			position = number_pair_of_string (aod stim "Position" "0.0, 0.0");
			color = number_triple_list_of_string (color_string (aod stim "ColorRGB" "1.0, 1.0, 1.0"));
			rotation = 
			(
				let st =  aod stim "RotationAngle" "0.0" in
				if st <> "0.0" then
					Some
					{
						rotation_angle = number_of_string st;
						rotation_center = number_pair_of_string  (aod stim "RelativeRotationCenter" "0.0,0.0")
					}
				else 
					None
			);
			scale_xyz =
			(
				let s = aod stim "ScaleXYZ" "1.0, 1.0, 1.0" in
				if (remove_spaces s) <> "1.0,1.0,1.0" then
					Some (number_triple_list_of_string s)
				else
					None
			);
			display_stimulus = number_of_string (aod stim "Display" "1.0");
			content_list = contentlist;
			(*content = to_displist contentlist;*)
			content_compiled = (*to_displist [] ()*)	
				(*planned: compile here for test reasons; this will be overwritten later!*)
				(*if compiled here, image is not working - program crashes!*)
				(to_displist contentlist) ()	(*this will be overwritten later*)
		} in
		(
			match aod stim "Compile" "auto" with
			"never" -> Stimulus_uncompiled (fun () -> st)
			| _ ->
				(
					match !compile_flag with
					true -> Stimulus_scenewide st
					| false -> Stimulus_uncompiled (fun () -> st)
						(*let var_number_frames_list =
							List.map
								(
									fun t -> 
										let v = snd t in
										v.number_of_frames
								)
								var_list
						in
						let no_inherit =
							not (List.mem None var_number_frames_list)
						in
						if no_inherit then
							let max_frames = 
								let m = ref 0 in
								List.iter
									(
										fun t ->
											if
									)
							in
							Stimulus_framewide 
						else
							Stimulus_uncompiled (fun () -> st)*)
				)
			)
	in	(*end of stimulus_of_xml*)
	
	let rec display_object_list_of_xml elem =
		let item_id () =	(*produce unique item IDs*)
			incr global_list_item_id;
			("ID" ^ (string_of_int !global_list_item_id))
		in
		let apply1 f =
			match (display_object_list_of_xml (List.hd (Xml.children elem))) with
			Stimulus_list l -> Stimulus_list (f l)
			| Image_list l -> Image_list (f l)
			| _ -> gl_failwith ("Function " ^(Xml.tag elem)^ " is applied to a non-list expression.") 
		in
		let apply2 f =
			let l1 = display_object_list_of_xml (List.hd (Xml.children elem))
			and l2 = display_object_list_of_xml (List.nth (Xml.children elem) 1) in
			match (l1, l2) with
			(Stimulus_list l1, Stimulus_list l2) -> Stimulus_list (f l1 l2)
			| (Image_list l1, Image_list l2) -> Image_list (f l1 l2)
			| _ -> gl_failwith ("Function " ^(Xml.tag elem)^ " is either applied to a non-list expression or to lists of different types.") 
		in
		(*the following 2 functions apply to assoc lists only!*)
		let list_difference l1 l2 =
			List.filter (fun t -> not (List.mem_assoc (fst t) l2)) l1
		and list_intersection l1 l2 =
			List.filter (fun t -> List.mem_assoc (fst t) l2) l1
		in
		match Xml.tag elem with
		"DisplayObjectList" -> 
			display_object_list_of_xml (List.hd (Xml.children elem))
		| "StimulusList" ->
			let stim_item st =
				(
					aod st "Identifier" (item_id ()),
					stimulus_of_xml st
				)
			in
			Stimulus_list (List.map stim_item (Xml.children elem))
		| "ImageList"->
			let im_item st =
				(
					aod st "Identifier" (item_id ()),
					image_of_xml st
				)
			in
			Image_list (List.map im_item (Xml.children elem))
		| "ListFromImageFiles"->
			let fstring = 
				remove_spaces (aod elem "FileNameList" "*.png, *.jpg") 
			in
			let stringlist = Str.split (Str.regexp ",") fstring in
			let list_of_pattern p =
				let dname = Filename.dirname p in
				let dircontent = 
					try
					Array.to_list (Sys.readdir dname)
					with Sys_error _ -> 
						gl_failwith ("FileNameList in ListFromImageFiles contains an invalid path: " ^ dname) 
				in
				let p_regexp = 
					let maskchar charstring =
						let rs = "\\" ^ charstring in
						Str.global_replace (Str.regexp rs) rs
					in
					let maskdot = maskchar "." (Filename.basename p) in
					let maskdol = maskchar "$" maskdot in
					let maskcirc = maskchar "^" maskdol in
					let maskplus = maskchar "+" maskcirc in
					let maskbrright = maskchar "[" maskplus in
					let maskbrleft = maskchar "]" maskbrright in
					let maskasterisk =
						Str.global_replace (Str.regexp "\\*") ".*" maskbrleft
					in
					let maskquest =
						Str.global_replace (Str.regexp "\\?") "." maskasterisk
					in
					Str.regexp (maskquest ^ "$")
				in
				let l = List.filter (fun t -> Str.string_match p_regexp t 0) dircontent in
				List.map (Filename.concat dname) l
			in
			let rec make_filelist fl l =
				match l with
				[] -> fl
				| _-> 
					let newfl = fl @ (list_of_pattern (List.hd l)) in
					make_filelist newfl (List.tl l)
			in
			let file_list = 
				let fl_unsorted = make_filelist [] stringlist in
				match (aod elem "Sort" "alphabetically") with
				"alphabetically" -> 
					List.sort String.compare fl_unsorted
				| "numerically" ->
					let sortfun x y =
						try
						let (x1, y1) = (Filename.chop_extension x, Filename.chop_extension y) in
						let (a, b) = (int_of_string x1, int_of_string y1) in
						compare a b
						with
							_ -> gl_failwith 
								"List of file names connot be sorted numerically. At least one argument is not a number." 
								
					in
					List.sort sortfun fl_unsorted
				| _ -> fl_unsorted
			in
			if List.length file_list < 1 then
				gl_failwith ("Function ListFromImageFiles was called with FileNameList \"" ^ fstring ^ "\", but no filename matches this list."); 
			let image_from_filename filename =
				let im = 
					Xml.Element ( "Image", [], 
					[
						Xml.Element ( "FromImageFile", ["Filename", filename], [] )
					] )
				in
				let im_id = 
					if bool_of_string (aod elem "IgnorePathNamesInIndices" "false") then
						Filename.basename filename
					else
						(*cut "current directory" (./):*)
						if Str.first_chars filename 2 = "./" then
							Str.string_after filename 2
						else
							filename
				in
				(im_id, image_of_xml im)
			in
			Image_list (List.map image_from_filename file_list)
		| "DeclaredList"->
			get_declared_object_list elem
		| "ListFromFunctionMapping" ->
		(
			(*let ar = Xml.attrib elem "Argument"*) 
			let arg_list = 
				let st =  Xml.attrib elem "ArgumentList" in
				Str.split (Str.regexp ",") (remove_spaces st)
			and identifier = Xml.attrib elem "ArgumentMatrixID"
			and compile_always = 
				bool_of_string (aod elem "CompileAlways" "true")
			and cont = List.hd (Xml.children elem) in
			List.iter
				(
				fun ar -> 
					Hashtbl.add 
						!Mathexpr.local_declarations 
						ar 
						(Mathexpr.Constant 0.0);
				)
				arg_list;
			try
			(*every row in mapping_matrix contains the arguments for each list item*)
			let mapping_matrix =
				let transpose mat =
					let (r, c) = Gsl_matrix.dims mat in
					let mt =  Gsl_matrix.create c r in
					Gsl_matrix.transpose mt mat;
					mt
				in
				let m = 
					transpose
					(
						match get_declared_matrix identifier with 
						MC m -> m
						| MF mf -> mf ()
					)
				in
				Gsl_matrix.to_arrays m
			in
			if List.length arg_list <> (Array.length mapping_matrix.(0)) then
				gl_failwith "ArgumentList of ListFromFunctionMapping has not the same length as its argument matrix has rows."; 
			(
				match Xml.tag cont with
				"StimulusFunction" ->
				(
					(*declare before evaluating f, then undeclare:*)
					let declare_temporarily argarr f =
						Array.iteri
							(fun i t ->
								Hashtbl.add
									!Mathexpr.local_declarations
									(List.nth arg_list i)
									(Mathexpr.Constant t))
							argarr;
						let r = f () in
						List.iter
							(fun ar -> 
								Hashtbl.remove 
									!Mathexpr.local_declarations 
									ar;)
							arg_list;
						r
					in
					let l = 
						let a =
							Array.map 
								(
									fun argarr ->
										let stimfun () =
										(
											match stimulus_of_xml cont with
											| Stimulus_uncompiled f -> 
												let new_stimfun () = 
													let st = f () in
													{
														position=st.position;
														color=st.color;
														rotation=st.rotation;	
														scale_xyz= st.scale_xyz;
														display_stimulus= st.display_stimulus;
														content_list= 
															List.map 
																(fun t -> fun () -> declare_temporarily argarr t) 
																st.content_list;
														content_compiled=st.content_compiled
													}
												in
												if compile_always then
													Stimulus_scenewide (new_stimfun ())
												else
													Stimulus_uncompiled new_stimfun
											| k -> k;
										) in
										let st = declare_temporarily argarr stimfun in
										((item_id ()), st)
								)
								mapping_matrix
						in
						Array.to_list a
					in
					List.iter
						(
							fun ar -> 
								Hashtbl.remove 
									!Mathexpr.local_declarations 
									ar;
						)
						arg_list;
					Stimulus_list l
				)
				| _ -> (*"ImageFunction"*)
				(
					(*let im_xml = List.hd (Xml.children cont) in*)
					let l = 
						let a =
							Array.map 
								(
									fun argarr ->
										Array.iteri
											(
												fun i t ->
													Hashtbl.replace
														!Mathexpr.local_declarations
														(List.nth arg_list i)
														(Mathexpr.Constant t)
											)
											argarr;
										Printf.printf "new: %f\n" argarr.(0); print_newline ();
										let im = image_of_xml cont in
										(
											let new_imfun im_fun = fun () ->
												Array.iteri
													(
														fun i t ->
															Hashtbl.add
																!Mathexpr.local_declarations
																(List.nth arg_list i)
																(Mathexpr.Constant t)
													)
													argarr;
												Printf.printf "new1: %f\n" argarr.(0); print_newline ();
												let new_im = im_fun () in
												List.iter
													(
													fun ar -> 
														Hashtbl.remove 
															!Mathexpr.local_declarations 
															ar;
													)
													arg_list;
												new_im
											in
											match im with
											Image i -> ((item_id ()), im)
											| Image_scenewide (fi, i) -> 
											(
												(item_id ()),
												Image_scenewide (new_imfun fi, i)
											)
											| Image_uncompiled uci -> 
											(
												(item_id ()),
												Image_uncompiled (new_imfun uci)
											)
											| _ -> failwith ""
										)
								)
								mapping_matrix
						in
						Array.to_list a
					in
					List.iter
						(
							fun ar -> 
								Hashtbl.remove 
									!Mathexpr.local_declarations 
									ar;
						)
						arg_list;
					Image_list l
				)
			)
			with
				Mathexpr.Invalid_declaration_type s -> 
					gl_failwith 
						("Function ListFromFunctionMapping uses declaration \"" ^ s ^ "\" which is not a vector/matrix or void matrix function.") 
						
		)
		| "ShuffleList" ->
			let rng = Mathexpr.rng_of_xml elem in
			let shuffle l = 
				let ar = Array.of_list l in
				Gsl_randist.shuffle rng ar;
				Array.to_list ar
			in
			apply1 shuffle
		| "ListTail" ->
			apply1 List.tl
		| "Sublist" ->
			let sublist sourcelist =
				let first_item = int_of_float (float_of_string' (aod elem "FirstItem" "1"))
				and last_item = int_of_float (float_of_string' (aod elem "LastItem" "5")) in
				if first_item < 1 || last_item > List.length sourcelist then
					gl_failwith 
						("Could not make sublist. FirstItem=" ^ (aod elem "FirstItem" "1") ^ " or LastItem=" ^(aod elem "LastItem=" "5")^ " are outside the possible range of items. Your list consists of " ^(string_of_int (List.length sourcelist))^ " items."); 
						
				let rec makelist l n =
					if n < first_item then
						l
					else
						makelist ((List.nth sourcelist (n-1)) :: l) (n-1)
				in
				makelist [] last_item
			in
			apply1 sublist
		| "ListFromIndexVector" ->
			let itemlist = Str.split (Str.regexp ",") (remove_spaces (Xml.attrib elem "IndexVector")) in
			let complement = bool_of_string (aod elem "Complement" "false") in
			let from_index_vector sourcelist =
				let get_item item =
					if List.mem_assoc item sourcelist then
						(item, List.assoc item sourcelist)
					else
						try 
						let i = int_of_float (float_of_string' item) in
						List.nth sourcelist (i-1)
						with
							_ -> gl_failwith ("IndexVector " ^(Xml.attrib elem "IndexVector")^ " contains item " ^ item ^ " which is not in list or out of range.") 
				in
				let rec makelist l = function
					| [] -> l
					| it::lt -> 
						makelist ((get_item it)::l) lt
				in
				let rl = makelist [] (List.rev itemlist) in
				if complement then
					list_difference sourcelist rl
				else
					rl
			in
			apply1 from_index_vector
		| "MatchRegExpOnIndexVector" ->
			let regexpstring = Xml.attrib elem "RegularExpression" in
			let rstringlist = Str.split (Str.regexp ",") (remove_spaces regexpstring) in
			let regexplist = List.map Str.regexp rstringlist in
			let complement = bool_of_string (aod elem "Complement" "false") in
			let from_regexp sourcelist =
				let index_vector_list = List.map fst sourcelist in
				(*
					find_string_matches result_list regexp_list string_list:
					give a list regexp_list of strings of regular expressions and
					return the matching subset of string_list appended to result_list
				*)
				let rec find_string_matches result_list regexp_list = function
					| [] -> (List.rev result_list)
					| current_string::restlist -> 
						let rec matches = function
							| [] -> false
							| h::t -> 
								(*return true if a substring of s matches:*)
								let matches_regexp re s =
									try
										ignore (Str.search_forward re s 0);
										true
									with Not_found -> false
								in
								if matches_regexp h current_string then
									true
								else
									matches t
						in
						find_string_matches
							(
								if (matches regexp_list) then 
									current_string::result_list
								else
									result_list
							)
							regexp_list
							restlist
				in
				(
					match find_string_matches [] regexplist index_vector_list with
					[] -> 
						gl_failwith ("MatchRegExpOnIndexVector: RegularExpression \"" ^ regexpstring ^ "\" does not match any of the list items.") 
					| matched_list ->
						let rl =
							List.filter
								( fun t -> List.mem (fst t) matched_list )
								sourcelist
						in
						if complement then
							list_difference sourcelist rl
						else
							rl
				)
			in
			apply1 from_regexp
		| "ListFromDeclaredVector" ->
		(
			let identifier = Xml.attrib elem "identifier" in
			let complement = bool_of_string (aod elem "Complement" "false") in
			try
			let index_vector =
				let m = 
					match get_declared_matrix identifier with 
					MC m -> m
					| MF mf -> mf ()
				in
				let flarr = Gsl_matrix.to_array m in
				Array.to_list (Array.map int_of_float flarr)
			in
			let from_index_vector sourcelist =
				let rec makelist l = function
					| [] -> l
					| it::lt -> 
						makelist ((List.nth sourcelist (it - 1))::l) lt
				in
				let rl = makelist [] (List.rev index_vector) in
				if complement then
					list_difference sourcelist rl
				else
					rl
			in
			apply1 from_index_vector
			with
				Mathexpr.Invalid_declaration_type s -> 
					gl_failwith 
						("Function ListFromDeclaredVector uses declaration \"" ^ s ^ "\" which is not a vector/matrix or void matrix function.") 
						
				| Invalid_argument _ | Failure "nth" -> 
					gl_failwith 
						("Function ListFromDeclaredVector applies an index vector with elements out of range.") 
						
		)
		| "ChooseItems" ->
			let rng = Mathexpr.rng_of_xml elem in
			let choose_items l = 
				let src = Array.of_list l in
				let nitems = (int_of_float (float_of_string' (aod elem "NumberOfItems" "5"))) in
				if nitems > Array.length src then
					gl_failwith ("Function ChooseItems is to choose " ^(aod elem "NumberOfItems" "5")^ " items, but the list to choose from has only " ^(string_of_int (Array.length src))^ " items");
				if nitems < 1 then
					gl_failwith "Function ChooseItems must at least choose 1 item"; 
				let dst = 
					Array.make nitems src.(0)
				in
				Gsl_randist.choose rng src dst;
				Gsl_randist.shuffle rng dst;
				Array.to_list dst
			in
			apply1 choose_items
		| "SampleItems" ->
			let rng = Mathexpr.rng_of_xml elem in
			let sample_items l = 
				let src = Array.of_list l in
				let nitems = (int_of_float (float_of_string' (aod elem "NumberOfItems" "5"))) in
				if nitems < 1 then
					gl_failwith "Function SampleItems must at least sample 1 item"; 
				let dst = 
					Array.make nitems src.(0)
				in
				Gsl_randist.sample rng src dst;
				Array.to_list dst
			in
			apply1 sample_items
		| "ReverseList" ->
			apply1 List.rev
		| "Unique" ->
			let rec unique = function 
				| [] -> []
				| x :: xt -> [x] @ (unique (List.filter (fun y -> y<>x) xt))
			in
			apply1 unique
		| "ConcatenateLists" ->
			let sourcelists = List.map display_object_list_of_xml (Xml.children elem) in
			(*check if all source lists are of the same type (ImageList or StimulusList):*)
			let check_equal_type l1 l2 =
				match (l1, l2) with
				(Stimulus_list _ , Stimulus_list _) -> true
				| (Image_list _ , Image_list _) -> true
				| _ -> false
			in
			if not (List.for_all (check_equal_type (List.hd sourcelists)) sourcelists) then
				gl_failwith "Function ConcatenateLists cannot concatenate an image list with a stimulus list or a list with a single item" ;
			let objectslist = 
				List.concat
				(
					List.map
						(
							function 
							| Stimulus_list l -> l 
							| Image_list l -> l
							| _ -> failwith ""
						)
						sourcelists
				)
			in
			(
				match List.hd sourcelists with
				Stimulus_list _ -> Stimulus_list objectslist
				| Image_list _ -> Image_list objectslist
				| _ -> gl_failwith "Function ConcatenateLists only concatenate lists" 
			)
		| "ListDifference" -> apply2 list_difference 
		| "ListIntersection" -> apply2 list_intersection
		| "ListAppend" ->
			let item_xml = List.nth (Xml.children elem) 1 in
			let sourcelist = display_object_list_of_xml (List.hd (Xml.children elem)) in
			let item =
				(
					aod item_xml "Identifier" (item_id ()),
					(
						match Xml.tag item_xml with
						"StimulusListItem" -> 
							stimulus_of_xml item_xml
						| _ -> image_of_xml item_xml
					)
				)
			in
			(
				match (sourcelist, Xml.tag item_xml) with
				(Stimulus_list l, "StimulusListItem") -> Stimulus_list (l @ [item])
				| (Image_list l, "ImageListItem") -> Image_list (l @ [item])
				| _ -> gl_failwith "ListAppend cannot append a stimulus to an image list or vice versa" 
			)
		| "ListPrepend" ->
			let item_xml = List.nth (Xml.children elem) 1 in
			let sourcelist = display_object_list_of_xml (List.hd (Xml.children elem)) in
			let item =
				(
					aod item_xml "Identifier" (item_id ()),
					(
						match Xml.tag item_xml with
						"StimulusListItem" -> 
							stimulus_of_xml item_xml
						| _ -> image_of_xml item_xml
					)
				)
			in
			(
				match (sourcelist, Xml.tag item_xml) with
				(Stimulus_list l, "StimulusListItem") -> Stimulus_list (item :: l)
				| (Image_list l, "ImageListItem") -> Image_list (item :: l)
				| _ -> gl_failwith "ListPrepend cannot prepend a stimulus to an image list or vice versa" 
			)
		| _ -> failwith "Not yet implemented..."
	in
	
	(*for processing stimulus and image lists:*)
	let object_list_reference_of_xml elem =
		let repeat = bool_of_string (aod elem "Repeat" "false") in
		let frames_per_item = number_of_string (aod elem "FramesPerItem" "10") in
		let identifier = Xml.attrib elem "Identifier" in
		match Xml.tag elem with
		"ProcessStimulusList" ->
			let object_ref_list =
				match get_declared_object_list elem with
				Stimulus_list sl ->
					List.map (fun t -> ref (snd t)) sl
				| _ -> gl_failwith ("ProcessStimulusList is to process declared object \"" ^identifier^ "\", but this is not a stimulus list") 
			in
			let stub = stimulus_of_xml elem in
			{
				object_stub = stub;
				frames_per_item = frames_per_item;
				object_list_repeat = repeat;
				object_reference_array = Array.of_list object_ref_list
			}
		| _ -> (*ProcessImageList*)
			(*implementation of this is horribly inefficient!*)
			let object_ref_list =
				match get_declared_object_list elem with
				Image_list il ->
					List.map (fun t -> ref (snd t)) il
				| _ -> gl_failwith ("ProcessImageList is to process declared object \"" ^identifier^ "\", but this is not an image list") 
			in
			let image_of_list_object pos =
				let listpos_xml = Xml.Element (
					"ItemAtPosition",
					["PositionInList", (string_of_int (pos + 1))],
					[]
				)
				in
				let from_list_xml = Xml.Element (
					"FromList",
					["Identifier", identifier],
					[listpos_xml]
				)
				in
				let image_xml = Xml.Element (
					"Image",
					Xml.attribs elem,
					[from_list_xml]
				)
				in
				ref (image_of_xml image_xml)
			in
			let refarr = 
				Array.init (List.length object_ref_list) image_of_list_object
			in
			{
				object_stub = !(refarr.(0));	(*not needed for images*)
				frames_per_item = frames_per_item;
				object_list_repeat = repeat;
				object_reference_array = refarr
			}
	in
	
	let declared_display_object_of_xml hashtbl elem =
		let identifier = Xml.attrib elem "Identifier" in
		let declared_contents =
			match Xml.tag elem with
			"DisplayObject" ->
			(
				let do_type = List.hd (Xml.children elem) in
				match Xml.tag do_type with
				"StimulusContents" ->
					Display_object (stimulus_of_xml do_type)
				| "ImageContents" -> 
					Display_object (image_of_xml do_type)
				| _ -> (*"FromList"*)
					Display_object (display_object_from_list do_type)
			)
			| _ -> (*DisplayObjectList*)
				display_object_list_of_xml elem
		in
		Hashtbl.add hashtbl identifier declared_contents;
		(identifier, declared_contents)
	in
	(*declare global display objects:*)
	let global_display_object_list_xml = 
		Xml.children (child_or_default input_structure "GlobalDisplayObjects")
	in
	
	(*global declarations:*)
	let global_declarations_list_xml = 
		Xml.children ( child_or_default input_structure "GlobalDeclarations") 
	in
	(*declare global declarations:*)
	let process_declarations math_hashtbl disp_obj_hashtbl =
		List.iter 
		(
			fun s -> 
				match Xml.tag s with
				"Declaration" ->
					let t = Mathexpr.declaration_of_xml s in
					Hashtbl.add math_hashtbl (fst t) (snd t);
				| "VariationInTime" -> 
					let defval = 
					Xml.Element (
						"NumericValue", ["Value", (aod s "InitialValue" "0.0")], []
					) in
					let defexpr =
					Xml.Element (
						"DefaultExpression", [], [defval]
					) in
					let declvar =
					Xml.Element (
						"Variable", [], [defexpr]
					) in
					let decl =
					Xml.Element (
						"Declaration",
						["Identifier", (Xml.attrib s "Identifier")],
						[declvar]
					) in
					let t = Mathexpr.declaration_of_xml decl in
					Hashtbl.add math_hashtbl (fst t) (snd t);
				| _ ->	(*display object*)
				(
					let (ident, decl_cont) = 
						declared_display_object_of_xml disp_obj_hashtbl s
					in
					(*represent the list length and indices in mathexpr;
					encode lengths with keys "@LENGTH@ key";
					encode indices with "@LISTINDEX@ identifier position":*)
					match decl_cont with
					| Stimulus_list l | Image_list l ->	
						(*length:*)
						Hashtbl.add 
							math_hashtbl
							("@LENGTH@ " ^ ident) 
							(Mathexpr.Constant (float_of_int (List.length l)));
						(*indices:*)
						Array.iteri
							(
								fun i t -> Hashtbl.add
									math_hashtbl
									("@LISTINDEX@ " ^ ident ^ " " ^ (string_of_int (i+1)))
									(Mathexpr.String_variable (ref (fst t)));
							)
							(Array.of_list l)
					| _ -> ()
				)	
		) 
	in
	process_declarations 
		Mathexpr.global_declarations 
		global_display_objects 
		(global_declarations_list_xml @ global_display_object_list_xml);
	
	(*List.iter 
		(declared_display_object_of_xml global_display_objects)
		global_display_object_list_xml;*)
	
	(*represent the list length in mathexpr;
	encode lengths with keys "@LENGTH@ key":*)
	(*let declare_list_lengths hashtbl objetcs_hashtbl =
		Hashtbl.iter
			(
				fun key -> function
					| Stimulus_list l | Image_list l ->	
						Hashtbl.add 
							hashtbl 
							("@LENGTH@ " ^ key) 
							(Mathexpr.Constant (float_of_int (List.length l)))
					| _ -> ()
			)
			objetcs_hashtbl
	in
	declare_list_lengths Mathexpr.global_declarations global_display_objects;*)
	
	let scene_of_xml tr =
		let local_declarations = Hashtbl.copy Mathexpr.global_declarations in
		(*let local_declaration_of_xml ld = 
			let locdec =
			(
				match (Xml.tag ld) with
				"Declaration" -> Mathexpr.declaration_of_xml ld
				| _ ->	
					let defval = 
					Xml.Element (
						"NumericValue", ["Value", (aod ld "InitialValue" "0.0")], []
					) in
					let defexpr =
					Xml.Element (
						"DefaultExpression", [], [defval]
					) in
					let declvar =
					Xml.Element (
						"Variable", [], [defexpr]
					) in
					let decl =
					Xml.Element (
						"Declaration",
						["Identifier", (Xml.attrib ld "Identifier")],
						[declvar]
					) in
					Mathexpr.declaration_of_xml decl
			) in
			Hashtbl.add local_declarations (fst locdec) (snd locdec);
		in*)
		let variation_of_xml v =
			let vname = (Xml.attrib v "Identifier") in
			let iv = float_of_string' (aod v "InitialValue" "0.0") 
			and fv = float_of_string' (aod v "FinalValue" "1.0") in
			let repeat = bool_of_string (aod v "Repeat" "true") in
			let nframes = 
			(
				let nf = aod v "NumberOfFrames" "Inherit" in
				match nf with
				"Inherit" -> None
				(*| _ -> Some (int_of_string nf)*)
				| _ -> Some (number_of_string nf)
			) in
			let varfun frames current_frame =
				let step = (fv -. iv) /. (float_of_int (frames - 1)) in 
				let i = float_of_int
				(
					if repeat then
						(*don't simplify this calculation - it's a workaround to some bug:*)
						((current_frame - 1) mod frames) + 1
					else
					(
						if current_frame > frames then
							frames
						else
							current_frame
					)
				) in
				match Mathexpr.get_declaration vname with
				Mathexpr.Variable r -> 
				(
					let given_function x =
						let fxml = child_or_default v "Function" in
						match Xml.children fxml with
						[] -> x
						| _ -> 
							Mathexpr.float_of_nom 
							(
								(Mathexpr.function_of_xml "" fxml) [|Mathexpr.Scal x|]
							)
					in
					r := given_function (iv +. (i -. 1.) *. step)
				)
				| Mathexpr.String_variable s ->
				(
					let given_function x =
						let fxml = child_or_default v "StringFunction" in
						match Xml.children fxml with
						[] -> string_of_float x
						| _ -> 
							Mathexpr.string_of_nom 
							(
								(Mathexpr.function_of_xml "" fxml) [|Mathexpr.Scal x|]
							)
					in
					s := given_function (iv +. (i -. 1.) *. step)
				)
				| _ -> gl_failwith (vname ^ " is not a variable and cannot be set in a variation of time.") 
				(*Printf.printf "step = %f, i = %f, newval = %f\n" step i (iv +. (i -. 1.) *. step);*)
			in
			{
				number_of_frames = nframes;
				variation_function = varfun
			}
		in
		
		(*converts SetVariable to variation:*)
		let set_variable_of_xml v =
			let identifier = Xml.attrib v "Identifier" in
			let cld = List.hd (Xml.children v) in
			let varref = 
			(
				try Hashtbl.find local_declarations identifier 
				with Not_found -> 
				gl_failwith ("Undeclared variable: " ^ identifier) 
			) in
			match Xml.tag v with
			"SetVariable" ->
				let reference = 
				(
					match varref with
					Mathexpr.Variable v -> v
					| _ -> gl_failwith (identifier ^ " is no numeric variable and cannot be set by SetVariable.") 
				) in
				let cont =
				(
					match Xml.tag cld with
					"IncreaseValue" -> 
						(F (fun () -> !reference +. 1.))
					| "DecreaseValue" -> 
						(F (fun () -> !reference -. 1.))
					| "SetToValue" ->	(*set to value*)
						(number_of_string (Xml.attrib cld "Value"))
					| _ ->	(*any mathematical expression*)
						F (
							fun () -> Mathexpr.float_of_nom 
								((Mathexpr.function_of_xml "" cld) [||])
						)
				) in
				let varfun frames current_frame =
					if current_frame < 2 then
						reference := float_of_number cont
				in
				{
					number_of_frames = None;
					variation_function = varfun
				}
			
			| "SetMatrixElementAtSceneStart" ->
				let mat =
				(
					try
					(
						match varref with
						Mathexpr.Matrix k -> k
						| _ -> gl_failwith (identifier ^ " is no matrix and cannot be modified by SetMatrixElementAtSceneStart.") 
					)
					with Not_found -> 
					gl_failwith ("Undeclared matrix: " ^ identifier) 
				) in
				let row_number_xml = List.hd (Xml.children (List.hd (Xml.children v))) 
				and col_number_xml = List.hd (Xml.children (List.nth (Xml.children v) 1)) in
				let row_and_col_function () =
					let (d1, d2) = Gsl_matrix.dims mat in
					let (r, c) =
					(
						(int_of_float (Mathexpr.float_of_nom ((Mathexpr.function_of_xml "" row_number_xml) [||]))) - 1,
						(int_of_float (Mathexpr.float_of_nom ((Mathexpr.function_of_xml "" col_number_xml) [||]))) - 1
					) in
					if 0 <= r && r < d1 && 0 <= c && c < d2 then
						(r, c)
					else
						gl_failwith ("SetMatrixElementAtSceneStart: Index out of bounds.") 
				in
				let cont_xml = List.nth (Xml.children v) 2 in
				let cont = 
				(
					match Xml.tag cont_xml with
					| "IncreaseValue" -> 
						fun () ->
							let (r, c) = row_and_col_function () in
							mat.{r, c} <- mat.{r, c} +. 1.
					| "DecreaseValue" -> 
						fun () -> 
							let (r, c) = row_and_col_function () in
							mat.{r, c} <- mat.{r, c} -. 1.
					| _ ->	(*any mathematical expression*)
						fun () -> 
							let (r, c) = row_and_col_function () in
							mat.{r, c} <- Mathexpr.float_of_nom ((Mathexpr.function_of_xml "" cont_xml) [||])
				) in
				let varfun frames current_frame =
					if current_frame < 2 then cont ()
				in
				{
					number_of_frames = None;
					variation_function = varfun
				}
			| "ModifyMatrixInSitu" ->	(*2 possibilities: 
								1) Specify a MatrixFunction -> is called with matrix as argument and modifies it in situ
								2) Specify a matrix expression which is the new value of the matrix;
								old and new matrix must have same dimensions*)
				let mat =
				(
					try
					(
						match varref with
						Mathexpr.Matrix k -> k
						| _ -> gl_failwith (identifier ^ " is no matrix and cannot be modified by ModifyMatrixInSitu.") 
					)
					with Not_found -> 
					gl_failwith ("Undeclared matrix: " ^ identifier) 
				) in
				let cont_xml = List.hd (Xml.children v) in
				let cont = 
				(
					match Xml.tag cont_xml with
					| "MatrixFunction" -> 
						fun () -> ignore ( (Mathexpr.function_of_xml "" cont_xml) [|Mathexpr.Mat mat|] )
					| _ ->	(*any matrix expression*)
						let (old_d1, old_d2) = Gsl_matrix.dims mat in
						fun () -> 
							let new_mat = Mathexpr.matrix_of_nom ((Mathexpr.function_of_xml "" cont_xml) [||]) in
							let (new_d1, new_d2) = Gsl_matrix.dims new_mat in
							if new_d1 <> old_d1 || new_d2 <> old_d2 then
								gl_failwith ("ModifyMatrixInSitu: old and new matrix have different dimensions"); 
							Gsl_matrix.memcpy ~src:new_mat ~dst:mat
				) in
				let varfun frames current_frame =
					if current_frame < 2 then cont ()
				in
				{
					number_of_frames = None;
					variation_function = varfun
				}
			| _ -> (*"SetStringVariable" *)
				let reference = 
				(
					match varref with
					Mathexpr.String_variable v -> v
					| _ -> gl_failwith (identifier ^ " is no string variable and cannot be set by SetStringVariable.") 
				) in
				let cont =
					fun () -> Mathexpr.string_of_nom 
						((Mathexpr.function_of_xml "" cld) [||])
				in
				let varfun frames current_frame =
					if current_frame < 2 then
						reference := cont ()
				in
				{
					number_of_frames = None;
					variation_function = varfun
				}
		in
		
		(*converte ExecuteProcess to variation:*)
		let execute_process_of_xml p =
			let start_frame = 
				int_of_float (float_of_string' (aod p "StartAtFrame" "1")) 
			in
			let cld = List.hd (Xml.children p) in
			match Xml.tag cld with
			"Interruptive" ->
				let pname = Xml.attrib cld "ExecuteString" in
				let process_fun = 
					try
					(match string_or_string_fun_of_string pname with
					SC s -> fun () -> s
					| SF f -> f)
					with Mathexpr.Undeclared_identifier i ->
						gl_failwith ("Process " ^ pname ^ " cannot be executed. Maybe you forgot to quote? For example: \"ls\" executes the program of this name, but ls looks for the defined string variable ls.");
				in
				let varfun frames current_frame =
					if current_frame = start_frame then
					(
						match Unix.system (process_fun ()) with
						Unix.WEXITED 127 ->
							gl_failwith ("Could not execute string \"" ^ (process_fun ()) ^ "\"")
						| _ -> ()
					)
				in
				{
					number_of_frames = None;
					variation_function = varfun
				}
			| _ -> (*NonInterruptive*)
				let file_or_default filenametag ?(inp=false) default_descr =
					match children cld filenametag with
					[] -> default_descr
					| h::t -> 
						let filename = Xml.attrib h "Filename" in
						if inp then
							Unix.openfile filename [Unix.O_RDONLY] 0o600
						else
							let flaglist =
								let append_or_new =
									if (bool_of_string (aod h "Append" "true")) then
										Unix.O_APPEND
									else Unix.O_TRUNC
								in
								append_or_new :: [Unix.O_CREAT; Unix.O_WRONLY]
							in
							Unix.openfile filename flaglist 0o600
				in
				let process_fun = 
					let pname = Xml.attrib cld "ProcessName" in
					try
					(match string_or_string_fun_of_string pname with
					SC s -> fun () -> s
					| SF f -> f)
					with Mathexpr.Undeclared_identifier i ->
						gl_failwith ("Process " ^ pname ^ " cannot be executed. Maybe you forgot to quote? For example: \"ls\" executes the program of this name, but ls looks for the defined string variable ls.");
				in
				let arg_array_fun () =
					let arg_of_str = function
						| SC s -> s
						| SF sf -> sf ()
					in
					try
					(Array.of_list 
						((process_fun ())::(List.map
							arg_of_str
							(string_or_string_fun_list_of_string 
								(aod cld "ProcessArguments" "")))))
					with Mathexpr.Undeclared_identifier i ->
						gl_failwith ("Process argument " ^ i ^ " was unquoted and is no string variable. Quote all arguments, e.g. \"a\", \"b\", ...");
				in
				let varfun frames current_frame =
					if current_frame = start_frame then
					(
						let in_descr = file_or_default "ProcessInputFromFile" ~inp:true Unix.stdin
						and out_descr = file_or_default "ProcessOutputToFile" Unix.stdout
						and err_decr = file_or_default "ProcessErrorsToFile" Unix.stderr in
						(ignore
							(Unix.create_process
								(process_fun ())
								(arg_array_fun ())
								in_descr
								out_descr
								err_decr))
					)
				in
				{
					number_of_frames = None;
					variation_function = varfun
				}
		in
		
		
		(* process declarations and variations in time: *)
		let loc_decl = child_or_default tr "LocalDeclarations" in
		(*set local_declarations in Mathexpr so that it can be found by the functions there:*)
		Mathexpr.local_declarations := local_declarations;
		
		(*declare local declarations:*)
		local_display_objects := Hashtbl.create 10;
		let local_display_object_list_xml = 	(*deprecated; only for backward compatibility*)
			Xml.children (child_or_default tr "LocalDisplayObjects")
		in
		process_declarations 
			local_declarations
			!local_display_objects
			((Xml.children loc_decl) @ local_display_object_list_xml);
		let var_list =
			let vlxml = children loc_decl "VariationInTime" in
			let var_to_set_list_xml = Xml.children (child_or_default tr "SetAtSceneStart") in
			let external_processes_list_xml = children tr "ExecuteProcess" in
			if List.length external_processes_list_xml > 0 && (not !allow_external_execution) then
				gl_failwith "SECURITY RISK: FlashDot tries to execute external processes but the flag -permit_exec was not set.";
			(List.map variation_of_xml vlxml) 
			@ (List.map set_variable_of_xml var_to_set_list_xml)
			@ (List.map execute_process_of_xml external_processes_list_xml)
		in
		
		(*** functions for variables, events etc.: ***)
		
		(*modify all DeclaredMatrix elements to "ModifyInSitu":*)
		let all_matrices_in_situ s =
			let rec replace_decl_mat st pos =
				try 
				let new_pos = 
					(Str.search_forward (Str.regexp "DeclaredMatrix *\\[\\([^]]*\\)\\]") st pos)
					+ 14
				in
				let found_contents = Str.matched_group 1 st in
				if not (Str.string_match (Str.regexp ".*ModifyInSitu.*") found_contents 0) then
				(
					let s1 = 
						if not (Str.string_match (Str.regexp ".*=.*") found_contents 0) then
							("Identifier=\"" ^ found_contents ^ "\"")
						else
							found_contents
					in
					let new_str =
						Str.global_replace (Str.regexp ("DeclaredMatrix *\\[" ^ found_contents ^ "\\]")) ("DeclaredMatrix["  ^ s1 ^ ", ModifyInSitu=\"true\"" ^ "]") st
					in
						replace_decl_mat new_str new_pos
				)
				else
					replace_decl_mat st new_pos	
				with Not_found -> st
			in
			replace_decl_mat s 0
		in
		let variables_to_log_list_of_string st =
			let strlist = separate_expressions (all_matrices_in_situ st) in
			(*convert expressions to functions:*)
			List.map
				(
					fun ss ->
						let no_leading_spaces = Str.global_replace (Str.regexp "^ +") "" ss in
						let s =
							if no_leading_spaces.[0] == '#' then 
								no_leading_spaces 
							else 
								("#" ^ no_leading_spaces)
						in
						let fct = fun () ->
						(
							try
							match number_of_string s with
							F f -> string_of_float (f ())
							| _ -> failwith ""
							with
								(*Mathexpr.Invalid_declaration_type s ->*)
								Failure flr ->
								(
									(*try*)
									match string_or_string_fun_of_string s with
									SF f -> f ()
									| _ -> failwith ""
									(*with 
										Failure f -> gl_failwith ("Error in VariablesToLog: " ^ f) 
										| _ -> gl_failwith ("Error in VariablesToLog: \"" ^ st ^ "\" is no valid expression or list of expressions.") *)
								)
								| Mathexpr.Undeclared_identifier i -> gl_failwith ("Error in VariablesToLog: \"" ^ i ^ "\" is not declared.") 
								| Mathexpr.Invalid_declaration_type i -> gl_failwith ("Error in VariablesToLog: \"" ^ i ^ "\" is of invalid type.") 
								| _ -> gl_failwith ("Error in VariablesToLog: \"" ^ st ^ "\" is no valid expression or list of expressions.") 
						) in
						(no_leading_spaces, fct)
				)
				strlist;
		in
		let matrices_to_log_list_of_string st = 
			let strlist = separate_expressions (all_matrices_in_situ st) in
			(*convert expressions to functions:*)
			List.map
				(
					fun ss ->
						let no_leading_spaces = Str.global_replace (Str.regexp "^ +") "" ss in
						let s =
							if no_leading_spaces.[0] == '#' then 
								no_leading_spaces 
							else 
								("#" ^ no_leading_spaces)
						in
						let fct =
						(
							try
							match matrix_of_string s with
							MF f -> f
							| _ -> failwith ""
							with
								Mathexpr.Undeclared_identifier i -> gl_failwith ("Error in MatricesToLog: \"" ^ i ^ "\" is not declared.") 
								| Failure f -> gl_failwith ("Error in MatricesToLog: " ^ f) 
								| _ -> gl_failwith ("Error in MatricesToLog: \"" ^ st ^ "\" is no valid expression or list of expressions.") 
						) in
						(no_leading_spaces, fct)
				)
				strlist;
		in
		let object_of_id_string s = 
		(
			try
			get_declared_contents s
			with 
				Mathexpr.Undeclared_identifier i -> gl_failwith 
					("Display object list identifier "^ i ^ " is not declared, so its index vector cannot be logged.") 
					
		) in
		let index_vectors_to_log_list_of_string st =
			let strlist = Str.split (Str.regexp ",") (remove_spaces st) in
			(*check whether declared:*)
			List.map
				(
					fun s ->
						let declared_object = object_of_id_string s in
						let ind_vec =
						(
							match declared_object with
							Stimulus_list sl -> List.map fst sl
							| Image_list il -> List.map fst il
							| _ -> gl_failwith ("The index vector of " ^ s ^ " is to be logged by an event, but it is not a display object list.") 
						) in
						(s, ind_vec)
				)
				strlist;
		in
		let list_index_to_log_of_xml x =
			let listid = Xml.attrib x "ListIdentifier"
			and number_of_item = number_of_string (Xml.attrib x "NumberOfItem") in
			let declared_object = object_of_id_string listid in
			let ind_vec =
			(
				match declared_object with
				Stimulus_list sl -> List.map fst sl
				| Image_list il -> List.map fst il
				| _ -> gl_failwith ("An index of " ^ listid ^ " is to be logged by an event, but it is not a display object list.") 
			) in
			let fct n =
				let pos = ((int_of_float (float_of_number n)) - 1) in
				if pos < 0 then
					 gl_failwith ("LogListIndex tries to log an item with a negative index (" 
					 ^ (string_of_int pos) 
					 ^ ") in list \"" ^ listid ^ "\". Did you forget a # in front of a variable name?");
				try
				List.nth ind_vec pos
				with _ -> gl_failwith ("LogListIndex tries to log item number " ^ (string_of_int pos) ^ " in list \"" ^ listid ^ "\", but it does not exist.") 
			in
			(listid, number_of_item, fct)
		in
		let button_of_xml b =
			let fos s =
				float_of_string (Xml.attrib b s)
			in
			
			let logfile_id = aod b "LogfileID" "standard_event_logfile" in
			(*check, if logfile_id is correct reference:*)
			if not (List.mem_assoc logfile_id logging) then
				gl_failwith ("LogfileID \"" ^ logfile_id ^ "\" does not exist."); 
			
			let (b_width, b_height) = (fos "Width", fos "Height") in
			let b_content b_state =
				let pixel_width = ((snd displ.horizontal_range) -. (fst displ.horizontal_range)) /. (float_of_int displ.width)
				and pixel_height = ((snd displ.vertical_range) -. (fst displ.vertical_range)) /. (float_of_int displ.height) in
				(*let button_proto f =
					f ~width:b_width ~height:b_height ~text:(Xml.attrib b "Text") ~font:(font_of_xml b ()) ~pixel_width:pixel_width ~pixel_height:pixel_height ~position:(float_pair_of_string (Xml.attrib b "Position")) ()
				in*)
				let bcont () = 
				(
					match b_state with
					Unpressed -> 
						Objects2D.button ~width:b_width ~height:b_height ~text:(Xml.attrib b "Text") ~font:(font_of_xml b ()) ~pixel_width:pixel_width ~pixel_height:pixel_height ~position:(float_pair_of_string (aod b "Position" "0.0, 0.0")) ()
					| Pressed -> 
						Objects2D.pressed_button ~width:b_width ~height:b_height ~text:(Xml.attrib b "Text") ~font:(font_of_xml b ()) ~pixel_width:pixel_width ~pixel_height:pixel_height ~position:(float_pair_of_string (aod b "Position" "0.0, 0.0")) ()
					| Mouseover ->
						Objects2D.mouseover_button ~width:b_width ~height:b_height ~text:(Xml.attrib b "Text") ~font:(font_of_xml b ()) ~pixel_width:pixel_width ~pixel_height:pixel_height ~position:(float_pair_of_string (aod b "Position" "0.0, 0.0")) ()
				) in
				let t = GlList.create `compile in
					bcont ();
				GlList.ends ();
				t
			in
			
			let sens_area =
				let (lx, ly) = float_pair_of_string (aod b "Position" "0.0, 0.0") in
				fun (x, y) -> 
					0. <= x -. lx && x -. lx <= b_width && 0. <= y -. ly && y -. ly <= b_height
			in
			let act =
				let a = 
					match children b "Action" with
					[] -> Xml.Element ("Nothing", [], [])
					| h::_ -> List.hd (Xml.children h)
				in
				match Xml.tag a with
				"Nothing" -> Nothing
				| "Exit" -> Exit
				| "JumpToNextPresentation" -> Jump_to_next_presentation
				| _ ->
					let queue_to_insert = 
						let cont = (List.hd (Xml.children a)) in
						let cld = Xml.children cont in
						let qt = Queue.create () in
						match Xml.tag cont with
						"Sequence" -> 
							sequence_of_xml qt cont;
							fun () -> qt
						| _ -> (*SequenceCase*)
							let (default, clist) = 
								let lr = List.rev cld in
								(
									List.hd lr, 
									List.map 
										(
											fun t -> 
												let s = Xml.children t in
												(
													List.hd s, 
													let q = Queue.create () in
													sequence_of_xml q (List.nth s 1);
													q
												)
										) 
										(List.tl lr)
								)
							in
							sequence_of_xml qt default;
							let bool_of_cond = function
								| Mathexpr.Scal 0.0 -> false
								| _ -> true
							in
							let rec find_condition = function
								[] -> qt
								| h::t -> 
									if bool_of_cond (Mathexpr.function_of_xml "" (fst h) [||]) then
										(snd h)
									else
										find_condition t
							in
							fun () -> find_condition clist
					in
					match aod a "StopCurrentPresentation" "true" with
					"true" -> Insert_sequence_stop queue_to_insert
					| _ -> Insert_sequence queue_to_insert
			in
			let list_indices_to_log_list =
				List.map list_index_to_log_of_xml (children b "LogListIndex")
			in
			(*let variables_to_set_list_xml = children b "SetVariableByEvent" in*)
			let variables_to_set_list_xml = 
				List.filter 
					(fun t -> Xml.tag t = "SetVariableByEvent" || Xml.tag t = "SetStringVariableByEvent" || Xml.tag t = "SetMatrixElementByEvent") 
					(Xml.children b) 
			in
			let variables_to_set_list = 
				let var_to_set_of_xml v = 
					let identifier = Xml.attrib v "Identifier" in
					let varref =
					(
						try Hashtbl.find local_declarations identifier 
						with Not_found -> 
						gl_failwith ("Undeclared variable: " ^ identifier) 
					) in
					match Xml.tag v with
					"SetVariableByEvent" ->
						let reference = 
						(
							match varref with
							Mathexpr.Variable v -> v
							| _ -> gl_failwith (identifier ^ " is no numeric variable and cannot be set by SetVariableByEvent.") 
						) in
						let cont_xml = List.hd (Xml.children v) in
						(
							match Xml.tag cont_xml with
							"FunctionOfEventComponent" ->
								let given_function x =
									let fxml = child_or_default cont_xml "Function" in
									reference := 
									(
										match Xml.children fxml with
										[] -> x
										| _ ->
											Mathexpr.float_of_nom
											(
												(Mathexpr.function_of_xml "" fxml) [|Mathexpr.Scal x|]
											)
									)
								in
								(
									match aod cont_xml "EventComponent" "TimeFromSceneOnset" with
									"MousePositionX" -> Set_by_mouse_position_x given_function
									| "MousePositionY" -> Set_by_mouse_position_y given_function
									| _ -> Set_by_time_from_onset given_function
								)
							| "IncreaseValue" -> 
								Set_to_number (fun () -> reference := !reference +. 1.)
							| "DecreaseValue" -> 
								Set_to_number (fun () -> reference := !reference -. 1.)
							| "SetToValue" ->	(*set to value*)
								Set_to_number 
								( 
									fun () -> 
										reference := float_of_number (number_of_string (Xml.attrib cont_xml "Value")) 
								)
							| _ ->	(*any mathematical expression*)
								Set_to_number 
								(
									fun () -> 
										reference := Mathexpr.float_of_nom ((Mathexpr.function_of_xml "" cont_xml) [||])
								)
						)
					| "SetMatrixElementByEvent" ->
						let mat =
						(
							try
							(
								match varref with
								Mathexpr.Matrix k -> k
								| _ -> gl_failwith (identifier ^ " is no matrix and cannot be modified by SetMatrixElementByEvent.") 
							)
							with Not_found -> 
							gl_failwith ("Undeclared matrix: " ^ identifier) 
						) in
						let row_number_xml = List.hd (Xml.children (List.hd (Xml.children v))) 
						and col_number_xml = List.hd (Xml.children (List.nth (Xml.children v) 1)) in
						let row_and_col_function () =
							let (d1, d2) = Gsl_matrix.dims mat in
							let (r, c) =
							(
								(int_of_float (Mathexpr.float_of_nom ((Mathexpr.function_of_xml "" row_number_xml) [||]))) - 1,
								(int_of_float (Mathexpr.float_of_nom ((Mathexpr.function_of_xml "" col_number_xml) [||]))) - 1
							) in
							if 0 <= r && r < d1 && 0 <= c && c < d2 then
								(r, c)
							else
								gl_failwith ("SetMatrixElementByEvent: Index out of bounds.") 
						in
						let cont_xml = List.nth (Xml.children v) 2 in
						(
							match Xml.tag cont_xml with
							"FunctionOfEventComponent" ->
								let given_function x =
									let fxml = child_or_default cont_xml "Function" in
									let (r, c) = row_and_col_function () in
									mat.{r, c} <- 
									(
										match Xml.children fxml with
										[] -> x
										| _ ->
											Mathexpr.float_of_nom
											(
												(Mathexpr.function_of_xml "" fxml) [|Mathexpr.Scal x|]
											)
									)
								in
								(
									match aod cont_xml "EventComponent" "TimeFromSceneOnset" with
									"MousePositionX" -> Set_by_mouse_position_x given_function
									| "MousePositionY" -> Set_by_mouse_position_y given_function
									| _ -> Set_by_time_from_onset given_function
								)
							| "IncreaseValue" -> 
								Set_to_number 
								(
									fun () -> 
										let (r, c) = row_and_col_function () in
										mat.{r, c} <- mat.{r, c} +. 1.
								)
							| "DecreaseValue" -> 
								Set_to_number 
								(
									fun () -> 
										let (r, c) = row_and_col_function () in
										mat.{r, c} <- mat.{r, c} -. 1.
								)
							| _ ->	(*any mathematical expression*)
								Set_to_number 
								(
									fun () -> 
										let (r, c) = row_and_col_function () in
										mat.{r, c} <- Mathexpr.float_of_nom ((Mathexpr.function_of_xml "" cont_xml) [||])
								)
						)
						
					| _ -> (*SetStringVariableByEvent*)
						let reference = 
						(
							match varref with
							Mathexpr.String_variable v -> v
							| _ -> gl_failwith (identifier ^ " is no string variable and cannot be set by SetStringVariableByEvent.") 
						) in
						let cld = List.hd (Xml.children v) in
						(
							match Xml.tag cld with
							"StringFunctionOfEventComponent" ->
								let given_function x =
									let fxml = child_or_default cld "StringFunction" in
									reference := 
									(
										match Xml.children fxml with
										[] -> string_of_float x
										| _ ->
											Mathexpr.string_of_nom
											(
												(Mathexpr.function_of_xml "" fxml) [|Mathexpr.Scal x|]
											)
									)
								in
								(
									match aod cld "EventComponent" "TimeFromSceneOnset" with
									"MousePositionX" -> Set_by_mouse_position_x given_function
									| "MousePositionY" -> Set_by_mouse_position_y given_function
									| _ -> Set_by_time_from_onset given_function
								)
							| _ ->	(*any mathematical expression*)
								Set_to_number 
								(
									fun () -> 
										reference := Mathexpr.string_of_nom ((Mathexpr.function_of_xml "" cld) [||])
								)
						)
				in
				List.map var_to_set_of_xml variables_to_set_list_xml
			in
			Button {
				buttonID = (Xml.attrib b "ButtonID");
				button_state = Unpressed;
				button_log_event = number_of_string (aod b "Log" "0.0");
				button_event_logfile_id = logfile_id;
				button_action = act;
				button_variables_to_set = variables_to_set_list;
				button_variables_to_log = variables_to_log_list_of_string (aod b "VariablesToLog" "");
				button_matrices_to_log = matrices_to_log_list_of_string (aod b "MatricesToLog" "");
				button_index_vectors_to_log = index_vectors_to_log_list_of_string (aod b "IndexVectorsToLog" "");
				button_list_indices_to_log = list_indices_to_log_list;
				button_log_before_set = bool_of_string (aod b "LogBeforeSet" "false");
				button_sensitive_area = sens_area;
				button_content = b_content;
				button_pressed_compiled = b_content Pressed;
				button_unpressed_compiled = b_content Unpressed;
				button_mouseover_compiled = b_content Mouseover;
			}
		in
		let event_of_xml ev =
			let eventtype = List.nth (Xml.children ev) 0 in
			let eventaction = 
				match children ev "Action" with
				[] -> Xml.Element ("Action", [], [Xml.Element ("Nothing", [], [])])
				| h::t -> h
			in
			let logfile_id = aod ev "LogfileID" "standard_event_logfile" in
			(*check, if logfile_id is correct reference:*)
			if not (List.mem_assoc logfile_id logging) then
				gl_failwith ("LogfileID \"" ^ logfile_id ^ "\" does not exist."); 
			
			let conts =
				match Xml.tag eventtype with
				"Key" ->
					Keyboard_event 
					(
						let keytype = List.hd (Xml.children eventtype) in
						match Xml.tag keytype with
						"ASCII" -> Ascii_key (Xml.attrib keytype "Sign").[0]
						| _ -> 
							let sk = Xml.attrib keytype "SpecialKey" in
							if List.mem sk ["Enter"; "Escape"; "Backspace"; "Tab"; "Space"] then
								Ascii_key
								(
									match sk with
									"Enter" -> char_of_int 13
									| "Escape" -> char_of_int 27
									| "Backspace" -> char_of_int 8
									| "Space" -> ' '
									| _ -> char_of_int 9	(*TAB*)
								)
							else
								Special_key
								(
									match sk with
									"F1" -> Glut.KEY_F1
									| "F2" -> Glut.KEY_F2
									| "F3" -> Glut.KEY_F3
									| "F4" -> Glut.KEY_F4
									| "F5" -> Glut.KEY_F5
									| "F6" -> Glut.KEY_F6
									| "F7" -> Glut.KEY_F7
									| "F8" -> Glut.KEY_F8
									| "F9" -> Glut.KEY_F9
									| "F10" -> Glut.KEY_F10
									| "F11" -> Glut.KEY_F11
									| "F12" -> Glut.KEY_F12
									| "CursorLeft" -> Glut.KEY_LEFT
									| "CursorRight" -> Glut.KEY_RIGHT
									| "CursorUp" -> Glut.KEY_UP
									| "CursorDown" -> Glut.KEY_DOWN
									| "PageUp" -> Glut.KEY_PAGE_UP
									| "PageDown" -> Glut.KEY_PAGE_DOWN
									| "Home" -> Glut.KEY_HOME
									| "End" -> Glut.KEY_END
									| _ -> Glut.KEY_INSERT
								)
					)
				| _ -> 
					Mouse_event {
						button = 
						(
							match aod eventtype "MouseButton" "Left" with
							"Left" -> Glut.LEFT_BUTTON
							| "Right" -> Glut.RIGHT_BUTTON
							| _ -> Glut.MIDDLE_BUTTON
							(*| _ -> Glut.OTHER_BUTTON 0*) (*what int value corresponds to which button?*)
						);
						up_or_down = 
						(
							match aod eventtype "PressedOrReleased" "Pressed" with
							"Pressed" -> Glut.DOWN
							| _ -> Glut.UP
						);
						sensitive_area = 
						(
							let area = match Xml.children eventtype with
								[] -> Xml.Element ("EntireWindow", [], [])
								| x::_ -> List.hd ( Xml.children x ) 
							in
							match Xml.tag area with
							"EntireWindow" -> fun (x, y) -> true
							| "Circular" ->
								let (cx, cy) = float_pair_of_string (Xml.attrib area "Center")
								and radius = float_of_string (Xml.attrib area "Radius") in
								fun (x, y) -> sqrt ( (x -. cx)**2. +. (y -. cy)**2. ) <= radius
							| "Rectangular" -> 
								let (lx, ly) = float_pair_of_string (Xml.attrib area "LowerLeftCorner")
								and w = float_of_string (Xml.attrib area "Width")
								and h = float_of_string (Xml.attrib area "Height") in
								fun (x, y) -> 
									0. <= x -. lx && x -. lx <= w && 0. <= y -. ly && y -. ly <= h
							| _ -> (*works only with CONVEX or STAR SHAPED polygons!*)
								let points_x = float_list_of_string (Xml.attrib area "PointsXValues")
								and points_y = float_list_of_string (Xml.attrib area "PointsYValues") in
								fun (x, y) ->
									let px = List.map (fun z -> z -. x) points_x
									and py = List.map (fun z -> z -. y) points_y in
									let p = List.combine px py in (*NOT tail recursive!*)
									let p1 = (List.tl p) @ [List.hd p] in (*NOT tail recursive!*)
									let pointpairs = List.combine p p1 in
									let angles = 
										List.map 
										(
											fun z -> 
												let ((ax, ay), (bx, by)) = z in
												acos ( (ax *. bx +. ay *. by) /. (sqrt (ax**2. +. ay**2.) *. sqrt (bx**2. +. by**2.)) )
										) pointpairs
									in
									let sum_angles = List.fold_left (+.) 0. angles in
									let pi = acos (-1.0) in
									sum_angles -. 2.*.pi <= 0.0001
						)
					}
			in
			let act =
				let a = (List.hd (Xml.children eventaction)) in
				match Xml.tag a with
				"Nothing" -> Nothing
				| "Exit" -> Exit
				| "JumpToNextPresentation" -> Jump_to_next_presentation
				| _ ->
					let queue_to_insert = 
						let cont = (List.hd (Xml.children a)) in
						let cld = Xml.children cont in
						let qt = Queue.create () in
						match Xml.tag cont with
						"Sequence" -> 
							sequence_of_xml qt cont;
							fun () -> qt
						| _ -> (*SequenceCase*)
							let (default, clist) = 
								let lr = List.rev cld in
								(
									List.hd lr, 
									List.map 
										(
											fun t -> 
												let s = Xml.children t in
												(
													List.hd s, 
													let q = Queue.create () in
													sequence_of_xml q (List.nth s 1);
													q
												)
										) 
										(List.tl lr)
								)
							in
							sequence_of_xml qt default;
							let bool_of_cond = function
								| Mathexpr.Scal 0.0 -> false
								| _ -> true
							in
							let rec find_condition = function
								[] -> qt
								| h::t -> 
									if bool_of_cond (Mathexpr.function_of_xml "" (fst h) [||]) then
										(snd h)
									else
										find_condition t
							in
							fun () -> find_condition clist
					in
					match aod a "StopCurrentPresentation" "true" with
					"true" -> Insert_sequence_stop queue_to_insert
					| _ -> Insert_sequence queue_to_insert
			in
			let list_indices_to_log_list =
				List.map list_index_to_log_of_xml (children ev "LogListIndex")
			in
			let variables_to_set_list_xml = 
				List.filter 
					(fun t -> Xml.tag t = "SetVariableByEvent" || Xml.tag t = "SetStringVariableByEvent" || Xml.tag t = "SetMatrixElementByEvent") 
					(Xml.children ev) 
			in
			let variables_to_set_list = 
				let var_to_set_of_xml v = 
					let identifier = Xml.attrib v "Identifier" in
					let varref =
					(
						try Hashtbl.find local_declarations identifier 
						with Not_found -> 
						gl_failwith ("Undeclared variable: " ^ identifier) 
					) in
					match Xml.tag v with
					"SetVariableByEvent" ->
						let reference = 
						(
							match varref with
							Mathexpr.Variable v -> v
							| _ -> gl_failwith (identifier ^ " is no numeric variable and cannot be set by SetVariableByEvent.") 
						) in
						let cont_xml = List.hd (Xml.children v) in
						(
							match Xml.tag cont_xml with
							"FunctionOfEventComponent" ->
								let given_function x =
									let fxml = child_or_default cont_xml "Function" in
									reference := 
									(
										match Xml.children fxml with
										[] -> x
										| _ ->
											Mathexpr.float_of_nom
											(
												(Mathexpr.function_of_xml "" fxml) [|Mathexpr.Scal x|]
											)
									)
								in
								(
									match aod cont_xml "EventComponent" "TimeFromSceneOnset" with
									"MousePositionX" -> Set_by_mouse_position_x given_function
									| "MousePositionY" -> Set_by_mouse_position_y given_function
									| _ -> Set_by_time_from_onset given_function
								)
							| "IncreaseValue" -> 
								Set_to_number (fun () -> reference := !reference +. 1.)
							| "DecreaseValue" -> 
								Set_to_number (fun () -> reference := !reference -. 1.)
							| "SetToValue" ->	(*set to value*)
								Set_to_number 
								( 
									fun () -> 
										reference := float_of_number (number_of_string (Xml.attrib cont_xml "Value")) 
								)
							| _ ->	(*any mathematical expression*)
								Set_to_number 
								(
									fun () -> 
										reference := Mathexpr.float_of_nom ((Mathexpr.function_of_xml "" cont_xml) [||])
								)
						)
					| "SetMatrixElementByEvent" ->
						let mat =
						(
							try
							(
								match varref with
								Mathexpr.Matrix k -> k
								| _ -> gl_failwith (identifier ^ " is no matrix and cannot be modified by SetMatrixElementByEvent.") 
							)
							with Not_found -> 
							gl_failwith ("Undeclared matrix: " ^ identifier) 
						) in
						let row_number_xml = List.hd (Xml.children (List.hd (Xml.children v))) 
						and col_number_xml = List.hd (Xml.children (List.nth (Xml.children v) 1)) in
						let row_and_col_function () =
							let (d1, d2) = Gsl_matrix.dims mat in
							let (r, c) =
							(
								(int_of_float (Mathexpr.float_of_nom ((Mathexpr.function_of_xml "" row_number_xml) [||]))) - 1,
								(int_of_float (Mathexpr.float_of_nom ((Mathexpr.function_of_xml "" col_number_xml) [||]))) - 1
							) in
							if 0 <= r && r < d1 && 0 <= c && c < d2 then
								(r, c)
							else
								gl_failwith ("SetMatrixElementByEvent: Index out of bounds.") 
						in
						let cont_xml = List.nth (Xml.children v) 2 in
						(
							match Xml.tag cont_xml with
							"FunctionOfEventComponent" ->
								let given_function x =
									let fxml = child_or_default cont_xml "Function" in
									let (r, c) = row_and_col_function () in
									mat.{r, c} <- 
									(
										match Xml.children fxml with
										[] -> x
										| _ ->
											Mathexpr.float_of_nom
											(
												(Mathexpr.function_of_xml "" fxml) [|Mathexpr.Scal x|]
											)
									)
								in
								(
									match aod cont_xml "EventComponent" "TimeFromSceneOnset" with
									"MousePositionX" -> Set_by_mouse_position_x given_function
									| "MousePositionY" -> Set_by_mouse_position_y given_function
									| _ -> Set_by_time_from_onset given_function
								)
							| "IncreaseValue" -> 
								Set_to_number 
								(
									fun () -> 
										let (r, c) = row_and_col_function () in
										mat.{r, c} <- mat.{r, c} +. 1.
								)
							| "DecreaseValue" -> 
								Set_to_number 
								(
									fun () -> 
										let (r, c) = row_and_col_function () in
										mat.{r, c} <- mat.{r, c} -. 1.
								)
							| _ ->	(*any mathematical expression*)
								Set_to_number 
								(
									fun () -> 
										let (r, c) = row_and_col_function () in
										mat.{r, c} <- Mathexpr.float_of_nom ((Mathexpr.function_of_xml "" cont_xml) [||])
								)
						)
						
					| _ -> (*SetStringVariableByEvent*)
						let reference = 
						(
							match varref with
							Mathexpr.String_variable v -> v
							| _ -> gl_failwith (identifier ^ " is no string variable and cannot be set by SetStringVariableByEvent.") 
						) in
						let cld = List.hd (Xml.children v) in
						(
							match Xml.tag cld with
							"StringFunctionOfEventComponent" ->
								let given_function x =
									let fxml = child_or_default cld "StringFunction" in
									reference := 
									(
										match Xml.children fxml with
										[] -> string_of_float x
										| _ ->
											Mathexpr.string_of_nom
											(
												(Mathexpr.function_of_xml "" fxml) [|Mathexpr.Scal x|]
											)
									)
								in
								(
									match aod cld "EventComponent" "TimeFromSceneOnset" with
									"MousePositionX" -> Set_by_mouse_position_x given_function
									| "MousePositionY" -> Set_by_mouse_position_y given_function
									| _ -> Set_by_time_from_onset given_function
								)
							| _ ->	(*any mathematical expression*)
								Set_to_number 
								(
									fun () -> 
										reference := Mathexpr.string_of_nom ((Mathexpr.function_of_xml "" cld) [||])
								)
						)
				in
				List.map var_to_set_of_xml variables_to_set_list_xml
			in
			{
				event_logfile_id = logfile_id;
				event_contents = conts;
				event_id = 
					if attrib_declared "EventID" ev then 
						Some (Xml.attrib ev  "EventID")
					else 
						None;
				log_event = number_of_string (aod ev "Log" "1.0");
				action = act;
				variables_to_set = variables_to_set_list;
				variables_to_log = 
					variables_to_log_list_of_string (aod ev "VariablesToLog" "");
				matrices_to_log = 
					matrices_to_log_list_of_string (aod ev "MatricesToLog" "");
				index_vectors_to_log = 
					index_vectors_to_log_list_of_string (aod ev "IndexVectorsToLog" "");
				list_indices_to_log = list_indices_to_log_list;
				log_before_set = 
					bool_of_string (aod ev "LogBeforeSet" "false")
			}
		in
		let scene_cursor_type =
			if not ( List.mem "CursorType" (fst (List.split (Xml.attribs tr))) ) then
				Glut.CURSOR_INHERIT
			else
				match (Xml.attrib tr "CursorType") with
				"RightArrow" -> 		Glut.CURSOR_RIGHT_ARROW
				| "LeftArrow" -> 		Glut.CURSOR_LEFT_ARROW
				| "Info" -> 			Glut.CURSOR_INFO
				| "Destroy" -> 		Glut.CURSOR_DESTROY
				| "Help" -> 		Glut.CURSOR_HELP
				| "Cycle" -> 		Glut.CURSOR_CYCLE
				| "Spray" -> 		Glut.CURSOR_SPRAY
				| "Wait" -> 		Glut.CURSOR_WAIT
				| "Text" -> 		Glut.CURSOR_TEXT
				| "Crosshair" -> 		Glut.CURSOR_CROSSHAIR
				| "UpDown" -> 		Glut.CURSOR_UP_DOWN
				| "LeftRight" -> 		Glut.CURSOR_LEFT_RIGHT
				| "TopSide" -> 		Glut.CURSOR_TOP_SIDE
				| "BottomSide" -> 	Glut.CURSOR_BOTTOM_SIDE
				| "LeftSide" -> 		Glut.CURSOR_LEFT_SIDE
				| "RightSide" -> 		Glut.CURSOR_RIGHT_SIDE
				| "TopLeftCorner" -> 	Glut.CURSOR_TOP_LEFT_CORNER
				| "TopRightCorner" -> 	Glut.CURSOR_TOP_RIGHT_CORNER
				| "BottomRightCorner" -> Glut.CURSOR_BOTTOM_RIGHT_CORNER
				| "BottomLeftCorner" -> Glut.CURSOR_BOTTOM_LEFT_CORNER
				| "None" -> 		Glut.CURSOR_NONE
				| "FullCrosshair" -> 	Glut.CURSOR_FULL_CROSSHAIR
				| _ -> 			Glut.CURSOR_INHERIT
		in
		
		let display_object_of_xml d =
			match Xml.tag d with 
			"Stimulus" -> stimulus_of_xml d
			| "Image" -> image_of_xml d
			| "Button" -> button_of_xml d
			| "ProcessStimulusList" -> Stimulus_list_reference (object_list_reference_of_xml d)
			| _ -> Image_list_reference (object_list_reference_of_xml d)
		in
		(*let (evlist, objectslist) = List.partition ( fun t -> Xml.tag t = "Event" ) (Xml.children tr) in*)
		let evlist = 
			let evl = List.filter ( fun t -> Xml.tag t = "Event" ) (Xml.children tr) in
			if bool_of_string (aod tr "ExitOnEscape" "true") then
				let stop_event =
					Xml.Element 
						("Event",
						["Log", "0"],
						[Xml.Element ("Key", [], [Xml.Element ("Special", ["SpecialKey", "Escape"], [])]);
						Xml.Element ("Action", [], [Xml.Element ("Exit", [], [])])])
				in
				stop_event::evl
			else evl
		and objectslist = 
			List.filter 
				( 
					fun t ->
						let obtags = ["Stimulus"; "Button"; "Image"; "ProcessStimulusList"; "ProcessImageList" ] in
						List.mem (Xml.tag t) obtags 
				) 
				(Xml.children tr)
		and bgcolor = 
			try color_list_of_string (color_string (aod tr "BackgroundColorRGB" "0.0, 0.0, 0.0")) 
			with Failure f -> gl_failwith f 
		in
		{
			background_color = bgcolor;
			cursor_type = scene_cursor_type;
			display_object_list = List.map display_object_of_xml objectslist;
			event_list = List.map event_of_xml evlist;
			local_declarations = local_declarations;
			variation_list = var_list
		};
	in	(*end of scene_of_xml...*)
	let errormsgscene = 
		error_message_scene 
			~canvas_width:displ.width
			~canvas_height:displ.height 
			~horizontal_range:displ.horizontal_range
			~vertical_range:displ.vertical_range
			"Runtime Error. You are probably using variables secifying colors, and one variable took a value which is invalid for colors."
	in
	let scenelist = List.map ( fun t -> (Xml.attrib t "SceneID", scene_of_xml t) ) xmlscenelist in
	(*check scenelist for uniqueness of SceneIDs:*)
	let sceneIDlist = fst ( List.split scenelist) in
	let rec unique = function
		| [] -> []
		| x :: xs -> [x] @ (unique (List.filter (fun a -> a<>x) xs))
	in
	if List.length sceneIDlist <> List.length (unique sceneIDlist) then
		gl_failwith "The scene IDs are not unique. At least one ID occurs more than once."; 
		
	let q = Queue.create () in
	List.iter (sequence_of_xml q) xmlseqlist;
	(*check whether each scene ID of the sequences is in sceneIDlist:*)
	let contained =
		Queue.fold (fun a b -> a && (List.mem b.sceneID sceneIDlist)) true q
	in
	if not contained then
		gl_failwith "At least one scene ID in the sequences is not declared." ;

	(*Return experiment:*)
	{
		display = displ;
		scene_list = errormsgscene::scenelist;
		sequence = q;
		logfile = (aod input_structure "LogFile" "logfile.xml");
		check_framerate = !frame_rate_must_be_checked || (bool_of_string (aod input_structure "CheckRefreshRate" "true"));
		event_log = logging;
		deviation_queue = Queue.create ();
		event_queue = Queue.create ()
	}
;;


(*evaluate mathematical expression [elem] (xml);
either print results (then return None) or return xml for graphical output:*)
let eval_mathexp output_file elem =
	let is_source = ref false in
	let evaluated_expression = 
		try
		(Mathexpr.function_of_xml "" elem) [||]
		with Failure s -> gl_failwith s 
	in
	match evaluated_expression with
	Mathexpr.Xml (Xml.Element ("EvaluateGraphically", [], h::t)) ->
		Some h	(*Evaluate graphical contents, e. g. "Experiment"*)
	| _ ->
	(
		let mathstring =
			let str_of_float f =
				let s1 = string_of_float f in
				if Str.last_chars s1 1 = "." then
					Str.string_before s1 ((String.length s1) - 1)
				else
					s1
			in
			let string_of_matrix  mat =
				let s = ref "\n" in
				let (r, c) = Gsl_matrix.dims mat in
				for i = 0 to r - 1 do
					for j = 0 to c - 1 do
						s := !s ^ (str_of_float mat.{i, j});
						if i < r-1 || j < c-1 then
							s := !s ^ ", "
					done;
					s := !s ^ "\n"
				done;
				!s
			in
			try
			match evaluated_expression with
			Mathexpr.Scal s -> str_of_float s
			| Mathexpr.Mat m -> string_of_matrix m
			| Mathexpr.Str s -> s
			| Mathexpr.Xml x -> (is_source := true; Fls.fls_string_of_xml x)
			with
				Failure s -> gl_failwith s 
		in
		match output_file with
		"" -> 
		(
			let number_of_rows =
				List.length (Str.split (Str.regexp_string "\n") mathstring)
			in
			let l = String.length mathstring in
			let fontstring =
				if l > 1000 || number_of_rows > 20 then
					(if !is_source then "Fixed8x13" else "Helvetica10")
				else
					(if !is_source then "Fixed9x15" else "Helvetica18")
			in
			let cut_mathstring = 
				let s1 = 
					if number_of_rows > 40 then
						let new_length = (l / number_of_rows) * 40 in
						(Str.string_before mathstring new_length) ^ " ... "
					else
						mathstring
				in
				if String.length s1 > 3000 then
					(Str.string_before s1 3000) ^ " ... "
				else
					s1
			in
			let mstring = 
				let ss = Str.global_replace (Str.regexp_string "\n") "\\n" cut_mathstring in 
				Str.global_replace (Str.regexp_string "\"") "\\\"" ss 
			in
			let maxheight = 1.3 in
			let h =
				if l<900 && number_of_rows < 8 then
					maxheight /. (float_of_int (10 - number_of_rows))
				else
					maxheight
			in
			let hs = string_of_float h in
			let lower_pos =
				string_of_float
				(
					((maxheight -. h) /. 2.) -. 0.65
				)
			in
			let s = 
			"(Experiment [CheckRefreshRate = \"false\"]
				(SimpleDisplayMode [FullScreen = \"false\"])
				(Scene [SceneID = \"textfield\"]
					(Stimulus [Position = \"-0.99, " ^ lower_pos ^ "\"]
						(TextField [Width = 1.9, Height = " ^ hs ^ ", Alignment = \"Left\", Font = \"" ^ fontstring ^ "\", Text = \"" ^ mstring ^ "\"]))
							(Button [ButtonID = \"Exit\", Position = \"-0.15, -0.74\", Width = \"0.3\", Height = \"0.07\", Text = \"Exit\"]
							(Action (Exit))))
				(Sequence
					(Presentation [Duration = -1, SceneID = \"textfield\"])))"
			in
			let x = List.hd (Fls.xml_list_of_fls_string s) in
			Some x
			(*let ex = experiment_of_xml x in
			Display.init ex;*)
		)
		| f -> 
		(
			let out = 
				if f = "-" then
					stdout
				else
					open_out f
			in
			Printf.fprintf out "%s" mathstring;
			close_out out;
			None
		)
	)	(*end of non-graphical evaluation*)
;;

(*display a simple opengl window; in order to test whether opengl is working on your platform*)
let gltestfun () =
	let rectangle ~x ~y ~width ~height =
		GlDraw.begins `polygon;
			GlDraw.vertex ~x ~y ();
			GlDraw.vertex ~x ~y:(y +. height) ();
			GlDraw.vertex ~x:(x +. width) ~y:(y +. height) ();
			GlDraw.vertex ~x:(x +. width) ~y ();
		GlDraw.ends ();
	in
	let dispfun () =
		GlClear.color (0.0, 0.0, 0.0);
		GlClear.clear [`color];
		GlDraw.color (1.0, 1.0, 1.0);
		GlMat.mode `projection;
		GlMat.load_identity ();
		GlMat.ortho ~x:(-1.0,1.0) ~y:(-1.0,1.0) ~z:(-1.0,1.0);
		GlMat.push ();
		GlMat.translate ~x:(-0.5) ~y:(-0.5) ();
		GlMat.push ();
		GlMat.translate ~x:(-0.25) ~y:(-0.25) ();
		rectangle ~x:(-0.2) ~y:(-0.2) ~width:(0.4) ~height:(0.4);
		GlMat.pop ();
		GlMat.pop ();
		
		GlMat.translate ~x:0.5 ~y:0.5 ();
		rectangle ~x:(-0.2) ~y:(-0.2) ~width:(0.4) ~height:(0.4);
		Gl.flush ()
	in
	(*ignore(Glut.init Sys.argv);*)
	Glut.initDisplayMode ~alpha:true ~depth:true () ;
	Glut.initWindowSize ~w:500 ~h:500 ;
	ignore(Glut.createWindow ~title:"lablglut & LablGL");
	Glut.displayFunc ~cb:(dispfun);
	(*Glut.mainLoop();
	exit 0*)
;;




let main () =
	ignore(Glut.init Sys.argv);
	(match operating_system with
	| MacOSX -> ignore 
		(Sys.command 
			("defaults write com.apple.glut GLUTSyncToVBLKey -bool " 
			^ !macosx_default_sync_var));
	| _ -> ());
	
	let inputfile = ref (Input_file (Filename.concat executable_dir "default.fld")) in
	let set_input_file filename = inputfile := (Input_file filename) in
	let encoding = ref "UTF-8" in
	let outputfile = ref "" in
	let error_outputfile = ref "" in
	let internal_dtd_file = ref None in	(*file containing internal DTDs for fls files*)
	let argv_array = ref [||] in
	let speclist =
	[
		(
			"-fld2fls", 
			Arg.String (fun filename -> Fls.fls_file_of_xml_file filename; exit 0),
			"\tconvert fld file to fls file"
		);
		(
			"-fls2fld",
			Arg.String 
			(
				fun filename -> 
					Fls.xml_file_of_fls_file 
						~encoding:!encoding 
						~internal_dtd:!internal_dtd_file 
						filename; 
					exit 0
			),
			"\tconvert fls file to fld file"
		);
		(
			"-stdin_fls",
			Arg.Unit (fun () -> inputfile := Stdin_fls),
			"\tread fls from standard input"
		);
		(
			"-stdin_fld",
			Arg.Unit (fun () -> inputfile := Stdin_fld),
			"\tread fld from standard input"
		);
		(
			"-o",
			Arg.Set_string outputfile,
			"\t\t-o filename : write result of evaluation of\n\t\tmathematical expression to file filename\n\t\t(default: output into flashdot window)\n\t\t-o - : send result to STDOUT (UNIX only)"
		);
		(
			"-o_stderr",
			Arg.Set_string error_outputfile,
			"\t-o_stderr filename : write errors to file filename\n\t\t-o_stderr - : send errors to STDOUT (UNIX only)\n\t\t-o_stderr -- : send errors to STDERR (UNIX default)\n\t\t-o_stderr -@ : graphical error output (Windows default)"
		);
		(
			"-logfile",
			Arg.Set_string log_file,
			"\t-logfile filename : overwrite the (first) event log file with filename"
		);
		(
			"-encoding",
			Arg.Set_string encoding,
			"\tspecify the encoding of the fld file; place before option fls2fld"
		);
		(
			"-internalDTD",
			Arg.Unit (fun () -> internal_dtd_file := Some (dtd_string_from_file default_dtd_file)),
			"\tadd internal DTD to fld file; place before option fls2fld"
		);
		(
			"-permit_exec",
			Arg.Set allow_external_execution,
			"\tpermit execution of external processes, i.e. run external programs\n\t\tpotential SECURITY RISK!\n\t\tset this option only if you are absolutely sure"
		);
		(
			"-gltest", 
			Arg.Unit gltestfun,
			"\tdisplay a simple opengl window;\n\t\tin order to test whether opengl is working on your platform"
		);
		(
			"-version",
			Arg.Unit (fun () -> Printf.printf "FlashDot version %s\n" flashdot_version; exit 0),
			"\tprint flashdot version and exit (UNIX only)"
		);
		(
			"-argv",
			Arg.Rest (fun x -> argv_array := Array.append !argv_array [|x|]),
			"\tregard all following strings as attribute macros\n\t\twhich replace \"ARGV1;\", ..., \"ARGVn;\"\n\t\t(this needs to be the last argument)"
		)
		
	] in
	Arg.parse speclist set_input_file "\nUsage: flashdot [options...] inputfile [options...]\n\nWhere options include:";
	standard_error_output := (
		match !error_outputfile with
		"" -> !standard_error_output
		| "-" -> Err_file_handler stdout
		| "--" -> Err_file_handler stderr
		| "-@" -> Err_graphical
		| error_file_name -> 
			try
			let e = open_out error_file_name in
			Err_file_handler e
			with _ -> gl_failwith ("Error output file " ^ error_file_name ^ " could not be opened")
	);
	let input_structure = parse_xml !inputfile in
	let rec process_input inp =
		match Xml.tag inp with
		"Experiment" -> 
			let clean_fls = Fls.convert_fls_tags inp in
			let expand_macros = Macros.expand_attribute_macros !argv_array clean_fls in
			let ex = experiment_of_xml expand_macros in
			Display.init ex;
		| "Expression" | "MathematicalExpression" -> 
		(
			Mathexpr.generate_declarations inp;
			let xexpr = List.hd (List.rev (Xml.children input_structure)) in
			match (eval_mathexp !outputfile xexpr) with
			None -> ()	
			| Some new_inp -> process_input new_inp
		)
		| "Scene" ->
			let sc =
				(*add event: if Escape then exit*)
				let stop_event =
					Xml.Element
					(
						"Event",
						["Log", "0"],
						[
							Xml.Element 
							(
								"Key", [],
								[Xml.Element ("Special", ["SpecialKey", "Escape"], [])]
							);
							Xml.Element 
							(
								"Action", [],
								[Xml.Element ("Exit", [], [])]
							)
						]
					)
				in
				Xml.Element
				(
					Xml.tag inp, Xml.attribs inp, stop_event :: (Xml.children inp)
				)
			in
			let scene_id = Xml.attrib sc "SceneID" in
			let ex = Xml.Element
			(
				"Experiment",
				["CheckRefreshRate", "false"],
				[
					Xml.Element
						("SimpleDisplayMode", ["FullScreen", "false"], []);
					sc;
					Xml.Element
					(
						"Sequence", [], 
						[
							Xml.Element 
							(
								"Presentation", 
								[("SceneID", scene_id); ("Duration", "-1"); (*"WakeUpOnEventMode", "true"*)], []
							)
						]
					)
				]
			) in
			process_input ex
		| "Stimulus" | "Image" | "Button" ->
			let sc =
				Xml.Element
				(
					"Scene",
					["SceneID", "tempscene"],
					[inp]
				)
			in
			process_input sc
		| s -> gl_failwith ("The first tag of your input file has to be \"Experiment\", \"Expression\", \"MathematicalExpression\", \"Scene\", \"Stimulus\", \"Button\" , or \"Image\", but it is \"" ^ s ^ "\"." ) 
	in
	process_input input_structure
;;

let _ = main ();;
